(** This module is a centralized place for managing case analysis.
 * Satisfiability procedures can announce cases, with the idea that if ALL of
 * the cases are proved (lead to a contradiction), then a contradiction arises. *)
module E = Errormsg
module L = Logic
module P = Proof
module U = Util
module S = Satproc
module IH = Inthash
module O = Options

open Satproc              (* message, etc. *)
open Pretty
open Edag

let debug = ref false     (* debug flag, settable from command line: "-debug CA" *)

let log fmt = ignore (E.log "CA: "); E.log fmt

 
(* in debug mode, print a log string *)
let debugLog (s: string) : unit =
begin
  if !debug then
    (ignore (log "%s\n" s));
end

let id = ref 0
let newId () : int = 
  incr id;
  !id



type clause = {
    cid: int; (** A unique identifier for this clause *)

    mutable cactive: bool; (** Whether the clause is active or not *)

    cases: case list;
    (** A case is announced when there is a proof of a disjunction of the
     * cases. Each case can be a disjunction of messages. Essentially the 
     * clause itself is given in DNF form. *)

    mkProof: proofgen;
    (** A way to generate a contradiction (proof of false) given a 
     * hypothetical proof of falsehood for each of the disjuncts (viewed as a 
     * conjunction). For each such proof we give the name of the hypothesis 
     * that the message body holds. This is invoked if the cases analysis 
     * succeeds. *)

    depth: int;
    (** The depth of a clause is one more than the depth of the last
        clause on which we have split, at the time when the clause was
        created. Clauses are created after splits because of matching only. *)

    corigin: corigin;
    (** The origin of the clause *)

    mutable cscore: int; 
    (** This is decremented every time the clause is _irrelevant_ to finding 
     * contradiction *)

    cgoal: bool; 
    (** If this clause is from a goal *)
  }

(** Keep a mapping from clause IDs to clause origin, for debugging *)
and corigin = 
    CONone
  | COInst of U.symbol (** Instantiation of an axiom *)


(** A clause is a disjunction of cases. A case is a conjunction of message 
 * bodies *)
and case = { 
    conjuncts: elemcase list;  (** A conjunction of messagebodies *)

    (** If we know that a case is false, it is not considered anymore. In 
     * that case, we store a hypothetical proof of falsehood, on the 
     * assumption that the case is true *)
    mutable cfalse: (Util.symbol -> Proof.proof) option;


    cpred: Logic.pred; (** The predicate that corresponds to the assumption 
                        * that this case is true. *)
  } 

and elemcase = {

    mbproof: case -> int -> Util.symbol -> Proof.proof; 
    (** The proof of the messagebody, given the host case, the index within 
     * it, and the hypothetical proof of the host case *)

    ecnegated: bool; (* Whether the ecshared is used negated here *)
    ecshared: elemcaseshared;
  }

(** We will hash the elemcaseshared, to make sure we share them *)
and elemcaseshared = { 
    mutable ecrefcnt: int; (* How many times it occurs in clauses *)
            echash: int; (* The hash value, based on the messagebody. *)

    mb: messagebody; (** This is the actual message body *)

    mutable determined: testResult;
  }
    
(** A proof generator for a clause, takes a list of hypothetical proofs of 
 * false by assuming each of the cases, and produces a proof of false *)
and proofgen =  (Util.symbol * Proof.proof) list -> Proof.proof


let d_sharedmb () (c: elemcase) = 
  (if c.ecnegated then text "NEG " else nil) ++
    (d_messagebody () c.ecshared.mb)

let d_conjuncts () (cl: elemcase list) = 
  align ++ 
    (docList ~sep:(line ++ text "AND ")
       (fun c -> d_sharedmb () c) () cl)
    ++ unalign


let d_case () (c: case) = d_conjuncts () c.conjuncts
  
let d_clause () (cd: clause) = 
  dprintf "%d:\n  @[   %a@]" cd.cid 
    (docList ~sep:(line ++ text "OR ") (d_case ())) cd.cases


let d_flatConjuncts () (cl: elemcase list) =
  if ((List.length cl) > 1) then (
    (text "(") ++
    (docList ~sep:(text " AND ")
       (fun c -> d_sharedmb () c) () cl) ++
    (text ")")
  )
  else
    (d_messagebody () (List.hd cl).ecshared.mb)

let d_flatClause () (cd: clause) =
  dprintf "%a"
    (docList ~sep:(text " OR ")
       (fun c -> d_flatConjuncts () c.conjuncts)) cd.cases

let positiveMessageBody (c: elemcase) : messagebody = 
  if c.ecnegated then 
    try 
      match c.ecshared.mb with 
      | Intern (bop, n1, n2) -> Intern (L.negateBinopRaise bop, n1, n2)
      | Extern _ -> assert false
    with L.CannotNegate _ -> 
      assert false
  else
    c.ecshared.mb

let memoFreeHyps: bool ref = ref true

(** Keep a hash table for the elemcaseshared *)
module ECH = H.Make(struct
                     type t = elemcaseshared
                     let equal (e: elemcaseshared) (e': elemcaseshared) = 
                       match e.mb, e'.mb with 
                         Intern (bop, n1, n2), Intern(bop', n1', n2') -> 
                           bop = bop' && n1.id = n1'.id && n2.id = n2'.id
                       | _ -> false
                     let hash e = e.echash 
                    end) 
let elemcaseTable : elemcaseshared ECH.t = ECH.create 113

let makeElemCaseShared (mb: messagebody) : elemcaseshared * bool = 
  (* First, see if we ought to negate the messagebody to increase the 
   * probability that we have it already *)
  let mb', neg = 
    match mb with 
      Intern (bop, n1, n2) -> begin
        match bop with 
          L.SetNe -> Intern (L.SetEq, n1, n2), true
        | L.SetLt -> Intern (L.SetGe, n1, n2), true
        | L.SetGt -> Intern (L.SetLe, n1, n2), true
        | _ -> mb, false
      end
    | Extern _ -> mb, false
  in
  (* let mb', neg = mb, false in *)
  let h = 
    match mb' with
      Intern (bop, n1, n2) -> H.hash (bop, n1.id, n2.id)
    | Extern p -> H.hash p
  in
  (* Now see if we have it already *)
  let res = 
    { mb = mb';
      ecrefcnt = 1;
      echash = h;
      determined = NotDetermined
    }
  in
  try 
    let res' = ECH.find elemcaseTable res in
    (* Increment the reference count *)
    res'.ecrefcnt <- 1 + res'.ecrefcnt;
    res', neg
  with Not_found -> begin
    (* Add it to the table *)
    ECH.add elemcaseTable res res;
    res, neg
  end

let releaseElemCaseShared (ecs: elemcaseshared) : unit = 
  if ecs.ecrefcnt <= 1 then begin
    (* We must deallocate it *)
    ECH.remove elemcaseTable ecs
  end else
    ecs.ecrefcnt <- ecs.ecrefcnt - 1
    

let elemOfMsg (msg: messagebody) : elemcase = 
  let mbprf (a: case) (idx: int) (casevar: U.symbol) : P.proof = 
    match a.conjuncts with 
      [ a ] -> P.Hyp casevar
    | al -> P.Ande (idx, P.Hyp casevar)
  in
  let ec, neg = makeElemCaseShared msg in
  { ecshared = ec;
    ecnegated = neg;
    mbproof = mbprf;
  }


let dropIrrelevantCases = ref true

(** Maintain the clauses as a mutable list. This way we can eliminate 
 * easily elements from the middle *)
type 'a mlist = { 
    mutable next: 'a mlist;  (** If next points to itself, then we have end of 
                           * list *)
    mutable data: 'a;
  } 
(* Make a dummyDlist *)
let singleML (d: 'a) : 'a mlist =   
  let c = { data = d;
            next = (Obj.magic ()) } in 
  c.next <- c;
  c

let consML (d: 'a) (next: 'a mlist) : 'a mlist = 
  { data = d;
    next = next }
  
let tailML (l: 'a mlist) : 'a mlist = 
  if l.next != l then l.next else raise (Failure "tailML")


(** Maintain allClauses as a skip list *)
type clistEntry = { 
            cl: clause;  (** The actual clause *)
    mutable nextactive: clistEntry list; (** The next active element *)
  }


let allClauses: clistEntry list ref = ref []

let init () =
  (debugLog "initialize")

let internalAddCaseAnalysis (cd: clause) : unit =
  if !debug then
    ignore (log "add new clause %a\n" d_clause cd);
  if !doProofTrace then
    ignore (E.logg "%sclause %d: %a\n"
                   (proofTraceInd ())
                   cd.cid
                   d_flatClause cd);
  allClauses := { cl = cd; nextactive = !allClauses} :: !allClauses;
  registerUndoRun 
    (fun _ -> 
      (* Remove the case from the priority queue *)
      if !debug then ignore (log "retracting clause %d\n" cd.cid);
      allClauses := List.tl !allClauses;
      (* Delete the references to the shared elemcases *)
      List.iter 
        (fun (c:case) -> 
          List.iter 
            (fun (ec: elemcase) -> releaseElemCaseShared ec.ecshared)
            c.conjuncts) 
        cd.cases
    )
    

(** Make a clause out of a predicate, along with a way to make a proof of 
 * false from a hypothetical proof of falsehood. The hypothesis is that the 
 * predicate itself is true. *)
let addUnitClauseOfPredicate 
                (p: L.pred) 
                (contraproof: U.symbol -> P.proof -> P.proof)
                ~(ingoal: bool)  : unit = 
  let c = 
    { cid = newId (); 
      depth = 0;
      corigin = CONone;
      cactive = true;  
      cases = [ { conjuncts = [ elemOfMsg (Extern p) ];
                  cpred  = p;
                  cfalse = None } ];
      mkProof = 
      (fun prfl -> 
        match prfl with 
          [ (h, pfalse) ] -> contraproof h pfalse
        | _ -> assert false);

      cscore = 0;

      cgoal = ingoal;
      
    }
  in
  internalAddCaseAnalysis c


let corigin: corigin IH.t = IH.create 113

let selupdSym = U.registerSymbolName "selupd"
let unknownSym = U.registerSymbolName "unknown"

let d_corigin () (co: corigin) = 
  match co with 
    CONone -> text "unknown"
  | COInst n -> text (U.symbolName n)

let getClauseOrigin (prf: P.proof) : corigin = 
  match prf with
  | P.AnnotE alleprfs 
  | P.Ore (_, _, P.AnnotE alleprfs, _) 
  | P.Impe (alleprfs, _) -> 
      let rec getInst(prf: P.proof) : U.symbol  = 
        match prf with 
          P.Alle (e, p) -> getInst p
        | P.Hyp h -> h
        | P.Alli ("sa", _) -> (* this must be the selupd *)
            selupdSym
        | _ -> (*
            ignore (E.warn "unknown proof of forall: %a\n" 
                      P.d_prf prf);
            *)
            unknownSym
        in
        let n = getInst alleprfs in 
        COInst n

  | _ -> (*
      ignore (E.warn "Unknown clause origin: %a\n" P.d_prf prf);
      *)
      CONone

(** This is the depth of the last clause we have split on *)
let currentSplitDepth = ref 0 

let maxMatchSplitDepth = ref 5
 
(** This is a simpler way to convert a predicate to a clause. We assume that 
 * the predicate is in DNF. *)
let addClauseOfPredicate (p: L.pred) (prf: P.proof) ~(ingoal:bool) = 
  (* Break a predicate into a list of disjunctions. Also return a proof 
   * generator that is given a proof of p, a list of hypothetical proofs of 
   * false with the hypotheses being the disjunctions, and return a proof of 
   * false *)
  let rec breakDisjunctions (p: L.pred) : 
      L.pred list * (P.proof -> (U.symbol * P.proof) list -> P.proof) = 
    match p with
      L.Or cl -> 
        cl,
        (fun pfp cases -> 
          (* We must have as many cases as we have cl *)
          assert (List.length cases = List.length cl);
          let p = P.Ore (L.Bool false, cl, pfp, cases) in 
          let p = 
            if !memoFreeHyps then begin
              mkFreeHyps p
            end else
              p
          in
          p)

    | L.Imply (p1, p2) -> 
        let p2l, p2p = breakDisjunctions p2 in
        let p1n = try L.negatePredRaise p1 
                  with L.CannotNegate _ -> assert false in 
        (p1n :: p2l),
        (fun pfp cases -> 
          match cases with 
            (v1,pf1false) :: restcases -> 
              (* Make a proof of p1 *)
              let pf1 = P.mkContraProof p1 v1 pf1false in 
              (* Make a proof of p2 *)
              let pf2 = P.Impe (pfp, pf1) in 
              (* Now make a proof of false *)
              p2p pf2 restcases
          | _ -> assert false)

    | _ -> [ p ], 
           (fun pfp cases -> 
             let s, p = match cases with [ (s, p) ] -> s, p | _ -> assert false in
             P.Cut (pfp, s, p))
  in
  let cases, mkProof = breakDisjunctions p in 
  (* Each case gets transformed into a Cases.case *)
  let makeCase (pc: L.pred) : case = 
    { conjuncts = begin
        let conjs: L.pred list = 
          match pc with 
            L.And conjs -> conjs
          | _ -> [ pc  ]
        in
        let makeElemCase (pe: L.pred) : elemcase = 
          let mb, pe_mb, _ = messageBodyOfPredicate pe in
          let ec, neg = makeElemCaseShared mb in
          { ecshared = ec;
            ecnegated = neg;
            mbproof = (fun c idx casevar ->
                match pc with 
                  L.And _ -> 
                    applyProofTrans pe_mb (P.Ande (idx, P.Hyp casevar))
                | _ -> applyProofTrans pe_mb (P.Hyp casevar));
          }
        in
        List.map makeElemCase conjs;
      end;

      cfalse = None;
      cpred = pc;
    }
  in
  let co = getClauseOrigin prf in 
  let cd = { cid = newId ();
             depth = 1 + !currentSplitDepth;
             corigin = co;
             cactive = true;
             cases = List.map makeCase cases;
             mkProof = mkProof prf;
             cscore = 0;
             cgoal = ingoal } 
  in 
  if !debugStats then 
    IH.add corigin cd.cid co;
  internalAddCaseAnalysis cd



(** Deactivate a clause *)
let deactivateClause (cd: clistEntry) : unit = 
  let cl = cd.cl in 
  assert (cl.cactive);
  cl.cactive <- false;
  registerUndoRun 
    (fun _ ->
      (* Add the case back *)
      if !debug then ignore (log "reactivating clause %d\n" cl.cid);
      cl.cactive <- true)

        
  (** Do a quick test of a conjunction of message bodies. Return true, false, 
   * or undecided. In the case of false, also set its proof of falsehood *)
let testCase (c: case) (cid: int) : bool option =
  assert (c.cfalse = None);
  let idx = ref (-1) in 
  let rec loopConj (alltruesofar: bool) = function 
      [] -> if alltruesofar then Some true else None
    | p :: rest -> begin
        incr idx;
        let ps = p.ecshared in 
        let det = 
          match ps.determined with 
            NotDetermined -> 
              let det' = testMessageBody ps.mb in 
              if det' != NotDetermined then begin
                if !doProofTrace then
                  ignore (E.logg "%s%a is %s, simplifying clause %d\n"
                                 (proofTraceInd ())
                                 d_messagebody ps.mb
                                 (match det' with Determined (b, _) -> string_of_bool b
                                                | _ -> "(oops)")
                                 cid);
                ps.determined <- det';
                registerUndoRun (fun _ -> ps.determined <- NotDetermined)
              end;
              det'
          | det' -> det'
        in
        match det with
          Determined(detval, prf) -> 
            if detval = p.ecnegated then begin
              c.cfalse <- 
                (let i = !idx in 
                Some (fun casevar -> 
                  let prfmb = p.mbproof c i casevar in 
                  P.G("not_FI", [ P.P prfmb; P.P prf ])));
              registerUndoRun (fun _ -> c.cfalse <- None);
              Some false
            end else 
              loopConj alltruesofar rest

        | NotDetermined -> loopConj false rest
    end
  in
  loopConj true c.conjuncts

exception DoThisClause of clistEntry * int


(** Keep here the unit clauses that we have found in the last scan. They are 
 * still active, so nothing is lost if we drop this list. The first one is 
 * the last added. *)
let unitClauses : clistEntry list ref = ref []
 
let useSubsumption = ref true

(** How many disjuncts to scan *)
let stopScanningAt = 2

(** Get the case analysis to be done next, along with how many active cases 
 * it has *)
let getFirstCaseAnalysis () : (clause * int) option =
  (* We scan all cases, and we try to clean them up. If we find disjuncts 
   * that are TRUE, we eliminate the whole case. If we find disjuncts that 
   * are FALSE, we simplify that cases. *)


  (** Scan the disjuncts and remember how many are still active. 
   * Return -1 if we must drop the clause. Stop scanning after we get the 
   * required number of active disjuncts. *)
  let rec testDisjuncts ~(cid:int) ~(stopat: int) (manykept: int) = function
      [] -> manykept
    | c :: restc -> 
        if manykept = stopat then 
          manykept (* Stop testing if we have reached our goal *)
        else if c.cfalse != None then 
          testDisjuncts ~cid:cid ~stopat:stopat manykept restc
        else begin 
          match testCase c cid with 
            Some true -> -1 (* drop this clause entirely *)
          | Some false -> testDisjuncts ~cid:cid ~stopat:stopat manykept restc
          | None -> (* keep it *)
              testDisjuncts ~cid:cid ~stopat:stopat (manykept + 1) restc
        end
  in

  (* Keep here the last active clause we see, in case we do not find any unit 
   * clauses *)
  let defaultClause: (clistEntry * int) option ref = ref None in 

  let nrActiveClauses = ref 0 in
  let nrInactive = ref 0 in 
  
  (** We leave this turned off, because it seems to increase the running time*)
  let skipInactiveClauses = false in

  let rec scanClauses (lastactive: clistEntry option) = function 
      [] -> ()
    | (cd :: rest) as this when cd.cl.cactive -> begin
        let cl = cd.cl in
        incr nrActiveClauses;
        (match lastactive with 
        | None -> () (* this is the first active clause *)
        | Some l -> 
            let old = l.nextactive in 
            l.nextactive <- this;
            registerUndoRun (fun _ -> l.nextactive <- old));

        (match testDisjuncts ~cid:cl.cid 
                             ~stopat:stopScanningAt 0 cl.cases with 
        | -1 ->             
            deactivateClause cd;
            if !debug then 
              ignore (log "drop tautology clause %a\n" d_clause cl)
                
        | 0 -> (* We derived a contradiction *)
            if !debug then 
              ignore (log "found empty clause %d\n" cl.cid);
            deactivateClause cd;
            raise (DoThisClause (cd, 0))
              
        | 1 -> (* A unit clause. Return this one *)
            if !debug then 
              ignore 
                (log "found unit clause %a\n" d_clause cl);
            (* will undo the whole list at once, further down. *)
            unitClauses := cd :: !unitClauses
              
        | n -> begin (* Not a singleton, keep going *)
            match !defaultClause with 
              None -> 
                defaultClause := Some (cd, n)
            | Some (oldcd, oldn) -> begin
                (* It turns out that it is very good to use the last
                   clause added, so we prefer to use the new one *)

                let usenew =  (cd.cl.depth <= oldcd.cl.depth) ||
                              (cd.cl.cscore >= oldcd.cl.cscore) in
                if usenew then defaultClause := Some (cd, n)
            end
        end

        );
        
        if skipInactiveClauses then begin
          scanClauses (Some cd) cd.nextactive;
        end else 
          scanClauses None rest
    end
    | _ :: rest -> incr nrInactive; scanClauses lastactive rest
  in

  let doUnitClause () : clause * int = 
    match !unitClauses with 
      cd :: rest -> 
        unitClauses := rest;
        deactivateClause cd;
        (* No need to undo this, because on undo, we just discard all the 
         * unitClauses *)
        cd.cl, 1

    | [] -> assert false
  in
  let res = 
    match !unitClauses with 
      [] -> begin (* No unit clauses remembered from last time we looked. 
                   * Look again, and see if we must do a clause *)
        let getOneImmediately: (clistEntry * int) option = 
          try scanClauses None !allClauses; None
          with DoThisClause (cls, n) -> Some (cls, n)
        in
(*
        ignore (log "Searched %d clauses, scanned %d active and %d inactive. Found %d unit clauses\n"
                  (List.length !allClauses)
                  !nrActiveClauses
                  !nrInactive
                  (List.length !unitClauses) 
                  );
*)
        (* Make sure to restore the unitClauses on undo *)
        if !unitClauses != [] then begin
          (* unitClauses := List.rev !unitClauses; *)
          registerUndoRun (fun _ -> unitClauses := []);
        end;
        match getOneImmediately with 
          Some (cd, n) -> Some (cd.cl, n)
        | None -> begin
            (* Look in the unitClauses *)
            if !unitClauses != [] then 
              Some (doUnitClause ())
            else begin 
              (* We get here if we did not find a unit clause to do *)
              match !defaultClause with 
                None -> 
                  if !debug then 
                    ignore (log "no more clauses to try\n");
                  None
                    
              | Some (cd, n) -> 
                  if !debug then 
                    ignore (log "found clause %a\n" d_clause cd.cl);
                  deactivateClause cd; 
                  assert (n >= 2 && n <= stopScanningAt); 
                  (* Because we stop scanning at 2 *)
                  (* We might have stoped testing all the disjuncts. Finish 
                   * testing them now *)
                  let activecount = 
                    if n = stopScanningAt then 
                      (* We might have stoped prematurely *)
                      testDisjuncts ~cid:cd.cl.cid 
                                    ~stopat:1000 0 cd.cl.cases
                    else
                      n
                  in
                  assert (activecount >= 2);
                  Some (cd.cl, activecount)
            end
        end
      end

    | _ -> Some (doUnitClause ())
  in
  (match res with 
    Some (c, _) -> (* It is already removed *) assert (not c.cactive)
  | _ -> ());
  res


(** Have here a function that converts a predicate into DNF. We use this to 
 * simplify the assumptions. Should descend into the AND and quantifiers **)
let convertToDNF (p: L.pred) : L.pred = 
  (* Finish a disjunction of conjuncts *)
  let toPred (l: L.pred list list) : L.pred = 
    let toConj (andl: L.pred list) = 
      match andl with
        [] -> L.Bool true
      | [ a ] -> a
      | _ -> L.And andl
    in
    match l with 
      [] -> L.Bool false
    | [ d ] -> toConj d
    | _ -> L.Or (List.map toConj l)
  in

  (* Return a disjunction of conjuncts that is equivalent to the predicate *)
  let rec loop (p: L.pred) : L.pred list list =  
    match p with 
      L.Bool b -> if b then [ [] ] else [ ]
    | L.PNez _ | L.PFun _ | L.All _ | L.Exist _ -> [ [ p ] ]
    | L.Not p -> loop (L.negatePred p)
    | L.Or orl -> 
        (* make them in the right order *)
        List.fold_right (fun p acc -> loop p @ acc) orl []

    | L.Imply (p1, p2) -> 
        (loop (L.negatePred p1)) @ loop p2

    | L.And andl -> 
        (* Convert each one separately. As you convert, deal with empty and 
         * singleton disjunctions. *)
        let processOneConjunct (conj: L.pred) 
                               ((acc: L.pred list list),
                                (acclen: int)) : L.pred list list * int = 
          let conj': L.pred list list = loop conj in 
          match conj' with 
            [] -> [], 0 (* Drop the whole acc thing *)
          | [ c1 ] -> (* Push this into the rest *)
              (List.map (fun acc1 -> c1 @ acc1) acc), acclen

          | _ -> 
              let conj'l = List.length conj' in 
              if acclen * conj'l > 32 then (* It is getting too big. Wrap the 
                                           * biggest one into a single 
                                           * predicate *)
                if conj'l > acclen then 
                  let conj1: L.pred = toPred conj' in 
                  List.map (fun acci -> conj1 :: acci) acc, acclen
                  else
                    let acc1: L.pred = toPred acc in 
                    List.map (fun conji -> conji @ [ acc1 ]) conj', conj'l
                else (* Still small, do the cross product *)
                  let res = 
                    List.fold_right
                      (fun acci acc -> 
                        List.fold_right
                          (fun conji acc -> (conji @ acci) :: acc)
                          conj'
                          acc)
                      acc
                      []
                  in
                  res, conj'l * acclen
        in
        let res, _ = List.fold_right processOneConjunct andl ([ [] ], 1) in
        res

    | _ -> E.s (E.unimp "toDNF: %a\n" L.d_pred p)
  in

  (* Skip the quantifiers *)
  let rec skipQuantifiersAnd (p: L.pred) : L.pred = 
    match p with 
      L.All (v, p) -> L.All (v, skipQuantifiersAnd p)
    | L.And pl -> 
        let rec map f = function [] -> [] 
              | h::t -> let h'=f h in h'::map f t in
        L.And (map skipQuantifiersAnd pl)

    | L.Exist (v, p) -> L.Exist (v, skipQuantifiersAnd p)

    | L.Not p -> skipQuantifiersAnd (L.negatePred p)

    | L.Pannot (a, p) -> L.Pannot (a, skipQuantifiersAnd p)
    | _ -> 
        toPred (loop p)
  in

  let res = skipQuantifiersAnd p in
(*
  ignore (E.log "Converted\n   %a\n into\n %a\n" L.d_pred p L.d_pred res);
*)
  res

(* Remember here the case analysis nesting depth, for debugging. for each 
 * depth, show how many cases are still to do. *)
let caseTree: int ref list ref = ref []
let pushClause (cd: clause) (activecount: int) = 
  let count = ref activecount in 
  if !count > 1 then
    pushWithUndo caseTree count

let nextCase () = 
  match !caseTree with 
    count :: _ -> decr count
  | _ -> assert false

let d_caseTree () = 
  (docList ~sep:(chr '.') (fun c -> num !c)) () (List.rev !caseTree)

type engineCallbacks = {
    egAboutToFail: unit -> bool; (** We call this function when we are about 
                                  * to fail. Return true if the is some 
                                  * progress done by this function. May raise 
                                  * Contradiction.  *)

    egPushHyp: Util.symbol -> Logic.pred -> unit;
      (** We call this function to announce a new hypothesis. This is used 
       * for proof checking. The hypothesis will be popped on undo. *)

    egBroadcast: Edag.messagebody -> Proof.proof -> string -> unit;
      (** We call this function to broadcast a message. This function may 
       * raise Contradiction *)

    egSnapshot: unit -> unit; (** Mark a state to which we must undo *)

    egUndo: unit -> unit; (* Do an undo *)
        
  }    

let irrelevantCases = ref 0
let allCases = ref 0 
let allCaseLeaves = ref 0 
let allActiveCaseLeaves = ref 0 
let caseLeavesProved = ref 0 
let caseLeavesDropEarly = ref 0 
let activeCaseLeavesDropEarly = ref 0 

(** For each instantiation, count the contra, unit and other clauses *)
type caseCounts = {
    mutable ccContra: int; (** Contradictory clauses *)
    mutable ccUnit: int; (** Total unit clauses **)
    mutable ccIUnit: int; (** Total number of irrelevant unit clauses *)
    mutable ccMulti: int; (** Total number of non-unit clauses *)
    mutable ccIMulti: int; (** Total number of irrelevant non-unit clauses *)

    mutable ccCost: float; (** For each non-contra clause, the product of the 
                          * active non-dropped clauses *)
    
  }
let countCases: caseCounts IH.t = IH.create 13
let getCaseCounts (cid: int) : caseCounts = 
  (* find the case info *)
  let n = 
    match (try IH.find corigin cid with Not_found -> CONone) with
      COInst n -> n
    | _ -> unknownSym
  in
  try IH.find countCases n
  with Not_found -> begin 
    let p = { ccContra = 0; ccUnit = 0; ccMulti = 0;
              ccIMulti = 0; ccIUnit = 0; ccCost = 1.0; } in
    IH.add countCases n p;
    p
  end

  
  
let caseLeavesFalsified = ref 0 
let unitCases = ref 0 
let contraCases = ref 0


  


let warnedNoRelevance = ref false


(* sm: ------ timed version of hasHyp ------ *)
let cycleSample () : int =
  (Stats.sample_pentium_perfcount_10 ())

let hasHyp_time = ref 0

let hasHyp_timed prf h : bool =
  let start:int = (cycleSample ()) in
  let res = (hasHyp prf h) in
  hasHyp_time := !hasHyp_time + (cycleSample ()) - start;
  res
(* ----------------------------------------- *)


(** Subsume a case. Return None if everything goes Ok. If you get a 
 * contradiction while subsuming, then return the proof of falsehood *)
let subsume (eg: engineCallbacks) 
            (case: case) (casevar: U.symbol) (pfalse: P.proof) : P.proof option =
  (* Subsume the negation of a predicate, given a proof of false
   hypothetical on the predicate being true *)
  let subsumePred (p: L.pred) (ph: U.symbol) (pfalse: P.proof) : P.proof option = 

    (* Extend the hyps
    eg.egPushHyp casevar case.cpred;
    *)
    
    (* The negation of this predicate *)
    let negp = try L.negatePredRaise p
    with L.CannotNegate _ -> assert false in
    let prf_negp = P.mkContraProof negp ph pfalse in
    (* Intern the predicate *)
    let negmb, negp_negmb, _ = messageBodyOfPredicate negp in 
    (* Proof of the negmb *)
    let prf_negmb = applyProofTrans negp_negmb prf_negp in 
    (* Broadcast the message. If there is a contradiction, it means that 
     * this case was irrelevant *)
    if !debug then 
      ignore (log "Subsume %a (from case %a)\n" L.d_pred negp d_case case);
    try
(*
   (match negmb with 
   Intern (L.SetEq, n1, n2) -> 
   ignore (E.log "will subsume %d = %d\n%a\n%a" 
   n1.id n2.id d_node n1 d_node n2)
   | _ -> ());
 *)
      eg.egBroadcast negmb prf_negmb "subsumption";
      None
    with S.Contradiction newpfalse -> begin
(*
   ignore (E.warn "Got contradiction after subsumption of %a\n"
   L.d_pred negp);
 *)
(*
   ignore (E.warn "pfalse is %a\nnewpfalse = %a\n"  P.d_prf pfalse
   P.d_prf newpfalse);
 *)
      Some newpfalse
    end
  in
  (* We optimize the case when there is only one conjunct *)
  match case.conjuncts with 
    [ _ ] -> subsumePred case.cpred casevar pfalse
  | _ when false -> begin 
       (* See how many NotDetermined conjuncts we have. We only do
          subsumption if there is only one *)
       let rec countActive (count: int) = function
           [] -> count
         | c :: restc -> begin 
           match c.ecshared.determined with 
             NotDetermined -> countActive (count + 1) restc
           | Determined(b, _) -> 
               if b = not c.ecnegated then 
                 countActive count restc
               else
                 (* We should not have false conjuncts *)
                 assert false
         end
       in
       let ca = countActive 0 case.conjuncts in
       if ca = 1 then begin
          (* Make a hypothesis variable *)
          let phvar = genNewHypothesisSymbol "SUBS" in
          (* Get the conjuncts *)
          let cjl: L.pred list = match case.cpred with 
              L.And cjl -> cjl
            | _ -> assert false
          in
          (* Scan the conjuncts again and construct a proof that they are all 
           * true *) 
          let pp: L.pred ref = ref (L.Bool true) in
          let rec makeConjPrf (cl: elemcase list) 
                              (cjl: L.pred list) : P.proof list =
            match cl, cjl with 
              [], [] -> []
             | (c :: restc), (cj :: restcj) -> begin 
                match c.ecshared.determined with 
                  NotDetermined -> pp := cj;
                                   (P.Hyp phvar) :: makeConjPrf restc restcj
                | Determined (b, ptrue) when b = not c.ecnegated -> ptrue :: makeConjPrf restc restcj
                | Determined _ -> assert false
              end
           | _, _ -> assert false
          in
          let prfl: P.proof list = makeConjPrf case.conjuncts cjl in
          assert (!pp <> L.Bool true);
          (* Make a proof of falsehood *)
          let pfalse' = P.Cut (P.Andi prfl, casevar, pfalse) in 
          (* Now subsume *) 
          subsumePred !pp phvar pfalse'     
       end (* if ca = 1 *) else
         None
  end
 | _ -> None


(* Do some case analysis and invoke a continuation with a proof of falsehood. 
 * If we return, then we have failed the case analysis. *)
let rec caseAnalysis (eg: engineCallbacks) 
                     (success: P.proof -> unit) : unit = 

  (* Try some case analysis. We get the cases from the Cases module. If you 
   * derive a contradiction, invoke the continuation. Otherwise, just return.*)
  let rec tryNextSetOfCases (success: P.proof -> unit) : unit =  
    match getFirstCaseAnalysis () with 
      None -> (* No more cases to try, and we still have not found a 
               * contradiction. We must ask everybody to try harder *)
        (* We write it like this to ensure tail-recursion *)
        let (someprogress: bool), (pfalse: P.proof option) = 
          try eg.egAboutToFail (), None
          with S.Contradiction pfalse -> true, (Some pfalse)
        in
        (match pfalse with 
          Some pfalse -> (* we found contradiction *) success pfalse
        | _ -> if someprogress then caseAnalysis eg success)
        (* We get here if we failed *)
            
    | Some (cd, activecount) -> begin
        (* We will do this clause. *)
        if !debugStats then begin
          let activecount' = 
            List.length 
              (List.filter (fun c -> c.cfalse == None) cd.cases)
          in
          if activecount <> activecount' then 
            ignore (E.bug "activecount = %d, computed count = %d\n"
                      activecount activecount');

          incr allCases;
          allCaseLeaves := !allCaseLeaves + (List.length cd.cases - 1);
          allActiveCaseLeaves := !allActiveCaseLeaves + (activecount - 1);
          (* See from which hyp it is coming fomr *)
          let co : corigin = 
            try IH.find corigin cd.cid
            with Not_found -> assert false
          in
          let cc = getCaseCounts cd.cid in 
          if activecount = 0 then cc.ccContra <- 1 + cc.ccContra
          else if activecount = 1 then cc.ccUnit <- 1 + cc.ccUnit
          else begin 
            cc.ccMulti <- 1 + cc.ccMulti;
            cc.ccCost <- (float_of_int activecount) *. cc.ccCost
          end;

          if activecount = 0 then 
            incr contraCases;
          if activecount = 1 then 
            incr unitCases;
        end;

        if !debug then begin 
          pushClause cd activecount;
          let msg = 
            if activecount = 0 then "splitcontra clause"
            else if activecount = 1 then "splitunit clause"
            else "splitting cases"
          in
          ignore (log "(%s: %a\n case tree:%t\n" msg
                    d_clause cd d_caseTree);
          
        end;

        if activecount > 1 && cd.depth > !currentSplitDepth then begin 
          setWithUndo currentSplitDepth cd.depth;
          if !debugStats && false then 
            ignore (log "currentSplitDepth: %d\n" !currentSplitDepth);
          if !currentSplitDepth >= !maxMatchSplitDepth 
              && !matchingDisabled == None then 
             setWithUndo matchingDisabled (Some "maxMatchSplitDepth"); 

        end;

        (* Must assert all disjuncts in turn. Construct a list 
         * of assumptions along with a proof of falsehood in 
         * each case. Return a proof of falsehood. *)
        let rec tryDisjuncts (sofar: (U.symbol * P.proof) list) 
                             (activeremaining: int)
                             (remaining: case list) : unit = 
          match remaining with 
            [] -> (* We have successfully tried all cases of this clause *)
              if !debug then 
                ignore (log "found contradiction)\n");
              assert (activeremaining = 0);
              let pfalse = 
                if !genProofs then 
                  cd.mkProof (List.rev sofar)
                else 
                  noProof                
              in
              (* We get here only if the case analysis was thought to be 
               * relevant *)
              if !debugStats then begin
                (* Verify relevance *)
                List.iter
                  (fun (case,prf) -> 
                    if not (hasHyp prf case) then begin
                      ignore (E.log "Case %s was falsely thought to be relevant\n"
                                (U.symbolName case));
                      if true then 
                        ignore (E.log "entire=%a\n notrelevant=%a\n"
                                  
                                  P.d_prf pfalse
                                  P.d_prf prf);
                      assert false;
                    end;
                  )
                  sofar
              end;
              
              success pfalse

          | a :: restd -> begin
              (* Make a hypothesis name for this disjunct *)
              let casevar = genNewHypothesisSymbol "CASE" in 
              match a.cfalse with 
                Some ptrans -> 
                  (* This case was already falsified *)
                  if !debug then
                    ignore (log "skipping falsified case(%s) %a\n" 
                              (U.symbolName casevar)
                              d_conjuncts a.conjuncts);
                  let prf = if !genProofs then ptrans casevar else noProof in
                  if !debugStats then 
                    incr caseLeavesFalsified;
                  tryDisjuncts ((casevar, prf) :: sofar) activeremaining restd

              | None ->  begin
                  if !debug then begin
                    ignore (log "trying case(%s) %a\n" 
                              (U.symbolName casevar) d_conjuncts a.conjuncts)
                  end;

                  eg.egSnapshot ();

                  (* Extend the hyps, but not the assumes, to find bugs in the 
                   * management of hyps *)
                  eg.egPushHyp casevar a.cpred;

                  (** Here is what we do once we succeed to get a proof of 
                   * falsehood after asserting this disjunct *)
                  let gotContradiction (pfalse: P.proof) : unit = 
                    if !doProofTrace then
                      ignore (E.logg "%scontradiction for case\n" 
                                (proofTraceInd ()));

                    (* Restore. Will restore the hyps as well. *)
                    if !debug then
                      ignore (log "got contradiction for this case(%s)\n"
                             (U.symbolName casevar));
                    eg.egUndo (); (* Retract the hypothesis casevar *)

                    (* See if the case was relevant. Change the proof 
                     * into a dummy one, if we are not checking proofs and if 
                     * we have scanned it *) 
                    let relevant', pfalse' = 
                      if !dropIrrelevantCases then begin 
                         (* Look for the symbol for real, leave the proofs 
                            alone. *)
                        (hasHyp pfalse casevar), pfalse
                      end else
                        true, pfalse
                    in
                    (* It could be that we think the case is relevant, but 
                     * subsumption will discover that it is not *)
                    let relevant'', pfalse'' = 
                      if relevant' then begin 
                        if !debug then 
                          ignore (log "case (%s) was relevant\n"
                                    (U.symbolName casevar));

                        if activeremaining > 0 && !useSubsumption then 
                          match subsume eg a casevar pfalse' with 
                            Some pfalse'' -> 
                              if !debug then 
                                ignore (log "case (%s) was irrelevant (discovered by subsumption)\n"
                                          (U.symbolName casevar));
                              false, pfalse''

                          | None -> relevant', pfalse'
                        else
                          relevant', pfalse'
                      end else
                        relevant', pfalse'
                    in
                    if relevant'' then begin
                      (* We continue with the remaining cases *)
                      tryDisjuncts ((casevar, pfalse') :: sofar) 
                                   (activeremaining - 1) restd
                    end else begin
                      (* Not relevant *)
                      if !debug then 
                        ignore (log "case(%s) was irrelevant%s. Skip the rest and raise contradiction)\n" 
                                  (U.symbolName casevar) 
                                  (if activecount = 1 then ":unit" else ":multi"));
                      if !doProofTrace then
                        ignore (E.logg "%sirrelevant case; bailing on remaining split of clause %d\n"
                                  (proofTraceInd ())
                                  cd.cid);


                      if !debugStats then begin 
                        incr irrelevantCases;
                        activeCaseLeavesDropEarly := 
                          !activeCaseLeavesDropEarly + (activeremaining - 1);
                        caseLeavesDropEarly := 
                          !caseLeavesDropEarly + (List.length restd);

                        let cc = getCaseCounts cd.cid in 
                        let activedone = activecount - (activeremaining - 1) in
                        (* Mark the irrelevant *)
                        if activecount = 0 then assert false;
                        if activecount = 1 then 
                          cc.ccIUnit <- 1 + cc.ccIUnit
                        else begin 
                          cc.ccIMulti <- 1 + cc.ccIMulti;
                          cc.ccCost <- cc.ccCost /. (float_of_int activecount);
                          cc.ccCost <- cc.ccCost *. (float_of_int activedone);
                        end;
                      end;

                      cd.cscore <- cd.cscore - 1;

                      (* If this is the first case, then the times must be 
                       * the same *)
                      success pfalse''
                    end
                  in
                  

                  (* Make a proof of false from asserting this disjunct *)
                  let pfalseo: P.proof option = 
                    try 
                      if !doProofTrace then
                        ignore (E.logg "%sconsidering a disjunct of clause %d: %a\n"
                                       (proofTraceInd ())
                                       cd.cid
                                       d_flatConjuncts a.conjuncts);
                      incr proofTraceDepth;
                      let idx = ref (-1) in 
                      let rec iter f = 
                        function [] -> () | h::t -> f h; iter f t in 
                      iter 
                        (fun c -> 
                          incr idx;
                          let cs = c.ecshared in
                          (* We should not have elements that are determined 
                           * to be false in here *)
                          assert 
                            (match cs.determined with 
                              Determined(b, _) when b = c.ecnegated -> false 
                            | _ -> true);
                          (* Broadcast only the messages that are not 
                           * determined *)
                          if cs.determined == NotDetermined then 
                            eg.egBroadcast 
                              (positiveMessageBody c)
                              (if !genProofs then 
                                c.mbproof a !idx casevar 
                              else noProof)
                              "trying case")
                        a.conjuncts;

                      (* We get here if there is no contradiction *)
                      None
                    with S.Contradiction pfalse -> 
                      if !debugStats then 
                        incr caseLeavesProved;
                      Some pfalse
                  in
                  match pfalseo with 
                    None -> 
                      (* If we do not get a contradiction after these 
                       * assertions, try remaining clauses. Once you finally 
                       * get a contradiction, undo this snapshot and try the 
                       * other cases *)
                      tryNextSetOfCases gotContradiction

                  | Some pfalse -> gotContradiction pfalse
              end
          end
        in
        tryDisjuncts [] activecount cd.cases
    end
  in
  tryNextSetOfCases success

let finish () = 
  if !debugStats then begin
    ignore (E.log "case tree internal nodes: %d\n" !allCases);
    ignore (E.log "case tree leaves: %d\n" !allCaseLeaves);
    ignore (E.log "    inactive leaves: %d\n" 
              (!allCaseLeaves - !allActiveCaseLeaves));
    ignore (E.log "    active leaves: %d\n" !allActiveCaseLeaves);
    ignore (E.log "caseLeavesProved: %d\n" !caseLeavesProved);
    ignore (E.log "activeCaseLeavesDropEarly: %d\n\n" 
              !activeCaseLeavesDropEarly);

    ignore (E.log "inactiveCaseLeavesDropEarly: %d\n" 
              (!caseLeavesDropEarly - !activeCaseLeavesDropEarly));
    ignore (E.log "caseLeavesDropEarly: %d\n" !caseLeavesDropEarly);

    ignore (E.log "irrelevantCases: %d\n" !irrelevantCases);
    ignore (E.log "unit internal nodes: %d\n" !unitCases);
    ignore (E.log "contra internal nodes: %d\n" !contraCases);
    ignore (E.log "count cases:\n");
(*
    (try 
      let cc = H.find countCases "Hproblem" in
      if cc.ccMulti = cc.ccIMulti then assert false
    with Not_found -> assert false);
*)
    (* Make a list of all clauses *)
    let l = ref [] in
    IH.iter (fun n cc -> l := (n,cc) :: !l) countCases;
    let sl = List.sort 
        (fun (_,cc1) (_,cc2) -> compare cc2.ccCost cc1.ccCost) !l in
    List.iter (fun (n,cc) -> 
      ignore (E.log "  %-20s: c:%d, u:%d (relu:%d), m:%d (relm=%d), cost=%8.1e\n"
                (U.symbolName n) cc.ccContra 
                cc.ccUnit 
                (cc.ccUnit - cc.ccIUnit)
                cc.ccMulti 
                (cc.ccMulti - cc.ccIMulti)
                cc.ccCost))
      sl;
      
  end
  (* sm: was testing.. *)
  ;ignore (E.log "hasHyp_time: %d\n" !hasHyp_time)


let init () =
  (* See if we can still drop the irrelevant cases *)
  if not !Edag.genProofs && !dropIrrelevantCases then begin
    ignore (E.warn "Turning off the irrelevant case optimization because we don't generate proofs\n");
    dropIrrelevantCases := false;
  end;
  ECH.clear elemcaseTable;

  if !memoFreeHyps && !useFreshHyps then begin 
    ignore (E.warn "Turning off the memoization of free hypotheses because you want fresh hypotheses");
    memoFreeHyps := false;
  end
  

let options = 
  [ 
    { O.optInUI = "";
      O.optRestart = false;
      O.optKind = O.OUnit;
      O.optExtra = (fun _ -> ());
      O.optCommandLine = "\n\t ****** Case analysis options \n\t\t";
      O.optHelp = "";
    };


    { O.optInUI = "";
      O.optRestart = false;
      O.optKind = O.OBool useSubsumption;
      O.optExtra = (fun _ -> ());
      O.optCommandLine = "-usesubsumption";
      O.optHelp = "use subsumption";
    };
    

    { O.optInUI = "";
      O.optRestart = false;
      O.optKind = O.OBool dropIrrelevantCases;
      O.optExtra = (fun _ -> ());
      O.optCommandLine = "-drop-irrelevant-cases";
      O.optHelp = "drop irrelevant cases";
    };

    { O.optInUI = "";
      O.optRestart = false;
      O.optKind = O.OBool memoFreeHyps;
      O.optExtra = (fun _ -> ());
      O.optCommandLine = "-memo-freehyps";
      O.optHelp = "memoize free hypotheses";
    };

    { O.optInUI = "";
      O.optRestart = false;
      O.optKind = O.OInt maxMatchSplitDepth;
      O.optExtra = (fun _ -> ());
      O.optCommandLine = "-max-match-split-depth";
      O.optHelp = "maximum split depth before matching is disabled";
    };
 
  ]

(* EOF *)
