module E = Errormsg
module I = Input
module L = Logic
module S = Satproc
module P = Proof
module U = Util
module CA = Cases
module O = Options

open Pretty
open Edag

let debug = ref false  (** Set with --debug EG *)


let timingFlag = ref false
 
let engineIdent = Util.registerSymbolName "EG"  (* This is the identifier of 
                                                 * the engine itself *)


let printAssumptions = ref false
let printWitness = ref true

let printInfix = ref true

let useLocationHyps = ref false


let proofFile: string ref = ref ""
    



(* Define new printers for expressions and predicates *)
let d_exp () (e: L.exp) = 
  L.printExp 
    (if !printInfix then L.infixLogicPrinter else L.defaultLogicPrinter)
    e

let d_pred () (p: L.pred) = 
  L.printPred 
    (if !printInfix then L.infixLogicPrinter else L.defaultLogicPrinter)
    p

let log fmt = ignore (E.log "EG: "); E.log fmt
let warn fmt = ignore (E.log "EG: "); E.warn fmt

let currLoc = ref E.locUnknown

type proofgen =
    {   start: string -> unit;
        startVC: string -> unit;
        assume: (unit -> string) -> Logic.pred -> unit;
        unassume: unit -> unit;
        pick: string -> unit;
        unpick: unit -> unit;
        conjunct: Logic.pred -> Proof.proof -> unit;
        choose: thechoice:int -> nrchoices:int -> unit;  
        finish: unit -> unit;
        finishVC: unit -> unit
      }


exception Failure

class type uiType = object
  method name: string
  method options: Options.optionDescr list
  method enterBlock: (Util.symbol * Logic.pred) list -> 
                     (string * string) list -> unit
  method exitBlock: unit

  (** Called when a goal is initiated *)
  method start_goal: Logic.pred -> unit
      
  (** Called when a goal is successfully proved *)
  method goal_success: Logic.pred -> Proof.proof -> unit

  (** Called when a goal fails, with a witness *)
  method goal_failure: Errormsg.location -> 
                       string -> Logic.pred -> Pretty.doc list -> unit

  (** Called when an assumption is encountered. Generate the name lazily *)
  method assume: (unit -> Util.symbol) -> Logic.pred -> unit

  (** Called when an assumption is retracted *)
  method unassume: unit

  (** Called to start the UI *)  
  method run: unit
end

let getParam (p: (string * string) list) (n: string) : string = 
  try snd (List.find (fun (n',_) -> n' = n) p)
  with Not_found -> ""

let printBlock () (p: (string * string) list) : doc = 
  dprintf "%s:%s  %s (true location %s:%s)" 
    (getParam p "file") (getParam p "line") 
    (getParam p "name") (getParam p "truefile")
    (getParam p "trueline")

let blocks : (string * string) list list ref = ref []
let enterBlock (params: (string * string) list) = 
  blocks := params :: !blocks
let exitBlock () = 
  match !blocks with 
    [] -> ()
  | _ :: r -> blocks := r

let useTrueLocation = ref false
let currentBlock () = 
  match !blocks with 
  | [] -> None
  | b :: _ -> 
      Some (dprintf "%s:%s %s" 
              (getParam b (if !useTrueLocation then "truefile" else "file"))
              (getParam b (if !useTrueLocation then "trueline" else "line"))
              (getParam b "name"))

let currentLoc () : (string * string) option = 
  match !blocks with 
  | [] -> None
  | b :: _ -> Some ((getParam b "file"), (getParam b "line"))

let currentHypName () : string = 
  match !blocks with 
  | [] -> ""
  | b :: _ -> 
      let h = getParam b "hypname" in 
      if h <> "" then h else
      getParam b "line" 


class defaultUI : uiType = 
  object

    method name = ""
    method options = []
        
    method enterBlock (assumes: (Util.symbol * Logic.pred) list)
                      (params: (string * string) list) = ()
    method exitBlock = ()

    method start_goal _ = ()
        
    method goal_success _ _ = ()
    method goal_failure (l: E.location) 
                        (why: string) (goal: L.pred) (w: doc list) = 
      if !printWitness then begin 
        let idx = ref (-1) in 
        ignore (E.log "Witness\n @[%a@]"
                  (docList ~sep:nil
                     (fun w -> 
                       incr idx;
                       dprintf "%d: %a\n" !idx insert w)) w)
      end;
      ignore (E.log "Failed because: %s\n" why);
      ignore (E.log "*****@!Goal was %a@! at %a\n from blocks (inside first):\n" 
                 L.d_pred goal E.d_loc l);
      List.iter (fun p -> ignore (E.log " %a\n" printBlock p))
        !blocks;
      ignore (E.log "\n\n")

    method assume _ _ = ()
    method unassume = ()
    method run = ()
end

let theUI = ref (new defaultUI)

let continue_on_error = ref false

let continue_without_asking = ref false

(*************)
                      (* The engine will make assertions and each assertion 
                       * will be assigned a unique id. The proofs of these 
                       * assertions are kept in a hash table. We keep an undo 
                       * stack where each element specifies *)


                     (* The engine state *)
let hyps: (U.symbol * L.pred) list ref = ref []
 
                        (* Current hypotheses (each with its name). This list 
                         * grows and shrinks with case analysis. It is used 
                         * for checking proofs deep inside the procedures. *)


let assumes: (U.symbol * L.pred) list ref = ref []
                        (* These are only the top level hypotheses. These are 
                         * used to check the top-level proofs. When we 
                         * are at top-level this must be the same as hyps, 
                         * but there were bugs with the management of hyps, 
                         * which endangered the soundness of proof checking *)


let vars: L.var list ref = ref [] 
                       (* Current variables *)

let satprocs: S.entries list ref = ref []

let proofgen: proofgen option ref = ref None



let pushAssumption (assumvar: U.symbol) (assumpred: L.pred) : unit = 
  let entry = (assumvar, assumpred) in 
  pushWithUndo hyps entry;
  pushWithUndo assumes entry;
  if !Cache.useProofCache then begin 
    Cache.pushAssumption assumvar assumpred;
  end
  
    
let invalidGoal = ref false (** True if we are in QUERY_INVALID *)
let failedInvalidGoal = ref false (** If we have failed to prove the goal, 
                                   * set this variable *)



let start (fname: string) (sp: S.entries list) pgen = 
  satprocs := sp;
  proofgen := pgen;
  (* Try to load the proof cache *)
  Cache.start fname;

  (match !proofgen with Some pg -> pg.start !proofFile | _ -> ())


let _  = 
  Edag.varsAndHyps := (fun _ -> !vars, !hyps)

(* in debug mode, print a log string *)
let debugLog (s: string) : unit =
begin
  if !debug then
    (ignore (log "%s\n" s));
end

let debugLogp (s: string) (p: L.pred) : unit =
begin
  if !debug then
    (ignore (log "%s: %a\n" s L.d_pred p));
end

let timeout = ref 10 

let setTimeOutAlarm (t: int option) = 
  let t' = 
    match t with 
      Some t -> t
    | _ -> !timeout
  in
  if t' > 0 then 
    ignore (Unix.alarm t')

let clearTimeOutAlarm () = 
  if !timeout > 0 then 
    ignore (Unix.alarm 0)
  
let finish () =
  if !Cache.useProofCache then 
    Cache.saveProofCache ();

  match !proofgen with Some pg -> pg.finish () | _ -> ()

let postAssumes (aboutToFail: bool) = 
  List.iter (fun sp -> sp.S.postAssume aboutToFail) !satprocs;
  Matching.collectMatches ()



let snapshot () : int = 
  if !debug then
    ignore (log "(snapshot\n");
  if !newMessages != [] then assert false;
  let id = Edag.snapshot () in 
  List.iter (fun sp -> sp.S.snapshot ()) !satprocs;
  Matching.snapshot ();
  id

                                        (* The global undo function *)
let undo (marko: int option) = 
  if !debug then
    ignore (log "undo)\n");
  (* Drop any accumulated messages. There were none at snapshot *)
  newMessages := []; 
  (* Do the undo in reverse order of snapshot *)
  List.iter (fun sp -> sp.S.undo ()) !satprocs;
  Edag.undo marko


let witness () : doc list = 
  IH.clear Edag.smallestMemoForClass;
  IH.clear Edag.smallestMemoForNode;
  theWitness := [];
  List.iter
    (fun sp -> 
      (* Have a timeout for witness generation *)
      setTimeOutAlarm (Some 5);
      (try sp.S.witness ()
      with Interrupted msg -> 
        accumWitness (dprintf "%s was interrupted (%s)" 
                        (U.symbolName sp.S.ident) msg));
      clearTimeOutAlarm ())
    
    !satprocs;
  (* Now get the witness from the parser *)
  !S.selectedParser.S.parserWitness accumWitness;

  (* Now reverse the witness, and remove duplicates *)
  let h: (doc, unit) H.t = H.create 13 in 
  let acc : doc list ref = ref [] in 
  List.iter
    (fun w -> 
      if H.mem h w then () else begin
        acc := w :: !acc;
        H.add h w ()
      end)
    !theWitness;
  
  !acc


let announceProof (pred: L.pred)
                  (prf: P.proof) : unit =
  assert (match !hyps, !assumes with 
    [], [] -> true
  | h :: _, a :: _ when h == a -> true
  | _ -> 
      ignore (warn "state.hyps = %a\nstate.assumes = %a\n"
                (docList ~sep:line 
                   (fun (h, p) -> dprintf "%s: %a"
                       (U.symbolName h) d_pred p)) !hyps
                (docList ~sep:line 
                   (fun (h, p) -> dprintf "%s: %a"
                       (U.symbolName h) d_pred p)) !assumes);
      false);

  (* Here we ought to check the proof *)
  if !P.checkProofs && prf != noProof then begin
    if !debug then
      ignore (E.log "Checking proof of %a\n" L.d_pred pred);
    ignore (P.check !vars !assumes prf pred);
  end;

  (match !proofgen with Some pg -> pg.conjunct pred prf | _ -> ())



(* counter for alpha conversion for quantified goals *)
let alphaCount: int ref = ref 0


let printFailMessage (l: E.location) 
                     (goal: L.pred) 
                     (whyfail: string) : unit =
  E.hadErrors := true;
  ignore (E.log "\n\n*****\nProof generation failure (%s) with assumptions\n" 
            whyfail);
  (if !printAssumptions then 
    List.iter
      (fun (n, p) -> ignore (E.log "  %s: %a\n" (U.symbolName n) L.d_pred p))
      (List.rev !assumes)
  else
    ignore (E.log "   ... assumptions suppressed. Use -assumptions to get them\n"));
  let w = witness () in 
  !theUI#goal_failure l whyfail goal w ;
  (* See if we must continue *)
  let doFail () =
    (* Invoke the GUI *)
    !theUI#run;
    
    (* Now turn off debugging so we don't see anything after this *)
    List.iter (fun sp -> sp.S.debug := false) !satprocs;

    if !Cache.useProofCache then 
      Cache.saveProofCache ();

    raise Failure
  in
  if !continue_on_error then begin
    let answer =
      if not !continue_without_asking then begin
        if !E.readingFromStdin then begin
          ignore (E.log "User interaction is disabled because goals are read from stdin\n");
          "y"
        end else begin
          print_string "*** More (y/Y) (RET to stop): ";
          read_line ()
        end
      end else 
        "y"
    in
    
    if answer = "" then 
      doFail ()
    else begin
      let a = String.sub answer 0 1 in 
      (if a = "Y" then 
        continue_without_asking := true
      else if a = "y" then 
        () 
      else
        doFail ());
      
      ignore (E.log "\n*** Continuing after error ***\n");
      (* Pretend that we succeeded *)
      ()
    end
  end else
    doFail ()


(** Sometimes we will find contradictory assumptions. In that case, we 
 * remember here a proof of FALSE, to be used for future proofs. We keep 
 * going and pretending to be proving the subgoals, because a proof of a 
 * certain form is expected, and also because it helps to maintain the 
 * invariants of the GUI *)
let proofOfFalsehood: P.proof option ref = ref None



(** This is a reimplementation of the main engine, in the form of a new 
 * assume function. The purpose is to unify the handling of assumptions, 
 * goals, and messages. *)
(** This function never raises Contradiction. Instead it sets 
 * "proofOfFalsehood" *)
let rec newAssume 
    ~(fromgoal:bool) (* Whether this is from a goal *)
    (a: L.pred)
    (aprf: P.proof) (* The proof of a *)
    (why: string) : unit = 
  
  (* Now look at the predicate *)
  let rec doAssumptionWithCheck (a: L.pred) (aprf: P.proof) : unit = 
    match !proofOfFalsehood with 
      Some pfalse -> 
        (* We had already a proof of falsehood. Nothing to do *)
        ()

    | None -> (* If this is a goal, we look in the cache *)
        if fromgoal then 
          ignore (Cache.checkProofCache 
                    a false
                    (fun a -> doAssumptionNoCheck a aprf; None))
        else
          doAssumptionNoCheck a aprf

  and doAssumptionNoCheck (a: L.pred) (aprf: P.proof) : unit = 
    try 
      match a with 
        L.Bool true -> ()
      | L.Bool false -> raise (S.Contradiction aprf)
      | L.PNez _ -> begin 
          (* Finally, assuming an atom *)
          let body, ptrans, _ = messageBodyOfPredicate a in 
          (* Make a message *)
          let msg = 
            { body = body;
              fromgoal = false;
              owner = 0;
              proof = applyProofTrans ptrans aprf } in
          broadcastThisMessage msg
      end
            
      | L.And al -> 
          let cnt = ref 0 in 
          (* Do not use List.iter because it confuses the profiler *)
          let rec iter f = function [] -> () | h::t -> f h; iter f t in 
          iter (fun p -> 
            doAssumptionWithCheck p (P.Ande (!cnt, aprf));
            incr cnt)
            al
            
      | L.Not p -> doAssumptionNoCheck (L.negatePred p) aprf
            
      | L.Pannot (_, p) -> doAssumptionNoCheck p aprf
            
      | (L.Imply _ | L.Or _) -> 
          CA.addClauseOfPredicate a aprf ~ingoal:fromgoal
            
      | L.All _ -> Matching.internForall a aprf
            
      | L.Exist (var, a') -> 
          (* Create a new symbol for the variable *)
          let param = newAlphaName var in
          (* Replace the deBruijn variable with the named variable *)
          let newPred: L.pred = 
            L.substPred
              (fun depth e -> 
                match e with 
                  L.Var d when d = depth -> Some (L.NVar param)
                | _ -> None)
              a'
          in
          (* pushWithUndo vars param; I removed this because it was 
           * adding the parameter with a wrong scope *)
          (* Turn it into a unit clause. This will help with proof 
           * generation *)
          CA.addUnitClauseOfPredicate newPred 
            (fun hyp pfalse -> P.ExistE(aprf, param, hyp, pfalse))
            ~ingoal:fromgoal
            
      | _ -> E.s (E.unimp "doAssumption: %a" L.d_pred a)
    with S.Contradiction pfalse -> 
      setWithUndo proofOfFalsehood (Some pfalse)

  and broadcastThisMessage (msg: message) : unit = 
    if not msg.fromgoal && !genProofs then 
      assert (proofChecker (text "before broadcast") 
                msg.proof (predicateOfMessageBody msg.body));

    (* Now activate the nodes in the message *)
    (match msg.body with 
      Intern (_, n1, n2) -> activateNode n1.root; activateNode n2.root
    | _ -> ());  
    
    (* Now send the message to everybody *)
    if !debug then 
      ignore (log "broadcast message %s:%a (%s)\n"
                (U.symbolName msg.owner)
                d_messagebody msg.body
                (if msg.fromgoal then "goal" else "assume"));
    let rec iter f = function [] -> () | h::t -> f h; iter f t in 
    iter  (fun sp -> sp.S.assume msg) !satprocs
  in
  ()

(** Broadcast a number of messages (from S.newMessages) and wait for 
 * fixpoint, or raise Contradiction. *)
let rec broadcastMessagesFixpoint () : unit = 

   (** Broadcast a message and accumulate some replies in Satproc.newMessages *)  
  let rec broadcastOneMessage (msg: message) : unit = 
    (** First make sure to internalize the body *)
    let rec internalizeBody (msg: message) : message = 
      match msg.body with 
        Intern _ -> msg
            
      | Extern p -> begin 
          let body, ptrans, _ = messageBodyOfPredicate p in
          match body with 
          | Extern ((L.Or _ | L.Imply _) as p) -> 
              (* In this case the proof transformer ought to be the identity *)
              assert (ptrans == None);
              
              CA.addClauseOfPredicate p msg.proof ~ingoal:false;
              { msg with body = Extern (L.Bool true)}
                
          | Extern (L.All _ as p) -> 
              assert (ptrans == None);
              Matching.internForall p msg.proof;
              
              { msg with body = Extern (L.Bool true) }
                
          | Extern (L.Exist(var, oldPred)) -> begin
              assert (ptrans == None);
              (* Create a new symbol for the variable *)
              let param = newAlphaName var in
              (* Replace the deBruijn variable with the named variable *)
              let newPred: L.pred = 
                L.substPred
                  (fun depth e -> 
                    match e with 
                      L.Var d when d = depth -> Some (L.NVar param)
                    | _ -> None)
                  oldPred
              in
              (* pushWithUndo vars param; I removed this because it was 
               * adding the parameter with a wrong scope *)
              (* Turn it into a unit clause. This will help with proof 
               * generation *)
              CA.addUnitClauseOfPredicate newPred 
                (fun hyp pfalse -> P.ExistE(msg.proof, param, hyp, pfalse))
                ~ingoal:false;

              (* Now drop the body *)
              { msg with body = Extern (L.Bool true) }
          end
                
          | Extern (L.And pl) -> 
              assert (ptrans == None);
              let conjid = ref (-1) in
              let rec iter f = function [] -> () | h::t -> f h; iter f t in 
              iter 
                (fun p -> 
                  incr conjid;
                  let id = !conjid in
                  let prf = P.Ande (id, msg.proof) in
                  broadcastEngineMessageFixpoint
                    (Extern p)
                    false
                    prf "relaying an Extern message")
                pl;

              (* Now discard this message *)
              { msg with body = Extern (L.Bool true) }

                    
          | Intern (_, n1, n2) -> 
              { msg with 
                body = body; 
                proof = applyProofTrans ptrans msg.proof }
                
          | _ -> 
              assert (ptrans == None);
              { msg with 
                body = body; 
                proof = applyProofTrans ptrans msg.proof }
      end
    in (* internalizeBody *)
    let msg' = internalizeBody msg in 

    if not msg.fromgoal && !genProofs then 
      assert (proofChecker (text "before broadcast") 
                           msg.proof (predicateOfMessageBody msg.body));

    (* Now activate the nodes in the message *)
    (match msg'.body with 
      Intern (_, n1, n2) -> activateNode n1.root; activateNode n2.root
    | _ -> ());  

    (* Now send the message to everybody *)
    if !debug then 
      ignore (log "broadcast message %s:%a (%s)\n"
                (U.symbolName msg'.owner)
                d_messagebody msg'.body
                (if msg'.fromgoal then "goal" else "assume"));
    let rec iter f = function [] -> () | h::t -> f h; iter f t in 
    iter  (fun sp -> sp.S.assume msg') !satprocs;

   in (* broadcast one message *)
   
   if !newMessages == [] then begin
     (* No messages. Try the postAssumes *)
     postAssumes false;
     if !newMessages == [] then 
       ()
     else 
       broadcastMessagesFixpoint ()
   end else begin
     let msgs = !newMessages in 
     newMessages := [];
     (* send all messages to everybody *)
     let rec iter f = function [] -> () | h::t -> f h; iter f t in 
     iter (fun msg -> broadcastOneMessage msg) msgs;
     (* And continue broadcasting *)
     broadcastMessagesFixpoint ()
   end


(* The global assume function. Might raise S.Contradiction. Might also raise 
 * Proved if the assumption was goal.  *)
and broadcastEngineMessageFixpoint 
    (pred: messagebody)
    (isgoal: bool)
    (prf: P.proof)
    (why: string) : unit =

  let p = 
    if isgoal then 
      match pred with 
      | Extern p -> begin 
          Extern (L.negatePredRaise p)
      end
      | Intern (bop, n1, n2) -> begin
          Intern (L.negateBinopRaise bop, n1, n2)
      end
    else
      pred
  in
  newMessageWhy
    { owner = engineIdent; 
      body = p;
      fromgoal = isgoal;
      proof = prf;
    } why;

  broadcastMessagesFixpoint ()


(** Prepare here the function used for case analysis. We must pass a lot of 
 * information to the case analysis, so we use a record *)
let engineCallbacks = {
  CA.egAboutToFail = (fun () -> 
    postAssumes true;
    if !newMessages != [] then begin 
      broadcastMessagesFixpoint ();
      true
    end else
      false);
  
  CA.egPushHyp = (fun casevar conjPrd -> 
    pushWithUndo hyps (casevar, conjPrd));

  CA.egBroadcast = (fun msg prf why ->
    broadcastEngineMessageFixpoint msg false prf why);
  
  CA.egSnapshot = (fun _ -> ignore (snapshot ()));

  CA.egUndo = (fun _ -> undo None);

}

let hypAlphaTable: (string, unit Alpha.alphaTableData ref) H.t = 
  H.create 13


                                        (* Might raise Failure *)
let rec proveConjunct
             (loc: E.location)
             (goal: L.pred) : unit =
  if !debug then
    ignore (E.log "EG: conjunct @[%a@]@!" L.d_pred goal);


  let proveAtom (goal: L.pred) = 
    let snap = snapshot () in 
    !theUI#start_goal goal;

    let assumvar = genNewHypothesisSymbol "C" in
    let assumpred = (L.negatePred goal) in
    pushAssumption assumvar assumpred;

    (** This is the main prover. Return a proof if successful. *)
    let mainProveAtom (goal: L.pred) : P.proof option = 
      setTimeOutAlarm None;

      (* the undo will restore the hyps *)               
      let prf = 
        try 
          (* Can raise Contradiction or Proved *)
          broadcastEngineMessageFixpoint 
            (Extern goal) true (P.Hyp assumvar) "goal";
          (* We get here when we are about to fail. It is time to try 
           * harder *)
          
          CA.caseAnalysis engineCallbacks
            (fun pfalse -> raise (S.Contradiction pfalse));
          
          if !doProofTrace then
            ignore (E.logg "%sfailed to prove the goal\n" (proofTraceInd ()));

          (* We get here when we cannot prove the goal *)
          if not !invalidGoal then begin
            printFailMessage loc goal "failed proof";
            None
          end else begin
            if !debug then
              ignore (log "Proof failed as expected\n");
            failedInvalidGoal := true;
            !theUI#goal_failure loc "as expected" goal [] ;
            None
          end
              
        with S.Contradiction pfalse -> 
          if !doProofTrace then
            ignore (E.logg "%scontradiction\n" (proofTraceInd ()));

          newMessages := [];
          
          Some (P.mkContraProof goal assumvar pfalse)
            
        | (Interrupted msg) as e ->
            newMessages := [];
            printFailMessage loc goal ("Interrupted: " ^ msg);
            None
      in
      clearTimeOutAlarm ();

      prf
    in (* mainProveAtom *)

    (** See if we have a proof of falsehood already *)
    let prf = 
      match !proofOfFalsehood with 
        Some pfalse -> Some (P.G("falsee", [P.F goal; P.P pfalse]))
      | None -> begin
          (* Maybe we have the proof in the cache *)
          Cache.checkProofCache goal !invalidGoal mainProveAtom
      end
    in
    (match prf with 
      Some prf ->
        !theUI#goal_success goal prf;
        announceProof goal prf
    | _ -> ());
    (debugLogp "proved" goal);
    undo (Some snap)

  in  (* proveAtom *)
 
  match goal with 
    L.Bool false | L.Or [] -> proveAtom goal

  | L.Bool true -> 
      (match !proofgen with Some pg -> pg.conjunct goal P.Truei | _ -> ())

  | (L.PNez _ | L.PFun _) -> proveAtom goal

  | L.And pl -> 
      (* We must prove each conjunct in turn. This is not tail-recursive ! *)
      let rec iter f = function [] -> () | h::t -> f h; iter f t in 
      iter (proveConjunct loc) pl


  | L.All(var, oldPred) -> begin
      ignore (snapshot ());
      (* Create a new symbol for the variable *)
      let param = newAlphaName var in
      (* Replace the deBruijn variable with the named variable *)
      let newPred: L.pred = 
        L.substPred
          (fun depth e -> 
            match e with 
              L.Var d when d = depth -> Some (L.NVar param)
            | _ -> None)
          oldPred
      in
      pushWithUndo vars param;
      
      (* wrap the to-be-created proof in an 'Alli' *)
      (match !proofgen with Some pg -> pg.pick param | _ -> ());
      
      (* prove the now-renamed predicate *)
      (proveConjunct loc newPred);
      
      (* finish the 'alli' *)
      (match !proofgen with Some pg -> pg.unpick () | _ -> ());
      
      undo None; 
      
    end

  | L.Pannot (_, p) -> 
      proveConjunct loc ( p)

  | L.Exist _ -> begin
      (* Assert its negation (a FORALL) *)
      ignore (snapshot ());
      assume (L.negatePred goal) "negated existential goal";
      (* Try to prove falsehood now *)
      proveConjunct loc (L.Bool false);
      undo None;
  end
  | L.Not p -> proveConjunct loc (L.negatePred p)

   (** If we assume the negation of this, it takes much longer *)
  | L.Imply(p, q) -> begin
      (* sm: prove an implication *)
      ignore (snapshot ());
      if !debug then 
        ignore (log "proving an implication %a\n" d_pred goal);

      if false then 
        (match p with 
          L.Bool _ | L.PNez _ -> ()
        | _ -> ignore (E.warn "Assuming non-atimic %a\n" L.d_pred p));
      assume p "LHS of implication goal";
      proveConjunct loc q;

      undo None; (* Retract the assumption *)
    end

  | L.Or pl -> 
      (* This is just like an implication. We assume the negation of all but 
       * the last one, and then we prove the last one  *)
      ignore (snapshot ());
      let rec loop = function
          [ ] -> assert false (* We caught this with false *)
        | [ p ] -> proveConjunct loc p
        | p :: rest -> 
            (* See if we are assuming conjunctions *)
            let negp = L.negatePred p in 
            if false then 
              (match negp with 
                L.Bool _ | L.PNez _ -> ()
              | _ -> 
                  ignore (E.warn "Assuming non-atomic predicate %a\n" L.d_pred negp));

            assume ~fromgoal:true negp "negated LHS of disjunctive goal";
            loop rest;
      in
      loop pl;

      undo None


  | _ -> E.s (E.unimp "EG:Proving unexpected predicate %a"
                L.d_pred goal)
      
    
   (* End the implication *)
and unassume () = 
  if !debug then
    ignore (E.log "EG: unassume@!");

  undo None


(** Add an assumption. If it was contradictory, we will remember that, so 
 * that when we prove things next time, we get a proof by contradiction. Does 
 * never raise Contradiction *)
and assume ?(fromgoal=false) (what: L.pred) (why: string): unit =

  let rec mkAssumption (what: L.pred) (prf: P.proof) = 
    match what with 
      L.Bool true -> ()

    | L.Bool false -> raise (S.Contradiction prf)

    | L.And pl -> 
        let cnt = ref 0 in 
        (* Do not use List.iter because it confuses the profiler *)
        let rec iter f = function [] -> () | h::t -> f h; iter f t in 
        iter (fun p -> 
          mkAssumption p (P.Ande (!cnt, prf));
          incr cnt)
          pl

   | p -> begin
        if !debug then
          ignore (E.log "EG: assume: @[%a@]@!" L.d_pred p);
      
        (** Communicate this to the oracle, to create a new axiom. We need a 
        * name  *)
        let rec mkName (prf: P.proof) : string = 
          match prf with
            P.Hyp v -> U.symbolName v
          | P.G(what, [P.P prf]) -> (mkName prf) ^ what
          | P.Ande (idx, prf) -> (mkName prf) ^ "@" ^ (string_of_int idx)
          | _ -> E.s (E.unimp "Eng.assume.mkName")
        in
        !theUI#assume 
          (fun _ -> U.registerSymbolName (mkName prf)) what;
        (match !proofgen with Some pg -> 
          pg.assume (fun _ -> mkName prf) what | _ -> ());
        registerUndoRun 
          (fun _ -> 
             !theUI#unassume; 
             (match !proofgen with Some pg -> pg.unassume () | _ -> ()));

   

        (* Do not broadcast assumptions if we already have a proof of 
         * falsehood *)
        if !proofOfFalsehood == None then begin
          setTimeOutAlarm None;
          broadcastEngineMessageFixpoint (Extern p) false prf why;
          clearTimeOutAlarm ()
        end
    end
  in

  (** See if we have a proof of falsehood already *)
  match !proofOfFalsehood with 
  | Some pfalse -> () (* Do nothing *)
  | None -> begin 
      (* Create a new assumption variable *)
      let assumvar = 
        let assumevarname = 
          if !useLocationHyps then 
            (* Try to generate a hypothesis name that is derived from the 
             * current location *)
            let lookupname = "H" ^ (currentHypName ()) in
            let undo = ref [] in
            let n = 
              fst (Alpha.newAlphaName 
                     hypAlphaTable (Some undo) lookupname ()) in 
            registerUndoRun 
              (fun _ -> Alpha.undoAlphaChanges hypAlphaTable !undo);
            n
          else
            "H"
        in
        genNewHypothesisSymbol assumevarname
      in
      (* Add to the hypotheses *)
      let what' = CA.convertToDNF what in 
      let what = what' in
      pushAssumption assumvar what;
      
      
      (* Make the assumption, but watch for a contradiction *)
      try
        mkAssumption what (P.Hyp assumvar)
      with S.Contradiction prf_false -> begin
        (* Drop the new messages *)
        newMessages := [];
        
        if !debug then 
          ignore (E.log "Found contradictory assumption\n");
        if !proofOfFalsehood == None then begin
          setWithUndo proofOfFalsehood (Some prf_false)
        end;
        (* Continue as if nothing has happened *)
      end
  end
    

let choose ~(thechoice: int) ~(nrchoices: int) = 
  if !debug then 
    ignore (E.log "EG: choose %d/%d\n" thechoice nrchoices);
  (match !proofgen with Some pg -> pg.choose thechoice nrchoices 
   | _ -> ())

    

(** Add a global assumption *)
let globalAssume (a: L.pred) (l: E.location) : unit = 
  currLoc := l;
  if !E.verboseFlag then
    ignore (log "GLOBASSUME at %a\n" 
              (fun _ b -> match b with Some d -> d | None -> text "unknown")
              (currentBlock ()));
  assume a "global assume"
  

let proveGoal 
    (name: string) 
    (nextElement: unit -> I.input) : unit = 

  let doOneElement (e: I.input) : unit = 
    match e with

    | I.EndGoal l ->
        currLoc := l
        
    | I.Choose(i,nr, l) -> 
        currLoc := l;
        choose i nr
          
    | I.Prove (p, l) -> begin
        (* try *) (* If I enable this then I do not see the exception trace 
                   * anymore *)
          currLoc := l;
          invalidGoal := false;
          if !debug then 
            ignore (log "Proving goal at %a\n" E.d_loc l);
          (if !E.verboseFlag then begin
            Cache.cacheHits := 0;
            Cache.cacheMisses := 0;
            ignore (log "GOAL   at %a " 
                      (fun _ b -> match b with Some d -> d | None -> text "unknown")
                      (currentBlock ()));
            Stats.timethis (proveConjunct l) p;
            ignore (E.log "  (%.0f ms, %d/%d cache hits)\n" 
                      (1000.0 *. !Stats.lastTime)
                      !Cache.cacheHits 
                      (!Cache.cacheHits + !Cache.cacheMisses));
          end else
            proveConjunct l p);
(*
        ignore (E.log "after goal\n");
        Edag.printGraph ();
*)

          if !E.readingFromStdin then (
            print_string "Ready\n";
            flush stdout;
           ) (*
        with e -> begin 
          !theUI#goal_failure l "uncaught exception"
            (L.PFun("uncaught exception", []))
            [];
          raise e
        end; *)
    end

    | I.ProveInvalid (p, l) -> 
        currLoc := l;
        invalidGoal := true;
        failedInvalidGoal := false;
        proveConjunct l p;
        if not !failedInvalidGoal then 
          E.s (E.error "Just proved invalid goal: %a\n" L.d_pred p)
          

    | I.Assume (ass, l) -> 
        currLoc := l;
        ignore (snapshot ());
        if !E.verboseFlag then
          ignore (log "ASSUME at %a\n" 
                    (fun _ b -> match b with Some d -> d | None -> text "unknown")
                    (currentBlock ()));
        assume ass "assumption"
          
    | I.UnAssume l -> 
        currLoc := l;
        undo None

    | I.Pick (v, l) -> 
        currLoc := l;
        ignore (snapshot ());
        pushWithUndo vars v;
        (match !proofgen with Some pg -> pg.pick v | _ -> ())

    | I.Block (args, l) ->
        currLoc := l;
        (* Print the top-level blocks *)
        !theUI#enterBlock !assumes args;
        enterBlock args;
        if false && !E.verboseFlag then
          (match currentBlock () with 
            None -> ()
          | Some d -> ignore (log "      at %a\n" insert d));
        ()

    | I.UnBlock l -> 
        currLoc := l;
        exitBlock ();
        !theUI#exitBlock

    | I.UnPick l -> begin
        currLoc := l;
        (match !proofgen with Some pg -> pg.unpick () | _ -> ());
        undo None
              
    end
    | I.Eof -> ()
    | _ -> E.s (E.bug "Unexpected input (%a) in goal" I.di e)
        

  in


  let rec doElements () =
    (* ignore (E.log "Looking for an element in goal %s\n" name); *)
    let elem = nextElement () in
(*
    if !E.hadErrors then 
      raise E.Error;
*)
    (* ignore (E.log "EG: Doing an element %a\n" I.delem elem); *)
    doOneElement elem;
    match elem with
      I.EndGoal _ -> () (* We are done *)
    | I.Eof -> ()
    | _ -> doElements () (* Keep going *)
  in

  if !debug then
    ignore (E.log "EG: start %s@!" name);
  L.resetVars (); (* Reset the variable counts, so that we can take best 
                   * advantage of caches *)
  (match !proofgen with Some pg -> pg.startVC name | _ -> ());

  doElements ();

  if !debug then
    ignore (E.log "EG: finished@!");

  (match !proofgen with Some pg -> pg.finishVC () | _ -> ());

  CA.finish ();

  ()

let options = [

  { O.optInUI = "";
    O.optRestart = false;
    O.optKind = O.OBool useLocationHyps;
    O.optExtra = (fun _ -> ());
    O.optCommandLine = "-locationhyps";
    O.optHelp = "use location-sesitive hyps";
  };

  { O.optInUI = "";
    O.optRestart = false;
    O.optKind = O.OBool useTrueLocation;
    O.optExtra = (fun _ -> ());
    O.optCommandLine = "-truelocation";
    O.optHelp = "use true location information";
  };

  { O.optInUI = "";
    O.optRestart = false;
    O.optKind = O.OBool useFreshHyps;
    O.optExtra = (fun _ -> ());
    O.optCommandLine = "-freshhyps";
    O.optHelp = "use fresh hypothesis names";
  };

  {
   O.optInUI = "";
   O.optRestart = false;
   O.optKind = O.OString proofFile;
   O.optExtra = (fun _ -> ());
   O.optCommandLine = "-o";
   O.optHelp = "the name of the output proof file";
 }

]

let init () = 
  (* Install a SIGALRM handler *)
  ignore (Sys.signal Sys.sigalrm 
            (Sys.Signal_handle (fun _ -> raise (Interrupted "Timeout"))));
  (* Install a CTRL-C handler *)
  ignore (Sys.signal Sys.sigint 
            (Sys.Signal_handle (fun _ -> raise (Interrupted "CTRL-C"))));

  assumes := [];
  hyps := [];
  vars := [];
  proofgen := None;
  blocks := [];
  
  if !useLocationHyps && not !useFreshHyps then begin
    ignore (E.log "Using fresh hypotheses names because you want location-sensitive names\n");
    setUseFreshHyps true;
  end else
    setUseFreshHyps !useFreshHyps

