module E = Errormsg
module L = Logic
module P = Proof
module S = Satproc
module H = Hashtbl

open Pretty

open Edag

let ccid = U.registerSymbolName "CC"
let debug = ref false

let log fmt = ignore (E.log "CC: "); E.log fmt
let bug fmt = ignore (E.log "CC: "); E.bug fmt

					(* EGraph events *)
      

let eqContras = ref 0 



type proofGen = unit -> P.proof * L.exp 
exception PathDone of proofGen list



					(* Proof Construction *)
let revOneThunk (pt: P.proof) : P.proof = P.Symm pt

(** The new equality proof generation *)
let genEqProof (a: node) (b: node) : P.proof =
  assert (!genProofs);
  if !debug then
    ignore (log "Generating EDAG EQ proof for %a =~ %a\n"
              d_nodeTop a d_nodeTop b);
  if a.root != b.root then E.s (bug "Generate EQ proofs");
  if a == b then 
    P.Refl a.exp
  else begin 
    let res = eqProofEqualNodes a b in 
    let res = eqproofToProof a.exp b.exp res in 
    assert (proofChecker (text "after genEqProof") res
              (L.eq a.exp b.exp));
    res
  end
  
  
let genCongrProof (a : node) (b : node) : P.proof =
  assert (!genProofs);
  if !debug then
    ignore (log "Generating EGraph congr proof for %a =~ %a\n"
              d_nodeTop a d_nodeTop b);

  if a.head != b.head || List.length a.args <> List.length b.args then 
    E.s (E.bug "no congruence proof possible for %a =~ %a\n"
           d_nodeTop a d_nodeTop b);

  let rec checkArgs (a_args: node list) (b_args: node list): P.proof list= 
    match a_args, b_args with 
      [], [] -> []
    | a' :: ta, b' :: tb -> (genEqProof a' b') :: checkArgs ta tb
    | _ -> E.s (E.bug "genCongrProof")
  in
  P.Cong(a.exp, b.exp, checkArgs a.args b.args)
	

(* Generate the proof that a and b are not equal *)
let genNeqProof (a : node) (b : node) 
                ((a',b',a'_ne_b') : neqwitness): P.proof = 
  assert (!genProofs && 
          ((a.root == a'.root && b.root == b'.root) ||
          (a.root == b'.root && b.root == a'.root)));
  if !debug then 
    ignore (log "Generating EGraph nonequality proof %d <> %d\n"
              a.id b.id);

  let res = 
    if a == a' && b == b' then
      a'_ne_b'
    else if a == b' && b == a' then 
      P.Symm a'_ne_b'
    else if a.root == a'.root then 
      P.G("neqtr", [ P.P (genEqProof a a');
                     P.E a'.exp;
                     P.P a'_ne_b'; 
                     P.E b'.exp;
                     P.P (genEqProof b' b) ])
    else if a.root == b'.root then 
      P.G("neqtr", [ P.P (genEqProof a b');
                     P.E b'.exp;
                     P.P (P.Symm a'_ne_b'); 
                     P.E a'.exp;
                     P.P (genEqProof a' b) ])
    else
      assert false
  in
(*
*)
  res
    
(*      
  (* We know that a = a' and b = b' *)
  let mkProof (a': node) (b': node) 
              (a'_ne_b': P.proof) = 
  in

  (* Scan the History and look for a'<>b' such that a=a' and b=b' (or 
   * viceversa) *)

  (* We try to find the first disequality that proves this (that is the last 
   * in the list) , because that proof is likely to use older assumptions *)
  let rec findIt (last: (node * node * P.proof) option) = function
      [] -> last
    | ((a',b',prf) as it) :: rest -> 
        if a'.root == a.root && b'.root == b.root then begin
          findIt (Some it) rest
        end else if a'.root == b.root && b'.root == a.root then begin
          findIt (Some it) rest
        end else
          findIt last rest
  in
  match findIt None !disequalities with 
    Some (a',b',prf) -> 
      if a'.root == a.root then 
        mkProof a' b' prf
      else
        mkProof b' a' (P.Symm prf)
  | None -> 
      E.s (E.bug "Cannot find proof that %d <> %d\n" a.id b.id)
*)  
        


let _ = 
  Edag.genEqProof := genEqProof;
  Edag.genNeqProof := genNeqProof


      (* Check for congruences of node "n" among the given list *)
let findCongruences (n: node)
                    (l: node circlist option) : node list = 
  foldCirclist
    (fun acc n' -> 
      if n.root != n'.root && congruent n.head n.args n' then 
        n' :: acc
      else
        acc)
    []
    l

(* Merge two nodes *)
let rec merge 
    (a1: node) 
    (a2: node)
    (a1_a2: P.proof) (** A proof that a1=a2 **)
    (ignoreone : bool) : unit = 
  let r1 = a1.root in let r1_a1 = a1.rootwhy in 
  let r2 = a2.root in let r2_a2 = a2.rootwhy in 

  if r1 != r2 then begin
    (* Not already in the same class *)
    if r1.size <= r2.size then (* r2 is the new root *)
      (* a1_a2 does not interfere with r2-a2, nor with r1-x *)
      let r2_a1 = Concat (r2_a2, One (false, a1_a2)) in
      (* Save r1_a1, before it changes *)
      let r1_a1 = a1.rootwhy in 
      mergeLocal r1 r2 
        (fun x1 -> (* proof of r2-x1. *)
          Concat (r2_a1, eqProofSymConcat r1_a1 x1.rootwhy))
    else (* r1 is the new root *)
      let r1_a2 = Concat (r1_a1, One (true, a1_a2)) in 
      (* save r2_a2 before it changes *)
      let r2_a2 = a2.rootwhy in 
      mergeLocal r2 r1 
        (fun x2 -> (* proof r1-x2 *)
          Concat (r1_a2, eqProofSymConcat r2_a2 x2.rootwhy))
  end; 

  (* Now announce the message. Do this after the actual merge to ensure that 
   * we can generate the proof *)
  if not ignoreone then begin 
    newMessage 
      { owner = ccid;
        fromgoal = false;
        body = Intern (L.SetEq, r1, r2);
        proof = if !genProofs then genEqProof r1 r2 else noProof;
      };
  end

					(* Merge Local is only called for 
                                         * roots. *)
and mergeLocal 
    (a: node)
    (newroot : node)

         (** Given a node x1 in class of a, build an eqproof for newroot-x1. 
          * We delay the proof generation so that we can build a short proof 
          * optimally. This function must be invoked before we change the 
          * root of x to point away from a. *)
    (mk_newroot_x: node -> eqproof)  =

  (* Do the union *)
  if !debug then
    ignore (log "Merging %d and %d\n" a.id newroot.id);
(*
  printGraph ();
*)

  (* See if these nodes are forbidden to be merged. This means that they 
   * are each in the other's forbid list *)
  (match forbiddenEquality newroot a with 
    Some w -> begin
      if !debug then
        ignore (log "Found contradiction when checking forbids\n");
      
      let pfalse = 
        if !genProofs then begin
          let newroot_ne_a = w () in
(*
          assert (proofChecker (text "after genNeqProof")
                    newroot_ne_a (L.ne newroot.exp a.exp));
*)
          P.mkEqNeq_FI newroot.exp a.exp
            (eqproofToProof newroot.exp a.exp (mk_newroot_x a))
            newroot_ne_a
        end else noProof 
      in 

      if !debugStats then 
        incr eqContras;
      
      assert (proofChecker (text "after finding contradiction during merge")
                pfalse (L.Bool false));
      raise (S.Contradiction pfalse)
    end
  | None -> ());

  (* Copy the active bit to the root, if necessary. We activate even a if 
   * necessary, because this way we will make sure to announce everybody 
   * that some nodes become active. Do this now, while a is a root. *)
  if newroot.active <> a.active then begin
    activateNode newroot;
    activateNode a;
  end;
  
  (** Announce the new forbids, before we join the forbids or the 
   * forbidCallbacks *)
  if newroot.forbid <> None then 
    iterCirclist registerCallback a.forbidCallbacks;
  
  if a.forbid <> None then 
    iterCirclist registerCallback newroot.forbidCallbacks;
  

  (** Callback to announce the new parents. Do this before we have joined the 
   * callbackNewParent *)
  iterCirclist 
    (fun n' -> callbackNewParent newroot n')
    a.parents;
  
  iterCirclist
    (fun n' -> callbackNewParent a n')
    newroot.parents;

  (* Reroot the entire equivalence class of a *)
  let rerootOne (x: node) = 
    (* Change reason first, before we change the root *)
    if !genProofs then begin
      let old = x.rootwhy in 
      registerUndoRun (fun _ -> x.rootwhy <- old);
      x.rootwhy <- mk_newroot_x x;
    end;
    x.root <- newroot;
  in
  iterEqc rerootOne a;
  
  (* Before we join the parents, collect the list of congruences. We must do 
   * this after we have joined the roots, to make sure that checking for 
   * congruences works *)
  let futureMerges: (node * node list) list = 
    foldCirclist 
      (fun acc (u: node) -> 
        let c = findCongruences u newroot.parents in
        (u, c) :: acc)
      []
      a.parents
  in
  

  (* Join the parent lists. Do this before we merge the parents callbacks *)
  newroot.parents <- concatenateCirclist newroot.parents a.parents;
  
  concatenateNewParentsCallbacks newroot a;
  
  concatenateMergeCallbacks newroot a;
  
  concatenateForbidCallbacks newroot a;
  
  
  (* Splice the equivalence classes together, after we have re-rooted one 
   * class. *)
  let old_eqca = a.eqc in a.eqc <- newroot.eqc; newroot.eqc <- old_eqca; 
  
  
  
  (* Splice the forbid lists. newroot gets the entire list *)
  newroot.forbid <- concatenateCirclist newroot.forbid a.forbid;
  newroot.forbidsize <- newroot.forbidsize + a.forbidsize;
  
  assert (watchNode 
            (dprintf "merge(%a, %a)\n" d_nodeTop newroot d_nodeTop a)
            newroot);
  
  (* Update the class size *)
  newroot.size <- newroot.size + a.size;
  

  (* Register the callbacks for the merge *)
  iterCirclist
    (fun f -> registerCallback f)
    newroot.mergeCallbacks;


  let rec iter f = function [] -> () | h::t -> f h; iter f t in
  iter
    (fun f -> registerCallback (fun _ -> f a; f newroot))
    !mergeCallbacks;

  (* Now register the undo operation *)
  registerUndoRun (fun _ -> 
    if !debug then 
      ignore (log "undoMerge for %d\n" a.id);
    let roota = a.root in (* This is the current root of a *)
    assert (newroot == roota);
    if a == roota then 
      E.s (E.bug "CC: undoMerge called for a root");
    
    (* Unsplice the equivalence classes *)
    let old_eqc = a.eqc in a.eqc <- roota.eqc; roota.eqc <- old_eqc;
    
    (* restore the class size *)
    roota.size <- roota.size - a.size;

    roota.forbidsize <- roota.forbidsize - a.forbidsize;
    
    (* Reroot the entire equivalence class. The proofs are done by a separate 
     * undo.  *)
    iterEqc (fun x -> x.root <- a) a;
    
    (* unsplice the forbid lists *)
    roota.forbid <- splitCirclist roota.forbid a.forbid;
    
    (* unsplice the parents lists *)
    roota.parents <- splitCirclist roota.parents a.parents;
    
    splitNewParentsCallbacks roota a;
    
    splitMergeCallbacks roota a;
    
    splitForbidCallbacks roota a;
    
    assert(watchNode 
             (dprintf "undoMerge(%a, %a)" d_nodeTop roota d_nodeTop a)
             a);
    assert(watchNode 
             (dprintf "undoMerge(%a, %a)" d_nodeTop roota d_nodeTop a)
             roota);
    
   );

  (* We now process the new merges. We must do this at the end, because the 
   * undo for a merge undoes all of the above at one time *)
  let rec iter f = function [] -> () | h::t -> f h; iter f t in
  iter
    (fun (n, c) -> 
      let rec iter f = function [] -> () | h::t -> f h; iter f t in
      iter 
        (fun n' -> 
          merge n n' 
               (if !genProofs then genCongrProof n n' else noProof) false)
        c)
    futureMerges;

  if !debug then begin
    ignore (log "After merge of %d and %d:\n" a.id newroot.id);
    printGraph ()
  end

					(* Check for congruences of node "n" 
                                         * among the given list *)
and checkCongruences (n: node)
                     (l: node circlist option) : unit = 
  let rec iter f = function [] -> () | h::t -> f h; iter f t in
  iter 
    (fun n' -> 
      merge n n' 
        (if !genProofs then genCongrProof n n' else noProof)
        false)
    (findCongruences n l)

let forbidMerge 
    (a: node)
    (b: node)
    (msg: message) : unit =

  let ra = a.root in
  let rb = b.root in
  if ra == rb then begin
    if !debug then 
      ignore (log "contradiction found while forbidding merge\n");
    if !debugStats then 
      incr eqContras;

    raise (S.Contradiction 
             (if !genProofs then 
               P.mkEqNeq_FI a.exp b.exp
                 (genEqProof a b)
                 msg.proof
             else
               noProof))
  end else if forbiddenEquality ra rb != None then begin
    () (* Already known *)
  end else begin
    if !debug then 
      ignore (log "forbidding merge of %d and %d\n" a.id b.id);
    let witness = (a,b,msg.proof) in 
    ra.forbid <- addToCirclist ra.forbid (rb, witness);
    ra.forbidsize <- 1 + ra.forbidsize;
    iterCirclist registerCallback ra.forbidCallbacks;

    rb.forbid <- addToCirclist rb.forbid (ra, witness);
    rb.forbidsize <- 1 + rb.forbidsize;
    iterCirclist registerCallback rb.forbidCallbacks;
(*
    ignore (log "after forbid\n"); printGraph ();
*)

    registerUndoRun (fun _ -> 
      ra.forbidsize <- ra.forbidsize - 1;
      rb.forbidsize <- rb.forbidsize - 1;
      ra.forbid <- removeFromCirclist ra.forbid None;
      rb.forbid <- removeFromCirclist rb.forbid None);
      

    assert (ra.active && rb.active);
  end




	

let assume (msg: message) = 
  runCallbacks ();

  (* Now process the message *)
  if !debug then 
    ignore (log "assume message %a\n" d_message msg);
  
  if msg.owner != ccid then begin
    (* Not our own message. *)

    match msg.body with 
      Intern (L.SetEq, n1, n2) -> merge n1 n2 msg.proof true
            
    | Intern(L.SetNe, n1, n2) -> 
        forbidMerge n1 n2 msg
            
    | _ -> ()
  end;

  runCallbacks ();

  ()
    

let postAssume _ : unit = 
  runCallbacks ()


module Witness = struct

    (* Go through the EDAG and find all SEL nodes. A cell corresponds to a 
     * class that is either the address of a SEL node, or contains a SEL 
     * node. Keep also track of memories. An edge between two cells is 
     * labeled with a memory. *)
  let repr: string IH.t = IH.create 13  (** map from class id to the 
                                         * representative expression *)
      
  let getReprNode (n: node) : string = 
    try IH.find repr n.root.id
    with Not_found -> assert false
        
  let getRepr (id: int) : string = 
    try IH.find repr id
    with Not_found -> assert false


  (** Accumulate here the result. We use a list to ensure that the order is 
   * Ok. *)
  let acc: doc list ref = ref [ ] 


  (* We keep track of edges for the pointer graph *)
  type edge = 
      { sel: node;
        start: L.exp; (** The class id of the start cell *)
        finish: L.exp; 
        memory: L.exp; (** the class id of the memory *)
      }

      
  let edges: edge list ref = ref []
  let addEdge (start: L.exp) (finish: L.exp) (memory: L.exp) 
              (sel: node) : unit = 
    if not (List.exists 
              (fun e -> 
                e.start = start && e.finish = finish &&
                e.memory = memory) !edges) then 
      edges := { sel = sel;
                 start = start; finish = finish; memory = memory } 
        :: !edges
                                                                       
                                                                       
                                                                       
            (* Keep track of the SEL nodes we have done *)
  let selNodes : unit IH.t = IH.create 13
      
      
  let sel = uniqueHead (UNINT (U.registerSymbolName "sel")) 
      
      
  let witness () = 

    if false then 
      Edag.printGraph ();

    (* Find all active SEL nodes *) 
    IH.iter 
      (fun h n -> 
        if n.root.active && h == sel then begin
          let addr, mem = match n.args with [a;m] -> a,m | _ -> assert false in
          IH.add selNodes n.id ();
          addEdge 
            (findSmallestForClass addr) 
            (findSmallestForClass n)
            (findSmallestForClass mem) n
        end)
      
      headToNode;

    (* For each edge add something to the result *)
    let rec iter f = function [] -> () | h::t -> f h; iter f t in
    iter 
      (fun edge -> 
        let sdotm = 
          sprint 1000 
            (dprintf "%a.%a" L.d_exp edge.start L.d_exp edge.memory) in
        accumWitnessPred (L.eq (L.Fun(sdotm,[])) edge.finish))

      !edges;
      
    (* Now scan all the active classes *)
    let rec iter f = function [] -> () | h::t -> f h; iter f t in
    iter 
      (fun n -> 
        if n == n.root && n.active then begin 
          (* Add equalities for each node to the class representative *)
          let rep = findSmallestForClass n in
          iterEqc 
            (fun n' -> 
              let n'exp = findSmallestForNode n' in 
              accumWitnessPred (L.eq n'exp rep))
            n;
          (* Add the disequalities for this class. Keep track of which 
           * equivalence classes we have already reported. *)
          let diseqcls: unit IH.t = IH.create 13 in 
          iterCirclist
            (fun (f, _) -> 
              let f' = f.root in
              if f'.id < n.id && f'.active && not (IH.mem diseqcls f'.id) then 
                begin 
                  IH.add diseqcls f'.id ();
                  let f'exp = findSmallestForClass f' in 
                  accumWitnessPred (L.ne f'exp rep)
                end)
            n.forbid
            
        end) 
      !edag;

    (* If the matching is disabled then mark that in the witness *)
    (match !matchingDisabled with
       Some why -> 
        accumWitnessPred (L.eq (L.Fun(("matching disabled: " ^ why), [])) L.trueExp)
      | None -> ())
    
end

let haveHead = Edag.uniqueHead (UNINT (U.registerSymbolName "have"))

(* A quick EDAG message tester *)
let edagMessageTester (msg: messagebody) : testResult = 
  match msg with 
    Extern (L.Bool true) -> Determined (true, P.Truei)
  | Extern (L.Bool false) -> Determined (false, P.Truei)
  | Intern (bop, n1, n2) -> begin
      match bop with 
        L.SetEq -> 
          if n1.root == n2.root then 
            Determined(true, if !genProofs then genEqProof n1 n2 else noProof)
          else begin
            match forbiddenEquality n1 n2 with 
              Some w -> 
                Determined(false, if !genProofs then w () else noProof)
            | _ -> NotDetermined
          end
                  
      | L.SetNe -> 
          E.s (E.bug "We should not have SetNe messages");
              
      | _ -> NotDetermined
  end
  | _ -> NotDetermined
  

let initialize () = 
  if !debug then ignore (log "Initializing EGraph...\n");
  undoStack := [];

  (** Add the true node to the EDAG *)
  trueNode := fst (insertExpNode L.trueExp);

  registerMessageTester edagMessageTester;

  ()




let entries = 
  { S.emptySatproc with 
    S.init  = initialize;
    S.finish = (fun _ -> 
      if !debugStats then 
        ignore (E.log "eqContra: %d\n" !eqContras));
    S.assume = assume;
    S.postAssume = postAssume;
    S.debug = debug;
    S.ident  = ccid;
    S.witness = Witness.witness;
    S.descr  = "congruence closure";
  } 
 
