(** This module handles injective functions **)
module E = Errormsg
module L = Logic
module P = Proof
module U = Util
module H = Hashtbl

open Satproc              (* message, etc. *)
open Edag


let injid = Util.registerSymbolName "INJ"

let debug = ref false 

let log (fmt: ('a,unit,Pretty.doc) format) : 'a = 
  ignore (E.log "INJ: ");
  E.log fmt

let warn (fmt: ('a,unit,Pretty.doc) format) : 'a = 
  ignore (E.log "INJ: ");
  E.warn fmt


(** For an injective function *)
let injectives: unit IH.t = IH.create 13

(* 
 * a stack based structure for snapshots and undos 
 *)
type undoElem = 
    Snapshot
  | AddInjective of U.symbol
        

let undoStack : undoElem list ref = ref [ ]

let id = ref 0
let newId () = 
  incr id;
  !id

let snapshot () = 
  undoStack := Snapshot :: !undoStack

let undo () = 
  let rec loop = function
      Snapshot :: rest -> undoStack := rest
    | AddInjective f :: rest -> IH.remove injectives f; loop rest
    | [] -> E.s (E.bug "INJ tried to pop an empty stack (too many undos?)")
  in
  loop !undoStack

let init () = 
  undoStack := []



(* take note of a fact *)
let assume (msg: message)     (* fact being asserted *)
           : unit = 
  if !debug then
    ignore (log "assume: %a\n" d_message msg);
  
  (* Scan the goal and look for "upd" *)
  (** A visitor that scans a predicate for "upd" *)
  let updates: L.exp list ref = ref [] in
  if msg.owner <> injid then begin
    match msg.body with
      Extern (L.All(v, L.PFun("injective", [ L.Fun(fname, args) ]))) -> 
        let fname_sym = U.registerSymbolName fname in
        IH.add injectives fname_sym ();
        undoStack := (AddInjective fname_sym) :: !undoStack

    | Extern (L.PFun("injective", _)) -> 
        ignore (log "invalid use of \"injective\": %a\n" d_messagebody msg.body)

      (* Watch for equalities *)
    | Intern(L.SetEq, n1, n2) when not msg.goal -> 
        iterEqc
          (fun n1' -> 
            match n1'.head with 
              UNINT f1 when IH.mem injectives f1 -> 
                (* Now scan the equivalence class of n2 for nodes with the 
                 * same function symbol *)
                iterEqc
                  (fun n2' -> 
                    match n2'.head with 
                      UNINT f2 when f1 = f2 -> 
                        if List.length n1'.args <> List.length n2'.args then 
                          ignore (warn "Injective function %s used with different number of arguments" (U.symbolName f1))
                        else begin
                          List.iter2 
                            (fun a1 a2 -> 
                              if a1.root != a2.root then 
                                newMessage
                                  { body = Intern (L.SetEq, a1, a2);
                                    owner = injid;
                                    mkProof = (fun _ -> assert false);
                                    goal = false; })
                            n1'.args n2'.args
                        end
                    | _ -> ())
                  n2

            | _ -> ())
          n1;

        ()

    | _ -> ()
        
  end

(* list of entry points exported as a satproc *)
let entries: Satproc.entries = {
  canEnable = (fun _ -> true);
  defaultEnabled = false;
  init      = init;
  snapshot  = snapshot;
  undo      = undo;
  assume    = assume;
  postAssume = (fun _ -> ());
  debug     = debug;
  ident     = injid;
  witness   = (fun w -> w);
  descr     = "injective";
  options   = [];
}
