open Pretty
module E = Errormsg
open Logic

type loc = E.location

type axiom = string * var list * pred list * pred
(** The input to the prover is presented as a sequence of "inputs".*)
type input =
    (** The next few kinds of inputs can appear only at top level *)
    StartGoal of string * loc           (* The start of a goal. Followed by a 
                                         * number of elements ended by 
                                         * some top-level element *)
  | TopChoose of int * int              (* A choose at the top level *)
  | Axiom  of axiom
  | Import of string
  | Decl   of string

  | Block of (string * string) list * loc
  | UnBlock of loc

  | Eof

  (** The next few kinds of inputs can appear only after a StartGoal **)

  | Prove of Logic.pred * loc (** Prove something. Not a conjunction, nor 
                               * implication, nor universal quantification. *)

  | ProveInvalid of Logic.pred * loc (** Attempt to prove something that 
                                      * should not be provable, for debugging *)
  | Assume of Logic.pred * loc (** Assume something. Left-hand of an 
                                * implication *)
  | UnAssume of loc     (** retract the last assumption *)

  | Pick of string * loc  (** Pick a variable. Universal quantification *)
  | UnPick  of loc    (** Retract the last variable picked *)

  | Choose of int * int * loc
  | Comment of string * loc
  | EndGoal of loc



(********* Pretty printing ***********)
let printLFSyntax = ref false

class lfLogicPrinterClass: logicPrinter = object (self)
    inherit defaultLogicPrinterClass

end

let lfLogicPrinter = new lfLogicPrinterClass

let di () = function
    StartGoal (name, l) -> dprintf "StartGoal(%s) @@%a" name E.d_loc l
  | EndGoal l -> dprintf "EndGoal @@%a" E.d_loc l
  | Prove (p, l) -> dprintf "Prove(@[%a@]) @@%a" d_pred p E.d_loc l
  | ProveInvalid (p, l) -> 
      dprintf "ProveInvalid(@[%a@]) @@%a" d_pred p E.d_loc l
  | Choose(i,nr,l) -> dprintf "Choose(%d, %d) @@%a" i nr E.d_loc l
  | Comment (s, l) -> dprintf "%% %s @@%a" s E.d_loc l
  | Block (sl, l) -> dprintf "Block [%a] @@%a" 
                    (docList
                        (fun (s, s') -> dprintf "%s=\"%s\"" s s'))
                    sl
                    E.d_loc l
  | UnBlock l -> dprintf "UnBlock @@%a" E.d_loc l
  | Assume (p,l) -> dprintf "Assume(@[%a@]) @@%a" d_pred p E.d_loc l
  | UnAssume l -> dprintf "UnAssume @@%a" E.d_loc l
  | Pick (v, l) -> dprintf "Pick(%s) @@%a" v E.d_loc l
  | UnPick l -> dprintf "UnPick @@%a" E.d_loc l
  | Decl s -> dprintf "Decl(%s)" s
  | Import s ->  dprintf "Import(%s)" s
  | Eof -> text "Eof"
  | Axiom (n, _, _, _) -> dprintf "Axiom(%s)" n
  | TopChoose (n, m) -> dprintf "TopChoose(%d, %d)" n m



