%{
module I = Input
module L = Logic
module E = Errormsg
open Pretty

let parse_error msg =
  E.parse_error msg


let bindings : (string * string) list ref = ref [] 
let pushBinding vname : string = 
  let v = L.newVar vname in
  bindings := (vname, v) :: !bindings;
  v

let rec popBindings n : unit = 
  if n = 0 then () else
  begin
    (match !bindings with
      [] -> E.s (E.bug "no binding to pop")
    | _ :: rest -> bindings := rest);
    popBindings (n - 1)
  end

let lookupVar name = 
  try
    let (_, v) = List.find (fun (n,_) -> n = name) !bindings in
    v
  with Not_found -> 
    (parse_error ("Variable " ^ name ^ " not bound"); "<error>")

let tryLookupVar name= 
  let (_, v) = List.find (fun (n,_) -> n = name) !bindings in
  v
  
(** Keep track of assumptions and universal quantification *)
type parenthesisKind = 
    Assume | Pick | Block
let toretract: parenthesisKind list ref = ref []

let openParenthesis (k: parenthesisKind) = 
  toretract := k :: !toretract
  
let closeParenthesis (l: I.loc) = 
  match !toretract with 
    Assume :: rest -> toretract := rest; I.UnAssume l
  | Block :: rest -> toretract := rest; I.UnBlock l
  | Pick :: rest -> toretract := rest; popBindings 1; I.UnPick l
  | _ -> E.s (E.bug "LfParse: empty retract list")

let rec fromList = function
    [] -> L.Bool true
  | [p] -> p
  | p :: rest -> L.And [p; fromList rest]

type lf = 
    LConst of string
  | LVar of string
  | LFun of (string * lf) list * lf    (* Bindings with types, the body *)
  | LApp of lf * lf list               (* The head and arguments *)
  | LPf  of L.pred                     (* PF applied to a predicate. A 
                                        * special case of LApp *)

let rec dlf () = function
    LConst s -> text s
  | LVar s -> text s
  | LFun (vars, body) -> 
      dprintf "Fun %a %a"
        (docList ~sep:nil (fun (v,t) -> dprintf "{%s : %a}" v dlf t))
        vars
        dlf body
  | LApp (head, args) -> dprintf "(%a @[%a@])" dlf head
        (docList ~sep:break (dlf ())) args
  | LPf p -> dprintf "(pf %a)" L.d_pred p

let makeAxiom name term = 
  match term with
    LFun (args, LPf conc) -> begin
      (* Separate the arguments into expression variables and hypotheses *)
      let rec separate vars hyps = function
          [] -> I.Axiom(name, List.rev vars, List.rev hyps, conc)
        | (aname, LConst "exp") :: rest -> 
            separate (aname :: vars) hyps rest
        | (aname, LPf a) :: rest when L.isAtomPred a -> 
            separate vars (a :: hyps) rest
        | rest -> begin
            (* ignore (E.log "Ignoring axiom %s (bad arg: %a)\n" name
                      dlf (LFun(rest, LPf (L.Atom conc)))); *)
            I.Decl name
        end
      in
      separate [] [] args
    end
  | LPf conc when L.isAtomPred conc -> I.Axiom (name, [], [], conc)
  | _ -> begin
      (* ignore (E.log "Ignoring axiom %s (not a Fun: %a)\n" name dlf term); *)
      I.Decl name
  end
  
%}


%token <string> Id Use STRING
%token <string * Input.loc> Comment
%token <int> Num
%token EXP PRED
%token <Input.loc> PF BLOCK RPAREN DOT 
%token LPAREN LBRACKET RBRACKET LBRACE RBRACE
%token COLON ARROW COMMA
%token PLUS MINUS TIMES
%token MOD DIV FLOOR CEIL
%token LT GT LEQ GEQ EQ NEQ
%token TRUE FALSE AND OR 
%token <Input.loc> IMPL ALL EXISTS CHOOSE
%token BEGINDECL ENDDECL
%token EOF STARTLF

%start main

%type <Input.input> main
%type <Logic.pred> pred
%type <Logic.pred list> predlist
%type <Logic.exp> exp
%type <Logic.exp list> exps
%type <string> bind

%type <lf> lfatom lfterm
%type <string> lfconstant
%type <(string * lf) list> lfbindings
%type <string * lf> lfbind
%type <lf list> lftermlist


%type <Logic.pred> atom
%type <Logic.pred> assumption goal
%%
 

/*(* This is the top-level. This parses only the beginning of goals *)*/
main:  
    Id COLON PF                      { I.StartGoal ($1, $3) }
  | lfconstant COLON lfterm DOT      { makeAxiom $1 $3 }
  | PF COLON lfterm DOT              { I.Decl "pf" }
  | Use				     { I.Import $1 }
  | EOF                              { I.Eof }

  | savepos goal                     { I.Prove ($2, $1) }
  | savepos LPAREN CHOOSE Num Num RPAREN     { I.Choose ($4, $5, $3) }
  | savepos LPAREN IMPL assumption    { openParenthesis Assume; 
                                       I.Assume ($4, $3) }
  | savepos LPAREN ALL lfbind
                                     { openParenthesis Pick; 
                                       I.Pick (fst $4, $3) }
  | RPAREN                           { closeParenthesis $1 }
  | Comment                          { let str, pos = $1 in 
                                       I.Comment (str, pos) }
  | savepos LPAREN BLOCK LBRACKET block_pars RBRACKET         
                                     { openParenthesis Block;
                                       I.Block ($5, $3) } 
  | error RPAREN                     { closeParenthesis $2 }
  | error DOT                        { I.EndGoal $2 }
  | DOT                              { I.EndGoal $1 }
;

savepos: 
                                     { E.getLocation () }
;

block_pars: 
     /*empty*/                          { [] }
  |  block_pars_ne                      { $1 }
  |  error                              { [] }
;
block_pars_ne:
  |  block_par                          { [ $1 ] }
  |  block_par COMMA block_pars_ne      { $1 :: $3 }
;
block_par:
     Id COLON STRING                    { ($1, $3) }
  |  Id COLON Num                       { ($1, string_of_int $3) }
  |  Id EQ STRING                       { ($1, $3) }
  |  Id EQ Num                          { ($1, string_of_int $3) }
;

assumption: 
    atom                                { $1 }
  | LPAREN AND assumption assumption RPAREN
                                        { L.And [$3; $4] }
  | LPAREN OR assumption assumption RPAREN 
                                        { L.Or  [$3; $4] }
  | LPAREN ALL lfbind assumption RPAREN { L.All (fst $3, $4) }


goal: 
    atom                                { $1 }

/*(* You should really not use /\ for goals, because this forces Kettle to 
   * read the whole thing. Just use sequences *)*/
  | LPAREN AND goal goal RPAREN         { L.And [$3; $4] }
  | LPAREN OR goal goal RPAREN          { L.Or [$3; $4] }
  | LPAREN EXISTS lfbind goal RPAREN
                                        { L.Exist(fst $3, $4) }
atom:
    FALSE				{ L.Bool false }
  | TRUE                                { L.Bool true  }
  | LPAREN EQ  exp exp	RPAREN 		{ L.eq $3 $4 }
  | LPAREN NEQ exp exp	RPAREN 		{ L.ne $3 $4 }
  | LPAREN GEQ exp exp	RPAREN 		{ L.geq $3 $4 }
  | LPAREN GT  exp exp	RPAREN 		{ L.gt $3 $4 }
  | LPAREN LEQ exp exp	RPAREN 		{ L.leq $3 $4 }
  | LPAREN LT  exp exp	RPAREN 		{ L.lt $3 $4 }
  | LPAREN Id exps RPAREN		{ L.pFun $2 $3 }
;
  

pred:
    TRUE				{ L.Bool true  }
  | FALSE				{ L.Bool false }

  | Id                                  { L.pFun "pvar" [] }

  | LPAREN ALL Id RPAREN                { L.pFun "pvarall" [] }

  | LPAREN ALL LPAREN bind predlist RPAREN RPAREN	
                                        { popBindings 1;
                                          L.All ($4, fromList $5) }
  | LPAREN EXISTS LPAREN bind predlist RPAREN RPAREN	
                                        { popBindings 1;
                                          L.Exist ($4, fromList $5) }
  | LPAREN AND predlist	RPAREN 		{ fromList $3 }
  | LPAREN OR predlist	RPAREN		{ L.Or   $3 }
  | LPAREN IMPL pred predlist	RPAREN  { L.Imply  ($3, fromList $4) }

  | LPAREN EQ  exp exp	RPAREN 		{ L.eq $3 $4 }
  | LPAREN NEQ exp exp	RPAREN 		{ L.ne $3 $4 }
  | LPAREN GEQ exp exp	RPAREN 		{ L.geq $3 $4 }
  | LPAREN GT  exp exp	RPAREN 		{ L.gt $3 $4 }
  | LPAREN LEQ exp exp	RPAREN 		{ L.leq $3 $4 }
  | LPAREN LT  exp exp	RPAREN 		{ L.lt $3 $4 }
  | LPAREN Id exps RPAREN		{ L.pFun $2 $3 }
;

predlist:
                                        { [] }
  | pred predlist                       { $1 :: $2 }

exp:
  | Num					{ L.integer $1 }
  | PLUS  exp exp			{ L.plus $2 $3 }
  | MINUS exp exp			{ L.minus $2 $3 }
  | TIMES exp exp			{ L.times $2 $3 }
  | Id					{ (* See if it is a variable *)
                                          try 
                                            L.NVar (tryLookupVar $1)
                                          with Not_found -> 
                                            L.Fun($1, []) 
                                        }
  | LPAREN Id exps RPAREN		{ try 
                                            ignore (tryLookupVar $2);
                                            parse_error ("Variable " ^ $2 ^ " is applied")
                                          with Not_found -> 
                                            L.Fun($2, $3) 
                                        }
  | DIV   exp Num			{ L.Fun("div", [L.integer $3; $2]) }
  | MOD   exp Num			{ L.Fun("mod", [L.integer $3; $2]) }
  | FLOOR exp Num			{ L.Fun("floor", [L.integer $3; $2]) }
  | CEIL  exp Num			{ L.Fun("ceil", [L.integer $3; $2]) }
  | LPAREN exp RPAREN			{ $2 }
;
exps:
    exp					{ [$1] }
  | exp exps				{ $1 :: $2 }
;

/*(**** Parsing of LF terms in normal form. Applications must be parenthesized.
 ***)*/
lfconstant: 
  | AND    {  "/\\"}
  | OR     { "\\/" }
  | IMPL   { "=>" }
  | ALL    { "all" }
  | EQ     { "=" }
  | NEQ    { "<>" }
  | GEQ    { ">=" }
  | GT     { ">" }
  | LEQ    { "<=" }
  | LT     { "<" }
  | FLOOR  { "floor" }
  | MOD    { "mod" }
  | CEIL   { "ceil" }
  | DIV    { "div" }
  | EXP    { "exp" }
  | PRED   { "pred" }
;

lfatom: 
    lfconstant                          { LConst $1 }
  | LPAREN lfconstant lftermlist RPAREN { LApp (LConst $2, $3) }
  | Id                                  { try 
                                              LVar (tryLookupVar $1) 
                                          with Not_found -> 
                                              LConst $1
                                        }
  | LPAREN Id lftermlist RPAREN         { LApp (LVar (lookupVar $2), $3) }

  | PF pred                             { LPf $2 }

  | LPAREN lfatom RPAREN                { $2 }

;

lfterm:
    lfbindings            { popBindings (List.length $1);
                            (let rec removeLast = function
                                [b] -> ([], b)
                              | a :: rest -> 
                                  let (args, b) = removeLast rest in
                                  (a :: args, b)
                              | [] -> 
                                  (parse_error "empty binding";
                                   raise (Failure "empty_binding"))
                            in
                            let (args, (_, body)) = removeLast $1 in
                            match args with 
                              [] -> body
                            | _ -> LFun (args, body)) 
                          } 
;

lftermlist:
  | lfterm                                        { [$1] }
  | lfterm lftermlist                             { $1 :: $2 }
; 

lfbindings: 
    lfatom                                        { [(pushBinding "Pi", $1)] }
  | lfbind lfbindings                             { $1 :: $2 }
;
 
lfbind:   
    LBRACE Id COLON lfterm RBRACE           { (pushBinding $2, $4) }
  | LBRACKET Id COLON lfterm RBRACKET       { (pushBinding $2, $4) }
  | LBRACKET Id RBRACKET                    { (pushBinding $2, LConst "val") }
  | LPAREN lfterm RPAREN ARROW              { (pushBinding "Pi", $2) }
  | lfatom ARROW                            { (pushBinding "Pi", $1) }

bind: 
    LBRACKET Id COLON EXP RBRACKET	{ pushBinding $2 }
;





