module E = Errormsg
module P = Proof
module L = Logic

open Pretty
open Edag  (* only for genProofs; too bad *)

(* This is an auxiliary module that can be useful to more procedures for 
 * linear arithmetic *)

  (** We represent facts using linear combinations of factors. **)
  type arithRes = 
      { iseq: bool; (** If this is an equality *)
        ct: int;  (** the constant part *)
        facts: (int * int) list; (** the weighted list of factors: (weight, 
                                  * factor index), ordered by increasing 
                                  * factor index. *)
        proofs: arithproof list; (* Weighted list of arithmetic proofs. The 
                                  * direction of the proofs is important 
                                  * because we are going to be using them for 
                                  * inequality proofs. We always have fact + 
                                  * ct - orig = 0.*)
      } 


  (** An equality proof, as needed for ArithEq and ArithGe. First is the 
   * factor, then whether it is an equality proof, and the actual proof thunk *)
  and arithproof = int * bool * proofthunk

  and proofthunk = 
      Proof of P.proof
    | ProofThunk of (unit -> P.proof)

        (** Some functions in this module are parameterized by some
            way to handle indices. I thought about using a functor, but I have not
            done that yet. Also, this way we can use one module for many kinds of 
            indices. *)

  type factorData = {
      toexp: int -> Logic.exp; (** Get the expression for the factor *)
      todoc: int -> Pretty.doc; (** Get the doc for it *)
    }


      (* Find weight of a factor *)
  let findFactorWeight (f: int) (c: arithRes) : int = 
    let rec loop = function
        [] -> 0
      | (w, f') :: rest -> 
          if f = f' then w else
          if f < f' then 0 else loop rest
    in
    loop c.facts


  (** Compute r.facts + r.ct *)
  let arithResToExp (fd: factorData) (r: arithRes) = 
    let ct = if r.ct <> 0 then Some (L.integer r.ct) else None in 
    let add (acc: L.exp option) (what: L.exp option) = 
      match acc, what with 
        None, _ -> what
      | _, None -> acc
      | Some a1, Some w1 -> Some (L.plus a1 w1)
    in
    let facts = 
      let rec loop (acc: L.exp option) = function
          [] -> acc
        | (w, f) :: rest -> 
            let ef = fd.toexp f in 
            let acc' = 
              if w = 1 then add acc (Some ef)
              else add acc (Some (L.times (L.integer w) ef))
            in
            loop acc' rest
      in
      loop None r.facts
    in
    match add ct facts with
      None -> L.zero
    | Some e -> e
          
  let d_arithRes (fd: factorData) () (r: arithRes) = 
    let ct = if r.ct <> 0 then num r.ct else nil in
    let facts = if r.facts <> [] then 
      docList ~sep:(text " + ")
        (fun (w, f) -> 
          if w <> 1 then dprintf "%d*%a" w insert (fd.todoc f) 
                    else fd.todoc f)
        ()
        r.facts
    else
      nil
    in
    (if facts == nil then 
      num r.ct
    else if r.ct = 0 then 
      facts
    else
      ct ++ text "+" ++ facts)

      ++ (if r.iseq then text " = 0" else text " >= 0")
        

  let eqProofsToPargs (prfs: arithproof list) : P.parg list = 
    let rec fold_left f acc = 
       function []-> acc | h::t -> fold_left f (f acc h) t in
     if !genProofs then 
       (fold_left
          (fun acc (f, iseq, p) -> 
            assert (f <> 0);
            assert (f > 0 || iseq);
            let pp = 
              match p with 
                Proof pp -> pp
              | ProofThunk pp -> pp ()
            in
            P.I f :: P.P pp :: acc)
          []
          prfs)
     else
       []

(** Compute a proof r.facts + r.ct >= 0 (or = 0) *)

  let proofOfLinArith (fd: factorData) (r: arithRes) : (unit -> P.proof) = 
    let e = arithResToExp fd r in
    (* Fold in the proofs from arithReg into mkProofs *)
    let rec fold_left f acc = 
      function []-> acc | h::t -> fold_left f (f acc h) t in
    let mkProof () : P.proof = 
      (* Build the proof of equalities from the [r] *)
      let peq = eqProofsToPargs r.proofs in 
      if r.iseq then 
        P.mkArithEq e L.zero peq
      else
        P.mkArithGeq e L.zero peq
    in
    mkProof

  (** Check the invariants of an arithRes *)
  let check (fd: factorData) (msg: doc) (r: arithRes) : bool = 
    let lastfactor: int option ref = ref None in
    let rec iter f = function [] -> () | h::t -> f h; iter f t in
    iter (fun (w, f) -> 
      assert (w <> 0);
      match !lastfactor with 
        None -> lastfactor := Some f
      | Some f' -> assert (f' < f); lastfactor := Some f') r.facts;

    (* Check the proof *)
    if !genProofs then 
      proofWellFormed 
        (dprintf "Linarith.check %a\n(%a)" (d_arithRes fd) r insert msg)
        (proofOfLinArith fd r ())
    else
      true


  let emptyRes ~(eq: bool) = { iseq = eq;
                               ct = 0; facts = []; 
                               proofs = []; }

  let isConst (r: arithRes) : int option = 
    if r.facts = [] then Some r.ct else None
      
  (* Accumulate a constant *)
  let accumConst (r: arithRes) (w: int) = { r with ct = w + r.ct }
      
  let scaleLincomb (w: int) (lc: (int * 'a) list) : (int * 'a) list = 
    let rec map f = function [] -> [] 
                     | h::t-> let h' = f h in h' :: map f t in 
    if w = 0 then [] 
    else map (fun (w', p) -> (w * w', p)) lc
  let scaleProofs (w: int) (lc: arithproof list) : arithproof list = 
    let rec map f = function [] -> [] 
                     | h::t-> let h' = f h in h' :: map f t in 
    if w = 0 then [] 
    else map (fun (w', eq, p) -> (w * w', eq, p)) lc
        
  (* Scale an arithRes *)
  let scale (w: int) (r: arithRes) = 
    { iseq = r.iseq; 
      ct = w * r.ct;
      facts = scaleLincomb w r.facts;
      proofs = scaleProofs w r.proofs;
    } 
      

  let rec combfacts 
      (fl1: (int * int) list) 
      (fl2: (int * int) list) : (int * int) list = 
    match fl1, fl2 with
      [], _ -> fl2
    | _, [] -> fl1
    | (w1, f1) :: rest1, (w2, f2) :: rest2 -> 
        if f1 = f2 then 
          if w1 + w2 = 0 then combfacts rest1 rest2
          else (w1 + w2, f1) :: combfacts rest1 rest2
        else if f1 < f2 then
          (w1, f1) :: combfacts rest1 fl2
        else
          (w2, f2) :: combfacts fl1 rest2


  let canonicalFacts (f: (int * int) list) = 
    let rec fold_left f acc = 
      function []-> acc | h::t -> fold_left f (f acc h) t in
    fold_left
      (fun acc f -> combfacts acc [ f ])
      []
      f
    
  (* Add two arithRes, maintaining their canonical forms. *)
  let add (r1: arithRes) (r2: arithRes) : arithRes = 
    { iseq = r1.iseq && r2.iseq;
      ct = r1.ct + r2.ct;
      facts = combfacts r1.facts r2.facts;
      proofs = r1.proofs @ r2.proofs
    }
