(* mlss.ml *)
(* Implementation of a fragment of set theory called "MLSS":
 * "multi-level syllogistic with singleton".  Based on:
 *
 *   Domenico Cantone and Calogero G. Zarba.
 *   A new fast tableau-based decision procedures for an
 *     unquantified fragment of set theory.
 *   In Ricardo Caferra and Gernot Salzer, editors. Automated
 *     Deduction in Classical and Non-Classical Logics,
 *     volume 1761 of Lecture Notes in Computer Science,
 *     pages 127-137, Springer, 2000.
 *   http://webloria.loria.fr/~zarba/papers/CZ00a.ps.gz
 *)

(* ----------------- imported concepts ----------------- *)
(* used with explicit qualification:
 * Inthash, Util, Engine, List *)
module E = Errormsg       (* E.log *)
module S = Satproc        (* S.Contradiction, S.entries *)
module L = Logic          (* pred, exp, etc. *)
module P = Proof          (* proof, etc. *)
open Edag                 (* node, circlist, etc. *)

(* from Logic *)
type pred = Logic.pred
type exp = Logic.exp
let d_exp = Logic.d_exp

(* e-dag node for "true" *)
let trueNode: node ref = Edag.trueNode
let trueExp: exp = Logic.trueExp


(* ----------------------- types ----------------------- *)
(* Record that a witness to the disequality of sets 'n1' and 'n2' has
 * been materialized, either as an explicit e-dag assertion, or a
 * pending case analysis.  The nodes here are *not* necessarily e-dag
 * roots (they are at the time of creation, but subsequent merging
 * might invalidate that). *)
type handledDis_t = {
  n1: node;
  n2: node;
}


(* ------------------ global variables ----------------- *)
(* --- set once, at program start --- *)
(* satproc id *)
let mlssid = Util.registerSymbolName "SET"

(* general debug flag, settable from command line: "-debug SET" *)
let debug = ref false

(* when true, print out debug info for the elemOf DFS *)
let debugDFS = false

(* when true, print debug info for the disequality processing *)
let debugDisequality = false

(* when true, debug the translation of subsetOf *)
let debugSubset = false


(* --- set once, during 'init' --- *)
(* function symbol ids *)
let elemOfCode: int ref = ref 0             (* "elemOf" *)
let subsetOfCode: int ref = ref 0           (* "subsetOf" *)
let strictSubsetOfCode: int ref = ref 0     (* "strictSubsetOf" *)
let unionCode: int ref = ref 0              (* "union" *)
let intersectionCode: int ref = ref 0       (* "intersection" *)
let setSubtractCode: int ref = ref 0        (* "setSubtract" *)
let singletonCode: int ref = ref 0          (* "singleton" *)

(* e-dag node for "emptySet" *)
let emptySetNode: node ref = ref Edag.dummyNode


(* --- these change over time --- *)
(* The variables in this section are dependent on the current
 * branch of case analysis, so are managed by calling
 * Edag.registerUndoRun (or variants). *)

(* Which disequalities have we already handled? *)
let handledDisequalities: handledDis_t list ref = ref []

(* This is set of e-dag nodes that assert or deny a subsetOf or
 * strictSubsetOf, that we have "handled" in the sense of translating
 * into more primitive operations.  The hashtable 'value' is ignored. *)
let handledSubsetNodes: int Inthash.t = (Inthash.create 13)


(* --------------------- utilities --------------------- *)
(* in debug mode, print a log string *)
let debugLog (s: string) : unit =
begin
  if !debug then
    (ignore (E.log "SET: %s\n" s));
end


let isEquality (p: L.pred) : bool =
begin
  (match p with
  | L.PNez(L.BinOp(L.SetEq, _, _)) -> true
  | _ -> false)
end


(* does any node in the equivalence class of 'n' satisfy 'f'? *)
let testInEqc (f : node -> bool) (n : node) : bool =
begin
  try
    (ignore (Edag.findEqc f n));
    true
  with _ ->
    false
end


let testCirclist (f: 'n -> bool)
                 (l: 'n circlist option) : bool =
begin
  try
    (ignore (Edag.findCirclist f l));
    true
  with _ ->
    false
end


(* make up an expression containing a fresh variable *)
let makeFreshExp (name: string) : exp =
begin
  let freshName:string = (newAlphaName name) in
  (L.NVar freshName)
end


(* broadcast a fact *)
let broadcast (fact: Logic.pred) (pf: P.proof) : unit =
begin
  (Edag.newMessage
    { body = (Extern fact);
      owner = mlssid;
      proof = pf;
      fromgoal = false; });
end

(* Generate a proof that 'n' equals TRUE. *)
let proofOfPred (n: node) : P.proof =
begin
  (!Edag.genEqProof n !trueNode)
end

(* Generate a proof that 'n' does not equal TRUE. *)
let proofOfNegPred (n: node) : P.proof =
  match forbiddenEquality n !trueNode with 
    Some w -> w ()
  | None -> E.s (E.bug "proofOfNegPred")
        

let applyElemOf (a1: exp) (a2: exp) : exp =
  (L.Fun ("elemOf", [a1;a2]))

let predElemOf (a1: exp) (a2: exp): pred =
  (L.eq trueExp (applyElemOf a1 a2))

let predNotElemOf (a1: exp) (a2: exp): pred =
  (L.ne trueExp (applyElemOf a1 a2))


let applyUnion (a1: exp) (a2: exp) : exp =
  (L.Fun ("union", [a1;a2]))


(* generate a generic proof using rule 'ruleName', main proof
 * component 'mainProof', and utilizing some introduced fresh
 * variables, 'freshVars' *)
let proofWithFreshVars (ruleName: string) (freshVars: exp list)
                       (mainProof: P.proof) : P.proof =
begin
  if !debug then ignore (E.log "SET: using %s\n" ruleName);
  (Proof.G (ruleName, (Proof.P mainProof) ::
                      (List.map (fun e -> (Proof.E e)) freshVars)))
end

(* ----------------------- init ------------------------- *)
(* initialize this satproc module *)
let init () : unit =
begin
  (debugLog "initialize");

  let getCode (s: string) : int = begin
    (Edag.uniqueHead (Edag.UNINT (Util.registerSymbolName s)))
  end in
  elemOfCode :=         (getCode "elemOf");
  subsetOfCode :=       (getCode "subsetOf");
  strictSubsetOfCode := (getCode "strictSubsetOf");
  unionCode :=          (getCode "union");
  intersectionCode :=   (getCode "intersection");
  setSubtractCode :=    (getCode "setSubtract");
  singletonCode :=      (getCode "singleton");

  emptySetNode := (fst (Edag.insertExpNode (Logic.Fun("emptySet",[]))));
end


(* ---------------- checkElemOfRelation ---------------- *)
type color = (*WHITE |*) GRAY | BLACK

exception FoundCycle of (node list *    (* cycle of elemOf links *)
                         node)          (* last set-typed node in the cycle *)

(* check for a cycle in the element membership graph *)
let checkElemOfRelation (trueEqc: node) : unit =
begin
  (* Keep track of which equivalence class ids have been explored.
   * Actually, WHITE is never used; it corresponds to an unmapped id. *)
  let visited : (color Inthash.t) = (Inthash.create 13) in

  let rec indent (depth: int) : string =
  begin
    if (depth == 0) then
      ""
    else
      "  " ^ (indent (depth-1))
  end in

  (* DFS from a node, looking for cycles *)
  let rec dfs (n: node) (depth: int) : unit =
  begin
    let eqc:node = n.root in

    if debugDFS then
      (ignore (E.log "SET: %sdfs(%d): at %a\n" (indent depth) eqc.id d_nodeTop n));
    let depth:int = (depth+1) in

    (* already visited? *)
    try
      match (Inthash.find visited eqc.id) with
      | GRAY -> (
          (* found a cycle! *)
          if debugDFS then (ignore (E.log "SET: %sdfs(%d): is gray -> cycle!\n" (indent depth) eqc.id));
          raise (FoundCycle ([], n (*not eqc*)));
        )
      | BLACK -> (
          (* already explored all nodes reachable from this one,
           * prune the search here *)
          if debugDFS then (ignore (E.log "SET: %sdfs(%d): is black\n" (indent depth) eqc.id));
          ()
        )
    with Not_found -> (
      (* not yet visited *)

      (* mark the node as being explored *)
      (Inthash.add visited eqc.id GRAY);

      (* recursively explore node's neighbors *)
      (explore eqc depth);

      (* done exploring *)
      (Inthash.add visited eqc.id BLACK);
    );
  end

  (* invoke 'dfs' on any nodes that 'eqc' is an element of *)
  and explore (eqc: node) (depth: int) : unit =
  begin
    let processParent (p: node) : unit =
    begin
      if false then (
        let argRootId (n: int) : int =
        begin
          if ((List.length p.args) > n) then
            (List.nth p.args n).root.id
          else
            -1
        end in

        if debugDFS then (ignore (E.log
          "SET: %sdfs(%d): p.root.id=%d trueEqc.id=%d\n\
          .                p.head=%d elemOfCode=%d\n\
          .                arg0.root.id=%d arg1.root.id=%d\n"
          (indent depth) eqc.id
          p.root.id trueEqc.id
          p.head !elemOfCode
          (argRootId 0) (argRootId 1)));
      );

      (* For 'p' to be interesting, several things must be true:
       *   - 'p' must be equal to "true"
       *   - 'p' must be an application of the "elemOf" f-symbol
       *   - 'eqc' must be the first argument of 'p'
       * Under these circumstances, 'eqc' is semantically an
       * element of the second argument to 'p'.
       *)
      if ((p.root == trueEqc) &&
          (p.head == !elemOfCode) &&
          ((List.hd p.args).root == eqc)) then (
        let lhs:node = (List.nth p.args 0) in
        let rhs:node = (List.nth p.args 1) in

        (* if the rhs is emptySet, then we have inclusion of
         * something in emptySet, which violates its emptiness *)
        if (rhs.root == (!emptySetNode).root) then (
          if debugDFS then (ignore (E.log "SET: %sdfs(%d): is elem of emptySet!\n" (indent depth) eqc.id));
          if !debug then ignore (E.log "SET: using emptySet_not_empty\n");
          raise (S.Contradiction (P.G("emptySet_not_empty",
                  [(P.P (proofOfPred p));                        (* lhs is in rhs *)
                   (P.P (!Edag.genEqProof rhs !emptySetNode))]   (* rhs is empty *)
                )))
        );

        (* recursively explore the second argument *)
        if debugDFS then (ignore (E.log "SET: %sdfs(%d): is elem of %d\n" (indent depth) eqc.id rhs.root.id));
        try
          (dfs rhs (depth+1))
        with FoundCycle (cycle, last) -> (
          (* accumulate 'p' into the growing cycle *)
          let cycle = (p :: cycle) in

          (* done? *)
          if (lhs.root == last.root) then (
            (* Make a proof out of the complete cycle.  The proof arguments
             * alternate between establishing the equality of two set terms
             * and establishing inclusion of one term as an element in
             * another (set) term. *)
            let f ((acc, prevNode) : (P.parg list * node)) (link: node) : (P.parg list * node) =
            begin
              let lhs:node = (List.nth link.args 0) in     (* element *)
              let rhs:node = (List.nth link.args 1) in     (* set *)
              let newAcc: P.parg list =
                (P.P (!Edag.genEqProof prevNode lhs))      (* equality *)
                  :: (P.P (proofOfPred link))              (* inclusion *)
                  :: acc in
              (newAcc, rhs)
            end in
            let ((pargs, last'): (P.parg list * node)) =
              (List.fold_left f ([],last) cycle) in
            assert(last == last');

            (* yield it *)
            if !debug then ignore (E.log "SET: using set_inclusion_cycle\n");
            raise (S.Contradiction (P.G("set_inclusion_cycle", pargs)))
          )
          else
            raise (FoundCycle (cycle, last))
        )
      );
    end in
    (iterCirclist processParent eqc.parents);
  end in

  (* iterate over all e-dag nodes *)
  let rec loop (list: node list) : unit =
  begin
    match list with
    | n :: rest -> (
        (try
          (dfs n 0);
        with (FoundCycle _) -> (
          (failwith "FoundCycle escaped from DFS!")
        ));
        (loop rest);
      )
    | _ -> ()
  end in
  (loop !edag);
end


(* --------------- checkForSetDisequalities ------------ *)
let containsElemOf (n: node) : bool =
begin
  (testInEqc (fun m -> m.head == !elemOfCode) n)
end

(* applies to both 'subsetOf' and 'strictSubsetOf' *)
let containsSubsetOf (n: node) : bool =
begin
  (testInEqc (fun m -> m.head == !subsetOfCode ||
                       m.head == !strictSubsetOfCode) n)
end

let containsSetFunctionSym (n: node) : bool =
begin
  let isSetFunctionSym (m: node) : bool =
  begin
    m.head == !unionCode ||
    m.head == !intersectionCode ||
    m.head == !setSubtractCode ||
    m.head == !singletonCode
  end in

  (testInEqc isSetFunctionSym n)
end

(* is any node in the equivalence class of 'n' the RHS of an
 * application of the 'elemOf' function? *)
let containsRHSOfElemOf (n: node) : bool =
begin
  let n = n.root in

  (* is 'node' an appliction of 'elemOf' with RHS 'n'? *)
  let isElemOf_n (parent: node) : bool =
  begin
    (parent.head == !elemOfCode) &&
    ((List.nth parent.args 1).root == n)
  end in

  (testCirclist isElemOf_n n.parents)
end


(* we do not have proper term types in Kettle (despite their presence
 * in the CVC input syntax), so infer which nodes are set-valued by
 * their relation to the set theory function and predicate symbols *)
let isSetNode (n: node) : bool =
begin
  if (n.root == !emptySetNode.root) then
    true
  else if (containsSetFunctionSym n) then
    true
  else if (containsRHSOfElemOf n) then
    true
  else
    false
end


(* do we know that 'elem' is not an element of 'set'? *)
let definitelyNotElemOf (elem: node) (set: node) : bool =
begin
  let elem = elem.root in
  let set = set.root in

  let processParent (parent: node) : bool =
  begin
    if ((parent.head == !elemOfCode) &&                    (* elemOf *)
        (parent.root != !trueNode.root) &&                 (* *not* known true *)
        ((List.nth parent.args 1).root == set) &&          (* RHS==set *)
        (Edag.forbiddenEquality parent !trueNode != None)) then (  (* known false *)
      (* extract the LHS, which is the non-element of 'set' *)
      let nonElem:node = (List.hd parent.args).root in

      (* is this 'elem'? *)
      if (elem == nonElem.root) then
        true      (* found it; known to not be member *)
      else
        false     (* this 'parent' is not evidence of nonmembership of 'elem' *)
    )
    else
      false       (* this 'parent' says nothing about nonmembership of 'set' *)
  end in
  (testCirclist processParent set.parents)
end


(* We have found that 'n1' and 'n2' are sets that have been asserted
 * to be unequal.  We will do one of three things:
 *   - Confirm that there is already a witness to their disequality,
 *     namely an element in one set but not in the other.
 *   - Discover that one is included in the other, and introduce
 *     witness in the larger set.  (Not yet implemented.)
 *   - Introduce a case split, where in one case the witness is in
 *     'n1' and in the other case the witness is in 'n2'.
 *)
let checkSetDisequality (n1: node) (n2: node) : unit =
begin
  let n1 = n1.root in
  let n2 = n2.root in

  (* Did we already consider this disequality?  This check is needed
   * because for those disequalities that we handle by introducing
   * a case split, it is not easy to tell that we did so if the split
   * has not yet actually been performed. *)
  let alreadyHandledThisOne () : bool =
  begin
    let processRecord (r: handledDis_t) : bool =
    begin
      (r.n1.root == n1 && r.n2.root == n2) ||
      (r.n1.root == n2 && r.n2.root == n1)
    end in

    if (List.exists processRecord !handledDisequalities) then
      true        (* already got it *)
    else (
      (* after going through this procedure, we *will* have handed
       * this one, so make a record of it *)
      let r:handledDis_t = {
        n1 = n1;
        n2 = n2
      } in
      (Edag.pushWithUndo handledDisequalities r);
      if debugDisequality then
        (ignore (E.log "SET: dis: considering: %a /= %a\n"
                       d_nodeTop n1
                       d_nodeTop n2));
      false       (* proceed with further processing *)
    )
  end in

  (* is there an element in 'large' that is known to not be in 'small'? *)
  let hasDisequalityWitness (small: node) (large: node) : bool =
  begin
    (* is 'lparent' an application of 'elemOf' asserting that its
     * LHS is in 'large' but not in 'small'? *)
    let processParentOfLarge (lparent: node) : bool =
    begin
      if ((lparent.root == !trueNode.root) &&                (* known true *)
          (lparent.head == !elemOfCode) &&                   (* elemOf *)
          ((List.nth lparent.args 1).root == large)) then (  (* RHS==large *)
        (* extract the LHS, which is the element of 'large' *)
        let elem:node = (List.hd lparent.args).root in

        (* is it known to *not* be an element of 'small'? *)
        if (definitelyNotElemOf elem small) then (
          if debugDisequality then
            (ignore (E.log "SET: dis: found witness of %a > %a: %a\n"
                           d_nodeTop large
                           d_nodeTop small
                           d_nodeTop elem));
          true     (* found our witness *)
        )
        else
          false
      )
      else
        false
    end in
    (testCirclist processParentOfLarge large.parents)
  end in

  (* introduce a case split to witness the disequality *)
  let introduceCaseSplit () : unit =
  begin
    (* make up name for the witness *)
    let witnessExp:exp = (makeFreshExp "set_dis_wit") in

    if debugDisequality then
      (ignore (E.log "SET: dis: splitting cases for %a /= %a, witness=%a\n"
                     d_nodeTop n1
                     d_nodeTop n2
                     d_exp witnessExp));

    (* assert the following fact:
         (elemOf(witness,n1) AND NOT(elemOf(witness,n2))) OR
         (elemOf(witness,n2) AND NOT(elemOf(witness,n1)))
     *)
    let a:exp = n1.exp in
    let b:exp = n2.exp in
    let fact: Logic.pred =
      (L.Or [(L.And [(predElemOf witnessExp a);
                     (predNotElemOf witnessExp b)]);
             (L.And [(predElemOf witnessExp b);
                     (predNotElemOf witnessExp a)])]) in
    let neqProof = 
      match forbiddenEquality n1 n2 with
        Some p -> p ()
      | None -> assert false
    in
    let pf: P.proof =
      (proofWithFreshVars "set_extensionality_split" [witnessExp]
                          neqProof) in
    (broadcast fact pf);
  end in

  if ((alreadyHandledThisOne ()) ||
      (hasDisequalityWitness n1 n2) ||
      (hasDisequalityWitness n2 n1)) then
    ()     (* nothing needs to be done *)
  (* TODO: optimization: consider possibility of known inclusion *)
  else
    (introduceCaseSplit ());
end


let checkForSetDisequalities () : unit =
begin
  (* examine every node *)
  let processNode (n1: node) : unit =
  begin
    (* examine the elements in its forbid list *)
    let processForbid (f: (node * (node * node * P.proof))) : unit =
    begin
      (* extract the forbidden node *)
      match f with
      | (n2,_) -> (
          (* is this a disequality among sets? *)
          if ((isSetNode n1) || (isSetNode n2)) then
            (checkSetDisequality n1 n2);
          )
    end in
    (iterCirclist processForbid n1.forbid)
  end in
  (List.iter processNode !edag)
end


(* ------------------- assertSubsetOf ------------------ *)
(* We only want to process each 'subsetOf' literal once. *)
let alreadyHandledSubsetOf (predNode: node) : bool =
begin
  if (Inthash.mem handledSubsetNodes predNode.id) then
    true        (* already got it *)
  else (
    (* after going through this procedure, we *will* have handed
     * this one, so make a record of it *)
    (Inthash.add handledSubsetNodes predNode.id predNode.id);
    (Edag.registerUndoRun (fun _ ->
      (Inthash.remove handledSubsetNodes predNode.id)));
    if debugSubset then
      (ignore (E.log "SET: sub: translating: %a\n"
                     d_nodeTop predNode));
    false       (* proceed with further processing *)
  )
end

(* 'predNode' is being asserted equal to 'true', and is known
 * to contain at least one 'subsetOf' application *)
let assertSubsetOf (predNode: node) : unit =
begin
  let processNode (n: node) : unit =
  begin
    if (n.head == !subsetOfCode &&
        (not (alreadyHandledSubsetOf n))) then (
      let a:exp = (List.nth n.args 0).exp in
      let b:exp = (List.nth n.args 1).exp in

      (* translate
           a is-subset-of b
         into
           exists c  such that  a union c  is  b
       *)
      let c:exp = (makeFreshExp "subset_wit") in
      let fact: Logic.pred = (L.eq (applyUnion a c) b) in
      let pf: P.proof = (proofWithFreshVars "subset_elim" [c] (proofOfPred n)) in
      (broadcast fact pf);
    )

    else if (n.head == !strictSubsetOfCode &&
             (not (alreadyHandledSubsetOf n))) then (
      let a:exp = (List.nth n.args 0).exp in
      let b:exp = (List.nth n.args 1).exp in

      (* translate
           a is-strict-subset-of b
         into
           exists c,x  such that  a union c  is  b  AND
                                  x is not in a  AND
                                  x is in b
       *)
      let c:exp = (makeFreshExp "ssubset_wit") in
      let x:exp = (makeFreshExp "ssub_dis_wit") in
      let fact: Logic.pred =
        (L.And [(L.eq (applyUnion a c) b);
                (predNotElemOf x a);
                (predElemOf x b)]) in
      let pf: P.proof = (proofWithFreshVars "ssubset_elim" [c;x] (proofOfPred n)) in
      (broadcast fact pf);
    )
  end in

  (Edag.iterEqc processNode predNode)
end

(* 'predNode' is being asserted NOT equal to 'true', and is known
 * to contain at least one 'subsetOf' application *)
let denySubsetOf (predNode: node) : unit =
begin
  let processNode (n: node) : unit =
  begin
    if (n.head == !subsetOfCode &&
        (not (alreadyHandledSubsetOf n))) then (
      let a:exp = (List.nth n.args 0).exp in
      let b:exp = (List.nth n.args 1).exp in

      (* translate
           a is-NOT-subset-of b
         into
           exists x  such that  x is in a  AND
                                x is not in b
       *)
      let x:exp = (makeFreshExp "nsubset_wit") in
      let fact: Logic.pred =
        (L.And [(predElemOf x a);
                (predNotElemOf x b)]) in
      let pf: P.proof = (proofWithFreshVars "nsubset_elim" [x] (proofOfNegPred n)) in
      (broadcast fact pf);
    )

    else if (n.head == !strictSubsetOfCode &&
             (not (alreadyHandledSubsetOf n))) then (
      let a:exp = (List.nth n.args 0).exp in
      let b:exp = (List.nth n.args 1).exp in

      (* translate
           a is-NOT-strict-subset-of b
         into
           a = b  OR
           exists x  such that  x is in a  AND
                                x is not in b
       *)
      let x:exp = (makeFreshExp "ssub_dis_wit") in
      let fact: Logic.pred =
        (L.Or [(L.eq a b);
               (L.And [(predElemOf x a);
                       (predNotElemOf x b)])]) in
      let pf: P.proof = (proofWithFreshVars "nssubset_elim" [x] (proofOfNegPred n)) in
      (broadcast fact pf);
    )
    else
      ()
  end in

  (Edag.iterEqc processNode predNode)
end


(* ----------------------- assume ---------------------- *)
let isEquality (op: Logic.binop) : bool = (op == Logic.SetEq)

let isDisequality (op: Logic.binop) : bool = op == Logic.SetNe


(* take note of a fact *)
let assume (msg: message) : unit =
begin
  (* equivalence class of "true" *)
  let trueEqc:node = (!trueNode).root in

  (* react to learning facts interpreted by this satproc *)
  (try
    (match msg.body with
    | Intern(op,n1,n2) -> (
        (* previously, I had been doing this only when I saw a new
         * 'elemOf' being asserted, but I also need to do it when
         * I see set-typed terms being merged, but detecting that
         * is a little difficult, so just do it always... *)
        (checkElemOfRelation trueEqc);

        (* normalize: put 'true' on the left *)
        let (trueNode,predNode) =
          if (n1.root == trueEqc) then (n1,n2) else
          if (n2.root == trueEqc) then (n2,n1) else

          (* we only respond to equalities with 'true', because
           * these are effectively predicate symbols *)
          raise Exit
        in

        (* assertion of 'subsetOf'? *)
        if ((isEquality op) &&
                 (containsSubsetOf predNode)) then (
          (assertSubsetOf predNode)
        );

        (* denial of 'subsetOf'? *)
        if ((isDisequality op) &&
                 (containsSubsetOf predNode)) then (
          (denySubsetOf predNode)
        );
      )
    | _ -> ())
  with Exit ->
    ());

  (checkForSetDisequalities ());
end


(* -------------------- hardcoded axioms --------------- *)
let union_axioms = "
  %    if s is not in union(t1,t2),
  %    then is s is not in t1 and s is not in t2
  ASSERT (FORALL x : set. (FORALL y : set. (FORALL z : set.
    TRIGGER( NOT(elemOf(x, union(y, z))),
      NOT(elemOf(x, union(y, z))) =>
        (NOT(elemOf(x,y)) AND NOT(elemOf(x,z)))
    ))));

  %    if s is in t1,
  %    then is s is in union(t1,t2)
  ASSERT (FORALL x : set. (FORALL y : set. (FORALL z : set.
    TRIGGER( have(union(y,z)) AND elemOf(x, y),
      elemOf(x, y) =>
        elemOf(x, union(y,z))
    ))));

  ASSERT (FORALL x : set. (FORALL y : set. (FORALL z : set.
    TRIGGER( have(union(z,y)) AND elemOf(x, y),
      elemOf(x, y) =>
        elemOf(x, union(z,y))
    ))));

  %    if s is in union(t1,t2),
  %    and s is not in t1,
  %    then is s is t2
  ASSERT (FORALL x : set. (FORALL y : set. (FORALL z : set.
    TRIGGER( elemOf(x, union(y,z)) AND NOT(elemOf(x, y)),
      elemOf(x, union(y,z)) AND NOT(elemOf(x,y)) =>
        elemOf(x, z)
    ))));

  ASSERT (FORALL x : set. (FORALL y : set. (FORALL z : set.
    TRIGGER( elemOf(x, union(y,z)) AND NOT(elemOf(x, y)),
      elemOf(x, union(y,z)) AND NOT(elemOf(x,z)) =>
        elemOf(x, y)
    ))));

  %    if s is in neither t1 nor t2,
  %    then is s is not in union(t1,t2)
  ASSERT (FORALL x : set. (FORALL y : set. (FORALL z : set.
    TRIGGER( NOT(elemOf(x, y)) AND NOT(elemOf(x, z)) AND have(union(y,z)),
      NOT(elemOf(x, y)) AND NOT(elemOf(x, z)) =>
        NOT(elemOf(x, union(y,z)))
    ))));
" (* end of union_axioms *)

let intersection_axioms = "
  %    if s is in intersection(t1, t2),
  %    then is s is in both
  ASSERT (FORALL x : set. (FORALL y : set. (FORALL z : set.
    TRIGGER( elemOf(x, intersection(y,z)),
      elemOf(x, intersection(y,z)) =>
        (elemOf(x,y) AND elemOf(x,z))
    ))));

  %    if s is not in t1,
  %    then is s is not in intersection(t1,t2)
  ASSERT (FORALL x : any. (FORALL y : set. (FORALL z : set.
    TRIGGER( have(intersection(y,z)) AND NOT(elemOf(x,y)),
      NOT(elemOf(x,y)) =>
        NOT(elemOf(x, intersection(y,z)))
    ))));

  ASSERT (FORALL x : any. (FORALL y : set. (FORALL z : set.
    TRIGGER( have(intersection(y,z)) AND NOT(elemOf(x,z)),
      NOT(elemOf(x,z)) =>
        NOT(elemOf(x, intersection(y,z)))
    ))));

  %    if s is not in intersection(t1, t2),
  %    and s is in t1,
  %    then is s is not in t2
  ASSERT (FORALL x : any. (FORALL y : set. (FORALL z : set.
    TRIGGER( NOT(elemOf(x, intersection(y,z))) AND elemOf(x,y),
      NOT(elemOf(x, intersection(y,z))) AND elemOf(x,y) =>
        NOT(elemOf(x, z))
    ))));

  ASSERT (FORALL x : any. (FORALL y : set. (FORALL z : set.
    TRIGGER( NOT(elemOf(x, intersection(y,z))) AND elemOf(x,z),
      NOT(elemOf(x, intersection(y,z))) AND elemOf(x,z) =>
        NOT(elemOf(x, y))
    ))));

  %    if s is in both t1 and t2,
  %    then is s is in intersection(t1,t2)
  ASSERT (FORALL x : any. (FORALL y : set. (FORALL z : set.
    TRIGGER( elemOf(x,y) AND elemOf(x,z) AND have(intersection(y,z)),
      elemOf(x,y) AND elemOf(x,z) =>
        elemOf(x, intersection(y,z))
    ))));
" (* end of intersection_axioms *)

let diff_axioms = "
  %    if s is in t1-t2,
  %    then is s is in t1 and not in t2
  ASSERT (FORALL x : any. (FORALL y : set. (FORALL z : set.
    TRIGGER( elemOf(x, setSubtract(y,z)),
      elemOf(x, setSubtract(y,z)) =>
        elemOf(x, y) AND NOT(elemOf(x,z))
    ))));

  %    if s is not in t1,
  %    then s is not in t1-t2
  ASSERT (FORALL x : any. (FORALL y : set. (FORALL z : set.
    TRIGGER( NOT(elemOf(x, y)) AND have(setSubtract(y, z)),
      NOT(elemOf(x, y)) =>
        NOT(elemOf(x, setSubtract(y, z)))
    ))));

  %    if s is in t2,
  %    then s is not in t1-t2
  ASSERT (FORALL x : any. (FORALL y : set. (FORALL z : set.
    TRIGGER( elemOf(x, z) AND have(setSubtract(y, z)),
      elemOf(x, z) =>
        NOT(elemOf(x, setSubtract(y, z)))
    ))));

  %    if s is not in t1-t2,
  %    but s is in t1,
  %    then s is in t2
  ASSERT (FORALL x : any. (FORALL y : set. (FORALL z : set.
    TRIGGER( NOT(elemOf(x, setSubtract(y,z))) AND elemOf(x, y),
      NOT(elemOf(x, setSubtract(y,z))) AND elemOf(x, y) =>
        elemOf(x, z)
    ))));

  %    if s is not in t1-t2,
  %    but s is not in t2,
  %    then s is not in t1
  ASSERT (FORALL x : any. (FORALL y : set. (FORALL z : set.
    TRIGGER( NOT(elemOf(x, setSubtract(y,z))) AND NOT(elemOf(x, z)),
      NOT(elemOf(x, setSubtract(y,z))) AND NOT(elemOf(x, z)) =>
        NOT(elemOf(x, y))
    ))));

  %    if s is t1,
  %    but s is not in t2,
  %    then s is in t1-t2
  ASSERT (FORALL x : any. (FORALL y : set. (FORALL z : set.
    TRIGGER( elemOf(x,y) AND NOT(elemOf(x,z)) AND have(setSubtract(y,z)),
      elemOf(x,y) AND NOT(elemOf(x,z)) =>
        elemOf(x, setSubtract(y,z))
    ))));
" (* end of diff_axioms *)

let singleton_axioms = "
  %    t1 is an element of {t1}
  ASSERT (FORALL x : set.
    TRIGGER( have(singleton(x)),
      elemOf(x, singleton(x))
    ));

  %    s is an element of {t1}
  %      is equivalent to
  %    s = t1
  ASSERT (FORALL x : any. (FORALL y : any.
    TRIGGER( elemOf(x, singleton(y)),
      elemOf(x, singleton(y)) =>
        x = y
    )));

  ASSERT (FORALL x : any. (FORALL y : any.
    TRIGGER( NOT(elemOf(x, singleton(y))),
      NOT(elemOf(x, singleton(y))) =>
        x /= y
    )));
" (* end of singleton_axioms *)


(* These are some of the "fulfilling rules".  Unlike the axioms
 * above, these ones *can* cause new terms to be created. *)
let case_split_axioms = "
  % if s is in union(t1,t2),
  % case split on whether s is in t1
  %
  % I would like to say 'AND NOT(have(...))' in the trigger, but
  % that does not appear to work (the rule does not get instantiated),
  % and if elemOf does exist then if this split is ever chosen, one
  % of the branches will be immediately refuted.
  ASSERT (FORALL x : any. (FORALL y : set. (FORALL z : set.
    TRIGGER( elemOf(x, union(y,z)), %AND NOT(have(elemOf(x,y))),
      elemOf(x,y) OR NOT(elemOf(x,y))
    ))));

  % if s is in t1,
  % and intersection(t1,t2) is around,
  % case split on whether s is in t2
  %
  % NOTE: I have so far been unable to come up with a formula whose
  % proof requires this rule, but I think that is just because I
  % do not understand the proof in the Cantone/Zarba paper well enough.
  ASSERT (FORALL x : any. (FORALL y : set. (FORALL z : set.
    TRIGGER( elemOf(x, y) AND have(intersection(y,z)),
      elemOf(x,z) OR NOT(elemOf(x,z))
    ))));

  % if s is in t1,
  % and t1-t2 is around,
  % case split on whether s is in t2
  %
  % NOTE: I have not even attempted to make a formula that requires
  % this split.
  ASSERT (FORALL x : any. (FORALL y : set. (FORALL z : set.
    TRIGGER( elemOf(x, y) AND have(setSubtract(y,z)),
      elemOf(x,z) OR NOT(elemOf(x,z))
    ))));
" (* end of case_split_axioms *)


let all_axioms =
  union_axioms ^
  intersection_axioms ^
  diff_axioms ^
  singleton_axioms ^
  case_split_axioms


(* ----------------- Satproc module entries ------------ *)
(* list of entry points exported as a satproc *)
let entries: S.entries = {
  S.emptySatproc with
  S.init               = init;
  S.defaultEnabled     = false;    (* must say "-enable SET" on command line *)
  S.globalAssumeString = (fun _ -> Some ("cvc", all_axioms));
  S.assume             = assume;
  S.debug              = debug;
  S.ident              = mlssid;
  S.descr              = "implementation of MLSS set theory";
}


(* EOF *)
