(****************************** Oracle proof generators ***)

module L = Logic
module E = Errormsg
module Eng = Engine
open Pretty

module Oracle : Oracle.ORACLE = Atioracle

type opgen = 
    { mutable oracle: Oracle.t;
      mutable outbits: Oracle.Bits.outbits;
    } 

let ostate =                            (* The global state of the oracle *)
  { oracle = Oracle.make ();
    outbits = Oracle.Bits.empty ()
  } 

let useOracleAnn = ref false

let proofChannel = ref stderr

let start (proofFile: string) = 
  proofChannel := 
    if proofFile = "" then stderr else 
    (try open_out proofFile with _ ->
      raise (Arg.Bad "Cannot open proof file"));


  if !useOracleAnn then begin
    Oracle.Bits.start "";
    ostate.outbits <- Oracle.Bits.empty ();
    ignore (fprintf !proofChannel
                "\n%%VERBATIM_\nANN_ARCHSTARTLF\nANN_ARCHGLOBAL(__oracle)");
  end

let finish () = 
  if !useOracleAnn then begin
    Oracle.Bits.dump !proofChannel ostate.outbits;
    ignore (fprintf !proofChannel 
              "\nANN_ARCHGLOBAL(__oracle_end)\n%%_VERBATIM\n")
  end;
  if !proofChannel != stderr then close_out !proofChannel


let startVC name = 
  let name' = 
    if String.length name >= 4 && String.sub name 0 3 = "Sat" then
      String.sub name 3 (String.length name - 3)
    else
      name
  in
  if not !useOracleAnn then begin
    Oracle.Bits.start name;
    ostate.outbits <- Oracle.Bits.empty ();
    ignore (fprintf !proofChannel
              "\nANN_ORACLE(%s,@!   Lprf%s)\n%%VERBATIM_\nANN_ARCHSTARTLF\nLprf%s:"
              name' name' name');
  end else 
    Oracle.Bits.putString ostate.outbits name'

let assume (mkname: unit -> string) (p: L.pred) = 
  match p with
    (L.PNez _ | L.Bool _ | L.PFun _) -> 
      let name = mkname () in
        Oracle.assume ostate.oracle name p
  | _ -> E.s (E.error "Oracles are not supported for non-atomic assumptions")

let unassume () = 
  Oracle.undo ostate.oracle

let pick v = 
  E.s (E.error "Oracles are not yet supported for universal quantification")

let unpick () = 
  E.s (E.error "Oracles are not yet supported for universal quantification")

let finishVC () = 
  if not !useOracleAnn then  begin
    Oracle.Bits.dump !proofChannel ostate.outbits;
    ignore (fprintf !proofChannel
              "\n%%_VERBATIM\n")
  end

let conjunct (p: L.pred) proof = 
  match p with 
    (L.PNez _ | L.Bool _ | L.PFun _) -> 
      Oracle.convert ostate.oracle ostate.outbits proof p
  | _ -> E.s (E.error "Oracles are not supported for non-atomic conjuncts")

let choose ~thechoice ~nrchoices = 
  if nrchoices > 1 then 
    Oracle.Bits.put ostate.outbits nrchoices thechoice


let oracleProofGen : Eng.proofgen = 
  { Eng.start = start; 
    Eng.startVC = startVC;
    Eng.assume = assume;
    Eng.unassume = unassume;
    Eng.pick = pick;
    Eng.unpick = unpick;
    Eng.finishVC = finishVC;
    Eng.finish = finish;
    Eng.conjunct = conjunct;
    Eng.choose = choose;
  } 

