// -*- Mode: caml -*-
%{ 
(* Kettle
 *)

(* Documentation *)


module E = Errormsg
open Pretty
open Logic
open Proof
open Prolog
module H = Hashtbl
module U = Util

let parse_error msg : 'a =           (* sm: c++-mode highlight hack: -> ' <- *)
  E.hadErrors := true;
  E.parse_error
    msg

let print = print_string

let currentLoc = ref E.locUnknown

(* An AST for rules *)
type rule_ast = 
    { name: string;
      mutable decls: (string * typ) list option; (* var : typ *)
      mutable vars: string list; (* The unification variables. These are in 
                                  * the order in which variables are 
                                  * encountered during parsing of variable 
                                  * declarations and of the rule body. *)
      mutable exvars: string list; (* Vars bound by existentials *)
      mutable goal: pred;
      mutable subgoal: subgoal_ast;
      loc: E.location;
    } 
and subgoal_ast = 
  | Comma of subgoal_ast list
  | Cut of E.location
  | Commit of E.location
  | Goal of goal_ast * E.location
  | True of E.location
  | False of E.location
  | If of string * exp list * pred option * 
          subgoal_ast * subgoal_ast * E.location
  | Case of string * exp list * 
            (string * pred option * subgoal_ast * E.location) list 
            * E.location
  | Choose of subgoal_ast list (** A disjunction *)

and goal_ast = 
  | Pred of pred
  | Terminal of pred
  | Extern of string * exp list * pred option
  | Is of exp * exp 


(** We declare here the types of the constructors *)
and declTyp = 
    TType               (** Type of types *)
  | TProp of typ list   (** Type of predicates, with some arguments *)
  | TCons of typ * typ list (** Type of constructors of a given type, given 
                             * some arguments *)

and typ = 
  | TVal                (** Type of machine values *)
  | TId of string       (** A type identifier *) 
  | TUnif of typ ref    (** This is used for infering types for constructors *)

(* For each name, keep a declared type along with a location for the 
 * declaration. If the location has "line" = -1, then the constant was not 
 * declared, but inferred. The declarations are indexed by the lowercase 
 * version of the name. *)
and declarations = (string, declTyp * E.location) Hashtbl.t

type rule_decl = 
    Rule of rule_ast
  | Constant of (string * E.location) list * declTyp
  | Tabled of (string * E.location) list

let d_typ () (t: typ) = 
  (* Watch for cycles. These should not occur, but they did during debugging *)
  let rec d_typ_rec (memo: typ ref list) (t: typ) = 
    match t with 
      TVal -> text "val"
    | TId s -> text s
    | TUnif r -> 
        if List.memq r memo then 
          E.s (E.bug "There is a cycle in the unification variables")
        else
          d_typ_rec (r :: memo) !r
  in
  d_typ_rec [] t


let rec d_declTyp () = function
    TType -> text "Type"
  | TProp args -> 
      dprintf "%aProp" (docList ~sep:nil (fun t -> d_typ () t ++ text "->")) args
  | TCons (res, args) -> 
      dprintf "%a%a" (docList ~sep:nil (fun t -> d_typ () t ++ text "->")) args 
        d_typ res


let d_declarations () (h: declarations) : doc = 
  let res = ref nil in
  H.iter (fun k (d, l) -> 
    res := !res ++ (dprintf "%s : %a\n" k d_declTyp d))
    h;
  !res

(** Unroll the unifiers in a type *)
let rec tUnroll (t: typ) = 
  match t with 
    TUnif r -> begin
      match !r with 
        TId "_" -> t
      | t' -> tUnroll t'
    end
  | _ -> t

let newUnifType () = TUnif (ref (TId "_"))

let polymorphicConstructors : (string, unit) H.t = 
  let h = H.create 13 in
  List.iter (fun s -> H.add h s ()) 
    [ "consMap"; "nilMap"; "selmap"; "updmap" ];
  h


  (** Type checking functions *)
(* "now" versions of exvars parameters give types of currently bound variables *)
(* As binders are encountered, the pre-generated types are shifted in from exvars *)
let rec typeCheckExp (decls: declarations) 
                     (uvars: typ list) (exNow: typ list)
                     (e: exp) (t: typ) : unit = 
  E.pushContext (fun _ -> dprintf "Expression %a" d_exp e);

  (match e with 
    Int _ | Float _ | NVar _ -> typeUnif t TVal
  | UVar idx -> begin
      try
        typeUnif t (List.nth uvars idx)
      with _ -> 
        E.s (E.error "Invalid UVar index")
  end 
  | Var idx -> begin
      try
        typeUnif t (List.nth exNow idx)
      with _ -> 
        E.s (E.error "Invalid Var index")
  end 
  | BinOp((SetEq|SetNe), e1, e2) -> (* Any types can be compared for equality*)
      let t' = newUnifType () in
      typeCheckExp decls uvars exNow e1 t';
      typeCheckExp decls uvars exNow e2 t';
      typeUnif t TVal
      
  | BinOp(_, e1, e2) -> 
      typeCheckExp decls uvars exNow e1 TVal;
      typeCheckExp decls uvars exNow e2 TVal;
      typeUnif t TVal
  | UnOp (_, e1) -> typeCheckExp decls uvars exNow e1 TVal; typeUnif t TVal
  | Fun (c, args) -> typeCheckFun decls uvars exNow c args (Some t)
  | Ite(b, e1, e2) -> typeCheckExp decls uvars exNow b TVal; 
                      let t' = newUnifType () in 
                      typeCheckExp decls uvars exNow e1 t';
                      typeCheckExp decls uvars exNow e2 t'
  );


  E.popContext ()


(** Type check a constructor application. The expected type is None for 
 * predicates *)
and typeCheckFun (decls: declarations) 
                 (uvars: typ list) (exNow: typ list)
                 (c: string) (args: exp list) (t: typ option) : unit = 
  match c with
    "&" -> ()
  | _ ->
      let argst = (* Get the argument types *)
	let c' = String.uncapitalize c in
	try
	  match t, H.find decls c' with 
            Some t', (TCons (rest, argst), _) -> 
              E.pushContext (fun _ -> dprintf "Result type of %s" c);
              typeUnif t' rest; 
              E.popContext ();
              argst

	  | None, (TCons _, _) -> 
              E.s (E.error "%s is not a predicate constructor" c)
	  | None, (TProp argst, _) -> argst
	  | Some _, (TProp _, _) -> 
              E.s (E.error "%s is not an expression constructor" c)
	  | _, _ -> E.s (E.bug "typeCheckFun")
	with Not_found -> begin
	  ignore (E.log "Using undeclared constructor %s at %a\n" c 
                    E.d_loc !currentLoc);
	  (* Add a constructor of the right arity *)
	  let argst = List.map (fun _ -> newUnifType ()) args in
	  let rest = newUnifType () in
	  let dtyp = 
            match t with 
              Some t' -> TCons (t', argst)
            | None -> TProp argst
	  in
	  (* Do not record declarations of polymorphic constructors *)
	  if not (H.mem polymorphicConstructors c') then 
            H.add decls c' (dtyp, E.locUnknown);
	  argst
	end
      in
      (* Now check the arguments *)
      if List.length args <> List.length argst then 
	E.s (E.error "Constructor %s used with %d arguments (expected %d arguments)"
               c (List.length args) (List.length argst));
      (* Now check the arguments *)
      let rec checkArgs (ith: int) (args: exp list) (argts: typ list) = 
	match args, argts with 
	  arg :: args, argt :: argts -> 
            E.pushContext (fun _ -> dprintf "%dth argument of %s" ith c);
            typeCheckExp decls uvars exNow arg argt;
            E.popContext ();
            checkArgs (ith + 1) args argts
	| [], [] -> ()
	| _ -> E.s (E.bug "typeCheckFun")
      in
      checkArgs 1 args argst

and typeCheckPred (decls: declarations) (uvars: typ list)
    (exvars: typ list) (exNow: typ list) (p: pred) : unit = 
  let tp = typeCheckPred decls uvars exvars exNow in
  match p with 
    Bool _ -> () 
  | And pl | Or pl -> List.iter tp pl
  | Imply (p1, p2) -> tp p1; tp p2
  | PNez e -> typeCheckExp decls uvars exNow e TVal
  | PFun (s, el) | NegPFun (s, el) -> typeCheckFun decls uvars exNow s el None
  | Exist (v, p) | All (v, p) ->
      let t, rest = List.hd exvars, List.tl exvars in
      typeCheckPred decls uvars rest (t :: exNow) p
  | Pannot (_, p) -> tp p
  | Not p -> tp p

and typeUnif (t1: typ) (t2: typ) : unit = 
  let t1' = tUnroll t1 in
  let t2' = tUnroll t2 in
  if t1' == t2' then 
    () 
  else begin
    let ok = 
      match t1', t2' with 
        TId s1, TId s2 -> s1 = s2
      | TVal, TVal -> true
      | TUnif r1, _ -> r1 := t2'; true
      | _, TUnif r2 -> r2 := t1'; true
      | _, _ -> false
    in
    if not ok then 
      E.s (E.error "Expecting type %a and found type %a\n" 
             d_typ t1 d_typ t2)
  end


let typeCheckType (decls: declarations) (t: typ) : unit = 
  match t with
    TVal -> ()
  | TId n -> begin
      try
        match H.find decls n with 
          TType, _ -> ()
        | _ -> raise Not_found
      with Not_found -> 
        E.s (E.error "Wrong type: %s (not a type constructor)." n)
  end
  | TUnif r -> E.s (E.bug "typeCheckType on a type with unifiers")

let typeCheckRule (decls: declarations) (r: rule_ast) : typ list * typ list = 
  currentLoc := r.loc;
(*  ignore (E.log "Typechecking rule %s\n" r.name); *)
  E.pushContext (fun _ -> dprintf "Rule %s at %a\n" r.name d_loc r.loc);
  (* First, type check the declarations, if any *)
  E.pushContext (fun _ -> dprintf "Variable declarations");
  (match r.decls with 
    Some dec -> 
      List.iter (fun (v, t) -> typeCheckType decls t) dec
  | _ -> ());
  (* Create the type environment for the unification variables *)
  let uvars : typ list = 
    List.map
      (fun vn -> 
        match r.decls with
          None -> (* No declarations *) newUnifType ()
        | Some decls -> begin (* Find its declaration *)
            try 
              let _, t = List.find (fun (v, _) -> v = vn) decls in
              t
            with Not_found -> 
              E.s (E.error "Using undeclared variable %s" vn)
        end)
      r.vars
  in
  let exvars : typ list = List.map (fun _ -> newUnifType ()) r.exvars in
  E.popContext ();
  E.popContext (); (* We replace the context now, with one that prints the 
                    * variable declarations *)
  E.pushContext (fun _ -> 
    let uvar_decls = ref nil in
    List.iter2 (fun v t -> uvar_decls := !uvar_decls ++ 
         (dprintf "       %s : %a\n" v d_typ t))
      r.vars uvars;
    dprintf "Rule %s at %a\nVariables:\n  @[%a@]\nConstants:\n  @[%a@]\n" 
      r.name d_loc r.loc insert !uvar_decls d_declarations decls);
    
  let tpl p l = 
    E.withContext (fun _ -> dprintf "Predicate %a (%a)" d_pred p d_loc l)
      (fun p -> typeCheckPred decls uvars exvars [] p) p in
  let tp p = 
    E.withContext (fun _ -> dprintf "Predicate %a" d_pred p)
      (fun p -> typeCheckPred decls uvars exvars [] p) p in
  let te = typeCheckExp decls uvars in
  let tel e t l = 
    E.withContext (fun _ -> dprintf "Expression %a (%a)" d_exp e d_loc l)
      (fun e -> typeCheckExp decls uvars [] e t) e in
  tpl r.goal r.loc;
  let rec checkSubgoal = function
      Cut _ | Commit _ | True _ | False _ -> ()
    | Comma sgl | Choose sgl -> List.iter checkSubgoal sgl
    | If (_, el, po, tl, fl, l) -> 
        List.iter (fun e -> tel e (newUnifType ()) l) el;
        (match po with None -> () | Some p -> tpl p l);
        checkSubgoal tl;
        checkSubgoal fl
    | Case (_, el, cases, l) -> 
        List.iter (fun e -> tel e (newUnifType ()) l) el;
        List.iter (fun (_, po, sgl, l) -> 
          (match po with None -> () | Some p -> tpl p l);
          checkSubgoal sgl)
          cases
    | Goal ((Pred p | Terminal p), l) -> tpl p l
    | Goal (Extern (_, el, po), l) -> 
        List.iter (fun e -> tel e (newUnifType ()) l) el;
        (match po with None -> () | Some p -> tpl p l)
    | Goal (Is (e1, e2), l) -> (* Ought to have the same type *)
        E.pushContext 
          (fun _ -> dprintf "Is (%a, %a) at %a" d_exp e1 d_exp e2 d_loc l);
        let t = newUnifType () in
        tel e1 t l;
        tel e2 t l;
        E.popContext ()
  in
  checkSubgoal r.subgoal;
  E.popContext ();
  uvars, exvars

  
let currentRule: rule_ast ref = 
  ref { name = "dummy"; vars = []; goal = Bool true;
	exvars = [];
        decls = None;
        subgoal = True (!currentLoc);
        loc = E.locUnknown }
    
let toVal (s: string) = 
  if s = "val" then TVal else TId s

let getVar (vname: string) = 
  let rec loop (idx: int) = function
      [] -> (* Must add it to the list *)
        !currentRule.vars <- !currentRule.vars @ [vname];
        UVar idx
    | v :: rest when v = vname -> UVar idx
    | _ :: rest -> loop (idx + 1) rest
  in
  let res = loop 0 !currentRule.vars in
(*
  ignore (E.log "Rule: %s var:%s idx=%d\n" !currentRule.name vname
            (match res with UVar idx -> idx));
*)
  res

let getExvar (vname: string) =
  let rec loop (idx: int) = function
      [] -> loop 0 !currentRule.vars
    | v :: rest when v = vname -> Var idx
    | _ :: rest -> loop (idx + 1) rest
  in
  let res = loop 0 !currentRule.exvars in
(*
  ignore (E.log "Rule: %s var:%s idx=%d\n" !currentRule.name vname
            (match res with UVar idx -> idx));
*)
  res



type pfgen = program -> subst -> proof list -> proof

(* Prepare an external procedure. *)
let doExtern (l: E.location) 
             (n: string) 
             (args: exp list)
    : (program -> subst -> (string * proof option) option) = 

  if !ignoreExterns then begin
    fun _ _ -> None
  end else begin
  let ext = 
    try H.find Prolog.externFunctions n 
    with Not_found -> 
      E.s (E.error "%s:%d: Cannot find extern function %s\n" 
             l.E.file l.E.line n)
  in
  (* Check the number of arguments *)
  let expectedArgs = 
    match ext with 
      Ext1 _ -> 1
    | Ext2 _ -> 2
    | Ext3 _ -> 3
    | Ext4 _ -> 4
    | Ext5 _ -> 5
    | ExtList _ -> -1
  in
  if expectedArgs > 0 && List.length args <> expectedArgs then 
    E.s (E.error "%s:%d: Extern %s expects %d arguments and is used here with %d\n" 
           l.E.file l.E.line n expectedArgs (List.length args));
  fun (p: program) (s: subst) -> 
    (* Apply the substitution to the arguments. We leave alone those 
     * arguments that still contain unification variables *)
(*
    if n = "getDescendantTags" then 
      ignore (E.log "args = @[%a@]@!s = %a@!"
                (docList (d_exp ())) args
                d_subst s);
*)
    let s_args = List.map (unrollUVars s) args in
(*
    if n = "getDescendantTags" then 
      ignore (E.log "s_args = @[%a@]"
                (docList (d_exp ())) s_args);
*)
    match ext, s_args with 
      Ext1 f, [a1] -> f p s a1
    | Ext2 f, [a1;a2] -> f p s a1 a2
    | Ext3 f, [a1;a2;a3] -> f p s a1 a2 a3
    | Ext4 f, [a1;a2;a3;a4] -> f p s a1 a2 a3 a4
    | Ext5 f, [a1;a2;a3;a4;a5] -> f p s a1 a2 a3 a4 a5
    | ExtList f, _ -> f p s s_args
    | _ -> E.s (E.bug "doExtern: nr args")
  end
      
(* For each program, keep a bunch of declarations *)
let theDeclarations : (int, declarations) H.t = H.create 3

let rec typToString = function
    TVal -> "val"
  | TId n -> n
  | TUnif r -> typToString !r

let rec declTypToString = function
    TType -> "Type"
  | TProp args -> List.fold_right (fun t s -> typToString t ^ " -> " ^ s) args "Prop"
  | TCons (ran, dom) -> List.fold_right (fun t s -> typToString t ^ " -> " ^ s) dom (typToString ran)

let binopToString = function
    Plus | Add -> "Zplus"
  | Minus | Sub -> "Zminus"
  | Times | Mult -> "Zmult"
  | Div -> "Zdiv"
  | BAnd -> "band"
  | BOr -> "bor"
  | BXor -> "bxor"
  | SetEq -> "seteq"
  | SetNe -> "setne"
  | SetGt -> "setgt"
  | SetAb -> "setab"
  | SetGe -> "setge"
  | SetAe -> "setae"
  | SetLt -> "setlt"
  | SetBl -> "setbl"
  | SetLe -> "setle"
  | SetBe -> "setbe"
  | _ -> "BINOP_UNIMPLEMENTED"

let rec expToString et = function
  | Int n -> "`" ^ Int32.to_string n ^ "`"
  | BinOp (bop, e1, e2) -> "(" ^ binopToString bop ^ " " ^ expToString et e1 ^ " " ^ expToString et e2 ^ ")"
  | Fun (n, []) -> n
  | Fun (n, es) -> List.fold_left (fun s e -> s ^ " " ^ expToString et e) ("(" ^ n) es ^ ")"
  | UVar n -> "_V" ^ string_of_int n
  | Var n -> let v, _ = List.nth et n in v
  | _ -> "EXP_UNIMPLEMENTED"

let rec predToString et etNow = function
    Bool true -> "True"
  | Bool false -> "False"
  | PNez e -> (match e with
    | BinOp (SetEq, e1, e2) -> "(" ^ expToString etNow e1 ^ " = " ^ expToString etNow e2 ^ ")"
    | BinOp (SetNe, e1, e2) -> "~(" ^ expToString etNow e1 ^ " = " ^ expToString etNow e2 ^ ")"
    | BinOp (SetLt, e1, e2) -> "(Zlt " ^ expToString etNow e1 ^ " " ^ expToString etNow e2 ^ ")"
    | BinOp (SetLe, e1, e2) -> "(Zle " ^ expToString etNow e1 ^ " " ^ expToString etNow e2 ^ ")"
    | BinOp (SetGt, e1, e2) -> "(Zgt " ^ expToString etNow e1 ^ " " ^ expToString etNow e2 ^ ")"
    | BinOp (SetGe, e1, e2) -> "(Zge " ^ expToString etNow e1 ^ " " ^ expToString etNow e2 ^ ")"
    | BinOp (bop, e1, e2) -> "`(" ^ binopToString bop ^ " " ^ expToString etNow e1 ^ " " ^ expToString etNow e2 ^ ") <> 0`"
    | _ -> "PNEZ_UNIMPLEMENTED")
  | PFun (n, es) -> 
      List.fold_left (fun s e -> s ^ " " ^ expToString etNow e) ("(" ^ n) es ^ ")"
  | NegPFun (n, es) -> 
      List.fold_left (fun s e -> "neg_" ^ s ^ " " ^ expToString etNow e) ("(" ^ n) es ^ ")"
  | And ps ->
      (match ps with
	[] -> "True"
      | p::ps -> List.fold_left (fun s p -> s ^ " /\\ "
	  ^ predToString et etNow p) ("(" ^ predToString et etNow p) ps) ^ ")"
  | Or ps ->
      (match ps with
	[] -> "True"
      | p::ps -> List.fold_left (fun s p -> s ^ " \\/ " ^ predToString et etNow p)
	    ("(" ^ predToString et etNow p) ps ^ ")")
  | Imply (p1, p2) -> "(" ^ predToString et etNow p1 ^ " -> " ^ predToString et etNow p2 ^ ")"
  | Exist (v, p) ->
      let t, rest = List.hd et, List.tl et in
      "(EX " ^ v ^ ":" ^ typToString t ^ " | " ^ predToString rest ((v, t) :: etNow) p ^ ")"
  | All (v, p) ->
      let t, rest = List.hd et, List.tl et in
      "(ALL " ^ v ^ ":" ^ typToString t ^ " | " ^ predToString rest ((v, t) :: etNow) p ^ ")"
  | Pannot _ | Not _ -> E.s (E.unimp "annot")

let rec subgoalToString et = function
    Goal (g, _) -> goalToString et g
  | Comma ss -> Some (List.fold_right (fun t s ->
      match subgoalToString et t with
	None -> s
      | Some ts -> ts ^ (if s <> "" then " -> " ^ s else s)) ss "")
  | True _ -> Some "True"
  | False _ -> Some "False"
  | Cut _ -> None
  | _ -> failwith "Kettle subgoalToString: subgoal type not implemented"

and goalToString et = function
    Pred p | Terminal p | Extern (_, _, Some p) -> Some (predToString et [] p)
  | Extern _ -> None
  | _ -> failwith "Kettle goalToString: goal type not implemented"

let ruleToString typs etyps r =
  let rec mkTyps n ts s =
    match ts with
      [] -> s
    | t::ts -> mkTyps (n+1) ts (s ^ "(_V" ^ string_of_int n ^ ":" ^ typToString t ^ ")") in
  match subgoalToString etyps r.subgoal with
    Some "True" | None -> mkTyps 0 typs "" ^ " " ^ predToString etyps [] r.goal
  | Some sg -> mkTyps 0 typs "" ^ " " ^ sg ^ " -> " ^ predToString etyps [] r.goal

let processRules (rl: rule_decl list) : rule list * lemma list
    * (in_channel -> out_channel -> bool) = 
  (* Find the declarations corresponding to this program *)
  let decls: (string, declTyp * E.location) H.t = 
    try H.find theDeclarations (programId !theProgram)
    with Not_found -> begin
      let decls = H.create 17 in
      H.add theDeclarations (programId !theProgram) decls;
      decls
    end
  in
  let lemmas : lemma list ref = ref [] in
  let toCheck : (string * string) list ref = ref [] in
  let processOneRule (r: rule_ast) : rule =  
    let vtyps, etyps = typeCheckRule decls r in

    setVarTypes r.name (List.map typToString vtyps);

    let intros =
      if List.length vtyps = 0 then
	""
      else
	"Intros until _V" ^ string_of_int (List.length vtyps - 1) ^ ".\n" in
    (match !coqFlag with
      CoqNone -> ()
    | _ -> toCheck := (r.name, "Goal " ^ ruleToString vtyps etyps r
		       ^ ".\n" ^ intros ^ "Apply " ^ r.name ^ ".\nAbort.\n") :: !toCheck);

    (* Compile a subgoal_ast into a subgoal *)
    let rec mkSubgoal (sg: subgoal_ast) : subgoal = 
      match sg with 
        Goal (Pred p, l) -> 
          sgPROVE p (fun acc p -> p :: acc)
      | Goal (Extern (n, args, po), l) -> 
          let f = doExtern l n args in
          sgCHECK (fun p (s: subst) (acc: proof list) -> 
            let thisId = goalId p in
            let thisLevel = level p in
            setLevel p (1 + thisLevel);  (* Prepare the level for the 
                                          * subgoals *)
            setGoalId p (1 + thisId);
            let id = 
              if !debug > 0 then 
                dprintf "%d.%d:" thisId thisLevel 
              else
                nil
            in
            if !debug > 0 then 
              ignore (E.log "KT:%a Extern %a@!" insert id 
                        d_spred (PFun(n, args), s));
            let res = 
              match po, f p s with
              | _, None -> 
                  if !debug > 0 then 
                    ignore (E.log "KT:%a Failed\n" insert id);
                  (* failure *) false, acc
              | None, Some (_, _) -> (* success, don't care about proof *)
                  if !debug > 0 then 
                    ignore (E.log "KT:%a Success %a\n" insert id 
                              d_spred (PFun(n, args), s));
                  true, acc
              | Some _, Some (_, Some prf) -> (* success, want proof *)
                  if !debug > 0 then 
                    ignore (E.log "KT:%a Success %a\n" insert id 
                              d_spred (PFun(n, args), s));
                  true, prf :: acc
              | Some _, Some (_, None) -> 
                  E.s (E.error "%a: Extern %s was supposed to return a proof"
                         E.d_loc l n)
            in
            setLevel p (level p - 1);
            res)

      | Goal (Is (e1, e2), l) -> 
          sgCHECK (fun p s acc -> 
            let thisId = goalId p in
            let thisLevel = level p in
            setLevel p (1 + thisLevel);  (* Prepare the level for the 
                                          * subgoals *)
            setGoalId p (1 + thisId);
            let id = 
              if !debug > 0 then 
                dprintf "%d.%d:" thisId thisLevel 
              else
                nil
            in
            if !debug > 0 then 
              ignore (E.log "KT:%a Is %a@!" insert id 
                        d_spred (PNez(BinOp(SetEq, e1, e2)), s));
            if Prolog.eunify p (e1, s) (e2, s) then begin
            if !debug > 0 then 
              ignore (E.log "KT:%a Success@!" insert id);
              setLevel p (level p - 1);
              true, (Proof.Refl (unrollUVars s e1)) :: acc
            end else begin
            if !debug > 0 then 
              ignore (E.log "KT:%a Failure@!" insert id);
              setLevel p (level p - 1);
              false, acc
            end)
      | Goal (Terminal  p, l) -> 
          sgASSUMPTION p (fun acc p -> p :: acc)

      | Cut l -> sgCUT
      | Commit l -> sgCOMMIT
      | True l -> sgTRUE
      | False l -> sgFALSE
      | Comma sgl -> sgSEQ (List.map mkSubgoal sgl)
      | If(exn, args, po, sgt, sgf, l) -> 
          (* Compile the subgoals *)
          let sgt' = mkSubgoal sgt in
          let sgf' = mkSubgoal sgf in
          let f = doExtern l exn args in
          sgIF (fun p (s: subst) (acc: proof list) -> 
            let thisId = goalId p in
            let thisLevel = level p in
            setLevel p (level p + 1);  (* Prepare the level for the subgoals *)
            setGoalId p (goalId p + 1);
            let id = 
              if !debug > 0 then 
                dprintf "%d.%d:" thisId thisLevel 
              else
                nil
            in
            if !debug > 0 then 
              ignore (E.log "KT:%a If %a@!" insert id 
                        d_spred (PFun(exn, args), s));
            let res = 
              match po, f p s with 
                _, None -> 
                  if !debug > 0 then 
                    ignore (E.log "KT:%a FALSE\n" insert id);
                  (* failure *) false, Unimplemented "F" :: acc
              | None, Some _ -> (* success, no need for proof *)
                  if !debug > 0 then 
                    ignore (E.log "KT:%a TRUE\n" insert id);
                  true, Unimplemented "T" :: acc
              | Some _, Some (_, Some prf) -> (* success, want a proof *) 
                  if !debug > 0 then 
                    ignore (E.log "KT:%a TRUE\n" insert id);
                  true, prf :: Unimplemented "T" :: acc
              | Some _, Some (_, None) -> 
                  E.s (E.error "%a: Extern %s returned success, but without a proof\n" E.d_loc l exn)
            in
            setLevel p (level p - 1);
            res)

            sgt'
            sgf'

      | Choose sgl -> 
          (* Compile the subgoals *)
          let sgl' = 
            let choiceNr = ref (-1) in
            List.map 
              (fun (sg_ast: subgoal_ast) -> 
                let sg: subgoal = mkSubgoal sg_ast in
                (* Prepend a goal that prints some info and adds the Magic 
                 * proof that tells us later where we're coming from *)
                incr choiceNr;
                let thisOne = !choiceNr in (* Read it now *)
                sgACCUM 
                  (fun p (s: subst) (acc: proof list) -> 
                    if !debug > 0 then begin
                      let id = dprintf "%d.%d:" (goalId p) (level p) in
                      ignore (E.log "KT:%a Choice #%d@!" insert id thisOne);
                    end;
                    Unimplemented ("Choose" ^ string_of_int thisOne) :: acc)
                &&& 
                sg)
              sgl
          in
          sgOR sgl'
          
      | Case(exn, args, cases, l) -> 
          (* Compile the cases *)
          let cases' = 
            List.map (fun (n, po, sg, l) -> 
                          (n, po, mkSubgoal sg, l)) cases in
          let f = doExtern l exn args in
          sgCHOICE (fun p s acc -> 
            let thisId = goalId p in
            let thisLevel = level p in
            setLevel p (1 + level p);  (* Prepare the level for the subgoals *)
            setGoalId p (1 + goalId p);
            let id = 
              if !debug > 0 then 
                dprintf "%d.%d:" thisId thisLevel 
              else
                nil
            in
            if !debug > 0 then 
              ignore (E.log "KT:%a Case %a@!" insert id 
                        d_spred (PFun(exn, args), s));
            (* Find which case we are taking *)
            let isFail, ncase, po = 
              match f p s with 
                None -> 
                  if !debug > 0 then 
                    ignore (E.log "KT:%a Failed\n" insert id);
                  (* failure *) true, "fail", None
              | Some (ncase, po) -> 
                  if ncase = "fail" then 
                    E.s (E.error "%a: Extern %s returned success with the reserved name \"fail\"\n" 
                           E.d_loc l ncase);
                  if !debug > 0 then 
                    ignore (E.log "KT:%a Success (case %s) %a@!" insert id ncase
                              d_spred (PFun(exn, args), s));
                  false, ncase, po
            in
            try
              (* Find the first matching case (or default) *)
              let (n', po', sg', l') = 
                List.find (fun (n', _, _, _) -> 
                  n' = ncase || (not isFail && n' = "*"))
                  cases'
              in
              (* Mark that we are taking the branch *)
              let acc' = Unimplemented ("C:" ^ n') :: acc in
              (* See if a proof is expected *)
              let sg, acc' = 
                match po', po with 
                  None, _ -> (* no proof expected *) sg', acc'
                | Some _, Some prf -> sg', prf :: acc'
                | Some _, None -> 
                    E.s (E.error "%a: Extern %s returned success %s, but without a proof\n" 
                           E.d_loc l exn ncase)
              in
              setLevel p (-1 + level p);
              sg, acc'
                
            with Not_found -> begin (* Get here if we did not find the case *)
              if not isFail then 
                ignore (E.warn "%a: Extern %s returned success %s, which does not match any of the cases\n" 
                          E.d_loc l exn ncase);
              setLevel p (-1 + level p);
              sgFALSE, Unimplemented ("C:false") :: acc
            end)
    in
      
    (* Create a proof generator, and lemmas *)
    let rec mkPrfGen 
        (sofar: pred list)  (* A list of the predicates whose proofs we have 
                             * seen so far, in reverse order *)
        (nrprfs: int)       (* How many proofs we have collected so far. This 
                             * can be larger than the length of "sofar" 
                             * because we collect Magic proofs for 
                             * conditionals *)
        (suffix: string)    (* The suffix to use for the lemmas that we 
                             * generate. These are formed from the names of 
                             * the branches we took so far *)
        (sgl: subgoal_ast list) 
        : pfgen = 
      let suffConcat (s: string) (n: string)  =
        if s = "" then n else s ^ "_" ^ n
      in
      match sgl with 
        [] -> (* We are at the end. Now we can make the lemma *)
          (* We must find the variables that occur only in our predicates. At 
           * the same time we will have to re-substitute their indices  *)
          let ourvars: int list ref = ref [] in (* Collect here the list of 
                                                 * variables that we actually 
                                                 * use. For each one we have 
                                                 * the index into r.vars *)
          List.iter 
            (fun p ->
              ignore (substPred 
                        (fun _ (e: exp) -> 
                          match e with 
                            UVar i -> 
                              if not (List.mem i !ourvars) then 
                                ourvars := i :: !ourvars;
                              None
                          | _ -> None)
                        p))
            (r.goal :: sofar);

          (* If we are using all of the rule variables, then we do not need 
           * to change anything *)

          let lname = if suffix = "" then r.name else r.name ^ "_" ^ suffix in

          let (l: lemma), (reordered: bool) = 
            if List.length !ourvars = List.length r.vars then begin
              { lname = U.registerSymbolName lname;
                lvars = r.vars; (* We just use the rule variables *)
                lhyps = List.rev sofar;
                lconc = r.goal },
              false
            end else begin
              assert(ignore (E.warn "Lemma %s does not use all variables of rule %s\n"
                               lname r.name);
                     true);
              (* We sort the variables in ourvars, to maintain as close a 
               * relationship with the declarations as possible *)
              ourvars := List.sort compare !ourvars;
          
              (* Now we rewrite all predicates to replace the indices. We map 
               * a rule index into a lemma index *)
              let newIndices: (int, int) H.t = H.create 13 in
              Util.list_iteri 
                (fun i i_rule -> H.replace newIndices i_rule i) !ourvars;
              
              let doSubst (_: int) (e: exp) : exp option = 
                match e with
                  UVar i_rule -> begin
                    try Some (UVar (H.find newIndices i_rule))
                    with Not_found -> 
                      E.s (E.bug "Cannot find the new index for %d\n" i_rule)
                  end
                | _ -> None
              in
              (* Rewrite the goal and the hypotheses *)
              let goal' = substPred doSubst r.goal in
              let sofar' = List.map (substPred doSubst) (List.rev sofar) in
              { lname = U.registerSymbolName lname;
                lvars = 
                List.map 
                  (fun i_rule -> 
                    try List.nth r.vars i_rule
                    with _ -> 
                      E.s (E.bug "Cannot find variable of index %d" i_rule))
                  !ourvars;
                lhyps = sofar';
                lconc = goal'},
              true
            end
          in
                
          lemmas := l :: !lemmas;
          (* Create the instantiations *)
          fun p (s: subst) (prf: proof list) -> 
            (* Produce the instantiations *)
            let insts = 
              Util.list_mapi 
                (fun i v -> 
                  let i_rule = 
                    if not reordered then i else 
                    (try List.nth !ourvars i  
                    with _ -> E.s (E.bug "Prolog_parse: nth"))
                  in
                  try
                    unrollUVarsRaise s (UVar i_rule)
                  with HasUVar ->
		    begin
                      ignore (E.warn "Proof fails only due to uninstantiated variable %s in rule %s" v (U.symbolName l.lname));
	              raise (Invalid_argument "instantiation")
		    end)
                l.lvars
            in
            (* Now produce the subgoals. We must filter the Magics *)
            let prf' = 
              List.filter
                (fun p -> 
                  match p with 
                    Unimplemented s when (s = "T" || s = "F" 
                              || (String.length s >= 3 && 
                                  String.sub s 0 2 = "C:")
                              || (String.length s >= 7 &&
                                  String.sub s 0 6 = "Choose")) -> false
                  | _ -> true) prf 
            in
            WithLemma(l.lname, insts, prf')
      | sg :: rest -> begin
          match sg with 
            Cut _ | Commit _ | True _ |
            Goal(Extern (_, _, None), _) -> 
              mkPrfGen sofar nrprfs suffix rest
          | False _ -> (* Done with this branch. Do no generate any lemmas *)
              fun p s acc -> 
                E.s (E.bug "Trying to generate proof of false")

          | Goal ((Pred p | Terminal p | Extern (_, _, Some p)), l) ->
              mkPrfGen (p :: sofar) (nrprfs + 1) suffix rest
          | Goal (Is (e1, e2), l) -> 
              mkPrfGen ((PNez(BinOp(SetEq, e1, e2))) :: sofar) 
                       (nrprfs + 1) suffix rest

          | Comma sgl -> mkPrfGen sofar nrprfs suffix (sgl @ rest)
          | If(n, _, po, sgt, sgf, l) -> begin
              let pft = 
                match po with 
                  None -> mkPrfGen sofar (nrprfs + 1) 
                                    (suffConcat suffix "T") (sgt :: rest) 
                | Some p -> mkPrfGen (p :: sofar) (nrprfs + 2) 
                                    (suffConcat suffix "T") (sgt :: rest)
              in
              let pff = mkPrfGen sofar (nrprfs + 1) 
                                 (suffConcat suffix "F") (sgf :: rest) in
              fun p (s: subst) (acc: proof list) -> 
                (* Must skip through acc to find the current one *)
                try
(*
                  ignore (E.log "%s:%d: Generating proof for IF. Position=%d. Found\n   @[%a@].\n"
                            l.file l.line nrprfs 
                            (docList ~sep:break (d_proof ())) acc);
*)
                  let this = 
                    try List.nth acc nrprfs 
                    with _ -> raise Not_found
                  in
                  match this with 
                    Unimplemented "T" -> pft p s acc
                  | Unimplemented "F" -> pff p s acc
                  | _ -> raise Not_found
                with Not_found -> 
                  E.s (E.bug "%a: Expecting Unimplemented T or Unimplemented F on position %d at call to extern %s. Found @[%a@]."
                         E.d_loc l nrprfs 
                         n (docList ~sep:break (d_prf ())) acc)
          end

          | Choose sgl -> begin
              (* Compile the choices *)
              let choices : pfgen list =
                let choiceNr = ref (-1) in
                List.map 
                  (fun sg -> 
                    incr choiceNr;
                    mkPrfGen sofar (1 + nrprfs)
                      (suffConcat suffix ("Choose" ^ string_of_int !choiceNr))
                      (sg :: rest))
                  sgl
              in
              fun p (s: subst) (acc: proof list) -> 
                try
                  (* See which branch we took *)
                  let this: proof = 
                    try List.nth acc nrprfs
                    with _ -> raise Not_found
                  in
                  let nstr : string = 
                    match this with
                      Unimplemented n when String.length n >= 7 
                                  && String.sub n 0 6 = "Choose" -> 
                             String.sub n 6 (String.length n - 6)
                    | _ -> raise Not_found
                  in
                  let n = 
                    try int_of_string nstr
                    with _ ->
                      ignore (E.warn "Choose proof not an integer");
                      raise Not_found
                  in
                  let pg: pfgen = 
                    try List.nth choices n
                    with _ -> 
                      ignore (E.warn "Not enough choices");
                      raise Not_found
                  in
                  pg p s acc
                with Not_found -> 
                  E.s (E.bug "Expecting Unimplemented proof on position %d at disjunction. Found @[%a@]."
                         nrprfs 
                         (docList ~sep:break (d_prf ())) acc)
          end

          | Case(n, _, cases, l) -> begin
              let cases' = 
                List.map (fun (n, po, sg, _) -> 
                  let sofar', nrprfs' = match po with
                    None -> sofar, nrprfs
                  | Some p -> p :: sofar, nrprfs + 1
                  in
                  (n, mkPrfGen sofar' (nrprfs' + 1) 
                               (suffConcat suffix n) (sg :: rest))) 
                  cases
              in
              fun p (s: subst) (acc: proof list) -> 
                try
                  let this = 
                    try List.nth acc nrprfs
                    with _ -> raise Not_found
                  in
                  let n' = 
                    match this with
                      Unimplemented n when String.length n >= 3 
                                  && String.sub n 0 2 = "C:" -> 
                                    String.sub n 2 (String.length n - 2)
                    | _ -> raise Not_found
                  in
                  (* Now find the Unimplemented *)
                  match List.filter (fun (n'', _) -> n''= n') cases' with
                    [(_, pf)] -> pf p s acc
                  | _ -> raise Not_found
                with Not_found -> 
                  E.s (E.bug "%a: Expecting Unimplemented (%a) on position %d at call to extern %s. Found @[%a@]."
                         E.d_loc l 
                         (docList (fun (n'', _) -> text n'')) 
                         cases'
                         nrprfs
                         n
                         (docList ~sep:break (d_prf ())) acc) 
          end
      end
    in
    let mkprf = mkPrfGen [] 0 "" [r.subgoal] in
    let plr = { Prolog.name = r.name;
                maxUnif = List.length r.vars - 1;
                conc = r.goal;
                sub  = mkSubgoal r.subgoal;
                init = [];
                mkprf = (fun p s acc -> 
                  let acc = List.rev acc in
(*
                  ignore (E.log "Produce a proof of %s with\n   @[%a@]\n"
                            r.name (docList ~sep:line (d_proof ())) acc);
*)
                  mkprf  p s acc);
                isassume = false;
              } in
    plr
  in
  let rules = ref [] in
  List.iter
    (fun rd -> 
      match rd with 
        Rule r -> rules := (processOneRule r) :: !rules
      | Tabled names -> 
          (* To announce that something is tabled, you create a special rule *)
          List.iter 
            (fun (x, _) -> 
              rules := 
                Prolog.tableConstantRule (String.capitalize x) :: !rules)
            names

      | Constant (names, dtyp) -> 
          (* Type check the declaration type *)
          E.withContext
            (fun _ -> dprintf "Constant declaration at %a"
                d_loc
                (match names with 
                  (_, l) :: _ -> l
                | _ -> E.s (E.bug "Empty constant declaration")))
            (fun _ ->
              (match dtyp with 
                TType -> ()
              | TProp args -> List.iter (typeCheckType decls) args
              | TCons (res, args) -> 
                  typeCheckType decls res; 
                  List.iter (typeCheckType decls) args))
            ();
          (** Add a new declaration to the program *)
          List.iter 
            (fun (name, loc) -> 
	      let cName =
		match dtyp with
		  TProp _ ->
		    if name = "" then
		      ""
		    else
		      String.make 1 (Char.uppercase (String.get name 0))
		      ^ String.sub name 1 (String.length name - 1)
		| _ -> name in
	      (match !coqFlag with
		CoqNone -> ()
	      | _ ->
		  toCheck := (name, "Goal " ^ declTypToString dtyp ^ ".\nApply " ^ cName ^ ".\nAbort.\n") :: !toCheck);
              try
                let old, oldloc = H.find decls name in
                if old <> dtyp then 
                  E.s (E.error "The name %s is already declared in the program"
                         name)
              with Not_found -> 
                H.add decls name (dtyp, loc))
            names)
    rl;
  let toCheck = List.rev !toCheck in
  let ver inch outch =
    let verOne b (n, s) =
      output_string outch s;
      flush outch;

      let rec untilAborted () =
	let l = input_line inch in
	(*ignore (E.log "::: %s\n" l);*)
	match l with
	  "Current goal aborted" -> ()
	| _ -> untilAborted () in

      let lines = ref "" in

      let rec findResult () =
	let l = input_line inch in
	(*ignore (E.log ">>> %s\n" l);*)
	match l with
	  "Subtree proved!" ->
	    (*ignore (E.log "Here we go!\n");*)
	    untilAborted ();
	    true
	| "Error: Unknown command of the non proof-editing mode" -> findResult ()
	| "Current goal aborted"
	| "Error: No focused proof (No proof-editing in progress)." ->
	    (match !coqFlag with
	      CoqVerbose -> ignore (E.log "**** %s failed:\n%s\n" n !lines)
	    | _ -> ignore (E.log "**** %s failed.\n" n));
	    false
	| _ ->
	    lines := !lines ^ l ^ "\n";
	    findResult () in

      findResult () && b in
    let res = List.fold_left verOne true toCheck in
    close_out outch;
    close_in inch;
    res in
  List.rev !rules, !lemmas, ver


%}

/*(* regular tokens *)*/
%token <string> VAR EXVAR
%token <string * Errormsg.location> ATOM
%token <string * Errormsg.location> LABEL
%token <int> INT
%token EOF COLON RPAREN DOT COLON_MINUS SEMICOLON
%token LBRACKET RBRACKET IF THEN ELSE ENDIF EXTERN BANG COMMA BAR
%token COMMIT IS CASE OF ENDCASE LCURLY RCURLY ARROW
%token TRUE FALSE TERMINAL
%token OR AND
%token EXISTS FORALL
%token <Errormsg.location> LPAREN
%token <Logic.binop> PBINOP BINOP

%token TYPE PROP CONSTANT TABLED

%type <Prolog.rule list * Proof.lemma list * (in_channel -> out_channel -> bool)> main 
%type <subgoal_ast> subgoal
%type <rule_decl list> rule_decl_list
%type <rule_decl> rule_decl
%type <Logic.pred> predterm

/*(* Make these right associative, so that we can build the lists *)*/
   
%right BAR
%right COMMA

%start main

%%
 
main:       /* (*A Kettle program is a sequence of rules *)*/
  rule_decl_list                  { processRules $1 }
;

rule_decl_list:
  /*(* empty *)*/                 { [] }
| rule_decl rule_decl_list        { $1 :: $2 }
;

rule_decl: 
  rulename vardecl predterm rule_subgoal DOT 
                              { !currentRule.goal <- $3;
                                !currentRule.decls <- $2;
                                !currentRule.subgoal <- $4;
                                Rule !currentRule 
                              }
| CONSTANT atom_list COLON decltyp DOT   
                              { Constant ($2, $4) } 
| TABLED atom_list DOT
                              { Tabled $2 }
;

atom_list: 
  ATOM                        { [ $1 ] }
| ATOM COMMA atom_list        { $1 :: $3 }
;


decltyp: 
             TYPE             { TType }
| typ_list                    { match $1 with 
                                  TCons(res, _), args -> TCons(res, args)
                                | TType, _ -> E.s (E.bug "kettle_parse") 
                                | TProp _, args -> TProp args
                              } 
;

typ_list:
| ATOM                        { TCons(toVal (fst $1), []), []  }
| PROP                        { TProp [], [] }
| ATOM ARROW typ_list         { let res, args = $3 in
                                res, toVal (fst $1) :: args
                              } 
;

var_list: 
  registervar                { [ $1 ] }
| registervar COMMA var_list { $1 :: $3 }
;

/*(* Register the variables as we see them in the declarations *)*/
registervar: 
 VAR                         { ignore (getVar $1); $1 }
;

rule_subgoal: 
  /* empty */                 { True !currentLoc }
| COLON_MINUS                 { True !currentLoc }
| COLON_MINUS subgoal         { $2 }
;

rulename: 
  ATOM COLON                  { let r = { name = fst $1; 
                                          decls = None;
                                          vars = [];
					  exvars = [];
                                          goal = Bool true;
                                          subgoal = True !currentLoc;
                                          loc = snd $1 } in
                                currentRule := r;
                              } 
;

vardecl:
  /* empty */                 { None }
| LBRACKET var_list COLON ATOM RBRACKET vardecl
                              { let tl = match $6 with 
                                   Some l -> l
                                 | None -> []
                                in
                                let t = toVal (fst $4) in
                                Some ((List.map (fun v -> (v, t)) $2) @ tl)
                              } 
;

                                
term: 
| ATOM                           { let an = fst $1 in
                                   if an = "" then 
                                     E.s (E.error "Empty atom");

                                   if String.sub an 0 1 = "&" then begin
                                     (* ignore (E.log "Found addrof %s\n" an);*)
                                     Fun ("&", [Fun (String.sub an 1 
                                               (String.length an - 1), [])])
                                   end else
                                      Fun(an, []) 
                                 }
| LABEL                          { let an = fst $1 in
                                   if an = "" then 
                                     E.s (E.error "Empty atom");

                                   Fun("&", [Fun (an, [])])
                                 }
| INT                            { Int (Int32.of_int $1) }
| VAR                            { getVar $1 }
| EXVAR                          { getExvar $1 }
| LPAREN BINOP term term RPAREN  { BinOp($2, $3, $4) } 
| LPAREN ATOM args RPAREN        { Fun(fst $2, $3) } 
;

register_exvar:
| EXVAR                          { !currentRule.exvars <- $1 :: !currentRule.exvars;
				   $1 }

predterm: 
| ATOM                              { PFun(String.capitalize (fst $1), []) }
| FALSE                             { Bool false }
| LPAREN ATOM args RPAREN           { PFun(String.capitalize (fst $2), $3) }
| LPAREN PBINOP term term RPAREN    { PNez(BinOp($2, $3, $4)) } 
| LPAREN OR pargs RPAREN            { Or $3 }
| LPAREN AND pargs RPAREN           { And $3 }
| LPAREN ARROW predterm predterm RPAREN
                                    { Imply ($3, $4) }
| LPAREN EXISTS register_exvar predterm RPAREN
                                    { Exist ($3, $4) }
| LPAREN FORALL register_exvar predterm RPAREN
                                    { All ($3, $4) }
;

pargs:   
/*(* empty *)*/                 { [] }
| predterm pargs                { $1 :: $2 }
;

args: 
  term                         { [$1] }
| term args                    { $1 :: $2 }
;


subgoal: 
| BANG                      { Cut (E.getLocation ()) } 
| COMMIT                    { Commit (E.getLocation ()) }
| TRUE                      { True (E.getLocation ()) }
| LPAREN subgoal RPAREN     
                            { $2 }
| predterm                  { Goal (Pred $1, E.getLocation ()) }

| LPAREN EXTERN ATOM args opt_type RPAREN { Goal(Extern (fst $3, $4, $5),
                                                 snd $3) } 
| LPAREN TERMINAL predterm RPAREN  { Goal (Terminal $3, 
                                           $1) }
| LPAREN IS term term RPAREN       { Goal (Is ($3, $4), $1) }
| IF LPAREN ATOM args opt_type RPAREN THEN subgoal ELSE subgoal ENDIF
                            { If (fst $3, $4, $5, $8, $10, $2) }
| CASE LPAREN ATOM args RPAREN OF cases ENDCASE   
                            { Case (fst $3, $4, $7, $2) }
| subgoal COMMA subgoal     { match $1, $3 with 
                                Comma l1, Comma l2 -> Comma (l1 @ l2)
                              | l1, Comma l2 -> Comma (l1 :: l2)
                              | Comma l1, l2 -> Comma (l1 @ [l2])
                              | l1, l2 -> Comma [l1; l2] 
                            }
| subgoal BAR   subgoal     { match $1, $3 with 
                                Choose l1, Choose l2 -> Choose (l1 @ l2)
                              | l1, Choose l2 -> Choose (l1 :: l2)
                              | Choose l1, l2 -> Choose (l1 @ [l2])
                              | l1, l2 -> Choose [l1; l2]
                            } 
;

cases: 
  case                      { [ $1 ] }
| case cases                { $1 :: $2 }
;

case: 
  casename opt_type ARROW subgoal SEMICOLON 
                            {
                              if fst $1 = "fail" then 
                                if $2 <> None then begin
                                  parse_error "The \"fail\" case should not have a proof";
                                end;
                              (fst $1, $2, $4, snd $1) 
                             }
;

casename: 
  ATOM                      { $1 }
| BINOP                     { if $1 = Mult then 
                                 ("*", E.getLocation ()) 
                              else begin
                                 parse_error "Invalid case name";
                                 ("*", E.getLocation ())
                              end  }
;

opt_type:
  /* empty */               { None }
| COLON predterm            { Some $2 }
;

