
(** An implementation of the Shostak procedure for linear arithmetic. First 
 * stage does only difference constraints **)

module E = Errormsg
module L = Logic
module P = Proof
module U = Util
module GA = GrowArray
module IH = Inthash
module O = Options

open Linarith

open Satproc              (* message, etc. *)
open Edag
open Pretty

let shid = Util.registerSymbolName "SH"

let debug = ref false

let fullAssert = true

let shContras = ref 0 

let log fmt = ignore (E.log "SH: "); E.log fmt
let warn fmt = ignore (E.log "SH: "); E.warn fmt
let error fmt = ignore (E.log "SH: "); E.error fmt
let unimp fmt = ignore (E.log "SH: "); E.unimp fmt
let bug fmt = ignore (E.log "SH: "); E.bug fmt


(** Joyeux noel et une bonne annee 2005. De Sylvia Necula, Paris le 5 
 * janvier, 2005. **)

(** We must keep track of all arithmetic nodes that are arguments of 
 * non-arithmetic nodes (to report equalities) and also of the non-arithmetic 
 * nodes that are argument to arithmetic nodes. We call these the interned 
 * nodes. Interned nodes are kept in arrays and are identified by small 
 * integer indices. *)

type internData = 
    { index: int;
      node: node; (** The node that this corresponds to *)
      mutable isderived: (internData * int * Linarith.arithproof list) option; 
      (* Among all interned nodes whose difference can be proved to be a 
       * constant, we keep only one active. The other ones are derived. If 
       * this is a derived node, then the index of the active node, the 
       * offset from it (the derived node is active + offset), and the 
       * arguments of the ArithEq proof that proves this. The directions of 
       * the eqproofs is important because we will use them to construct 
       * inequality proofs. We always prove n - act - offset = 0. *)

      mutable derivatives: internData list;
      (** For an active node only (not isderived), keep a list of derivative 
       * nodes *)
    }


let initSize = 32
(** Keep here the data for all the interned nodes *)
let indexData: internData GA.t = 
  GA.make initSize (GA.Elem { node = dummyNode;
                              index = -1;
                              isderived = None;
                              derivatives = [] })

(** This is the largest used index *)
let max_index = ref (-1)

(** Keep here a pointer to the index of 0 *)
let zeroIndex = ref (-1)


(** Save here the constraints that we cannot process directly as difference 
 * constraints. *)
let savedConstraints: arithRes list ref = ref []

  
let nodeOfIndex (idx: int) = 
  (GA.get indexData idx).node
let expOfIndex (idx: int) = (nodeOfIndex idx).exp

let isActiveIndex (idx: int) =
  (GA.get indexData idx).isderived = None

(** Given an index i, find (j, off, prf) such that j is active, i = j + off 
 * and prf proves that i - j - off = 0 (the direction of the proof is 
 * important) *)
let rec findActiveIndex (idx: int) : internData * int * arithproof list = 
  let id = GA.get indexData idx in 
  let rec loop (id: internData) :  internData * int * arithproof list = 
    match id.isderived with 
      None -> id, 0, []
    | Some (act, off, plist (* : id - act - off = 0  *)) -> 
        let res, off', plist' (* : act - res - off' = 0 *) = loop act in 
        res, off' + off, plist @ plist'
  in

  loop id


(** Keep a map from node ID to the index of the interned node *)
let internedNodes: internData IH.t = IH.create 13

(** For each ID of an equivalence class root, keep track of one node in the 
 * class that is interned *)
let eqClassInterns: internData IH.t = IH.create 13

let eqClassHasIntern (rootid: int) : internData option = 
  try Some (IH.find eqClassInterns rootid)
  with Not_found -> None
  
(** See if a node is interned *)
let isInternedOpt (n: node) : internData option = 
  try Some (IH.find internedNodes n.id)
  with Not_found -> None

let isSomebodyInClassInterned (n: node) : internData option = 
  try 
    let id = IH.find eqClassInterns n.root.id in
    Some id
  with Not_found -> None

(** Keep a list of nodes for which we have processed the arithmetic inside *)
let processedNodes: unit IH.t = IH.create 13

let markProcessed (n: node) : bool = 
  try IH.find processedNodes n.id; true
  with Not_found -> begin
    let id = n.id in 
    IH.add processedNodes id ();
    registerUndoRun (fun _ -> IH.remove processedNodes id);
    false
  end

let abs n = if n < 0 then - n else n

(** Find the GCD of two numbers *)
let gcd (n1: int) (n2: int) : int = 
  if n1 <= 0 || n2 <= 0 then E.s (bug "gcd called with negative factors");
  let rec gcint (n1: int) (n2: int) : int = 
    if n1 = n2 then n1 
    else 
      let k = n1 / n2 in 
      let r = n1 - n2 * k in
      if r = 0 then 
        n2
      else gcint n2 r
  in
  if n1 >= n2 then 
    gcint n1 n2
  else
    gcint n2 n1

(** Keep here the actual matrix *)
type edgeData = { 
    min: int;  (** This is the MIN over the constraints of xi - xj **)
    start: int;  (** This min is attained through transitivity through the 
                  * edge (start, finish) with the given proof **)
    finish: int;

    proof: arithproof list;
    (** Make a proof that owner(start) >= owner(finish) + min, up to arithmetic 
     * simplification *)
  } 
  

(** The matrix, one row at a time *)
let matrix: edgeData option GA.t GA.t = 
  GA.make initSize 
    (GA.Susp 
       (fun _ -> 
         GA.make (GA.num_alloc_index indexData)
           (GA.Elem None)))


let smatrix (i: int) (j: int) : edgeData option = 
  GA.get (GA.get matrix i) j

let d_owner () (idx: int) = 
  assert (idx <= !max_index);
  text "f" ++ num idx ++ 
    text "(" ++ d_nodeTop () ((GA.get indexData idx).node)
                                         ++ text ")"


let shostakFactorData = 
  { Linarith.toexp = (fun i -> expOfIndex i);
    Linarith.todoc = (fun i -> d_owner () i);
  }

let proofOfLinArith = Linarith.proofOfLinArith shostakFactorData
let arithResToExp = Linarith.arithResToExp shostakFactorData
let d_arithRes = Linarith.d_arithRes shostakFactorData 
let check = Linarith.check shostakFactorData
let d_factor () f = shostakFactorData.Linarith.todoc f

(** Print the state *)
let onlyactive = true
let printState () = 
  for i = 0 to !max_index do 
    if not onlyactive || isActiveIndex i then begin 
      GA.iteri
        (fun j edo -> 
          match edo with 
            Some ed when j <> i && (isActiveIndex j || not onlyactive) -> 
              ignore (E.log " %a - %a >= %d (through %d-%d)\n" 
                        d_owner i d_owner j ed.min 
                        ed.start 
                        ed.finish)
          | _ -> ())
        (GA.get matrix i)
    end
  done;
  if true || not onlyactive then begin 
    for i = 0 to !max_index do
      let i', off', _ = findActiveIndex i in 
      if i <> i'.index then 
        ignore (E.log "  Index %a is derived as %a + %d\n" 
                  d_owner i d_owner i'.index off')
    done
  end;
  ignore (E.log " with saved constraints:\n%a"
            (docList ~sep:nil
               (fun c -> dprintf "  %a\n" d_arithRes c)) !savedConstraints);
  
  ()
  

(** Intercept findActiveIndex and check the proofs that it returns *)
let findActiveIndex (idx: int) : internData * int * arithproof list = 
  let (id, off, prfs) as res = findActiveIndex idx in
  assert(let idxexp = expOfIndex idx in 
         let pargs = eqProofsToPargs prfs in
         proofWellFormed (dprintf "findActiveIndex %a = %a + %d" 
                            d_owner idx d_owner id.index off)
           (P.mkArithEq idxexp
              (L.plus id.node.exp (L.integer off)) pargs));
  res

(** Make a list of proofs (all with coefficient 1) that added together prove 
 * that owner(n1) - owner(n2) >= M(n1,n2) **)

let proofListEdges = ref [] (* for debugging *)
let rec makeGeqProofList (n1: int) (n2: int) : arithproof list = 
  if !debug && false then 
    ignore (log "makeGeqProofList: %a and %a\n" d_owner n1 d_owner n2);
  assert(if List.exists (fun (s,f) -> s=n1 && f=n2) 
      !proofListEdges then false else begin
        proofListEdges := 
          (n1, n2) :: !proofListEdges;
        true
      end);
  match smatrix n1 n2 with 
    Some ed -> 
      let pre = 
        if ed.start <> n1 then 
          makeGeqProofList n1 ed.start
        else
          []
      in
      let this = 
        if ed.start <> ed.finish then begin
          ed.proof
        end else
          []
      in
      let post = 
        if ed.finish <> n2 then 
          makeGeqProofList ed.finish n2
        else
          []
      in
      pre @ this @ post
          
        
  | None -> 
      E.s (bug "makeGeqProof but no edge between %a-%a"
             d_owner n1 d_owner n2)

(** Define a new entry point, for debugging *)
let makeGeqProofList n1 n2 = 
  assert(proofListEdges := []; true);
  makeGeqProofList n1 n2


(** Make a proof that proves (owner(n1) + off1) >= (owner(n2) + off2)  *)
let makeGeqProof (n1: int) (off1: int) (n2: int) (off2: int) : P.proof = 
  let geql = makeGeqProofList n1 n2 in 
  let e1 = expOfIndex n1 in 
  let e1' = if off1 = 0 then e1 else L.plus e1 (L.integer off1) in 
  let e2 = expOfIndex n2 in 
  let e2' = if off2 = 0 then e2 else L.plus e2 (L.integer off2) in 
  P.mkArithGeq e1' e2' (eqProofsToPargs geql)
                    
  
(** Make a proof of equality between e1 = e2 *)
let makeEqualityProof (e1: L.exp) (e2: L.exp) 
                      (fgeq: int) (pfgeq: arithproof list) 
                      (fleq: int) (pfleq: arithproof list) = 
  (* Perhaps one of the lists contains only equality proofs *)
  let rec onlyEq = function
      [] -> true
    | (_, true, _) :: rest -> onlyEq rest
    | _ -> false
  in
  if onlyEq pfgeq then 
    P.mkArithEq e1 e2 ~fact:fgeq (eqProofsToPargs pfgeq)
  else if onlyEq pfleq then 
    P.mkArithEq e1 e2 ~fact:fleq (eqProofsToPargs pfleq)
  else
    P.G("eqi", [ P.P (P.mkArithGeq e1 e2 ~fact:fgeq 
                        (eqProofsToPargs pfgeq));
                 P.P (P.mkArithGeq e2 e1 ~fact:fleq
                        (eqProofsToPargs pfleq)) ])
  
               


(** Handle a trivial constraint. such a constraint is either redundant or is 
 * a contradiction. Raise Contradiction in the latter case *)
let handleTrivialConstraint (r: arithRes) : unit = 
  assert (r.facts == []);
  if r.ct = 0 || (not r.iseq && r.ct >= 0) then begin
    if !debug then
      ignore (log "constraint %a is redundant\n" d_arithRes r);
    ()
  end else begin
    if !debug then
      ignore (log "constraint %a is contradictory\n" d_arithRes r);
    (* We negate everything if r.ct > 0 *)
    let sgnrct = if r.ct > 0 then 1 else -1 in 
    let prfs = scaleProofs (- sgnrct) r.proofs in 
    let prf = 
      (* Make a proof that 0 >= - r.ct *)
      if !genProofs then begin
(*
        let p = 
          P.mkArithGeq L.zero (L.integer (sgnrct * r.ct))  
            (eqProofsToPargs prfs) in 

        P.G("geqLt_FI", [ P.P p;
                          P.P (P.mkArithGeq 
                                 (L.integer (sgnrct * r.ct))
                                 (L.plus L.zero L.one) 
                                 []) ])
*)
        P.mkArithContra (eqProofsToPargs prfs)

      end else 
        noProof
    in
    if !debugStats then 
      incr shContras;
    raise (Contradiction prf)
  end

(* We should not need to emit disequalities, since testMessage can now test 
 * very well if a message is true or false *)
let emitDiseq = false

(** Mark a node relationship. prf : n1 - n2 - off21. This procedure must be 
 * called only when the entries corresponding to n1 and n2 already reflect 
 * the equality. *)
let setDerivative (n1: internData) 
                  (n2: internData) 
                  (off21: int) 
                  (eqproofs: arithproof list) : unit = 
  if fullAssert then 
    assert (check (text "start setDerivative")
              { iseq = true;
                ct = - off21;
                facts = canonicalFacts [ (1, n1.index);(-1, n2.index) ];
                proofs = eqproofs });
                
  if !debug then 
    ignore (log "setDerivative %a = %a + %d\n" 
              d_owner n1.index d_owner n2.index off21);

  let act1, off1, prfs1 (* : n1 - act1 - off1 = 0 *) 
      = findActiveIndex n1.index in
  let act2, off2, prfs2 (* : n2 - act2 - off2 = 0 *) 
      = findActiveIndex n2.index in

  if act1 == act2 then begin 
    (* Already in the same set of derivatives *)
    if off1 <> off2 + off21 then assert false;
    if !debug then 
      ignore (log "   already in the same derivative set\n");
  end else begin 
    (* Now we know that 
          prfs2 + eqproofs - prfs1 : act1 - act2 - off2 - off + off1 = 0
    *)
    let totprfs = prfs2 @ eqproofs @ scaleProofs (-1) prfs1 in 
    let totoff = off2 + off21 - off1 in 

    if fullAssert then 
      assert (check (text "in setDerivative, before equalities")
                { iseq = true;
                  ct = - totoff;
                  facts = canonicalFacts [ (1, act1.index); 
                                              (-1, act2.index)];
                  proofs = totprfs });

    (* Not in the same set. First collect the proofs of equalities between 
     * the two derivative sets. We do only one node from act1 and from act2 
     * for each distinct offset (from act1). *)
    let noffsets: int list ref = ref [] in
    let rec iter f = function [] -> () | h::t -> f h; iter f t in
    iter 
      (fun nd -> 
        let ndact, ndoff, ndprfs (* : nd - act1 - ndoff *) 
            = findActiveIndex nd.index in 
        assert (ndact == act1); (** This was supposed to be derived from act1*)

        if not (List.mem ndoff !noffsets) then begin
          noffsets := ndoff :: !noffsets;

          (* Keep here the offsets from act2 that we can compared for 
           * disequality already *)
          let diseq2 = ref [] in
          (* Now see if we find a node derived from act2 that is equal to nd *)
          let rec iter f = function [] -> () | h::t -> f h; iter f t in
          iter
            (fun bd -> 
              let bdact, bdoff, bdprfs (* : bd - act2 - bdoff = 0 *) 
                  = findActiveIndex bd.index in 
              assert (bdact == act2);

              if false then 
                ignore (log "   nd = %a, bd= %a\n" 
                          d_owner nd.index d_owner bd.index);
              if nd.node.root != bd.node.root && 
                 totoff + ndoff - bdoff = 0 then begin 
                   if !debug then 
                     ignore (log "discovered that %a = %a\n" 
                               d_owner nd.index d_owner bd.index);
                   
                   newMessage
                     { body = Intern (L.SetEq, nd.node, bd.node);
                       owner = shid;
                       fromgoal = false;
                       Edag.proof = 
                        P.mkArithEq nd.node.exp bd.node.exp
                         (eqProofsToPargs
                            (ndprfs @ 
                             totprfs @ 
                             scaleProofs (-1) bdprfs))
                     }
                 end;

              (* Now see if we ought to announce the disequality *)
              if emitDiseq && totoff + ndoff - bdoff <> 0 && not (List.mem bdoff !diseq2) then begin
                diseq2 := bdoff :: !diseq2;
                if forbiddenEquality nd.node bd.node == None then begin
                  if !debug then 
                    ignore (log "Discovered disequality %a != %a\n" 
                              d_owner nd.index d_owner bd.index);
                  newMessage
                    { body = Intern (L.SetNe, nd.node, bd.node);
                      owner = shid;
                      fromgoal = false;
                      Edag.proof = 
                        P.ArithNe(nd.node.exp, bd.node.exp, 1,
                                  eqProofsToPargs
                                    (ndprfs @ 
                                     totprfs @ 
                                     scaleProofs (-1) bdprfs))
                    }
                end
              end;
)
            (act2 :: act2.derivatives)
        end)
      (act1 :: act1.derivatives);

    (* Merge act1 and act2 and this should take care of the merge *)

    (* If one of the nodes is a constant, make the other one derived *)
    let doMarkDerived 
        (n: internData)
        (base: internData) (off: int) 
        (prfs: arithproof list) : unit = 
      assert(proofWellFormed (dprintf "markNodeAsDerived %a = %a + %d"
                                d_owner n.index d_owner base.index off)
               (P.mkArithEq n.node.exp
                  (L.plus base.node.exp (L.integer off)) 
                  (eqProofsToPargs prfs)));
      
      let old_nder = n.isderived in 
      
      if !debug then 
        ignore (log "mark node %a as derived %a + %d\n" 
                  d_owner n.index d_owner base.index off);
      n.isderived <- Some (base, off, prfs);
      
      let old = base.derivatives in
      (* Take all the derivatives of n, and n itself and add them to the 
       * derivatives of act *)
      base.derivatives <- n :: n.derivatives @ old;
      registerUndoRun 
        (fun _ -> n.isderived <- old_nder; 
          base.derivatives <- old)
    in

    if act1.index <> !zeroIndex then 
      doMarkDerived act1 act2 totoff totprfs
    else
      doMarkDerived act2 act1 (- totoff) (scaleProofs (-1) totprfs)
    
  end

(** Implement Fourier-Motzkin elimination. It is expensive but we use it only 
 * when Shostak is about to give up *)
module Fourier = struct
  
  let log fmt = ignore (E.log "FM: "); E.log fmt
  let warn fmt = ignore (E.log "FM: "); E.warn fmt
  let error fmt = ignore (E.log "FM: "); E.error fmt
  let unimp fmt = ignore (E.log "FM: "); E.unimp fmt
  let bug fmt = ignore (E.log "FM: "); E.bug fmt


  (** smallest factors that when multiplied by n1 and n2 give the same number*)
  let smallestFactors (n1: int) (n2: int) : int * int =
    let g = gcd (abs n1) (abs n2) in 
    (n2 / g), (n1 / g)


  (** Find the gcd of the list of factors *)
  let gcdFactors (f: (int * int) list) : int = 
    match f with 
      [] -> 1
    | (w, _) :: rest -> begin
        let rec loop (sofar: int) = function
            [] -> sofar
          | (w, _) :: rest -> 
              (* if sofar = 1 then 1 else *) loop (gcd sofar (abs w)) rest
        in
        loop (abs w) rest
    end

  (** Find the variable to eliminate. Pick the one with the lowest product of 
   * positive and negative constraints *)
  let findNextFactor (cl: arithRes list) : int = 
    let facts : ((int * int) ref) IH.t = IH.create 13 in 
    let foundFactor (w, f) = 
      let entry = 
        try IH.find facts f with Not_found -> 
          begin
            let r = ref (0, 0) in 
            IH.add facts f r;
            r
          end
      in
      let p, n = !entry in 
      let r' = if w > 0 then p + 1, n else p, n + 1 in
      entry := r'
    in
    let rec iter f = function [] -> () | h::t -> f h; iter f t in
    iter
      (fun c -> 
        (* each factor occurs at most once *)
        iter foundFactor c.facts)
      cl;
    (* now find the factor with smallest product *)
    let loop f entry
        (((bestsofar: int), (sofarprod: int)) as sofar) = 
      let p, n = !entry in 
      let pn = p * n in 
      if pn < sofarprod then (f, pn) else sofar
    in
    let best, bestprod = IH.fold loop facts ((-1), 10000000) in 
    if  best < 0 then assert false;
    if !debug then 
      ignore (log "will eliminate factor %a (score = %d)\n" 
                d_factor best bestprod);
                
    best


      (** Tighten a constraint for integers *)
  let integerTightening (r: arithRes) : arithRes = 
    (** See if we can reduce it *)
    let g = gcdFactors r.facts in 
    if g = 1 then 
      r (** There is no GCD of the factors *)
    else begin 
      (* We have prf: g * facts' + r.ct >= 0. *)
      (* Thus prf: g * facts' + g * r.ct' + rest >= 0. 
         As long as 0 <= rest < g, we can derive that facts' + r.ct' *) 
      let rest = 
        (* The mod operator works strangely for negative numbers *)
        if r.ct > 0 then r.ct mod g 
        else begin
          let m = (- r.ct) mod g in 
          if m = 0 then 0 else g - m 
        end
      in
      assert (rest >= 0);
      assert (rest < g);
      assert ((r.ct - rest) mod g = 0);

      (* Even if everything is divisible, divide everything though, to keep 
       * numbers small, and to allow the detection of equalities *)

      (* First take care of the expression *)


      let rec map f = function [] -> [] 
                            | h::t-> let h' = f h in h' :: map f t in 
      let r2 = 
        { r with 
          ct = (r.ct - rest) / g;
          facts = map (fun (w, f) -> (w / g, f)) r.facts;
        }
      in


      (* There is a special case for equalities *)
      if r.iseq && rest <> 0 then begin
        (* this is an equality. All the facts are divisible by g, but the 
         * constant is not. *)
        if !debug then 
          ignore (log "found contradiction during integer tightening\n");
        let pfalse = 
          if !genProofs then begin 
            let e2 = arithResToExp r2 in
            (* First show that e2 >= 0 *)
            let e2prfs = eqProofsToPargs r2.proofs in
            let prf_e2_pos = 
              (fun _ -> P.mkArithGeq e2 L.zero ~fact:g e2prfs) in
            (* Multiply this proof by g and get g * e2 >= 0. Subtract from 
             * this the proof of r, and we get a proof that - rest >= 0, 
             * which we use to prove false. *)
(*
            P.G("geqLt_FI",
                [ P.P (P.mkArithGeq (L.integer (- rest)) L.zero
                         (P.I g :: P.P (prf_e2_pos ()) :: 
                          (eqProofsToPargs (scaleProofs (-1) 
                                              r.proofs))));
                  P.P (P.mkArithGeq L.zero 
                         (L.plus (L.integer (- rest)) L.one)
                         [])]) 
*)
            P.mkArithContra (P.I g :: P.P (prf_e2_pos ()) ::
                             (eqProofsToPargs (scaleProofs (-1) r.proofs)))
          end else noProof
        in
        if !debugStats then 
          incr shContras;
        raise (Contradiction pfalse)

      end;

      (* Now we adjust the proofs as well. We'll turn all the proofs into a 
       * single one, to take advantage of the integer tightening *)
      let r3 = 
        { r2 with 
          proofs = 
          [ (1, r.iseq, 
             ProofThunk (fun _ -> 
               let e2 = arithResToExp r2 in
               let e2prfs = eqProofsToPargs r2.proofs in
               if r.iseq then
                 P.mkArithEq e2 L.zero ~fact:g e2prfs
               else
                 P.mkArithGeq e2 L.zero ~fact:g e2prfs))]
        }
      in
      assert (check (text "after integerTightening") r3);
      if !debug then 
        ignore (log "Constraint %a was tightened to %a\n"
                  d_arithRes r d_arithRes r3);
      r3
    end (* integerTightening *)

  let checkConstraintForDerivatives (c: arithRes) : unit = 
    assert (c.iseq);
    let setDer (i1: int) (i2: int) = 
      let n1 = GA.get indexData i1 in 
      let n2 = GA.get indexData i2 in 
      setDerivative n1 n2 (- c.ct) c.proofs
    in
    match c.facts with 
      [ (1, f) ] -> setDer f !zeroIndex
    | [ (-1, f) ] ->  setDer !zeroIndex f
    | [ (1, p); (-1, n) ] -> setDer p n 
    | [ (-1, p); (1, n) ] -> setDer n p 
    | _ -> () (* nothing to announce *)

  type mergeResult = 
      SomeRes of arithRes (** A simplified and merged result *)
    | FoundEquality of arithRes (** It was simplified to a trivial 
                                 * constraint, but we also found a new 
                                 * equality *)
    | Trivial 



      (** Merge two constraints with given factors and simplify the result. 
       * Report the merged and simplified constraint. If None, it means that 
       * it was siplified to a trivial constraint. May also raise a 
       * Contradiction. *)
  let mergeConstraints (c1: arithRes) (f1: int) 
                       (c2: arithRes) (f2: int) : mergeResult = 
    assert ((f1 > 0 || c1.iseq) && (f2 > 0 || c2.iseq));
    if false && !debug then 
      ignore (log "Merging constraints\n       %d: %a\n  and  %d: %a\n"
                f1 d_arithRes c1 f2 d_arithRes c2);
    (* Find the smallest factors *)
    (* Add the constraints *)
    let f1c1 = scale f1 c1 in 
    let f2c2 = scale f2 c2 in 
    let r' = add f1c1 f2c2 in 
    let r'' = integerTightening r' in 
    if !debug then 
      ignore (log "Merged constraints\n       %d: %a\n  and  %d: %a\n  into %a\n"
                f1 d_arithRes c1 f2 d_arithRes c2 d_arithRes r'');
    (* See if this is a trivial constraints *)
    if r''.facts = [] then begin
      if r''.ct = 0 then begin 
        (* If both of the constraints are inequalities, then report a new 
         * equality *)
        if not c1.iseq && not c2.iseq then begin
          (* Now we must make the proof of equality *)
          let peq () =
            let e = arithResToExp c1 in
            makeEqualityProof e L.zero 1 c1.proofs
                                       f1 f2c2.proofs;
          in
          let res = { c1 with iseq = true; proofs = [ (1, true, 
                                                       ProofThunk peq) ]} in
          assert (check (text "mergeConstraints: new equality") res);
          checkConstraintForDerivatives res;
          FoundEquality res
        end else 
          Trivial
      end else begin
        handleTrivialConstraint r'';
        Trivial
      end
    end else begin 
      if fullAssert then 
        assert (check (text "after merge, before checkForDerivatives")
                  r'');
      if r''.iseq then checkConstraintForDerivatives r'';
      SomeRes r''
    end

        (** Will remember here, in reverse order, the variables that we have 
         * eliminated and the list of constraints at that point. With each 
         * variable remember the constraints as they were before elimination. *)
  let eliminated: (int * arithRes list) list ref = ref []
      
  (** Eliminate equalities in a set of constraints *)
  let eliminateEq (cl: arithRes list) : arithRes list = 
    if !debug then 
      ignore (log "Eliminate EQ constraints from \n%a"
                (docList ~sep:nil (fun r -> dprintf "   %a\n" d_arithRes r))
                cl);

    (* separate the equalities from inequalities *)
    let eq : arithRes list ref = ref [] in
    let ineq : arithRes list ref = ref [] in
    let rec iter f = function [] -> () | h::t -> f h; iter f t in
    iter 
      (fun c -> if c.iseq then eq := c :: !eq else ineq := c :: !ineq)
      cl;

    (* Repeat as long as there are equalities *)
    let rec loop () = 
      match !eq with 
        [] -> ()
      | ec :: reste -> 
          eq := reste;
          (* Pick first variable from ec *)
          let ecw, f = try List.hd ec.facts with _ -> assert false in
          if !debug then 
            ignore (log "Eliminating variable %a from %a\n" 
                      d_factor f d_arithRes ec);
          eliminated := (f, [ec]) :: !eliminated;

          (* now take all the others and eliminate the variable *)
          let elimOne (r: arithRes) : arithRes option = 
            let rw = findFactorWeight f r in
            if rw = 0 then 
              Some r
            else
              let ecm, rm = smallestFactors ecw rw in
              assert (ecm * ecw = rm * rw);
              (* Must negate one of the factors. But the one for r must be 
               * positive, because it may be an inequality *)
              let ecm, rm = if rm < 0 then ecm, - rm else - ecm, rm in
              assert (rm > 0);
              assert (ecm * ecw + rm * rw = 0);
              match mergeConstraints ec ecm r rm with 
                SomeRes r' -> Some r'
              | Trivial -> None
              | FoundEquality r -> 
                  ignore (warn "ignoring equality %a\n" d_arithRes r);
                  None
          in
          let elimList (cl: arithRes list) : arithRes list = 
            let rec fold_left f acc = 
              function []-> acc | h::t -> fold_left f (f acc h) t in
            fold_left (fun acc c -> 
              match elimOne c with 
                None -> acc
              | Some c' -> c' ::acc)
              []
              cl
          in
          (* do the other equalities *)
          eq := elimList reste;
          (* do the inequalities *)
          ineq := elimList !ineq;
          (* continue as long as there are equalities *)

          if !debug then 
            ignore (log "After eliminating %a:\n%a"
                      d_factor f
                      (docList ~sep:nil 
                         (fun r -> dprintf "   %a\n" d_arithRes r))
                      (!eq @ !ineq));
            
          loop ()
    in
    loop ();
    !ineq

  (** Eliminate inequalities. Either raise Contradiction, or return some new 
   * equalities  *)
  let rec eliminateGeq (cl: arithRes list) : arithRes list = 
    (* We do not handle the equalities yet *)
    assert (not (List.exists (fun r -> r.iseq) cl));

    (** Keep here some equalities that we discover while we do elimination *)
    let newEquals: arithRes list ref = ref [] in

    (* Before we start to eliminate, do some integer tightening *)
    let rec map f = function [] -> [] 
                            | h::t-> let h' = f h in h' :: map f t in 
    let cl = map integerTightening cl in 

    let rec iterate (cl: arithRes list) : unit = 
      if cl <> [] then begin 
        if !debug then 
          ignore (log "Eliminate GEQ constraints from\n%a"
                    (docList ~sep:nil (fun r -> dprintf "   %a\n" d_arithRes r))
                    cl);
(*
        List.iter (fun c -> assert (check (text "elim GEQ") c)) cl;
*)

        (* find the best variable to eliminate *)
        let best: int = findNextFactor cl in 
        
        (* Eliminate best. First, separate the constraints in three classes: 
         * positive, negative, irrelevant *)
        let pos: (int * arithRes) list ref = ref [] in
        let neg: (int * arithRes) list ref  = ref [] in
        let irr: arithRes list ref  = ref [] in
        
        let rec iter f = function [] -> () | h::t -> f h; iter f t in
        iter (fun c -> 
          let w = findFactorWeight best c in
          if w = 0 then
            irr := c :: !irr
          else if w > 0 then 
            pos := (w, c) :: !pos
          else
            neg := (w, c) :: !neg)
          
          cl;
        
        eliminated := (best, cl) :: !eliminated;

        (* now take each pos and each neg and merge them, put the results in 
         * irr *)
        let rec iter f = function [] -> () | h::t -> f h; iter f t in
        iter 
          (fun (pw, p) -> 
            iter 
              (fun (nw, n) -> 
                let pf, nf = smallestFactors pw (- nw) in 
                match mergeConstraints p pf n nf with 
                  Trivial -> ()
                | SomeRes c -> irr := c :: !irr
                | FoundEquality c -> 
                    newEquals := c :: !newEquals
              )
              !neg
          )
          !pos;
        
        (* Loop over with irr *)
        iterate !irr
      end
    in

    iterate cl;
    !newEquals
    

  (** this is the main elimination procedure *)
  let eliminate (cl: arithRes list) : unit = 
    eliminated := [];

    let origcl = ref cl in 
    let rec loop () = 
      (* eliminate the equalities *)
      let cl' = eliminateEq !origcl in 
      (* eliminate the inequalities, and collect some new equalities *)
      
      let neweq = eliminateGeq cl' in 

      (* If we got here without raising an inequality, then we are about to 
       * fail. But perhaps we have found some equalities, try again *)
      if neweq <> [] then begin 
        if !debug then 
          ignore (log "About to fail, but retrying with some new equalities\n");
        origcl := neweq @ !origcl;
        loop ()
      end
    in
    loop ()

end


let surfaceOnly = ref false (** If we want to process surface arithmetic only *)


(** Intern a node and return its index *)
let internNode (n: node) : int = 
  assert (IH.mem processedNodes n.id);

  match isInternedOpt n with 
    Some i -> i.index 
  | None -> begin
      let idx = 1 + !max_index in 
      max_index := idx;
      
      if !debug then 
        ignore (log "interning node %a at index %d\n" d_nodeTop n idx);
      
      let ndata = { index = idx; node = n; 
                    isderived = None; derivatives = [] } in
      GA.setg indexData idx ndata; 
      IH.add internedNodes n.id ndata;


      (* Now initialize the row in the matrix *)
      let row = GA.getg matrix idx in
      for j = 0 to idx do
        GA.set row j None
      done;
      (* And the column *)
      for i = 0 to idx do 
        let ri = GA.get matrix i in 
        GA.setg ri idx None
      done;
      GA.set row idx (Some { min = 0;
                             start = idx; 
                             finish = idx; 
                             proof = [] });
      
      registerUndoRun (fun _ -> 
        IH.remove internedNodes n.id;
        assert (!max_index = idx);
        decr max_index;);
      
      (* Now see if we have somebody from the same class already interned *)
      let rootid = n.root.id in
      (match eqClassHasIntern rootid with 
        None -> 
          (* Mark this as an interned node *)
          IH.add eqClassInterns rootid  ndata;
          registerUndoRun (fun _ -> IH.remove eqClassInterns rootid);
          if !debug then
            ignore (log "   first node in its class to be interned\n")

      | Some rootint -> 
          (* Find the active class for that root *)
          let act, off, prfs (* : rootint - act - off = 0 *) 
              = findActiveIndex rootint.index in
          assert (n.root == rootint.node.root);

          setDerivative ndata act off 
            ((1, true, 
              ProofThunk (fun _ -> (!genEqProof) n rootint.node)) :: prfs);
      );

      idx
  end

  
(** Simplify a constraint to use only active nodes. Return a pointer-equal 
 * constraint if not changed *)
let simplifyDerivedNodes (r: arithRes) : arithRes = 
  let res = ref r in (* Accumulate here the result *)
  let changed = ref false in (* See if we have made changes *)
  let rec iter f = function [] -> () | h::t -> f h; iter f t in
  iter
    (fun (w, f) -> 
      let fact, off, prf (* : f - fact - off = 0 *) = findActiveIndex f in
      if fact.index <> f then begin
        changed := true;
        (* If fact == zeroIndex, then we add a constant, not a factor *)
        let ct', facts' = 
          if fact.index == !zeroIndex then 
            w * off, [ (-w, f) ]
          else
            w * off,
            canonicalFacts [ (-w, f); (w, fact.index) ]
        in 
        res := add !res { iseq = true;
                             ct = ct';
                             facts = facts';
                             proofs = scaleProofs (- w) prf;
                           };
      end)
    r.facts;

  if !changed then begin
    !res
  end else r

(* Intern a factor node and accumulate it to the results *)
let internNodeAndAccum (r: arithRes) (f: int) (n: node) : arithRes = 
  (* Here r.proofs := r.facts + r.ct - orig = 0 *)
  let idx = internNode n in 

  let r1 = { iseq = true;
             ct  = 0;
             facts = [ (f, idx) ];
             proofs = [] } in
  (* There may be derived nodes inside *)
  simplifyDerivedNodes (add r r1)


(** Add an edge to the matrix. It must be the case that owner(n1) - owner(n2) 
 * >= ct, with arithmetic and the [mkProof] **)
let addEdge (n1: int) (n2: int) (ct: int) 
            (prfs: arithproof list) : unit = 
  if !debug then 
    ignore (log "Adding an edge %a - %a >= %d\n" d_owner n1 d_owner n2 ct);
  (* Check that the proof is as needed *)
  assert (proofWellFormed (dprintf "addEdge %a - %a >= %d"
                             d_owner n1 d_owner n2 ct)
            (P.mkArithGeq
               (L.minus (expOfIndex n1) (expOfIndex n2))
                (L.integer ct)
                  (eqProofsToPargs prfs)));
                                             
  (* See if we have a contradiction *)
  (match smatrix n2 n1 with 
    Some ed when ed.min + ct > 0 -> 
      if !debug then 
        ignore (log "Found contradiction\n");
      let prf = 
        if !genProofs then begin
          (* make a proof that n2 - n1 >= ed.min *)
          let prf_n2_n1 = makeGeqProofList n2 n1 in
          (* make a proof that n1 - n2 >= ct *)
          let prf_n1_n2 = prfs in
          assert(proofWellFormed 
                   (dprintf "Shostak message for %a and %a"
                      d_owner n1 d_owner n2)
                   (P.mkArithGeq (expOfIndex n1)
                      (L.plus (expOfIndex n2) (L.integer ct))
                      (eqProofsToPargs prf_n1_n2)));
                               
          (* Make a proof that 0 >= ed.min + ct *)
(*
          let p = P.mkArithGeq L.zero
              (L.integer (ed.min + ct))
              (eqProofsToPargs 
                 (prf_n2_n1 @ prf_n1_n2)) in

          P.G("geqLt_FI", [ P.P p;
                            P.P (P.mkArithGeq (L.integer (ed.min + ct)) 
                                   (L.plus L.zero L.one) []) ])
*)
          P.mkArithContra (eqProofsToPargs
                             (prf_n2_n1 @ prf_n1_n2))
        end else 
          noProof
      in
      if !debugStats then 
        incr shContras;

      raise (Contradiction prf)

  | _ -> ());


  (** Remember here the list of changed entries. We store (i, j, off), where 
   * i - j >= off. *)
  let possibleEqualities : (int * int * int) list ref = ref [] in 
  let setEdge  (rowidx: int) 
               (row: edgeData option GA.t) 
               (colidx: int) 
               (min: int)
               (old: edgeData option) = 
    GA.set row colidx  
      (Some { min = min; 
              start = n1; 
              finish = n2; 
              proof = prfs });
    registerUndoRun (fun _ -> GA.set row colidx old);
    possibleEqualities := (rowidx, colidx, min) :: !possibleEqualities;
  in
    

  (** See if this is redundant **)
  let row_n1 = GA.get matrix n1 in 
  let _ = 
    match GA.get row_n1 n2 with
      Some ed when ed.min >= ct -> 
        if !debug then 
          ignore (log "Constraint is redundant\n")
            
    | old_n1_n2 -> begin 
        (* not contradictory and not redundant *)
        
        (* Relax this node *)
        setEdge n1 row_n1 n2 ct old_n1_n2; 
        
        (* For each node i with an edge to n1 *)
        let row_n2 = GA.get matrix n2 in 
        for i = 0 to !max_index do 
          let rowi = GA.get matrix i in 
          match GA.get rowi n1 with 
            Some i_n1 when isActiveIndex i -> 
              (* For each node j with an edge from n2 *)
              for j = 0 to !max_index do
                if j <> i then begin 
                  match GA.get row_n2 j with 
                    Some n2_j -> begin 
                      let new_i_j = i_n1.min + ct + n2_j.min in 
                      match GA.get rowi j with 
                        (* Tighten the edge with a larger value. Do not 
                         * tighten unless you really increase the value. This 
                         * way we avoid cycles. *)
                        Some i_j when not (isActiveIndex j) || 
                                      i_j.min >= new_i_j ->      ()
                      | old_i_j -> 
                          (* Relax if old edge is None, or (isActive j && 
                           * strictly increasing) *)
                          setEdge i rowi j new_i_j old_i_j
                    end
                  | _ -> ()
                end
              done
          | _ -> ()
        done; 
        ()
    end 
  in

  (* Now look for equalities *)
  let rec iter f = function [] -> () | h::t -> f h; iter f t in
  iter
    (fun (i, j, i_j) ->
      if false then 
        ignore (E.log "looking at possible equality %a = %a + %d\n"
                  d_owner i d_owner j i_j);
      (* We know that own(i) >= own(j) + i_j *)
      let rowj = GA.get matrix j in 
      match GA.get rowj i with 
        Some d when d.min = - i_j -> 
          (* Now we know that own(j) >= own(i) - i_j also *)
          let mkProof () = 
            let prf = 
              if !genProofs then begin
                let ie = expOfIndex i in 
                let je = expOfIndex j in 
                let j_i_j = 
                  if i_j = 0 then je else L.plus je (L.integer i_j) 
                in 
                let pfgeq = makeGeqProofList i j in 
                let pfleq = makeGeqProofList j i in 
                makeEqualityProof ie j_i_j 1 pfgeq 1 pfleq
(*
                let prf_n1_n2 = makeGeqProof i 0 j i_j in 
                let prf_n2_n1 = makeGeqProof j i_j i 0 in 
                (* Go through the proofs and find common ones, with opposite 
                 * signs *)
                P.G("eqi",  [ P.P prf_n1_n2; P.P prf_n2_n1 ])
*)
              end else
                noProof
            in
            prf
          in
          setDerivative (GA.get indexData i) 
                        (GA.get indexData j) 
                        i_j 
                        [ (1, true, ProofThunk mkProof) ]


      | _ -> ())
    !possibleEqualities


let handleShostakConstraint (r: arithRes) : bool = 
  (* See if these are of a form we need *)
  match r.facts with 
    [] -> (* This is either redundant or a contradiction *)
      handleTrivialConstraint r; 
      true

          (* Now some specific forms that Shostak understands *)

  | _ -> begin
      (* do a difference *)
      let diff (np: int) (nm: int) (ct: int) (prfs: arithproof list) = 
        if not r.iseq then 
          addEdge np nm (- r.ct) r.proofs
        else begin 
          addEdge np nm (- r.ct) r.proofs; 
          addEdge nm np    r.ct  (scaleProofs (-1) r.proofs); 
        end
      in
      match r.facts with 
        (1, np) :: [] -> diff np !zeroIndex (- r.ct) r.proofs; true
      | (-1, nm) :: [] -> diff !zeroIndex nm (- r.ct) r.proofs; true
      | (1, np) :: (-1, nm) :: [] -> diff np nm (- r.ct) r.proofs; true
      | (-1, nm) :: (1, np) :: [] -> diff np nm (- r.ct) r.proofs; true
      | _ -> false
  end



let handleConstraint (r: arithRes) : unit = 
  let r' = Fourier.integerTightening (simplifyDerivedNodes r) in
  if not (handleShostakConstraint r') then begin
    if !debug then 
      ignore (log "saving constraint for later\n");
    let oldsaved = !savedConstraints in 
    savedConstraints := r' :: oldsaved;
    registerUndoRun (fun _ -> savedConstraints := oldsaved);
  end


      (* Now we must do elimination. We must add in the constraints from the 
       * graph and then we forget about the graph *)
let moveGraphToSavedConstraints () = 
  for i = 0 to !max_index do
    if isActiveIndex i then begin
      let rowi = GA.get matrix i in 
      for j = 0 to !max_index do 
        if i <> j && isActiveIndex j then begin
          match GA.get rowi j with 
            Some d -> 
              (* Computes the facts and constant. We rely on the fact 
               * that the only constant active is 0, and its index is 
               * smallest *)
              let facts = 
                assert (i >= !zeroIndex && j >= !zeroIndex && i <> j);
                if i < j then 
                  if i = !zeroIndex then 
                    [ (-1, j) ]
                  else
                    [ (1, i); (-1, j) ]
                else
                  if j = !zeroIndex then 
                    [ (1, i) ]
                  else
                    [ (-1, j); (1, i) ]
              in
              let r = 
                { iseq = false; 
                  facts = facts;
                  ct = - d.min; 
                  proofs = makeGeqProofList i j 
                } 
              in
              assert (check (text "constraint extracted from matrix") r);
              savedConstraints := r :: !savedConstraints
                                        
          | _ -> ()
        end
      done;
    end
  done

let processSavedConstraints () : unit = 
  (* Go over the constraints and simplify some derived ones *)
  if !debug then begin
    ignore (log "Shostak is about to give up in state:\n");
    printState ();
  end;
           
  if !savedConstraints <> [] then begin

    (* First, simplify the derived ones because we might have created some 
     * more derived nodes since we simplified them. *)
    let saved = !savedConstraints in 
    savedConstraints := [];
    (* Before we do anything, we make sure that we can restore the state of 
     * the savedConstraints *)
    registerUndoRun (fun _ -> savedConstraints := saved);

    let rec iter f = function [] -> () | h::t -> f h; iter f t in
    iter 
      (fun r -> 
        let r' = Fourier.integerTightening (simplifyDerivedNodes r) in
        if r' == r || not (handleShostakConstraint r') then 
          savedConstraints := r' :: !savedConstraints)
      saved;

    if !savedConstraints <> [] then begin
      if !debug then begin
        ignore (log "After simplifying derived nodes:\n");
        ignore (E.log "  with saved constraints:\n%a"
                  (docList ~sep:nil
                     (fun c -> dprintf "  %a\n" d_arithRes c)) 
                  !savedConstraints);
      end;

      moveGraphToSavedConstraints ();


      Fourier.eliminate !savedConstraints
    end
  end

(** Keep here the integer nodes that we have encountered *)
let integerNodes: node IH.t = IH.create 13

(*****
 *****
 *****   Scan EDAG nodes (from messages) and find the nodes that we must 
 *****   intern 
 *****
 *****)

(** Scan a node and its arguments and process the arithmetic inside *)
let rec processInterestingArithmetic (n: node) : unit = 
  (* We do not process a node twice *)
  if not (markProcessed n) then begin 
    match fSymOfHead n.head with 
      INT _ | BINOP (L.Plus|L.Minus|L.Times) -> 
        (* This is an interesting arithmetic node. Evaluate its arithmetic, 
         * intern the factors, and processInterestingArithmetic inside the 
         * factors *)
        let res = evalArithmeticNode (emptyRes ~eq:true) 1 n in
        (* Now we have res.eqproofs: res.facts + res.ct - n *)

        if !debug then 
          ignore (log "evaluated interesting node %a as: %a\n"
                    d_nodeTop n d_arithRes res);


        (** Now intern n itself. Compute res - n. Now it must be the case 
         * that res_i is equivalent to 0, modulo arithmetic and the proofs in 
         * res. *)
        let res_i = internNodeAndAccum res (-1) n in 

        (* Check the proof *)
        assert(check (dprintf "interesting node (eq) n=%a res=%a"
                        d_nodeTop n d_arithRes res_i)
                 res_i);

        (* Now mark this equality *)
        handleConstraint res_i

    | _ -> (* Not arithmetic, scan its arguments *)
        let rec iter f = function [] -> () | h::t -> f h; iter f t in 
        iter processInterestingArithmetic n.args
  end

(** Evaluate an arithmetic node. See the comments for arithRes for the 
 * property that this function establishes.  *)
and evalArithmeticNode (acc: arithRes) (w: int) (n: node) : arithRes = 
  let isFactor () = 
    (* This is a factor. process the interesting arithmetic that it might 
     * have inside, then intern it, and accumulate to the result *)
    if not !surfaceOnly then processInterestingArithmetic n;
    internNodeAndAccum acc w n
  in
    
  match fSymOfHead n.head with 
    INT i -> accumConst acc (w * i)
  | BINOP op -> begin
      let a1, a2 = 
        match n.args with [a1; a2] -> a1, a2
        | _ -> E.s (bug "invalid BINOP node")
      in
      match op with 
        L.Plus -> evalArithmeticNode (evalArithmeticNode acc w a1) w a2
      | L.Minus -> evalArithmeticNode (evalArithmeticNode acc w a1) (- w) a2
      | L.Times -> begin
          match fSymOfHead a1.head, fSymOfHead a2.head with 
            INT i1, _ -> evalArithmeticNode acc (w * i1) a2
          | _, INT i2 -> evalArithmeticNode acc (w * i2) a1
          | _ -> isFactor ()
      end
      | _ -> isFactor ()
  end
  | _ -> isFactor ()

let evalArithMinus (a1: node) (a2: node) (ct: int) : arithRes = 
  evalArithmeticNode 
    (evalArithmeticNode 
       (accumConst (emptyRes ~eq:false) (- ct)) 1 a1) (- 1) a2

  

(** We get called back when a node that used to be root has changed roots. 
 * This node used to be a root. *)
let checkRootChange (n: node) = 
  if n.id == n.root.id then () 
  else begin 
    (* This node has changed root. See if we had anybody interned in its old 
     * class. *)
    match eqClassHasIntern n.id with 
      None -> (* nobody was interned in that class. Ignore *)
        ()

    | Some nintern -> begin 
      (* We had at least one interned node in that class. See if we had nodes 
       * interned in the new class *)
        if !debug then 
          ignore (log "Callback because node %a (interned at %d) has changed root to %a\n" 
                    d_nodeTop n nintern.index d_nodeTop n.root);

        let rootid = n.root.id in 
        match eqClassHasIntern rootid with 
          None -> 
            (* The class this was merged to had nobody interned *)
            if !debug then 
              ignore (log "   new root had nobody interned\n");
            IH.add eqClassInterns rootid nintern;
            registerUndoRun (fun _ -> IH.remove eqClassInterns rootid)

        | Some ointern -> 
            assert (nintern.node.root == ointern.node.root);
            (* Make the nintern node a derivative of ointern *)

            if !debug then 
              ignore (log "   new root is interned as %a\n" 
                        d_owner ointern.index);

            (* Assert the equality, to ensure that they can be equal, and to 
             * propagate new equalities. Try to use evalMinus, to detect 
             * the case when the fact is already known  *)
            let r = evalArithMinus nintern.node ointern.node 0 in 
 

            (* Prepare a proof generator for the equalities *)
            let mkeqprf = 
              ProofThunk (fun _ -> 
                assert !genProofs;
                (!genEqProof) nintern.node ointern.node)
            in
            (*
               ignore (log "   eval = %a\n" d_arithRes r);
             *)
            (* Add the proof *)
            let r' = { r with iseq = true; 
                              proofs = (1, true, mkeqprf) :: r.proofs } in 

            handleConstraint r';

    end
  end



(** Process a node. Record the node *)
let assume (msg: message) : unit  = 
  (* maybe at this point we should send everything to Simplex ? *)
  
  if msg.owner != shid then begin
    if !debug then
      ignore (log "seeing message %a (goal: %b)\n" 
                d_messagebody msg.body msg.fromgoal);

    let body' = msg.body in

    (** Process an inequality: n1 - n2 >= ct, with a proof of an equivalent 
     * inequality (up to simple arithmetic simplification) **)
    let processGeq 
        (n1: node) (n2: node) (ct: int) 
        (proof: P.proof) = 
      let res = evalArithMinus n1 n2 ct in 
      
      if !debug then 
        ignore (log "processGeq: n1=%a, n2=%a, res = %a\n" 
                  d_nodeTop n1 d_nodeTop n2 d_arithRes res);
      
      assert (
      let e = L.minus (L.minus n1.exp n2.exp) (L.integer ct) in 
      proofChecker 
        (text "Shostak.processGeq") 
        (P.mkArithGeq e L.zero [ P.I 1; P.P proof])
        (L.geq e L.zero));

      (* Add in the proof of the inequality *)
      handleConstraint { res with 
                         proofs = (1, false, Proof proof) :: res.proofs };
      
      
      if !debug then begin
        ignore (log "Shostak state after Geq:\n");
        printState ()
      end
    in

    match body' with 
      Intern (L.SetGe, n1, n2) -> 
        processGeq n1 n2 0  msg.Edag.proof
          
    | Intern (L.SetGt, n1, n2) -> 
        processGeq n1 n2 1 
          (P.G("gtgeq", [ P.P msg.Edag.proof ]))


    | Intern (L.SetLe, n1, n2) -> 
        processGeq n2 n1 0 
          (P.G("leqgeq", [ P.P msg.Edag.proof ]))

    | Intern (L.SetLt, n1, n2) -> 
        processGeq n2 n1 1 
          (P.G("ltgeq", [ P.P msg.Edag.proof ] ))


    | Intern (bop, n1, n2) -> (* including equalities *)
        processInterestingArithmetic n1;
        processInterestingArithmetic n2;
        if !debug then begin
          ignore (log "Shostak state after uninteresting message:\n");
          printState ()
        end

    | _ -> ()
  end;

  ()



(** Test whether we know the answer for a message *)
let testMessage (msg: messagebody) : testResult  = 

    match msg with 
    | Extern _ -> NotDetermined
    | Intern (bop, n1, n2) -> begin
        match isSomebodyInClassInterned n1, isSomebodyInClassInterned n2 with 
          Some i1, Some i2 -> begin
            
            (** A thunk that can make the proof that sign * (n1 - i1 - n2 + 
             * i2) = 0. *)
            let mkTransProof (sign: int) : arithproof list = 
              (* Adjust i1prf for the difference between i1a.node and n1 *)
              let n1_i1 = 
                if i1.node != n1 then 
                  [(sign, true,
                    Proof (!genEqProof n1 i1.node))] else []
              in 
              (* Adjust i2prf for the difference between i2a.node and n2 *)
              if i2.node != n2 then 
                (sign, true, Proof (!genEqProof i2.node n2)) 
                :: n1_i1 else n1_i1
            in

            (* Test an inequality: i1 - i2 >= ct. Invoke the ctrue 
             * continuation with an appropriate series of proofs, if true. 
             * Invoke the cfalse continuation with a series of proofs for i2 
             * - i1 >= - ct + 1, if false. *)
            let testGeq (i1: internData) (i2: internData) (ct: int) 
                (ctrue: arithproof list -> testResult)
                (cfalse: arithproof list -> testResult) : testResult = 

              let i1a, i1c, i1prf (* i1 - i1a - i1c = 0 *) 
                  = findActiveIndex i1.index in 
              let i2a, i2c, i2prf (* i2 - i2a - i2c = 0 *) 
                  = findActiveIndex i2.index in 

              (* Look in the matrix *)
              assert (i1a != i2a || ((smatrix i1a.index i2a.index != None)
                                       && (smatrix i2a.index i2a.index != None)));
              match smatrix i1a.index i2a.index with 
                Some ed when ed.min + i1c - i2c >= ct -> 
                  ctrue (makeGeqProofList i1a.index i2a.index @ i1prf @ scaleProofs (-1) i2prf)
              | _ -> begin 
                  match smatrix i2a.index i1a.index with 
                    Some ed when ed.min + i2c - i1c >= - ct + 1 -> 
                      cfalse (makeGeqProofList i2a.index i1a.index @ i2prf @ scaleProofs (-1) i1prf)

                  | _ -> NotDetermined
              end
            in

            (** Test an equality: i1 - i2. If true, then call ctrue with the 
             * needed proofs. If false, then call cfalse with the proofs for 
             * i2 - i1 > 0 or i1 - i2 > 0. In the latter case the boolean 
             * argument will be false *)
            let testEq 
                (i1: internData) (i2: internData) 
                (ctrue: arithproof list -> testResult)
                (cfalse: (bool -> arithproof list -> testResult)) : testResult = 
              
              let i1a, i1c, i1prf (* i1 - i1a - i1c = 0 *) 
                  = findActiveIndex i1.index in 
              let i2a, i2c, i2prf (* i2 - i2a - i2c = 0 *) 
                  = findActiveIndex i2.index in 
(*
              ignore (log "Looking at i1=%a i1a=%a i1c=%d\n" 
                        d_owner i1.index d_owner i1a.index i1c);
              ignore (log "Looking at i2=%a i2a=%a i2c=%d\n" 
                        d_owner i2.index d_owner i2a.index i2c);
              printState ();
*)
              if i1a == i2a then begin
                if i1c = i2c then 
                  ctrue (i1prf @ scaleProofs (-1) i2prf)
                else 
                  if i1c > i2c then 
                    cfalse false (i1prf @ scaleProofs (-1) i2prf)
                  else
                    cfalse true (i2prf @ scaleProofs (-1) i1prf)
              end else begin 
                (* Look in the matrix *)
                match smatrix i1a.index i2a.index with 
                  Some ed when ed.min + i1c - i2c > 0 -> 
                    cfalse false (makeGeqProofList i1a.index i2a.index @ i1prf @ scaleProofs (-1) i2prf)
                | _ -> begin 
                    match smatrix i2a.index i1a.index with 
                      Some ed when ed.min + i2c - i1c > 0 -> 
                        cfalse true (makeGeqProofList i2a.index i1a.index @ i2prf @ scaleProofs (-1) i1prf)
                          
                    | _ -> NotDetermined
                end
              end
            in
              
            let res = 
              match bop with 
               L.SetGe -> 
                 testGeq i1 i2 0
                   (fun pftrue -> 
                     Determined(true, 
                                P.mkArithGeq n1.exp n2.exp
                                  (eqProofsToPargs (mkTransProof 1 @ pftrue))))
                   (fun pfalse -> 
                     Determined(false, 
                                P.G("geqlt", [ P.P (P.mkArithGeq n2.exp
                                                      (L.plus n1.exp L.one)
                                                      (eqProofsToPargs (pfalse @ 
                                                                                mkTransProof (-1))))])))
              | L.SetGt -> 
                  testGeq i1 i2 1
                    (fun pftrue -> 
                      Determined(true, 
                                 P.G("geqgt", [ P.P (P.mkArithGeq n1.exp
                                                       (L.plus n2.exp L.one)
                                                       (eqProofsToPargs (mkTransProof 1 @ pftrue)))])))
                    (fun pfalse -> 
                      Determined(false, 
                                 P.G("geqleq", [ P.P (P.mkArithGeq
                                                        n2.exp n1.exp
                                                        (eqProofsToPargs (pfalse @ 
                                                                                  mkTransProof (-1))))])))
              | L.SetLe -> 
                  testGeq i2 i1 0
                    (fun pftrue -> 
                      Determined(true, 
                                 P.G("geqleq", 
                                     [ P.P (P.mkArithGeq
                                              n2.exp n1.exp
                                              (eqProofsToPargs (mkTransProof (-1) @ pftrue)))])))
                    (fun pfalse -> 
                      Determined(false, 
                                 P.G("geqgt", 
                                     [ P.P (P.mkArithGeq
                                              n1.exp (L.plus n2.exp L.one)
                                              (eqProofsToPargs (pfalse @ 
                                                                        mkTransProof 1)))])))
              | L.SetLt -> 
                  testGeq i2 i1 1
                    (fun pftrue -> 
                      Determined(true, 
                                 P.G("geqlt", 
                                     [ P.P (P.mkArithGeq n2.exp
                                              (L.plus n1.exp L.one)
                                              (eqProofsToPargs (mkTransProof (-1) @ pftrue)))])))
                    (fun pfalse -> 
                      Determined(false, 
                                 P.mkArithGeq n1.exp n2.exp
                                   (eqProofsToPargs (pfalse @ 
                                                     mkTransProof 1))))
                    
              | L.SetEq -> 
                  testEq i1 i2
                    (fun pftrue -> 
                      Determined(true, 
                                 P.mkArithEq n1.exp n2.exp
                                   (eqProofsToPargs (pftrue @ mkTransProof 1))))
                    (fun i2_gt_i1 pfalse -> 
                      let p =
                        if i2_gt_i1 then 
                          P.Symm (P.G("gtNeq", 
                                      [ P.P (P.mkArithGeq n2.exp
                                               (L.plus n1.exp L.one) 
                                               (eqProofsToPargs (pfalse @ mkTransProof (- 1))))]))
                        else 
                          P.G("gtNeq", 
                              [ P.P (P.mkArithGeq n1.exp
                                       (L.plus n2.exp L.one)
                                       (eqProofsToPargs (pfalse @ mkTransProof 1)))])
                      in
                      Determined(false, p))

              | L.SetNe -> 
                  testEq i1 i2
                    (fun pftrue -> 
                      Determined(false, 
                                 P.mkArithEq n1.exp n2.exp
                                   (eqProofsToPargs (pftrue @ mkTransProof 1))))
                    (fun i2_gt_i1 pfalse -> 
                      let p =
                        if i2_gt_i1 then 
                          P.Symm (P.G("gtNeq", 
                                      [ P.P (P.mkArithGeq n2.exp
                                               (L.plus n1.exp L.one)
                                               (eqProofsToPargs (pfalse @ mkTransProof (- 1))))]))
                        else 
                          P.G("gtNeq", 
                              [ P.P (P.mkArithGeq n1.exp
                                       (L.plus n2.exp L.one)
                                       (eqProofsToPargs (pfalse @ mkTransProof 1)))])
                      in
                      Determined(true, p))


              | _ -> NotDetermined
            in
            (* Check the result *)
            if fullAssert then 
              assert 
                (let p = (L.PNez(L.BinOp(bop, n1.exp, n2.exp))) in 
                match res with 
                  NotDetermined -> true
                | Determined(true, ptrue) -> 
                    assert (proofChecker 
                              (text "Shostak.testMessage: IsTrue") 
                              ptrue p); 
                    true
                | Determined(false, pfalse) -> 
                    assert (proofChecker
                              (text "Shostak.testMessage: IsFalse") 
                              pfalse  
                              (L.negatePred p));
                    true);

            res
          end (* isInterned *)
        | _ -> NotDetermined
    end   


let postAssume (aboutToFail: bool) = 
  if aboutToFail then
    processSavedConstraints ()

let init () = 
  IH.clear internedNodes;
  IH.clear processedNodes;
  IH.clear eqClassInterns;
  savedConstraints := [];
  GA.reset_max_init_index indexData;
  for i = 0 to GA.max_init_index matrix do 
    GA.reset_max_init_index (GA.get matrix i)
  done;
  GA.reset_max_init_index matrix;

  if not emitDiseq then registerMessageTester testMessage;

  IH.clear integerNodes;

  (* Intern zero *)
  let nzero, _ = insertExpNode L.zero in
  ignore (markProcessed nzero); (* mark the node processed *)
  zeroIndex := internNode nzero;

  (* Now register the callback with the EDAG *)
  addGlobalMergeCallback checkRootChange;

  ()

let canEnable (proposed: U.symbol list) : bool = 
  (* Do not enable if Simplex is enabled *)
  if List.mem (U.registerSymbolName "SX") proposed then begin
    ignore (warn "Will not enable Shostak because Simplex is enabled\n");
    false
  end else 
    true

let witness () = 

  let debugWitness = false in

  if debugWitness then 
    printState ();

  let fdata = { 
    toexp = (fun idx -> findSmallestForNode (nodeOfIndex idx));
    todoc = (fun i -> assert false)
  } in
  (* Emit all the savedConstraints. We are here only if we have done 
   * elimination *)

  (* If we do not have savedConstraints, it means that we might not have done 
   * Fourier elimination. This means that all constraints are in the graph. 
   * In order to have a uniform witness extraction, pretend that we do 
   * elimination *)
  let origSavedConstraints = !savedConstraints in

  if !savedConstraints == [] then begin
    if !debug then 
      ignore (log "Will do FM elimination to get the witness\n");
    moveGraphToSavedConstraints ();
    (try Fourier.eliminate !savedConstraints
    with (Contradiction _) as e ->  
      ignore (warn "Fourier.eliminate raised %s when looking for witness\n"
                (Printexc.to_string e));
      assert false);
  end;

  (* Restore the saved constraints *)
  savedConstraints := origSavedConstraints;

  (* Now process the eliminated list and figure out some values for the 
   * variable *)
  let witnessTable: float IH.t = IH.create 13 in (** Keep here the witness *)

  (* Add zero to it *)
  IH.add witnessTable !zeroIndex 0.0;

  let prepareWitness (f: int) (v: float) : unit = 
    let v' = 
      if floor v = v then 
        L.integer (truncate v) 
      else
        L.Fun(Printf.sprintf "%.2f" v, [])
    in
    match fdata.toexp f with 
      L.Int _ | L.BinOp((L.Plus|L.Minus), _, _) -> ()
    | ef -> 
        let p = L.eq ef v' in 
        accumWitnessPred p;
        if debugWitness then 
          ignore (log "found witness %a\n" L.d_pred p); 
        ()
  in

  let rec iter l f = match l with [] -> () | h::t -> f h; iter t f in
  iter
    !Fourier.eliminated
    (fun (v, constr) -> 
      if debugWitness then 
        ignore (log "Recovering witness for %a from\n  @[%a@]\n"
                  d_owner v 
                  (docList ~sep:line (d_arithRes ())) constr);
      let min: float option ref = ref None in (* Smallest value *)
      let updMin (v: float) = 
        match !min with 
        | Some v' when v' > v -> ()
        | _ -> min := Some v
      in
      let max: float option ref = ref None in (* Largest value *)
      let updMax (v: float) = 
        match !max with 
        | Some v' when v' < v -> ()
        | _ -> max := Some v
      in
      
      (* Find a witness for v, from the given constraints. Evaluate each 
       * constraint and keep track of how it constraints the value of v *)
      let doOneConstraint (c: arithRes) : unit = 
        (* Evaluate the facts. So far we have a*v+b *)
        let rec evalFacts (a: int) (b: float) = function
            [] -> (a, b)
          | (w, v') :: restf -> 
              if v' = v then 
                evalFacts (a + w) b restf
              else begin
                let valv' = 
                  try IH.find witnessTable v'
                  with Not_found -> begin 
                    (* We have a constraint, for which we do not have a 
                     * witness. Choose 0 *)
                    IH.add witnessTable v' 0.0;
                    prepareWitness v' 0.0;
                    0.0
                  end
                in
                evalFacts a (b +. (float_of_int w) *. valv') restf
              end
        in
        let a, b = evalFacts 0 (float_of_int c.ct) c.facts in 
        (* Now we know that a*v+b >= 0 *)
        if a = 0 then 
          (* Not a relevant constraint. But check it anyway *)
          assert (if c.iseq then b = 0.0 else b >= 0.0)
        else begin 
          (* A relevant constraint *)
          let bound = -. b /. (float_of_int a) in 
          if a > 0 || c.iseq then updMin bound;
          if a < 0 || c.iseq then updMax bound;
        end
      in
      iter constr doOneConstraint;
      let valv = 
        match !min, !max with 
          Some min, Some max -> 
            if min > max then assert false;
            let amin = abs_float min in 
            let amax = abs_float max in 
            if amin <= amax then min else max
        | Some min, None -> min
        | None, Some max -> max
        | _ -> assert false
      in
      prepareWitness v valv;
      assert (not (IH.mem witnessTable v));
      IH.add witnessTable v valv);



  (* Add the equalities that we have in the Shostak graph *)
  for i = 0 to !max_index do 
    let id = GA.get indexData i in
    let ei = findSmallestForNode id.node in 
    match id.isderived with 
    | Some (base, off, _) when off <> 0 -> begin
        (* base must be in the index table *)
        try 
          let valbase = IH.find witnessTable base.index in
          prepareWitness id.index (valbase +. (float_of_int off))
        with Not_found -> begin
          match fdata.toexp id.index with
            L.Int _ | L.BinOp((L.Plus|L.Minus), _, _) -> ()
          | ide  ->
              let p = 
                L.eq ide
                  (L.plus (fdata.toexp base.index) (L.integer off)) in 
              if debugWitness then 
                ignore (log "Found expression witness: %a\n" 
                          L.d_pred p);
              accumWitnessPred p; 
        end
          
    end
    | _ -> ()
  done;

  ()

(* list of entry points exported as a satproc *)
let entries: Satproc.entries = {
  emptySatproc with 
  canEnable = canEnable;
  init      = init;
  Satproc.finish    = (fun _ -> 
    if !debugStats then begin
      ignore (E.log "shContra: %d\n" !shContras);
    end);
  assume    = assume;
  postAssume = postAssume;
  debug     = debug;
  ident     = shid;
  witness   = witness;
  descr     = "shostak";
  options   = 
  [ 
(*
    { O.optInUI = "";
      O.optRestart = false;
      O.optKind = O.OUnit;
      O.optExtra = (fun _ -> ());
      O.optCommandLine = "\n\t ****** Shostak options \n\t\t";
      O.optHelp = "";
    };
    { O.optInUI = "";
      O.optRestart = false;
      O.optKind = O.OBool WithSimplex.enable;
      O.optExtra = (fun _ -> ());
      O.optCommandLine = "-fallback-arith";
      O.optHelp = "do not fall-back on powerful arithmetic";
    };
*)
  ]
}
