module L = Logic
module S = Satproc
module E = Errormsg
module P = Proof
module U = Util
module IH = Inthash

open Pretty
open Edag

let sxid = Util.registerSymbolName "SX" (* Set in the initializer *)

let init_rows = 8
let printLiveOnly = ref false
let doPrint = ref false
let pTableau = ref true
let threshHold = 0.00001
let intHeurRounds = ref 4


let log fmt = ignore (E.log "SX: "); E.log fmt
let nZero = ref dummyNode
let nOne = ref dummyNode

      
    (* These are the row/column owners *)
type owner =  { mutable node  : ownode;   (* Whether a node in the graph owns 
                                           * this row/column *)
	              oid    : int;	(* This is only for printing *)
	              ospos  : simplPos ref; (* this position is shared with 
                                              * the ref cell from the node, 
                                              * if the "node" is a node *)
	        mutable restr : whyrestr option; (* If the row/column is 
                                                * restricted to be >= 0, and 
                                                * why *)

                mutable killedby: owner option; (* If killed, then the owner 
                                                 * that killed it *)
              }

and ownode = 
    NoNode (* No owner *)
  | Node of node (* A given EDAG node own *)
  | Slack of int * (int * node) list (* Slack (c + i1*n1 + ...) *)

(** Why is a node restricted to be >= 0 ? *)
and whyrestr =
    BYEQ  of message * L.exp * L.exp * (unit -> P.proof) 
        (* Restricted because of an equality. The restriction is e1 >= e2. A 
         * proof of e1 = e2 can be obtained with the given thunk. *)

  | BYGEQ of message * L.exp * L.exp * (unit -> P.proof) 
        (* Restricted because of an inequality message. The restriction is e1 
         * >= e2, and a proof of it can be obtained with the given thunk. *)


  | GEQINT of owner * L.exp * (unit -> P.proof) 
        (* An owner that corresponds to "e", and a proof that e >= 0, due to 
         * an integer restriction *)

      
let dummyOwner : owner = { node = NoNode;
			   oid   = -1;
			   ospos = ref (Row (-1));
                           killedby = None;
			   restr = None }
      
    (* This is the current Simplex state *)
let max_rows: int ref = ref init_rows
let max_cols: int ref = ref init_rows


(* Column 0 of the tableau corresponds to constants. *)
(* For each row of the tableau, the owner of the row can be written as the 
 * weighted combination of the column owners given the coefficients in the 
 * row *)
(** The first few rows and columns in a tableau are dead, meaning that 
 * their owners are known to be equal to 0 *)
let data    : float array ref =
      ref (Array.make (!max_rows * !max_cols) 0.0)
let rows    : owner array ref =
      ref (Array.make !max_rows dummyOwner)
let cols    : owner array ref =
      ref (Array.make !max_cols dummyOwner)
      
let nr_rows : int ref = ref 0  (** Number of rows active in the tableau *)
let nr_cols : int ref = ref 1
let livecol : int ref = ref 1  (** The first live column *)

let liverow : int ref = ref 0  (** The first live row *)

let divisors: float list ref = ref []
      
    (* Owner functions *)
let nodeZeroOwner : owner ref = ref dummyOwner
let newNodeOwner (n : node) : owner =
  { node = Node n;
    ospos = n.spos;   (* We share the ref-cells *)		
    oid   = n.id;
    killedby = None;
    restr = None }
      
let newSlackOwner (ct: int) (facts: (int * node) list) : owner =
  { node = Slack (ct, facts);
    ospos = ref NotSimpl;
    oid   = newNodeId ();
    killedby = None;
    restr = None }
      
      
let eqOwner (o1: owner) (o2: owner) = o1.oid = o2.oid
      
let getPos (o : owner) : simplPos = ! (o.ospos)
let setPos (o : owner) (pos : simplPos) = o.ospos := pos
      
let setOnPos (own : owner) (pos: simplPos) : unit =
  own.ospos := pos;
  match pos with
    Row i -> (!rows).(i) <- own
  | Col j -> (!cols).(j) <- own
  | _ -> E.s (E.bug "setOnPos: not Simplex")


  
let removeOwner (own : owner) : unit =
  let _ =
    match ! (own.ospos) with
      Row i -> (!rows).(i) <- dummyOwner
    | Col j -> (!cols).(j) <- dummyOwner
  | _ -> E.s (E.bug "removeOwner: not Simplex")
  in
  own.ospos := NotSimpl;
  own.restr <- None

let ownedByRow i : owner = (!rows).(i)
let ownedByCol j : owner = (!cols).(j)

let ownedBy = function 
    Row i -> ownedByRow i
  | Col j -> ownedByCol j
  | _ -> E.s (E.bug "ownedBy: not Simplex")
      
      
let getNode (own : owner) = own.node
let getOwner (n : node) : owner =
  match ! (n.spos) with
    Row i -> ownedByRow i
  | Col j -> ownedByCol j
  | _ -> E.s (E.bug "getOwner: not Simplex")

let isRestr (own : owner) = own.restr <> None
	
let getWhyrestr (o : owner) : whyrestr = 
  match o.restr with
    Some h -> h
  | _ -> E.s (E.bug "getWhyrestr: not restricted")

let isRestrRow i : bool = isRestr (!rows).(i)
let isRestrCol j : bool = isRestr (!cols).(j)

let rec for_a (f: int -> 'a -> 'a) (start: 'a) (fst, last) = 
  if fst > last then 
    start
  else
    for_a f (f fst start) (fst + 1, last)

      (*
      
let printOwner ({node=NoNode,id,restr,...}:owner) : unit = 
  (print "T";ProverUtils.iprint id;
   if isSome(!restr) then print "*" else print " ")
  | printOwner ({node=List _,id,restr,...}:owner) : unit = 
    (print "S";ProverUtils.iprint id;
     if isSome(!restr) then print "*" else print " ")
  | printOwner ({node=Node n,restr,...}:owner) =
    (printNodeId n;
     if isSome(!restr) then print "*" else print "")
      *)

let d_owner () (own: owner) : doc = 
  match own.node with 
    NoNode -> dprintf "T%d" own.oid
  | Node n -> 
      let restr = if own.restr <> None then "*" else "" in
      if n.args = [] then 
        dprintf "%s%s" (headToString n.head) restr
      else 
        dprintf "%d%s" n.id restr

  | Slack _ -> dprintf "S%d" own.oid

(* Build the expression corresponding to an owner *)
let buildOwnExpression (own: owner) : L.exp = 
  match own.node with 
    Node t ->  t.exp
        
  | Slack (eint, facts) -> begin
      let addExpr = function 
        | (z, e) when z = L.zero -> e
        | (e, z) when z = L.zero -> e
        | (L.Int n1, L.Int n2) -> L.integer (L.to_int n1 + L.to_int n2)
	| (e1, e2) -> L.plus e1 e2
      in
      
      let subExpr = function 
	| (e, z) when z = L.zero -> e
	| (L.Int n1, L.Int n2) -> L.integer (L.to_int n1 + L.to_int n2)
	| (e1, e2) -> L.minus e1 e2
      in
      let rec fold_left f acc = 
        function []-> acc | h::t -> fold_left f (f acc h) t in
      addExpr (L.integer eint,
               fold_left 
                 (fun acc (f, n) -> 
                   let toadd = 
                     if f = 1 then n.exp else L.times (L.integer f) n.exp
                   in
                   addExpr (toadd, acc))
                 L.zero
                 facts)
  end
  | _ -> 
      ignore (log "Cannot find expression for %a\n" d_owner own);
      E.s (E.bug "simplex: buildExpr")
  

let buildOwnLinComb (own: owner) : int * (int * node) list = 
  match own.node with 
    Node t -> 0, [(1, t)]
  | Slack (eint, facts) -> eint, facts
  | _ -> assert false

let d_whyrestr () = function 
    BYEQ (msg, _, _, _) -> dprintf "BYEQ: %a" d_message msg
  | BYGEQ (msg, e1, e2, _) -> dprintf "BYGEQ: %a" d_message msg
  | GEQINT (own, _, _) -> dprintf "GEQINT: %a" d_owner own
      
let d_rowOwner () i = d_owner () (!rows).(i)
let d_colOwner () j = d_owner () (!cols).(j)
    
      
    (* Tableau functions *)
let initTableau () =
  let dt = !data in
  let mc = !max_cols in
  let mr = !max_rows in
  for i = 0 to mr - 1 do
    for j = 0 to mc - 1 do 
      dt.(j + i * mc) <- 0.0
    done
  done
(*
  ProverUtils.for
    (fn i ->
      ProverUtils.for
	(fn j ->
	  RealArray.update(dt, j + i * mc, 0.0))
	  (0,mc-1))
      (0,mr-1)
*)  
    
      (* Index in the tableau *)
let stab i j = 
  let res = (!data).(j + i * !max_cols) in
(*  ignore (log "Reading (%d, %d) = %f\n" i j res); *)
  res
  
let utab i j (v : float) = 
(*  ignore (log "Setting entry (%d, %d) to %f\n" i j v); *)
  (!data).(j + i * !max_cols) <- v
       
(*
let copy_realarray ~(src : float Array.array)
		   ~(si  : int)
		   ~(dst : float Array.array RealArray.array,
		     di  : int,
		     len : int} =
  let loop (_,_,0) = ()    | loop (i,j,l) = 
    (RealArray.update(dst,j,RealArray.sub(src, i));
     loop (i+1,j+1,l-1))
  in
  loop (si,di,len)
*)    
    
    (* Swap the positions of two variables *)
let swapPos (p1: simplPos) (p2: simplPos) =
  let o1 = ownedBy p1 in
  let o2 = ownedBy p2 in
  setOnPos o1 p2;
  setOnPos o2 p1
    
let swapRow i1 i2 =
  let _ = swapPos (Row i1) (Row i2) in
  for j = 0 to !nr_cols - 1 do
    let t1 = stab i1 j in
    utab i1 j (stab i2 j);
    utab i2 j t1
  done
(*  
	ProverUtils.for
	(fn j ->
	 let
	   let t1 = stab(i1, j)
	 in
	   (utab(i1, j, stab(i2, j));
	    utab(i2, j, t1))
	 end)
	(0, !nr_cols - 1)
      end
*)    

let swapCol j1 j2 =
  let _ = swapPos (Col j1) (Col j2) in
  for i = 0 to !nr_rows - 1 do 
    let t1 = stab i j1 in
    utab i j1 (stab i j2); 
    utab i j2 t1
  done
(*
	ProverUtils.for
	(fn i ->
	 let
	   let t1 = stab(i, j1)
	 in
	   (utab(i, j1, stab(i, j2));
	    utab(i, j2, t1))
	 end)
	(0, !nr_rows - 1)
      end
*)

    
					(* Hist stuff *)
type histEvent =
    HistKill of owner * owner  (** We killed o1 because of o2 *)
  | HistAdd  of owner
					(* Sometimes it is convenient to 
                                         * rerestrict a node. Then remember 
                                         * the old restriction *)
  | HistRestr of owner * whyrestr option

  | HistMark of int

  | HistSnapshot


let d_histEvent () = function
    HistKill (o1, by) -> dprintf "Kill(%a, %a)" d_owner o1 d_owner by
  | HistRestr (o, _) -> dprintf "Restr(%a)" d_owner o
  | HistAdd o -> dprintf "Add(%a)" d_owner o
  | HistMark m -> dprintf "Mark(%d)" m
  | HistSnapshot -> text "*"

let history  : histEvent list ref = ref []

(** Keep track of (o1, o2) when we killed o1 because of o2 *)
let killers: (owner * owner) list ref = ref []

let initHistory () = history := []
let addToHistory (ev : histEvent) = 
  history := ev :: !history
(*
  registerUndoRun 
    (fun _ -> 
      match !undoList with 
        _ :: t -> undoList := t
      | _ -> assert false)
*)

let printHistory () = 
  ignore (E.log "History: %a\n"
            (docList (fun e -> d_histEvent () e)) !history)

let markId = ref 0 
let registerUndoMarker () : int = 
  incr markId;
  history := HistMark !markId :: !history;
  !markId

let findKillerOf corpse_id =
  let rec loop = function 
      [] -> E.s (E.bug "Cannot find killer of %d" corpse_id)
    | (o1, killer) :: _ when o1.oid = corpse_id -> killer
    | _ :: t -> loop t
  in
  loop !killers 


let d_killList () : doc  =
  docList
    (function (o1, o2) -> dprintf "%a by %a" d_owner o1 d_owner o2)
    ()
    !killers


    (* Printing the tableau *)
let printSimplexState (msg: string) =
  let nc = (!nr_cols) - 1 in
  let livecol = !livecol in
  let nr = (!nr_rows) - 1 in
  let liverow = !liverow in
  let startrow = if !printLiveOnly then liverow else 0 in
  let startcol = if !printLiveOnly then livecol else 1 in
  ignore (E.log "%s\nTableau status: (%d rows of max %d, %d col of max %d)@!" 
            msg
            (nr + 1) !max_rows (nc + 1) !max_cols);
  ignore (E.log "        %a" d_colOwner 0);
  for j = startcol to nc do
    if j = livecol then ignore (E.log " | ");
    ignore (E.log "      %a" d_colOwner j);
  done;
  ignore (E.log "\n");
  (* Print the tableau *)
  for i = startrow to nr do 
      if i = liverow then begin
        ignore (E.log "-------");
        for j = startcol to nc do 
          ignore (E.log "---------");
        done;
      end;
    (* Print the constant column first *)
    ignore (E.log "\n%a  %5.2f" d_rowOwner i (stab i 0));
    for j = startcol to nc do
      if j = livecol then ignore (E.log "  | ");
      ignore (E.log "   %5.2f" (stab i j))
    done;
    (* If we did not print the first live col, print now the delimiter *)
    if livecol > nr then 
      ignore (E.log " | ");
    ignore (E.log "\n")
  done;
  (* If we did not print the first live row, print now the delimiter *)
  if liverow > nr then begin
    ignore (E.log "-------");
    for j = startcol to nc do 
      ignore (E.log "---------");
    done;
    ignore (E.log "\n");
  end;
  ignore (E.log "Legend:\n");
  let printLegend (own: owner) =
    let dexp = 
      match own.node with 
        NoNode -> nil
      | _ -> L.d_exp () (buildOwnExpression own)
    in
    if isRestr own then
      ignore (E.log "%a=%a (%a)\n" 
                d_owner own insert dexp d_whyrestr (getWhyrestr own))
    else begin
      if dexp <> nil then 
        ignore (E.log "%a=%a\n" d_owner own insert dexp)
      else
        ()
    end
  in
  for j = startcol to !nr_cols - 1 do
    printLegend (ownedByCol j)
  done;
  for i = startrow to !nr_rows - 1 do 
    printLegend (ownedByRow i)
  done;
  ignore (E.log "Divisors so far: %a\n"
            (docList (fun x -> real x)) !divisors);
  ignore (E.log "Kill history: %a\n\n"
            insert (d_killList ()))


      
    (* Extend the tableau *)
let  extendTableau (new_rows, new_cols) =
  let new_r = (!max_rows) * (1+((new_rows-1) / (!max_rows))) in
  let new_c = (!max_cols) * (1+((new_cols-1) / (!max_cols))) in
  if new_r > (!max_rows) || new_c > (!max_cols) then begin
    if !doPrint then
      ignore (log "Extending Simplex tableau from %d to %d rows and from %d to %d columns\n" !max_rows new_r !max_cols new_c);
    
	(* Allocate *)
    let new_d    = Array.make (new_r * new_c) 0.0 in
    let new_rows = Array.make new_r dummyOwner in
    let new_cols = Array.make new_c dummyOwner in
    (* Copy the data *)
    for i = 0 to !max_rows - 1 do 
      Array.blit (!data) (i * !max_cols) new_d (i * new_c) !max_cols
    done;
	    
    (* Copy the rows *)
    Array.blit !rows 0 new_rows 0 !max_rows;

    (* Copy the columns *)
    Array.blit !cols 0 new_cols 0 !max_cols;
    
    max_rows := new_r;
    max_cols := new_c;
    data     := new_d;
    rows     := new_rows;
    cols     := new_cols
  end
      
let addNewRow (own : owner) =
  if !doPrint then
    ignore (log "Adding a new row for %a. rows before: %d\n"
              d_owner own !nr_rows);
  
  let old_rows : int = !nr_rows in
  extendTableau (old_rows + 1, !nr_cols);
  nr_rows := 1 + old_rows;
  setOnPos own (Row old_rows);
  addToHistory (HistAdd own);
  old_rows
    
    
let addNewColumn (own : owner) =
  if !doPrint then
    ignore (log "Adding a new column for %a. Columns before: %d\n"
              d_owner own !nr_cols);
	  
  let old_cols : int = !nr_cols in
  extendTableau (!nr_rows, old_cols + 1);
  nr_cols := 1 + old_cols;
  setOnPos own (Col old_cols);
  addToHistory (HistAdd own);
  old_cols
    
    
    
    (* Search for a pivot. There are multiple purposes for which this is 
     * called *)
type pivotPurpose =
    MaxRow of int 
  | MinRow of int 
  | TranspCol of int
      
	
let d_pivotPurpose () = function
    MaxRow i -> dprintf "Maximize row %a" d_rowOwner i
  | MinRow i -> dprintf "Minimize row %a" d_rowOwner i
  | TranspCol j -> dprintf "Transpose column %a" d_colOwner j
	
type pivotResult =
    Pivot of int * int | Tie of int * int |
    NoPivotExtreme | PivotUnbounded of int * int

let d_pivot () = function
    Pivot (i, j) -> dprintf "Pivot(%a,%a)" d_rowOwner i d_colOwner j
  | Tie (i, j) -> dprintf "Tie(%a,%a)" d_rowOwner i d_colOwner j
  | NoPivotExtreme -> text "No pivot: extreme"
  | PivotUnbounded (i, j) -> 
      dprintf "Unbounded(%a,%a)" d_rowOwner i d_colOwner j
	
	
	
let findPivot (purpose : pivotPurpose) : pivotResult  =
  if !doPrint then begin
    ignore (log "Looking for a pivot to %a in state:" 
              d_pivotPurpose purpose);
    printSimplexState ""
  end;

	(* Collects the min and max in a column v *)
	(* min = Max{i restr, Tiv < 0} (Ci/Tiv)*)
	(* max = Min{i restr, Tiv > 0} (Ci/Tiv)*)
  let minMax (v : int) : (float * int * float * int * int list) =
    let isrc = isRestrCol v in
	    
    (* Iterate over live rows, starting at the end. *)
    let rec loop i ((min, rmin, max, rmax, nonres) as acc) = 
      if i < !liverow then
	acc
      else
        let iv = stab i v in
	let civ = stab i 0 /. iv in
	if iv = 0.0 then
	  loop (i-1) acc
	else begin
	  if isRestrRow i then
	    if iv > 0.0 then
	      if civ < max then
		loop (i-1) (min, rmin, civ, i, nonres)
	      else
		loop (i-1) acc
	    else (* iv < 0 *)
	      if civ > min then
		loop (i-1) (civ, i, max, rmax, nonres)
	      else
		loop (i-1) acc
	  else		(* not restricted *)
	    if civ <= 0.0 || not isrc then
	      loop (i-1) (min, rmin, max, rmax, i :: nonres)
	    else
	      loop (i-1) acc
        end
    in

    let (min, rmin, max, rmax, nonres) as res = 
      loop (!nr_rows - 1) (- 1000000.0, -1, 1000000.0, -1, []) 
    in
					(* Filter non-restricted with 
                                         * respect to min-max *)
    let nonres' =
      List.filter 
        (fun i ->
          let tiv = (stab i 0) /. (stab i v) in
	  min <= tiv && tiv <= max)
        nonres
    in
    if !doPrint then begin
      ignore (log "MinMax for column %a finds@!"  d_colOwner v);
      if rmin < 0 then 
        ignore (E.log "\tMin=*\n")
      else 
        ignore (E.log "\tMin=%f in row %a\n" min d_rowOwner rmin);
      if rmax < 0 then 
        ignore (E.log "\tMax=*\n")
      else 
        ignore (E.log "\tMax=%f in row %a\n" max d_rowOwner rmax);
      ignore (E.log "\tUnres: %a\n" 
                (docList (d_rowOwner ())) nonres');
    end;

    (min, rmin, max, rmax, nonres')
  in (* end of minMax *)
					(* Find a pivot for transpose *)
  let findTranspose (v : int) : pivotResult =
    let (min,rmin,max,rmax,nonres) = minMax v in
    (* Prefer restricted rows *)
    if rmax >= 0 then
      Pivot (rmax, v)
    else
      if rmin >= 0 then
        Pivot (rmin, v)
      else			(* Get a non-restricted one *)
        (match nonres with
	  nr :: _ -> Pivot (nr, v)
        | _ -> NoPivotExtreme)
  in (* end of findTranspose *)
					(* Find a pivot for maximize *)
  let findMaximize (z : int) : pivotResult =
    
					(* Try all columns *)
    let ((crow, ccol, cscore) : (int * int * float)) =
      let rec loopCols j ((_, _, champscore) as acc) = 
        if j = !nr_cols then 
          acc (* We are done *)
        else begin
          let tzv = stab z j  in
          if tzv = 0.0 ||  (tzv < 0.0 && isRestrCol j) then
            loopCols (j + 1) acc
          else begin
            let (min, rmin, max, rmax, _) = minMax j in
            (* The score for this column *)
            let (score, rscore) =
	      if tzv > 0.0 && rmin >= 0 then
	        (-. tzv *. min, rmin)
	      else
	        if tzv < 0.0 && rmax >= 0 then
	          (-. tzv *. max, rmax)
	        else	  (* This is unbounded. Increase the goal enough *)
	          (-. (stab z 0) +. 1.0, z)
            in
            let (crow, ccol, cscore) as newchamp = 
	      if champscore < score then
	        (rscore, j, score)
	      else
	        acc
            in
            if !doPrint then begin
              ignore (E.log "Champion after column %a is" d_colOwner j);
              if crow >= 0 then
                ignore (E.log "(%a,%a) with score %f\n" 
                          d_rowOwner crow d_colOwner ccol score)
              else
                ignore (E.log "*\n")
            end;
        
            loopCols (j + 1) newchamp
          end
        end
      in
      loopCols !livecol (-1, -1, -1.0)
    in
    let res =
      if crow < 0 then
        NoPivotExtreme
      else
        if crow = z then
	  PivotUnbounded (crow, ccol)
        else
	  if cscore <= 0.0 then
	    Tie (crow, ccol)
	  else
	    Pivot (crow, ccol)
    in
    if !doPrint then
      ignore (E.log "Found %a\n" d_pivot res);
    res

  in
  match purpose with
    TranspCol n -> findTranspose n
  | MaxRow z -> findMaximize z
  | _ -> E.s (E.bug "findMaximize")
					(* Verify that the state is 
                                         * satisfiable. This is only for 
                                         * debugging  *)
let checkSatisf () =
  let rec loopRows i acc = 
    if i = !nr_rows then 
      acc
    else
      loopRows (i + 1) (acc && (not (isRestrRow i) || (stab i 0) >= 0.0))
  in
  if not (loopRows 0 true) then begin
    ignore (log "Simplex state is not satisfiable:\n");
    printSimplexState "";
    E.s (E.bug "simplex not satisfiable")
  end
     
					(* Perform a pivot operation *)
let pivotTableau row col =
  if !doPrint then
    ignore (log "Pivoting tableau on (%a,%a)\n" 
              d_rowOwner row d_colOwner col);
  let nr_rows = !nr_rows in
  let nr_cols = !nr_cols in
	 
  let arc = stab row col in
  if abs_float arc <> 1.0 then
    divisors := (abs_float arc) :: (!divisors);

  (* Do everything but row and column *)
  for i = 0 to nr_rows - 1 do 
    let aic = stab i col in
    for j = 0 to nr_cols - 1 do 
      if i = row || j = col then
	()
      else
	utab i j ((stab i j) -.
		  (stab row j) *. aic /. arc)
    done
  done;
            
  (* Do the column col *)
  for i = 0 to nr_rows - 1 do 
    utab i col ((stab i col) /. arc)
  done;
       
  (* Do the row *)
  for j = 0 to nr_cols - 1 do 
    if j = col then
      utab row col (1.0 /. arc) 
    else
      utab row j ((-. (stab row j)) /. arc)
  done;
       
  (* Change the rows and cols metadata*)
  swapPos (Row row) (Col col);

  if !doPrint && !pTableau then begin
    printSimplexState "After pivoting"
  end
   
   
   (* Minimizes or maximizes a row *)
 
type extremeRes = Extreme of float | Unbounded | Stopped | Cycle
     
let rowToExtreme 
    (row : int)
    (purpose : pivotPurpose)
    (stopat : float) : extremeRes =
       
  let max_cycle = 3 in
	 
  let finished = match purpose with
    MaxRow _ -> (fun v -> v > stopat)
  | MinRow _ -> (fun v -> v < stopat)
  | _ -> E.s (E.bug "rowToExtreme")
  in 
  let rec loop = function 
      0 -> Cycle
    | cycle -> 
	if finished (stab row 0) then
	  Stopped
	else
	  (match findPivot purpose with
	    NoPivotExtreme -> Extreme (stab row 0)
	  | PivotUnbounded (i, j) -> (pivotTableau i j;
				      Unbounded)
	  | Pivot (i, j) -> (pivotTableau i j;
			     loop (max_cycle))
	  | Tie (i, j) -> (pivotTableau i j;
			   loop (cycle - 1)))
  in
  if !doPrint then
    ignore (E.log "%s the value of %a upto %f\n"
              (match purpose with MaxRow _ -> "Maximizing" 
                                 | _ -> "Minimizing")
	      d_rowOwner row stopat);
  let res =  loop (max_cycle) in
  res
   
   
   (* Transpose a column into a row, by pivoting the tableau. Return the new 
    * position of the column. A column that is identical to 0 may be left in 
    * place.  *)
let transposeColumn (col : int) : simplPos =
  match findPivot (TranspCol col) with
    NoPivotExtreme -> Col col
  | PivotUnbounded _ -> E.s (E.bug "transposeColumn: unexpected")
  | Pivot (row, col) -> (pivotTableau row col; Row row)
  | _ -> E.s (E.bug "transposeColumn") 
	 


let undoKill (a : owner) =
  if !doPrint then
    ignore (log "Undoing Kill of %a\n" d_owner a);
(*
  assert (match !history with 
            HistKill (a', _) :: _ when a.oid = a'.oid -> true
          | _ -> false); 
*)
  match getPos a with
    Row i ->
      if i < !liverow then begin
	decr liverow;
	if i < !liverow then swapRow i !liverow;
      end else
	()			(* Not dead. Ignore *)
  | Col j ->
      if j < !livecol then begin
	decr livecol;
	if j < !livecol then swapCol j !livecol;
      end else ()
  | NotSimpl -> if !doPrint then ignore (E.log "   - undo kill ignored\n")

let addKillToHistory (a: owner) (what: owner) = 
  a.killedby <- Some what;
  history := HistKill (a, what) :: !history;
  killers := (a, what) :: !killers

                                    
(*
  registerUndoRun 
    (fun _ -> 
      undoKill a)
*)


let undoRestr (a: owner) (oldh: whyrestr option) = 
  assert (match !history with 
            HistRestr (a', _) :: _ when a.oid = a'.oid -> true
          | _ -> false); 
  if !doPrint then 
    ignore (log "Undoing restriction of %a\n" d_owner a);
  a.restr <- oldh

(** This function undoes adding an owner. *)
let undoAdd (a : owner) =
  if !doPrint then
    ignore (E.log "Undoing add %a\n" d_owner a);

  (* Perhaps we are undoing an add directly, which case we must undo all 
   * events belonging to the add *)
  let rec loop () = 
    match !history with
      HistAdd a' :: rest -> 
        (* There is a bug here, but I do not understand it. *)
        if a'.oid <> a.oid && false then begin (* this is an add for a different node *)
          loop ();
          (* put it back *)
          history := HistAdd a' :: !history;
        end else
          history := rest

    | HistKill (a', _) :: rest -> 
        if a'.oid <> a.oid then 
          E.s (E.bug "undoAdd: you still have undoKill in the list");
        undoKill a';
        history := rest;
        loop ()
    | HistRestr (a', oldh) :: rest -> 
        if a'.oid <> a.oid then 
          E.s (E.bug "undoAdd: you still have undoRestr in the list");
        undoRestr a' oldh;
        history := rest;
        loop ()
    | _ -> E.s (E.bug "undoAdd: empty undo list")
  in
  loop ();

  (* Undo the kill, if it is killed *)
  (* match a.killedby with 
    Some killer -> undoKill a
  | None -> () *);

  (* We can throw away a row; we do that by putting it last and decrementing 
   * the number of active rows. *)
  let undoAddRow i =
    (nr_rows := !nr_rows - 1;
     if i < !nr_rows then swapRow i !nr_rows else ();
     for j = 0 to !nr_cols - 1 do 
       utab !nr_rows j 0.0
     done;
     removeOwner a)
  in
  match getPos a with
    Row i -> undoAddRow i
  | Col j -> begin
      (* We cannot throw away a column, except if it is identical to 0. We 
       * must transpose the column into a row first and then throw the row. *)
      match transposeColumn (j) with
        Row i -> undoAddRow i
      | Col col -> (* It must be that the column is identically 0 *)
          assert (for i = 0 to !nr_rows - 1 do 
                    assert (stab i col = 0.0)
                 done; true);
	  (nr_cols := !nr_cols - 1;
	   if j < !nr_cols then swapCol j !nr_cols;
	   removeOwner a)
      | _ -> E.s (E.bug "undoAdd: node is not Simplex")
  end
  | NotSimpl ->  if !doPrint then ignore (E.log "   - undo add ignored\n")

let undoMarkOpt (mo: int option) =
  if !doPrint then 
    ignore (log "Undoing Simplex state\n");
	   
  let rec loop () =
    match !history with 
      [] -> E.s (E.bug "Simplex: empty undo stack")
    | HistSnapshot :: rest -> history := rest
    | HistMark m :: rest -> begin 
        history := rest;
        match mo with 
          Some m' when m = m' -> () (* stop here *)
        | _ -> loop ()
    end

    | HistKill (a, _) :: rest -> undoKill a; history := rest; loop ()
    | HistAdd a :: rest -> undoAdd a; history := rest; loop ()
    | HistRestr (a, oldh) :: rest -> 
        undoRestr a oldh; history := rest; loop ()
  in
  loop ();
  if !doPrint then begin
    printSimplexState "Simplex state after undo:\n"
  end

    
let snapshot () = 
  addToHistory HistSnapshot;
  registerUndoRun (fun _ -> undoMarkOpt None)
      

let restrictOwn (own : owner) (h : whyrestr) =
  let idzero = !nodeZeroOwner.oid in
  (* Restrict a node and recall the last restriction *)
  let doRestr () = 
    if !doPrint then 
      ignore (log "Simplex marking the restriction of %a\n" d_owner own);
    
    addToHistory (HistRestr (own, own.restr));
    own.restr <- Some h
  in
  (* Do not restrict twice *)
  if idzero <> own.oid then begin
	  (* Always prefer equality restrictions. If not equality then do not 
           * bother re-restricting *)
    match h with			
      BYEQ _ -> doRestr ()
    | _ -> 
	if own.restr = None then
	  doRestr ()
	else begin
	  if !doPrint then 
            ignore (log "Simplex: %a was already restricted\n" d_owner own);
        end
  end
	

   (* Collect all the rows and columns that are manifestly equal to this one. 
    * Accumulates two lists of pairs. In the first list, we have the  
    * identical rows (match in all columns). In the second list we have the 
    * equal rows (match in all live columns). *)
let collectEqualitiesWithRow 
    (i : int) 
    (fst_row : int) (* first row to compare *)
    (init : ((owner * owner) list * (owner * owner) list))
    : ((owner * owner) list * (owner * owner) list) =
  
  if !doPrint then
    ignore (log "Collecting equalities with row %a :\n" d_rowOwner i);
  
  let thisRowOwner = ownedByRow i in
	 
       (* Compare rows *)
  let (id_rows, eq_rows) = 
    for_a
      (fun i' ((sofar_id, sofar_eq) as sofar) ->
	if i <> i' && (stab i 0) = (stab i' 0) then begin
	  (* Compares two rows, but only if they agree on the constant column 
          *)
	  let rec comp_row j max : bool =
	    j >= max ||
	    ((stab i j) = (stab i' j) && comp_row (j+1) max)
          in
					(* Two rows are equal if they have 
                                         * the same entries in the live 
                                         * columns. *)
	  let equal = comp_row !livecol !nr_cols in
          (* Two rows are identical if they have the save entries in all 
           * columns *)
	  let ident = equal && comp_row 1 !livecol in
          if !doPrint && (equal || ident) then
            ignore (E.log "%a" d_rowOwner i');

	  if equal then begin
	    let newPair = (ownedByRow i', thisRowOwner) in
	    if ident then begin
	      if !doPrint then ignore (E.log "(ID), ");
	      (newPair :: sofar_id, sofar_eq)
	    end else begin
	      if !doPrint then ignore (E.log ", ");
	      (sofar_id, newPair :: sofar_eq)
            end
	  end else 
	    sofar

	end else
	  sofar)
      init
      (fst_row, (!nr_rows) - 1)
  in

  (* Identifies an equality between this row and a column, when the row is 
  * identically zero, except a 1 in a column *)
  let eq_col : owner option =
    if stab i 0 = 0.0 then begin
      let nrOnes, lastOne = 
        for_a (* for all live columns, acc = (the number of columns with 1 in 
               * them, and the last such column) *)
	  (fun j acc -> 
            match stab i j with 
              0.0 -> acc (* No changes *)
            | 1.0 -> begin
                match acc with 
                  (0, _) -> (1, j)
                | _ -> (2, 0)
            end
            | _ -> (* No equality *) (2, 0))

	  (0, 0)
	  (!livecol, (!nr_cols) - 1) 
      in
      if nrOnes = 1 then begin
        if !doPrint then 
          ignore (E.log "%a," d_colOwner lastOne);
        Some (ownedByCol lastOne)
      end else
        None
    end else
      None
  in
  let res =
    match eq_col with
      None -> (id_rows, eq_rows)
    | Some colown ->
        (* If the dead columns are all 0, then this is identical *)
	if for_a (fun j acc -> 
          acc && (stab i j) = 0.0) 
            true (1, !livecol - 1)
	then begin
	  if !doPrint then 
            ignore (E.log "(ID)");
	  ((colown, thisRowOwner) :: id_rows, eq_rows)
	end else
	  (id_rows, (colown, thisRowOwner) :: eq_rows)
	       
  in
  if !doPrint then ignore (E.log "\n");
       
  res
   
   
   (* Collecting the equalities with ZERO entailed by a row. Only applies if 
    * the row is maximized to ZERO *)
exception NotMaximized
let collectColumnEqualities (row : int) : (owner * owner) list =
  (* The position of zero *)
  let zeroOwner      = !nodeZeroOwner in
  let thisRowOwner   = ownedByRow row in
	 
  let rec loop j acc =
    if j >= (!nr_cols) then
      acc
    else
      if (stab row j) <> 0.0 then
	if isRestrCol j then
	  if stab row j < 0.0 then
	    loop (j+1) ((ownedByCol j, zeroOwner) :: acc)
	  else
	    raise NotMaximized
	else
	  raise NotMaximized
      else
	loop (j+1) acc
  in
  let res = 
    if stab row 0 >= 1.0 || stab row 0 < 0.0 ||
       not (isRestrRow row) 
    then []
    else begin
      try
        loop (!livecol) []
      with NotMaximized -> []
    end
  in
  let rec iter f = function [] -> () | h::t -> f h; iter f t in
  iter (fun (own, _) -> addKillToHistory own thisRowOwner) res;

  if !doPrint && res <> [] then
    ignore (log "Equalities with ZERO entailed by row %a: @[%a@]\n" 
              d_rowOwner row
              (docList
                 (fun (own, _) -> d_owner () own)) res);
  res
    
   
   
				(* Add UndoKill for killed rows *)
let killRow i (own: owner) (killer: owner) =
     				(* Never kill restricted rows, because they 
                                 * must be taken into account when looking 
                                 * for pivots *)
     
  if not (isRestr own) then
    if i = !liverow then begin
      if !doPrint then 
        ignore (log "Killing first live row %a\n" d_rowOwner i);
      addKillToHistory own killer;
      liverow := !liverow + 1
    end else begin
      if i > !liverow then begin
	if !doPrint then 
          ignore (log "Killing row %a\n" d_rowOwner i);
	swapRow i !liverow;
	addKillToHistory own killer;
	liverow := 1 + (!liverow)
      end;
    end
  else begin
    if !doPrint then
      ignore (log "Trying to kill restricted row %a. Do not kill!\n"
                d_owner own);
  end
     

(** Keep here a pointer to the function that generates proofs. This is a 
 * forward reference *)
let pMkEqProof: (node -> node -> P.proof) ref = ref (fun _ -> assert false)

   (* Kills columns and rows, as necessary and packages the equalities for 
    * use in a message *)
let packageOneEquality 
    ((own1, own2) : owner * owner) : unit =

  let killCol j ownj ownkiller =
    if eqOwner ownkiller !nodeZeroOwner then
      if j = !livecol then begin
	if !doPrint then 
          ignore (log "Killing first live column %a\n" d_colOwner j);
	 livecol := !livecol + 1
      end else begin
	if j > !livecol then begin
	  if !doPrint then 
            ignore (log "Killing column %a\n" d_colOwner j);
	  swapCol j !livecol;
	  livecol := (!livecol) + 1
	end
      end                       (* Only kill columns killed by 0 *)
    else
      E.s (E.bug "simplex killCol")
  in
	     
  let makeEQ (own1 : owner) (own2: owner) : unit = 
    match (getNode own1, getNode own2) with
      (Node n1, Node n2) when n1.root != n2.root -> 
        newMessage
          { owner = sxid;
            fromgoal = false;
            proof = if !genProofs then (!pMkEqProof) n1 n2 else noProof;
            body = Intern (L.SetEq, n1, n2);
          }

    | _ -> ()


  in					(* Do not kill restricted owners, 
                                         * unless killed by restricted owners. 
                                         * Also do not kill 0. 0 is always 
                                         * restricted  *)

					(* Columns are only killed if they 
                                         * are = 0 *)
  if eqOwner own1 !nodeZeroOwner then begin
    (match getPos own2 with
      Row i -> killRow i own2 own1
    | Col j -> killCol j own2 own1
    | _ -> E.s (E.bug "packageOneEquality"));
    makeEQ own1 own2 
  end else begin
    if eqOwner own2 !nodeZeroOwner then begin
      (match getPos own1 with
	Row i -> killRow i own1 own2
      | Col j -> killCol j own1 own2
      | _ -> E.s (E.bug "packageOneEquality"));

      makeEQ own1 own2
    end else begin				(* None is restricted *)
      match (getPos own1, getPos own2) with
	(Col j, Row i) ->
	  if not (isRestrRow i) || isRestrCol j then
	    killRow  i own2 own1;
					(* Do not kill this row *)
	  makeEQ own1 own2

      | (Row i1, Row i2) ->
	  let kill1 () = killRow i1 own1 own2 in
	  let kill2 () = killRow i2 own2 own1 in

	  let _ = 
	    match (isRestrRow i1, isRestrRow i2, i1 < i2) with
	      (false, false, true) -> kill1 ()
	    | (false, false, false) -> kill2 ()
	    | (false, true, _) -> kill1 ()
	    | (true, false, _) -> kill2 ()
	    | (true, true, true) -> kill1 ()
	    | (true, true, false) -> kill2 ()
	  in
	  makeEQ own1 own2

      | _ -> E.s (E.bug "packageOneEquality: not Simplex")
    end
  end
   
   
   (* Search the entire tableau for now *)
let minimizeTableau (fst_row: int) 
                    (last_row : int) : unit = 
  (* For all restricted rows that are maximized at 0, kill the columns that 
   * contain non-zeros *)
       
  if !doPrint then
    ignore (log "Minimizing tableau\n");
	 
  (* For several rows *)
  (* Scan all rows and kill the columns equal to zero *)
  let eq_cols =
    for_a
      (fun i _  ->
        let rec iter f = function [] -> () | h::t -> f h; iter f t in
	iter packageOneEquality (collectColumnEqualities i);
        ())
      ()
      (fst_row, last_row)
	 
  in
       (* Now compare all rows *)
  let rec iter f = function [] -> () | h::t -> f h; iter f t in
  iter packageOneEquality
    (for_a
       (fun i acc ->
	 (* There should not be identical rows *)
         match collectEqualitiesWithRow i (i+1) ([], acc) with
           ([], new_acc) -> new_acc
         | _ -> E.s (E.bug "minimizeTableau"))
       []
       (0 (* !liverow *) , !nr_rows - 1))


(************************************************************************
 ******
 ****
 ****     PROOF CONSTRUCTION 
 ****
 *****
 ************************************************************************)


(* Our proofs are linear combinations of restricted owners. *)
type linComb = (float * owner) list

   
   (* Proof Construction *)
   (* Normalize the real coefficients to integers *)
let threshHold_2 = threshHold /. 2.0

let isInt (x : float) : int option =
  let xi = floor (x +. threshHold_2) in
  if abs_float (x -. xi) < threshHold then
    Some (int_of_float xi) 
  else 
    None
   
let makeInt (x : float) : int =
  match isInt x with
    None -> E.s (E.bug "Should be an integer (%f)" x)
  | Some i -> 
      (* ignore (log "makeInt %f -> %d\n" x i); *)
      i

let updateGCD (x : float) (sofar : float) : float =
  (* Finds a number that multiplied by this yields an integer. Uses the list 
   * of divisors *)
  let rec scanDivisors (divs: float list) (sofar: float) =
    match isInt (x *. sofar) with
      None -> begin
        match divs with 
          [] -> begin
            ignore (log "Cannot turn %f into an integer\n" (x *. sofar));
	    E.s (E.bug "No GCD was found")
          end
        | d :: rest -> scanDivisors rest (d *. sofar)
      end
    | _ -> sofar
  in
  scanDivisors (List.rev !divisors) sofar
   
   
let updateGCDList (xl : (float * 'a) list) (sofar : float) : float =
  let rec fold_left f acc = 
    function []-> acc | h::t -> fold_left f (f acc h) t in
  abs_float (fold_left (fun acc (x,_) -> updateGCD x acc) sofar xl)
     

					(* There are several proof types *)

type proofType =
     PR_EQ of node * node  (* n1 = n2 *)

  |  PR_GEQ of int * L.exp * L.exp * int
   (* PR_GEQ(i, e1, e2, diff). ownRow(i) = e2 - e1 + diff. Row i is maximized 
    * at "m" such that floor(m) <= diff. Prove that e1 >= e2. *)

let d_proofType () = function 
    PR_EQ (t1, t2) -> dprintf "%a = %a" d_nodeTop t1 d_nodeTop t2
  | PR_GEQ (n, e1, e2, diff) -> 
      dprintf "%a >= %a (diff=%d, %a (ownRow(%d)) = e2 - e1 + diff)" 
        L.d_exp e1 L.d_exp e2 
        diff 
        d_owner (ownedByRow n)
        n

    

					(* Copy a row in another *)
let copyInRow drow srow fact =
  for j = 0 to !nr_cols - 1 do 
    utab drow j (fact *. (stab srow j))
  done

    (* Accumulate two rows: r1 += fact * r2 *)
let accumulateInRow (r1 : int) (r2 : int) (fact : float) =
  for j = 0 to !nr_cols - 1 do 
    utab r1 j ((stab r1 j) +. fact *. (stab r2 j))
  done


  (* Given a GCD, and a linear combination of owners, compute the linear 
   * combination of proofs (multiplies all factors bu GCD and turns into 
   * integers; resulting coefficients must be positive for the GEQ proofs). 
   * This is actually the point where we call the S.proveIt to obtain proofs 
   * of messages. We remember if the proof is of an equality, the e1 and e2 
   * and the proof. *)
let linCombToProofList (gcd: float) 
                       (lc: linComb) 
    : (int * bool * L.exp * L.exp * P.proof) list = 
  List.fold_left
    (fun acc (c, own) -> 
      if c = 0.0 then (* Skip this *)
        acc
      else begin
        if not (isRestr own) then 
          E.s (E.bug "Sx: proof contains non-restricted owner (%f * %a)"
                 c d_owner own);
        let coeff = makeInt (gcd *. c) in
        let iseq, e1, e2, prf = 
          match getWhyrestr own with
	  | BYEQ (msg, e1, e2, proveIt) -> true, e1, e2, proveIt ()
                
	  | BYGEQ (_, e1, e2, proveIt) -> 
              if coeff < 0 then 
                E.s (E.bug "Sx: proof contains negative coefficient (%f * %a)"
                       c d_owner own);
              false, e1, e2, proveIt ()
                
	  | GEQINT (own, e, proveIt) -> 
              (false,  
               e, 
               L.zero,
               proveIt ())
        in
        (coeff,iseq,e1,e2,prf) :: acc
      end)
    []
    lc




(** Prepare a list of proofs for use with the ArithGeq proof rules. Return 
 * also the expression that is proved to be =0 or >=0 *)
let prepareArithProofList
    (eqprf: bool) 
    (prfl: (int * bool * L.exp * L.exp * P.proof) list) 
    : L.exp * P.parg list = 
  List.fold_left
    (fun (accexp, acc) (f,iseq,e1,e2,prf) -> 
      if eqprf && not iseq then assert false;
      if not iseq && f < 0 then assert false;
      let accexp' = L.plus (L.times (L.integer f) (L.minus e1 e2)) accexp in
      (accexp', P.I f :: P.P prf :: acc))
    (L.zero, [])
    prfl
        
      


  (** Make a GEQ proof *)
let makeGeqProof 
    (e1: L.exp) 
    (e2: L.exp) 
    (diff: float) 
    (geql: linComb) : P.proof = 
  (* Makes a proof of e1 >= e2 *)
  (* It must be true that 
          - diff < 1
          - geql contains only EQ or GEQ restricted owners. The GEQ ones 
            have negative coefficients
          - eqCheck(e1 - e2 + diff = geql)
  *)
  
  if !doPrint then 
    ignore (log "makeGeqProof (%a >= %a)\n diff = %f\n @[%a@]\n"
              L.d_exp e1 L.d_exp e2 diff
              (docList ~sep:(text " + " ++ break) 
                 (fun (c, own) -> dprintf "%f*%a" c d_owner own))
              geql);
 
  if diff >= 1.0 then 
    E.s (E.bug "makeGeqProof called with wrong diff = %f\n" diff);
  
  (* Compute the factor to make diff and geql integers *)
  let fact = updateGCDList geql (updateGCD diff 1.0) in
  
  (* Scale the diff, and make it an integer *)
  let diff' = makeInt (fact *. diff) in
  let fact' = makeInt fact in
  
  (* diff' <= fact' - 1 *)
  
    (* Compute the list of proofs *)
  let geql : (int * bool * L.exp * L.exp * P.proof) list = 
    linCombToProofList fact geql in

  (* Look for the trivial case *)
  let prf = 
    let _, prf_geq = prepareArithProofList false geql in 
    P.mkArithGeq e1 e2 ~fact:fact' prf_geq
  in
  assert(proofWellFormed (dprintf "makeGeqProof") prf);
  prf
  
let rec generateProof (ptype : proofType) =
  if !doPrint then
    ignore (log "Simplex generates proof for %a\n" d_proofType ptype);
    
  (* Check if a column is owned by an equality *)
  let restrictedByEQ j =
    let ownj = ownedByCol j in
    if isRestr ownj then
      (match getWhyrestr (ownedByCol j) with
	BYEQ _ -> true
      | _ -> false)
    else begin
      ignore (log "Owner %a should have been restricted\n" d_colOwner j);
      E.s (E.bug "Simplex 1152: owner should be restricted")
    end
  in


  (* Start with a row that is manifestly maximized. Its all non-zero 
   * coefficients are in restricted columns (in which case they are 
   * negative), or in dead columns. But each of the dead columns are killed 
   * by some other row. Rewrite this row to have only negative coefficients, 
   * even in the dead columns. At end, the owner(row) = row + linComb 
   * returned by rewrite row. *)
  let rewriteRow (row : int) : linComb = 
    (* Compute the columns that have positive coefficients in columns that 
     * are not restricted by equalities *)
    let posColumns = 
      for_a
	(fun j acc ->
	  if stab row j > 0.0 && not (restrictedByEQ j) then
	    j :: acc
	  else
	    acc)
	[]
	(1, !nr_cols - 1)
	     
    in
    
    (* Make one pass over the row. Accumulate extra linear combinations. *)
    let onePass (extra: linComb) : linComb = 
      for_a
        (fun j (extra: linComb) -> 
          let coeff = stab row j in
          if coeff <= 0.0 || (restrictedByEQ j) then 
            extra
          else begin 
            (* positive coefficient in non-EQ column. Must be killed by a row*)
	    (* Find the killer of this row *)
	    let ownj = ownedByCol j in
	    let killer : owner = 
              match ownj.killedby with 
                Some k -> k
              | _ -> assert false
            in
            let krow = 
              match ! (killer.ospos) with 
                Row krow -> krow
              | _ -> E.s (E.bug "doRewrite. Column %a not killed by a row" 
                            d_owner ownj)
            in
	    let new_coeff = coeff /. (stab krow j) in
	    let absstab = abs_float (stab krow  j) in
	    if absstab <> 1.0 then
	      divisors := absstab :: (!divisors);
            
	    accumulateInRow row krow (-. new_coeff);
            
            (* Add the killer to the "extra" list *)
            let rec addKill killer (coeff: float) (lst: linComb) : linComb = 
              match lst with 
                [] -> [(coeff, killer)]
              | (c, own) :: rest ->
	          if own.oid = killer.oid then
	            (c +. coeff, own) :: rest
	          else
	            (c, own) ::  (addKill killer coeff rest)
            in
            addKill killer new_coeff extra
          end)
        extra
        (1, !nr_cols - 1)
    in
    
    (* Repeat one pass until nothing changes. This should always terminate, 
     * but let's put a timeout  *)
    let rec repeat (iterations: int) (extra: linComb) = 
      if iterations > 10 then 
        E.s (E.bug "SX:rewriteRow appears to be looping");
      let extra' = onePass extra in
      if extra != extra' then 
        repeat (iterations + 1) extra'
      else
        extra'
    in
    repeat 0 []
  in (* rewriteRow *)



  (* Compare two rows. Collect the columns that are restricted by equalities, 
   * and in in which r1 + r2 = 0 and r1 <> 0. Set those columns to 0 in r1 
   * and r2. *)
  let compareRowsExtractEqualities (r1: int) (r2: int) : linComb =
    for_a
      (fun j acc ->
	let coeff1 = stab r1 j in
	let coeff2 = stab r2 j in
	let own = ownedByCol j in
	if coeff1 <> 0.0 && coeff1 +. coeff2 = 0.0 
                         && restrictedByEQ j then begin
	  utab r1 j 0.0;
          utab r2 j 0.0; 
          (coeff1, own) :: acc
	end else
	  acc)
      []
      (1, !nr_cols - 1)
  in
       

  (* Called on a row that is manifestly maximized, and with only negative 
   * coeffients, even in the dead columns. Compute the linear combination 
   * that shows the row to be <= constant_column. *)
  let collectRowProof (row : int) : linComb = 
    for_a
      (fun j acc -> 
       (stab row j, ownedByCol j) :: acc)
      []
      (1, !nr_cols - 1)
  in 



					(* The row owned by ZERO *)
  let  rowZero = 
    match getPos (!nodeZeroOwner) with
    | Row i -> i
    | _ -> E.s (E.bug "rowZero")
  in
	 
   
    
                     
  let proof : P.proof =
    match ptype with
      PR_EQ (n1, n2) -> begin
	if n1.id = n2.id then
	  P.Refl n1.exp
	else begin
          let o1 = getOwner n1 in
          let o2 = getOwner n2 in

          (* Take a snapshot, so that we can undo adding the rows *)
          let mark = registerUndoMarker () in

          (* Add a new scratch row to contain n1 - n2 *)
          let own_r1 = newSlackOwner 0 [ (1, n1); (-1, n2) ] in
          let r1 = addNewRow own_r1 in
          (* Initialize r1 = n1 - n2 *)
	  let _ = 
	    match (getPos o1, getPos o2) with
	      (Col j, Row i) ->
		copyInRow r1 i (-1.0);
		utab r1 j (stab r1  j +. 1.0)	
		       
	    | (Row i1, Row i2) ->
		copyInRow r1  i1 1.0;
		accumulateInRow r1 i2 (-1.0)
            | _ -> E.s (E.bug "PR_EQ")
          in
		     
	  (* Add another row r2 = - r1 = n2 - n1 *)
	  let own_r2 = newSlackOwner 0 [ (1, n2); (-1, n1) ] in
	  let r2 = addNewRow own_r2 in
		   
	  copyInRow r2 r1 (- 1.0);
		   
	  if !doPrint then
	    printSimplexState "Before rewrite row:\n";
		
          let extra1: linComb = rewriteRow r1 in
          let extra2: linComb = rewriteRow r2 in

	  if !doPrint then
	    printSimplexState "After rewrite row:\n";
		   
	  (* Now extract the EQ part, and clear some entries. *)
	  let eql : linComb = compareRowsExtractEqualities r2 r1 in

	  if !doPrint then begin
	    printSimplexState "After compare rows:\n"
          end;

          let leql = (collectRowProof r1) @ extra1 in
          let geql = (collectRowProof r2) @ extra2 in

	  undoMarkOpt (Some mark); (* Undo the adding of the two rows *)


					(* Collects the gcd from all lists *)
	  let gcd  = 
            updateGCDList leql 
              (updateGCDList geql (updateGCDList eql 1.0)) in
		   
          let fact = makeInt gcd in
          let eql   = linCombToProofList (-. gcd) eql in
          let geql  = linCombToProofList (-. gcd) geql in
          let leql  = linCombToProofList (-. gcd) leql in

	  if n1.exp = n2.exp then 
	    P.Refl n1.exp
	  else if eql = [] && geql = [] && leql = [] then 
            P.ArithEq(n1.exp, n2.exp, 1, [])
          else begin

            let ege, ege_prf = prepareArithProofList false geql in 
            let _, ele_prf = prepareArithProofList false leql in 
            let _, eq_prf   = prepareArithProofList true  eql in

            (* First we prove that ege is = 0. Recall that ege + ele = 0 can 
             * be proved with just arithmetic. *)
            let ege_eq0 = P.G("eqi", 
                              [ P.P (P.mkArithGeq ege L.zero ege_prf);
                                P.P (P.mkArithGeq L.zero ege ele_prf)]) 
            in
            P.mkArithEq n1.exp n2.exp ~fact:fact 
              (P.I 1 :: P.P ege_eq0 :: eq_prf)
                        
          end

        end
      end   

          (* row = e2 - e1 + diff. row is maximized at m and 
           *      floor(m) <= diff. 
           * Prove that e1 >= e2 
           *)
 | PR_GEQ (row, e1, e2, diff) -> begin

     (* Take a snapshot, so we can undo *)
     let mark = registerUndoMarker () in 

     (* Add a new scratch row, which is equal to "row" - diff = e2 - e1. 
      * It should be maximized at "m - diff". *)
     let own_r1 = newSlackOwner 0 [] in
     let r1 = addNewRow own_r1 in
     copyInRow r1 row 1.0;
     utab r1 0 (stab r1 0 -. (float diff));
     
     if !doPrint then printSimplexState "In PR_GEQ, before collectProof";
     
     (* We need to show that r1 <= m - diff *)
     let geql = (collectRowProof r1) @ (rewriteRow r1) in
     let diff' = stab r1 0 in (* diff < 1 *)
     
     (* r1 = diff' + geql *)
     (* e1 - e2 + diff' = - geql *)
     
     undoMarkOpt (Some mark); (* Undo the adding of the row *)
     
     
     makeGeqProof e1 e2 diff' 
       (List.map (fun (c, o) -> (-. c, o)) geql)
 end
       (* Now build the proof *)
  in
  if !doPrint then
    ignore (log "Simplex finished generating proof for %a\n @[%a@]@!" 
              d_proofType ptype
              P.d_prf proof);
  proof


and whyrestrToGeqProof whyrestr = 
  match whyrestr with
    BYEQ (msg, e1, e2, proveIt) ->  P.G("eqgeq", [P.P (proveIt ())]), e1, e2
  | BYGEQ (_, e1, e2, proveIt) -> proveIt (), e1, e2
  | GEQINT (own, e1, proveIt) -> 
      proveIt (), e1, L.zero



(** Build an expression corresponding to a linear combination *)
let buildExprFromList (l: linComb) : L.exp =
  L.simplifyArith 
    (L.buildExpLinearComb
       (List.map 
          (fun (f, own) -> 
            (makeInt f, buildOwnExpression own))
          l))


let buildLinCombFromList (l: linComb) : int * (int * node) list = 
  let ct = ref 0 in 
  let res = 
    List.fold_left 
      (fun acc (f, own) -> 
        let c, l = buildOwnLinComb own in 
        ct := c + !ct;
        acc @ l)
      []
      l
  in
  !ct, res
    
(** Called when "row" is maximized to < 0.0 and is also restricted to be >= 0 
 * by "h". Does not return *) 
let foundContradiction (row: int) (h: whyrestr) : 'a = 

  (* Maybe the row was restricted due to assuming the negation of an 
   * inequality goal. In that case, we can try to use the maximization proof 
   * to construct the proof of the goal directly
  (match h with
    BYGEQ(msg, _, _, mkprf) when msg.goal -> begin
      match msg.body with 
        Intern (L.SetGe, n1, n2) -> 
          raise (S.Proved (generateProof (PR_GEQ(row, n1.exp, n2.exp, -1))))

      | Intern (L.SetGt, n1, n2) -> 
          raise (S.Proved 
                   (P.G("geqgt", [P.P (generateProof(PR_GEQ(row, 
                                                            n1.exp, 
                                                            L.plus n2.exp L.one, 
                                                            -1)))])))
      | Intern(L.SetLe, n1, n2) -> 
          raise (S.Proved 
                   (P.G("geqleq", [P.P (generateProof(PR_GEQ(row, 
                                                             n2.exp, n1.exp, -1)))])))

      | Intern(L.SetLt, n1, n2) -> 
          raise (S.Proved 
                   (P.G("geqlt", [P.P (generateProof(PR_GEQ(row, 
                                                            n2.exp, 
                                                            L.plus n1.exp L.one,
                                                            -1)))])))
      | _ -> ()
    end
  | _ -> ());
 *)
  (* If we got here we cannot prove the goal directly *)
      
  let buildProof: P.proof = 
    (* h gives us a proof that ownRow(i) >= 0. We actually get a 
    * proof of e1 >= e2, for e1, e2 such that 
    *    e1 - e2 = ownRow(i). 
    *)
    let (geq_prf, e1, e2) = whyrestrToGeqProof h in
    if !doPrint then begin 
      ignore (log "whyrestrToGeqProof (e1>=e2):@!  e1=%a@!  e2=%a@! geq_prf=%a@!"
                L.d_exp e1 L.d_exp e2 P.d_prf geq_prf);
      
      ignore (E.log "h=%a\n" d_whyrestr h);
    end;
    
    (* Now make a proof that ownRow(i) <= -1, or more precisely 
    * that: e2 >= e1 + 1, so that we can use geqLt_FL. *)
    let leq_prf = generateProof (PR_GEQ (row, e2, 
                                         L.plus e1 L.one, 
                                         -1)) in
    P.mkArithContra [ P.I 1 ; P.P geq_prf ;
                      P.I 1 ; P.P leq_prf ]
  in
  raise (S.Contradiction buildProof)
  


let _ = 
  pMkEqProof := (fun n1 n2 -> generateProof (PR_EQ(n1, n2)))



    (* Restricts a node to be positive *)
 
    (* Add the expression "eplus + eint - eminus", with the given owner. *)
let rec addSomeExpr
    (eplus : node) (* eplus + eint - eminus *)
    (eint : int)
    (eminus : node)
    (own : owner) : owner =


  (* First, scan the subexpressions and intern them in the Simplex tableau. 
   * We do this first, in case it turns out this node is redundant and we 
   * want to eliminate it. It is best to do that if the node is last *)
  let rec loopIntern (e: node) = 
    (* The nodes that are not of direct interest to Simplex *)
    let notInteresting () = 
      match ! (e.spos) with 
        Col j -> ()
      | Row i -> ()
            (* New to the Simplex tableau *)
      | _ -> 
          let j = addNewColumn (newNodeOwner e) in
          List.iter scanExpression e.args
    in
    match fSymOfHead e.head, e.args with
      INT n, _ -> ()
    | BINOP L.Times, [a1; a2] -> begin
        match fSymOfHead a1.head, fSymOfHead a2.head with 
          INT _, _ -> loopIntern a2
        | _, INT _ -> loopIntern a1
        | _, _  -> notInteresting ()
    end
    | BINOP (L.Plus|L.Minus), [a1; a2] -> 
        loopIntern a1; loopIntern a2
    | _ -> notInteresting () 
  in    
  loopIntern eplus; loopIntern eminus;

  
  (* Create a marker, in case we need to undo this add *)
  let mark = registerUndoMarker () in 

  (* Create a new row in the table *)
  let new_row : int = addNewRow own in
	 
  let add_to j (fact : float) =
    utab new_row j ((stab new_row j) +. fact)
  in
  add_to 0 (float eint); (* Add the eint *)
	 
	 
  (* This scans the expressions *)
  let rec loop (e: node) (fact: float) = 
    let notInteresting () = 
      match ! (e.spos) with 
        Col j -> 
          add_to j fact 
            
            
      | Row i -> 
          for j = 0 to !nr_cols - 1 do
	    add_to j (fact *. (stab i j))
          done
            
            (* New to the Simplex tableau *)
      | _ -> 
          let j = addNewColumn (newNodeOwner e) in
          add_to j fact;
          List.iter scanExpression e.args
    in
    match fSymOfHead e.head, e.args with 
      INT n, _ -> add_to 0 ((float n) *. fact) 

    | BINOP L.Times, [a1; a2] -> begin
        match fSymOfHead a1.head, fSymOfHead a2.head with 
          INT n, _ -> loop a2 ((float n) *. fact)
        | _, INT n -> loop a1 ((float n) *. fact)
        | _, _ -> notInteresting ()
    end

    | BINOP L.Plus, [a1; a2] -> 
        loop a1 fact;
        loop a2 fact
      
    | BINOP L.Minus, [a1; a2] -> 
        loop a1 fact;
        loop a2 (-. fact)
      
      (* The rest are not of direct interest to Simplex *)
    | _ -> notInteresting ()
  in    
  
  loop eplus 1.0;
  if eminus != !nZero then
    loop eminus (-1.0);

  (* See if this row is equal to some other one *)
  let (id_row, eq_row) = collectEqualitiesWithRow new_row !liverow ([], []) in

  List.iter packageOneEquality eq_row;

  match id_row with
    [] -> begin 
      (match getNode (own) with
        Slack _ -> begin
          own.node <- Slack (eint, [(1, eplus); (-1, eminus) ]);
	  if !doPrint then
	    ignore (log "Setting expression of slack %a to %d + %a%a\n" 
                      d_owner own eint d_nodeTop eplus
                      insert
                      (if eminus != !nZero then
                        dprintf " - %a" d_nodeTop eminus
                      else
                        nil))
        end
      | _ -> ());
      own
    end
	
  | [(new_own, _)] -> begin
      (* This row is already equal to some existing node *)
      match getNode own with
	Node _ -> 
          List.iter packageOneEquality id_row;
          own

      | _ -> 
         (* We must undo the adding of the row. But perhaps the row that we 
          * have added has been killed as well. I turned this off because the 
          * undoing was being done in a strange way  *)

          (match !history with
            HistAdd o :: _ when o == own -> ()
          | _ -> begin
              ignore (E.log "must remove owner %a.Mark is %d\n" 
                        d_owner own mark); 
              printHistory ();  
              
              assert true
          end);

          (* undoMarkOpt (Some mark); *)
          undoAdd own;
	  new_own
  end
					(* There should not exist more than 1 
                                         * row identical with this one *)
  | l -> 
      printSimplexState "";
      ignore (E.log "%a" (docList (fun (x, _) -> d_owner () x))
                         l);
      E.s (E.bug "addSomeExpr: unexpected")
   
   (* Scan expressions and when a Simplex symbol is found call addExpression. 
    * This makes sure that all Simplex expressions (+, -, x and constants) 
    * have a name in the Simplex tableau *)
and scanExpression  (e: node) : unit = 
  match fSymOfHead e.head with 
  | BINOP L.Plus| BINOP L.Minus| BINOP L.Times | INT _ -> 
      addExpression e

  | _ ->
      List.iter scanExpression e.args 
       
   (* We are within a Simplex expression. Recurse down until the first 
    * uninterpreted symbol. From there scanExpression *)
and skipSimplex (nd: node) : unit = 
  match fSymOfHead nd.head with
    BINOP L.Plus | BINOP L.Minus | BINOP L.Times | UNINT _-> 
      List.iter skipSimplex nd.args
  | _ -> ()
	 
   (* This adds the expression to the Simplex tableau *)
and addExpression (nd: node) : unit = 
  match !(nd.spos) with 
    Row _ -> ()
  | Col _ -> ()
  | NotSimpl -> 
      ignore (addSomeExpr nd 0 !nZero (newNodeOwner nd))
       

(** Add a restriction that "eplus + eint - eminus >= 0" for the reason given 
 * by "whyrestr". Return the (restricted) owner for the combination. *)
let rec restrict 
    (countdownIntHeur: int) (* If "countdown > 0", then we use some integer 
                              * heuristics *)
    (eplus : node) 
    (eint : int)
    (eminus : node)
    (h : whyrestr) : owner =
       
  (* Create a new slack *)
  let new_slack : owner = newSlackOwner eint [ (1, eplus); (-1, eminus)] in
  (* Add the slack to the tableau *)
  let own = addSomeExpr eplus eint eminus new_slack in

  doRestrict countdownIntHeur own h;
  own
   

and doRestrict 
     (countdownIntHeur: int) 
     (own : owner) (h : whyrestr) : unit = 
  if !doPrint then 
    ignore (log "Restricting row %a\n" d_owner own);

  let opos = getPos own in
  let isDeadOwner = 
    match opos with
      Row i -> i < !liverow
    | Col j -> j < !livecol
    | _ -> E.s (E.bug "doRestrict")
  in
  if isDeadOwner then begin
    if !doPrint then 
      ignore (log "Trying to restrict dead owner %a. Reviving it!\n " 
                d_owner own);
    undoKill own
  end;
  
  let new_acc = 
    match opos with
      Col j -> 
	if j < !livecol then 
          () (* Still dead *)
	else begin
	  match transposeColumn j with
	    Row row -> doRestrict countdownIntHeur own h
		(* Could not transpose. Restrict it! *)
	  | Col _ -> 
              restrictOwn own h

          | _ -> E.s (E.bug "doRestrict: NotSimpl")
        end
		
	      
    | Row i -> begin
	match rowToExtreme i (MaxRow i) 0.001 with
	  Extreme max ->
	    if max < 0.0 then begin
              if !doPrint then
                ignore (log "Simplex has detected a contradiction. Row %a is maximized at %f\n" d_rowOwner i max);
    
              
              foundContradiction i h
	    end else begin
	      if max <  1.0 then begin
		restrictOwn own h;
		minimizeTableau !liverow (!nr_rows - 1)
	      end else begin
		restrictOwn own h
              end
            end

	| _ -> 
            restrictOwn own h
    end
    | _ -> E.s (E.bug "doRestrict")
  in
  checkSatisf ();

  (* we get here if we could not find a contradiction *)
  if countdownIntHeur > 0 then 
    integerRestriction countdownIntHeur;

(*
  if !doPrint then 
    ignore (log "Equalities after restiction: %a\n"
	      (docList (fun x -> d_message () x)) acc');
*)  
  ()
   

					(* Heuristics for integers *)
and integerRestriction (countdownIntHeur: int) : unit = 
  if !doPrint then 
    ignore (log "\n**** Simplex is trying the integer heuristic (%d rounds left)***\n\n" countdownIntHeur);

  (* Invoked on a row whose constant is not an integer *)
  let doOneRow (row : int) : unit = 
    if !doPrint then
      ignore (log "Restricting row %a to integers\n" d_rowOwner row);
	     
    (* Split the non-zero coefficients in "row" into those whose columns are 
     * restricted to be equal to 0, less than 0 (restricted column with 
     * negative coefficient), greater than 0, and the 
     * non-restricted columns *)
    let (eq: linComb), (min: linComb), (max: linComb), (other: linComb) = 
      for_a
	(fun j ((eq, min, max, any) as acc) ->
	  let coeff : float = stab row j in
	  let own : owner = ownedByCol j in
	  if coeff = 0.0 then
	    acc
	  else
	    if j < !livecol then
	      ((coeff, own) :: eq, min, max, any)
	    else begin
	      match own.restr with
		None -> (eq, min, max, (coeff, own) :: any)
	      | Some _ ->
		  if coeff < 0.0 then
		    (eq, (coeff, own) :: min, max, any)
		  else
		    (eq, min, (coeff, own) :: max, any)
	    end)
	([],[],[],[])
	(1, !nr_cols - 1)
    in 
	     
    let gcd_other = updateGCDList other 1.0 in
    let ct = stab row 0 in (* This is the non-integer constant *)

					(* Do the min part *)
    let makeIntOwnList fact acc lst =
      List.fold_left 
        (fun acc (x, own) ->
	  (makeInt (x *. fact), own) :: acc)
        acc
        lst
    in
    let makeOwnList fact acc lst =
      List.fold_left 
        (fun acc (x, own) -> (x *. fact, own) :: acc) 
        acc
        lst
    in
    
    begin
      match max with
	[] -> ()
      | _ -> begin
          (* There is some "max": *)
          (* row = ct + min + max + eq + other *)
          
	  let gcd = updateGCDList min gcd_other in
          (* If we multiply the row by gcd we make the "min" and "other" 
           * integers *)

          (* We have that 
              gcd * row - gcd * (min + other) - gcd * ct = gcd * (eq + max)
          *)

          (* gcd * row - gcd * (min + other) has only integer coefficients *)
          let ceil_ct = ceil (ct *. gcd) in

            
          let diff = ceil_ct -. ct *. gcd in
          (* We strengthen to: 
             gcd * row - gcd * (min + other) - ceil_ct >= 0 
          *)

          (* See if it is worth increasing the constant *)
	  if diff /. gcd > threshHold then begin
	    if !doPrint then
	      ignore (log "Strengthening lower bound for row %a by %f\n"
                        d_rowOwner row (diff /. gcd));
            
            (* 
               let e1 = gcd * row - gcd * (min + other) - ceil_ct
             *)

            let gcd_min_other = 
              List.map 
                (fun (c, o) -> (-. c *. gcd, o))
                (min @ other)
            in
            let e1 = 
              buildExprFromList
                ((-. ceil_ct, ownedByCol 0) ::
                 (gcd, ownedByRow row) ::
                 gcd_min_other)
            in

            (** Add a row for e1 *)
            let new_slack : owner = 
              let ct, f = 
                buildLinCombFromList 
                  ((gcd, ownedByRow row) ::  gcd_min_other) in
              newSlackOwner 
                (ct + makeInt (-. ceil_ct))
                f
            in
            let new_row = addNewRow new_slack in

            accumulateInRow new_row row gcd;
            List.iter 
              (fun (coeff, own) -> 
                match !(own.ospos) with 
                  Col j -> utab new_row j coeff
                | _ -> E.s (E.bug "integerRestriction"))
	      gcd_min_other;
            utab new_row 0 ((stab new_row 0) -. ceil_ct);


            (* Compute gcd * (eq + max) *)
            let gcd_eq_max: linComb  = 
              List.map (fun (c, o) -> (c *. gcd, o))
                (eq @ max)
            in


            (* We have that 
                eqCheck(e1 + diff = gcd_eq_max)
                gcd_eq_max >= 0
                diff < 1
             *)
            

            (* Restrict e1 >= 0 *)
            doRestrict
              (countdownIntHeur - 1) (* to avoid looping *)
              new_slack
              (GEQINT (new_slack, 
                       e1,
                       (fun _ -> 
                         makeGeqProof e1 L.zero
                           diff gcd_eq_max)))
	  end
      end
    end;

    begin
      match min with
	[] -> ()
      | _ -> begin
          
	  let gcd = updateGCDList max gcd_other in

          (* We have that 
              gcd * row - gcd * (max + other) - gcd * ct = gcd * (eq + min)
          *)
          
	  let floor_ct = floor (ct *. gcd) in

          let diff = ct *. gcd -. floor_ct in
          (* We strengthen to: 
             gcd * row - gcd * (max + other) - floor_ct <= 0 
          *)


	  if diff /. gcd > threshHold then begin
	    if !doPrint then
	      ignore (log "Strengthening upper bound for row %a by %f\n"
                        d_rowOwner row (diff /. gcd));


             (* 
               let e1 = - gcd * row + gcd * (max + other) + floor_ct
               We have e1 = - gcd * (eq + min) - diff
               We have that e1 >= 0
             *)

            (* Compute gcd * (max + other) *)
            let gcd_max_other: linComb  = 
              List.map (fun (c, o) -> (c *. gcd, o))
                (max @ other)
            in

            let e1 = 
              buildExprFromList
                ((floor_ct, ownedByCol 0) ::
                 (-. gcd, ownedByRow row) ::
                 gcd_max_other)
            in

            (** Add a row for e1 *)
            let new_slack : owner = 
              let ct, f = 
                buildLinCombFromList 
                  ((-. gcd, ownedByRow row) ::  gcd_max_other) in
              newSlackOwner 
                (ct + makeInt floor_ct)
                f
            in
            let new_row = addNewRow new_slack in

            accumulateInRow new_row row (-. gcd);
            List.iter 
              (fun (coeff, own) -> 
                match !(own.ospos) with 
                  Col j -> utab new_row j coeff
                | _ -> E.s (E.bug "integerRestriction"))
	      gcd_max_other;
            utab new_row 0 ((stab new_row 0) +. floor_ct);



            (* Compute - gcd * (eq + min) *)
            let gcd_eq_min: linComb  = 
              List.map (fun (c, o) -> (-. c *. gcd, o))
                (eq @ min)
            in


            (* We have that 
                eqCheck(e1 + diff = gcd_eq_min)
                gcd_eq_min >= 0
                diff < 1
             *)
            
            (* Restrict e1 >= 0 *)
            doRestrict
              (countdownIntHeur - 1) (* to avoid looping *)
              new_slack
              (GEQINT (new_slack, 
                       e1,
                       (fun _ -> 
                         makeGeqProof e1 L.zero
                           diff gcd_eq_min)))
          end 
      end
    end;
  in (* doOneRow *)
					(* Looks at all rows whose constant 
                                         * field is not an integer *)
  for_a
    (fun i _ ->
      match isInt (stab i 0) with
	None ->
	    				(* Add a restricted row *)
	  doOneRow i;
	  killRow i (ownedByRow i) !nodeZeroOwner

		   
      | Some _ -> ())
    ()
    (0, !nr_rows - 1)
   


	 
let processEquality 
    (countdownIntHeur: int)
    (e1 : node)
    (e2 : node)
    (msg : message) (* proves e1 = e2 *)
    : unit = 

  let own = 
    restrict countdownIntHeur e1 0 e2 (BYEQ (msg, e1.exp, e2.exp, 
                                             (fun _ -> msg.proof)))
  in
  
  match getPos(own) with
    Row _ -> 
      ignore (restrict countdownIntHeur e2 0 e1 
                (BYEQ (msg, e2.exp, e1.exp, 
                       (fun _ -> P.Symm msg.proof))))


  | Col j ->
      addKillToHistory own !nodeZeroOwner;

      packageOneEquality (own, !nodeZeroOwner);
      
      minimizeTableau 
        !liverow (!nr_rows - 1)

       
  | _ -> E.s (E.bug "processEquality")

let processDisEquality 
    (countdownIntHeur: int) 
    (e1 : node)
    (e2 : node)
    (msg : message) (* proves e1 <> e2 *)
    : unit =
					(* Assert e1 < e2 and see if 
                                         * contradiction *)

  
  (* Restrict e1 < e2, or more precisely, e2 - 1 >= e1 *)
  let restrictLt (mkprf: unit -> P.proof) : unit =
    ignore 
      (restrict countdownIntHeur e2 (-1) e1 
         (BYGEQ(msg, 
                e2.exp,
                (L.plus e1.exp  L.one), 
                mkprf)))
  in
  
  (* Restrict e2 < e1, or more precisely, e1 - 1 >= e2 *)
  let restrictGt (mkprf: unit -> P.proof) : unit =
    ignore (restrict countdownIntHeur e1 (-1) e2 
              (BYGEQ(msg, e1.exp, (L.plus e2.exp L.one), mkprf)))
  in

  (* In order to assert e1 <> e2, we first assert e1 < e2. If we get a 
   * contradiction, then we assert e1 > e2 *)
  let assume_lt = U.registerSymbolName (L.newVar "AssumeLt") (* e1 < e2 *) in
  let finished =
    if !doPrint then
      ignore (log "Simplex tries to assert %a < %a when doing disequality\n"
                d_nodeTop e1 d_nodeTop e2);
    
    let mark = registerUndoMarker () in
    try

      restrictLt 
        (fun _ -> P.G("ltgeq", 
                      [P.P (P.Hyp assume_lt)]));
      undoMarkOpt (Some mark);

      false
    with S.Contradiction prf_false -> begin
      newMessages := [];

      (* We got a contradiction from the assumption that e1 < e2 *)
      undoMarkOpt (Some mark);
      if !doPrint then
	ignore (log "Simplex converts disequality to %a > %a\n"
                  d_nodeTop e1 d_nodeTop e2);
      (* Make a proof that e1 >= e2 *)
      let geqprf = 
        P.mkContraProof 
          (L.geq e1.exp e2.exp)
          assume_lt
          prf_false
      in
      restrictGt (fun _ -> P.G("neqGt", [ P.P geqprf; 
                                          P.P msg.proof ]));
      true (* We are done *)
    end
  in
  let finished =
    if finished then
      true
    else begin
      let assume_gt = U.registerSymbolName (L.newVar "AssumeGt") (* e1 > e2 *) in
      if !doPrint then
	ignore (log "Simplex tries to assert %a > %a when doing disequality\n" 
                  d_nodeTop e1 d_nodeTop e2);
      let mark = registerUndoMarker () in
      try
        restrictGt ((fun _ -> P.G("gtgeq", 
                                  [P.P (P.Hyp assume_gt)])));
        undoMarkOpt (Some mark);
        
        false
      with (S.Contradiction prf_false) -> begin 
        newMessages := [];
        undoMarkOpt (Some mark);
	if !doPrint then
	  ignore (log "Simplex converts disequality to %a < %a\n"
                    d_nodeTop e1 d_nodeTop e2);
        (* We got a contradiction on the fact that e1 >= e2 + 1. This leads 
         * to a proof that e2 >= e1 *)
        let geqprf = 
          P.mkContraProof 
            (L.geq e2.exp e1.exp)
            assume_gt
            prf_false
        in
        restrictLt 
          (fun _ -> P.G("neqGt", 
                        [ P.P geqprf;
                          P.P (P.G("neqSym", 
                                   [P.P msg.proof ]))]));
	true
      end
    end
  in
  if not finished then begin 
    scanExpression e1; 
    scanExpression e2
  end


	 
				    (* This is the main message processor *)
let assume (msg: message) : unit =
  if !doPrint then begin
    ignore (log "**********************\n");
    ignore (log "Simplex processes message %a\n"
              d_message msg);
  end;
  if msg.owner != sxid then begin
    (* see if we must negate the goal *)
    let body' = msg.body in
    (* see if we want to use integer heuristics. Only for goals *)
    let countdownIntHeur = if msg.fromgoal then !intHeurRounds else 0 in
    match body' with 
      Intern (L.SetEq, n1, n2) -> 
        processEquality countdownIntHeur n1 n2 msg

    | Intern(L.SetNe, n1, n2) ->             
        processDisEquality countdownIntHeur n1 n2 msg

    | Intern(L.SetGe, n1, n2) -> 
        ignore (restrict countdownIntHeur 
                  n1 0 n2  
                  (BYGEQ(msg, n1.exp, n2.exp, 
                         (fun _ -> msg.proof))))
          
    | Intern(L.SetGt, n1, n2) ->
	ignore (restrict countdownIntHeur n1 (-1) n2  
                  (BYGEQ(msg, n1.exp, (L.plus n2.exp L.one), 
                         (fun _ -> P.G("gtgeq", [P.P msg.proof])))))
  
    | Intern (L.SetLe, n1, n2) ->
	ignore (restrict countdownIntHeur n2 0 n1  
                  (BYGEQ(msg, n2.exp, n1.exp, 
                         (fun _ -> P.G("leqgeq", [P.P msg.proof])))))

    | Intern (L.SetLt, n1, n2) -> 
        ignore (restrict countdownIntHeur  n2 (-1) n1  
                  (BYGEQ(msg, n2.exp, (L.plus n1.exp L.one),
                         (fun _ -> P.G("ltgeq", [P.P msg.proof])))))
          
    | Extern (L.Bool true) -> ()
          
          
    | Intern (bop, n1, n2) -> 
        (** Look for something of interest *)
        scanExpression n1;
        scanExpression n2

    | Extern (L.PNez at as p) -> 
        (* Scan the subexpressions for something of interest *)
        let an, _ = insertExpNode at in
        scanExpression an

    | _ -> () (* Ignore non atoms *)
  end

exception Skip of bool
let witness () : unit = 
  let debugWitness = false in 

  if debugWitness then printSimplexState "witness";

  let accum (n: node) (ct: float) = 
    let fact = updateGCD ct 1.0 in 
    let newwit = 
      L.eq (if fact = 1.0 then n.exp else
      (L.times n.exp (L.integer (int_of_float fact))))
        (L.integer (int_of_float (fact *. ct))) in 
    if debugWitness then ignore (log "Simplex witness: %a\n"  L.d_pred newwit);
    accumWitnessPred newwit
  in
  (* We scan each row. We do not report rows that are now owned by Nodes. We 
   * do not report nodes that are of the form "f1 * S1 + f2 * S2 + N", where 
   * Si are null-restricted slacks. These are equalities between nodes, which 
   * the EDAG will report. *)
  (* All rows are equal to the value in first column *)

  (** keep track of interesting columns *)
  let interestingCols: unit IH.t = IH.create 13 in 
  for i = 0 to !nr_rows - 1 do 
    (* Skip the row if owned by NoNode or by Zero *)
    let rowown = (ownedByRow i) in 
    let skip = 
      match rowown.node with 
        Node n when n == !nZero -> true
      | NoNode -> true
      | _ -> false in
    if not skip then begin 
      let skip = 
        let plusOne = ref false in (* We found the +1 coeff for a node *)
          
        (* Scan the columns. Throw Not_found if this row is interesting *)
        try 
          for j = 1 to !nr_cols - 1 do 
            let colown = ownedByCol j in 
            let skipBecause ?(skip=true) msg = 
              if debugWitness then 
                ignore (log "%s row %a because col %a: %s\n" 
                          (if skip then "Ignore" else "Keep") 
                          d_owner rowown d_owner colown msg);
              raise (Skip skip)
            in
            let s = stab i j in 
            if s = 0.0 then 
              ()
            else begin
              (* non-zero entry *)
              match colown.node with 
                NoNode -> skipBecause "NoNode entry"
              | Slack _ -> begin 
                  match colown.restr with 
                    Some (BYEQ _) -> ()
                  | _ -> skipBecause ~skip:false "non nullSlack"
              end
              | Node _ -> 
                  if s <> 1.0 then skipBecause "non +1 entry";
                  if !plusOne then skipBecause "multiple +1 entries";
                  plusOne := true
            end
          done;
          if stab i 0 <> 0.0 then
            false
          else if !plusOne then begin 
            if debugWitness then 
              ignore (log "Ignore row %a because has exactly one +1 entry\n"
                        d_owner rowown);
            true
          end else
            false (* Do not skip it. It is interesting *)
        with Skip v -> v
      in
      if not skip then begin 
        (* Scan the row and mark the columns that appear in it as interesting*)
        for j = 1 to !nr_cols - 1 do 
          if stab i j <> 0.0 then 
            IH.add interestingCols j ()
        done;
        (* If owned by a Node, we must announce an equality *)
        match rowown.node with 
          NoNode -> ()
        | Node n -> 
            accum n (stab i 0)
        | Slack _ -> ()
      end
    end
  done;
  (* Now do the interesting columns *)
  for j = 1 to !nr_cols - 1 do 
    if IH.mem interestingCols j then begin 
      match (ownedByCol j).node with
        NoNode -> ()
      | Node n -> accum n 0.0
      | Slack _ -> ()
    end;
  done

  
 
let initialize () = 
  if !doPrint then ignore (log "Initializing Simplex...\n");

  if !nZero == dummyNode then 
    nZero := fst (insertExpNode L.zero);
  if !nOne == dummyNode then 
    nOne := fst (insertExpNode L.one);

	 
  let zeroOwn = newNodeOwner !nZero in

  initTableau ();
  nr_rows := 1;
  nr_cols := 1;
  livecol := 1;
  liverow := 0;
  divisors := [];
  killers := [];
  initHistory ();
  
  setOnPos zeroOwn (Row 0);
  setOnPos (newNodeOwner !nOne) (Col 0);

  nodeZeroOwner := zeroOwn;
  zeroOwn.restr <- 
     Some (BYEQ ({ owner = sxid; 
                   fromgoal = false;
                   proof = P.Refl L.zero;
                   body = Intern (L.SetEq, !nZero, !nZero);
                 },
                  L.zero,
                  L.zero,
                  (fun _ -> P.G("eq0", []))))


let entries = 
  { S.emptySatproc with 
    S.init     = initialize;
    S.defaultEnabled = false;
    S.snapshot = snapshot;
    S.assume   = assume;
    S.debug    = doPrint;
    S.ident    = sxid;
    S.witness  = witness;
    S.descr    = "Simplex";
  } 
