#include "main.h"
#include <time.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <getopt.h>

#include "input.h"
#include "hash.h"

#include <sys/times.h>
#include <unistd.h>

#define safeArithPrime 268435399
#define maxFastPrime 32749
#define biggestPrime 2147483647

//#define SMALL_PRIME

#ifdef SMALL_PRIME
 #define randPrime 9973
#else
 #define randPrime safeArithPrime
#endif

double invPrime;

#ifdef DEBUG
 #define repetitions 1
#else 
 #define repetitions 40
#endif

#define timing 0
#define debug 0

#define optimization 1

#define preCompute 0 
//#define nrRandom 16384
//#define nrRandom 32768
#define nrRandom 8192
int nrPoints;

int test=0; //if (test != 0) then run just once 

#ifdef DEBUG
#define ENTER(a) printf("Entering %s\n",a)
#define LEAVE(a) printf("Leaving %s\n",a)
#define MARK(i) printf("Here %d\n",i)
#else
#define ENTER(a) 
#define LEAVE(a)
#define MARK(a)
#endif

#define enter(a) ENTER(a)
#define leave(a) LEAVE(a) 
#define mark(i) MARK(i)

#define MARK(i)

#define min(a,b) a<b?a:b
#define max(a,b) a>b?a:b

// Return Values
#define FALSE_GOAL 0        /* Goal is false */
#define PROVED_GOAL 1       /* Goal is true */
#define INCONSISTENT_GOAL 2 /* Goal is inconsistent */

// Exit Codes
#define FEW_POINTS 3        /* Too few points */
#define UNIMP_PRED 20       /* Unimplemented Predicate */
#define UNIMP_EXP 21        /* Unimplemented Expression */

typedef enum { VAR_TERM, FUNC_TERM } term_type;

typedef struct {
  term_type tag;
  union {
    char* str;
    NewValue* v;
  } u;
} term;



struct difflist {
  value v;
  struct groupNode* gn;
  struct difflist* next;
};

struct adjustment {
  value af;
  int   j;
  int   k;
  element w;
  PHASH tmp_func_assignment;
};


/* --------- GLOBAL VARIABLES -----------------*/
int nrAdjusts;        // number of adjustments performed
int maxFrom;
int nrProcessed;      // number of equalities processed
int nrAdjustNewTerms; // number of cheap adjustment performed

int firsttime = 1; // for printing debugging info 
struct difflist *pendingDiffs;


// HashTables
PHASH var_assignment;
PHASH funcHash; 

/* to convert function terms into terms with a single function
   term. a random value is stored for each function symbol
*/
PHASH randWeight_func;

/* to convert function terms into terms with a unary function term
   a random value is stored for each arity
*/
element randWeight_arg[16]; 


/* --------- PROTOTYPES OF FUNCTIONS -------------*/

bool process_assumption (struct exp* e);
bool verify_goal(struct exp*);
int prove_goal(struct pred *p);
void compute_adjustment_factors(value diff, value af,
                                int* jp, int* kp, element* wp, int from);
void adjust_value(value v, value af, int j, int k, element r, value result);
bool adjust_values(value diff, struct groupNode* gn);
bool adjust_values2 (value diff, struct groupNode* gn);
element get_randWeight_func(char *str);
void addPendingDiffs(value diff, struct groupNode* gn);
void print_NewValues();
void print_pendingDiffs();

// The following functions operate on elements
element get_rand();
element add(element, element);
element sub(element, element);
element divide(element, element);
element join(element, element, element);

// The following functions operate on values
NewValue* rand_NewValue();
void add_value(value, value, value, int);
void sub_value(value, value, value, int);
void mult_value(value, value, value, int);
void divide_value(value, value, value, int);
void join_value(value, value, value, value, int);
bool is_constant_value(value, int);
bool is_equal_value(value, value, int);
bool is_equal_zero(value, int);

#ifdef SCOTT_DEBUG
void checkHeap();
#endif


/*--------------- GROUPS --------------------------*/
/*

   All values that are "only" function arguments belong to a set of groups, namely 
   the group of the top level terms of which they are composed. For example, 
   if there is a term F(x+y) then the set of groups of the term (x+y+z) 
   = \{ G(x) \} \cup \{ G(y) \}.

   Any other value belongs to one group.

*/

int nrGroups;

struct groupNode* getGroupRep(struct groupNode* gn)
{
  struct groupNode *gn1 = gn, *gn2;
  _ASSERT(gn != NULL); 
  if (!gn->rep) { return gn; }

  while (gn1->rep) {
    gn1=gn1->rep;
  }

  while(gn->rep) {
    gn2 = gn->rep;
    gn->rep = gn1;
    gn = gn2;
  }
  return gn1;
}

struct groupNode* newGroup()
{
  struct groupNode* gn;
  gn = (struct groupNode*)malloc(sizeof(struct groupNode));
  gn->g = nrGroups++;
  gn->from=0;
  gn->rep=NULL;
  return gn;
}

void setGroup(struct groupNode* gn1, struct groupNode* gn2)
{
  struct groupNode *gnr1, *gnr2;
  gnr1 = getGroupRep(gn1);
  gnr2 = getGroupRep(gn2);
  if (gnr1 != gnr2) {
    gnr1->rep=gnr2;
    gnr2->from=max(gnr2->from,gnr1->from);
    maxFrom = max(maxFrom, gnr2->from);
  }
}





/*--------------- FUNCTIONS BEGIN ------------------*/
inline element inverse(element a1)
{
  unsigned long a=a1,b=randPrime,q;
  unsigned long n0;
  int x,x0=1,x1=0;
  _ASSERT(a1 < randPrime);
  do {
    // q = a/b; r = a - q*b; a = b; b = r; 
    n0 = a;
    __asm__ ("divl %4"				
             : "=a" ((unsigned int) (q)),   	
               "=d" ((unsigned int) (n0))	
             : "0" ((unsigned int) (n0)),	
               "1" ((unsigned int) (0)),	
               "rm" ((unsigned int) (b)));
    a = b; 
    b = n0;

    
    x = x0 - x1*q; x0 = x1; x1 = x;
  }
  while (b != 0);
  if (x0<0) { x0 = x0+randPrime; }
  return x0;
}


inline element pmod(int i)
{
  i = i % randPrime;
  if (i < 0) i = i + randPrime;
  return (element) i;
}

inline element add(element i1, element i2)
{
  int i = i1+i2;
  int r = i-randPrime;
  if (r >= 0) return r; else return i;
}

inline element sub(element i1, element i2)
{
  int i = i1-i2;
  if (i < 0) i += randPrime;
  return i;
}

// Define this if you want to do the unsigned division
#define USE_UNSIGNED
static inline unsigned int mult_ll(int i1, int i2)
{
  unsigned long res;
#ifdef USE_UNSIGNED  
  unsigned long long c;
  if(i1 < 0) { i1 += randPrime; }
#else
           long long c;
#endif  
  c = (unsigned long long)i1*i2;
  res = (unsigned long)(c % randPrime);
#ifdef USE_UNSIGNED  
  if(res & 0x80000000) { res += randPrime; }
#endif  
  return res;
}


inline element mult_asm(element i1, element i2)
{
  union {
    struct DWstruct {long low, high;} s;
    long long ll;
  } nn;
  unsigned long n0, n1;
  unsigned long res; 
  unsigned long q0;
  _ASSERT(i1 < randPrime && i2 < randPrime);

  nn.ll =  (unsigned long long)i1*i2;
  n0 = nn.s.low;
  n1 = nn.s.high;


  /* We must ensure the n1 (the high 32-bit of i1*i2) is less than 
   * randPrime. This is crucial because the "divl" instruction will raise an 
   * exception if the quotient does not fit on 32 bits. */
  _ASSERT(n1 < randPrime);

   __asm__ ("divl %4"				
            : "=a" ((unsigned int) (q0)),   	
            "=d" ((unsigned int) (n0))	
            : "0" ((unsigned int) (n0)),	
            "1" ((unsigned int) (n1)),	
            "rm" ((unsigned int) (randPrime)));
   res = n0;
   
#ifdef DEBUG
  {
    unsigned long res_ll = mult_ll(i1, i2);
    _ASSERT(res == res_ll);
  }
#endif
  return res;
}

// #define mult mult_ll
#define mult mult_asm


inline element divide(element i1, element i2)
{
  int i;
#ifdef DEBUG
  _ASSERT(i2 != 0)
#endif  
  i = inverse(i2);
  return mult(i1,i);
}

inline element join(element i1, element i2, element w)
{
  return add(mult(sub(i1, i2), w), i2);
}


inline element get_rand()
{
  int t;
  t = rand ();
#ifdef SMALL_PRIME
  t = t % randPrime;
#else  
  t = t >> 4;
#endif
  _ASSERT(t < randPrime);
  return (element) t;
}
		       
__inline NewValue* rand_NewValue()
{
  int i,found;
  HASH_ENTRY* phe;
  NewValue* result = newValue();
  value data = result->data;
  for(i=0; i<nrPoints; i++) {
    data[i] = get_rand();
  }
  result->rep=NULL; result->func=NULL;
  phe=HashLookupOrAdd(funcHash,result,&found);
  phe->key=result;
  return result;
}


__inline void mult_value(value v1, value v2, value result, int from)
{
  int i;
  for (i=from; i<nrPoints; i++)
    result[i] = mult(v1[i],v2[i]);
}


__inline void add_value(value v1, value v2, value result, int from)
{
  int i;
  for (i=from; i<nrPoints; i++)
    result[i] = add(v1[i],v2[i]);
}


__inline void sub_value(value v1, value v2, value result, int from)
{
  int i;
  for (i=from; i<nrPoints; i++)
    result[i] = sub(v1[i],v2[i]);
}


__inline void join_value(value v1, value v2, value w, value result, int from)
{
  int i;
  for (i=from; i<nrPoints; i++)
    result[i] = join(v1[i],v2[i],w[i]);
}

__inline void divide_value(value v1, value v2, value result, int from)
{
  int i;
  for (i=from; i<nrPoints; i++)
    result[i] = divide(v1[i],v2[i]);
}

__inline bool is_constant_value(value v, int from)
{
  int i,c = v[from];
  for (i=from+1+1; i<nrPoints; i++)
    if (c != v[i]) return 0;
  return 1;
}


__inline bool is_equal_value(value v1, value v2, int from)
{
  int i;
  for (i=from; i<nrPoints; i++) {
    if (v1[i] != v2[i]) return 0;
  }
  return 1;
}


__inline bool is_equal_zero(value v1, int from)
{
  int i;
  for (i=from; i<nrPoints; i++) {
    if (v1[i] != 0) return 0;
  }
  return 1;
}
		       

// HASHTABLES STUFF

void print_value(value v, int from)
{
  int i;
  printf("(");
  for (i=from;i<nrPoints-1;i++)
    printf("%lu,",v[i]);
  printf("%lu)",v[nrPoints-1]);
}

void f_print_var(HASH_KEY str, void* v, void* closure)
{
  NewValue *nv = getValueRep((NewValue*)v);
  struct groupNode *gn = getGroupRep(nv->group);
  printf("%s=",(char*)str);print_value(nv->data, gn->from); printf("[%d]",gn->g);
  printf("\n");
}

void f_print_func(HASH_KEY key, void* data, void* closure)
{
  NewValue *v, *arg;
  struct groupNode *gnv, *gnarg;
  _ASSERT(key != NULL);
  arg = getValueRep((NewValue*)key);
  v = getValueRep(arg->func);
  if (v) {
    gnv = getGroupRep(arg->group);
    gnarg = getGroupRep(v->group);
    printf("F");
    print_value(arg->data, gnarg->from);
    printf("[%d]",gnarg->g);
    printf("="); print_value(v->data, gnv->from);
    printf("[%d]",gnv->g);
    printf("\n");
  }
}

void dump_state()
{
  void *closure;
  printf("Dumping State\n");
  IterHash(var_assignment,f_print_var,(void*)&closure);
  IterHash(funcHash,f_print_func,(void*)&closure);
  printf("\n");
  //print_NewValues();
  print_pendingDiffs();
  printf("\n");
}


bool NewValue2int(HASH_KEY key)
{
  int i = nrPoints-1;
  value v = ((NewValue*)key)->data;
  _ASSERT(i >= 2);
  return (v[i]+v[i-1]+v[i-2]);
}



bool NewValueCompare(HASH_KEY v1, HASH_KEY v2) {
  int from;
  from = maxFrom;
  return ! memcmp(&((((NewValue*)v1)->data)[from + 1]),
                  &((((NewValue*)v2)->data)[from + 1]),
                  sizeof(element) * (nrPoints - from - 2));
}



void initialize()
{
  nrAdjusts=0;
  nrAdjustNewTerms=0;
  nrProcessed=0;
  preallocateHashes ();
  var_assignment = NewStringPointerHash();
  funcHash = NewHash(NewValue2int, NewValueCompare);
  maxFrom=0;
  nrGroups=0;
}



void cleanup()
{
  FreeHash(var_assignment);
  FreeHash(funcHash);
  releaseAllValues();
}



element get_randWeight_func(char *str)
{
  int found;
  HASH_ENTRY *e = HashLookupOrAdd(randWeight_func, str, &found);
  if(! found) {
    e->data = (void*)get_rand();
#if DEBUG > 2
    e->data = (void*)(strlen(str));
#endif
  }
  return (element)e->data;
}





term newTerm;
element factor;

// sets pgn to be the rep of the groups of the top level terms in the expression
void eval_expr(struct exp* pe, element g, int infunc, value result, struct groupNode** pgn)
{
  int i,j,found,from;
  struct groupNode* gn_arg;
  NewValue *v;
  value arg;
  NewValue *tmp;
  element a;
  struct explist *el;
  HASH_ENTRY* phe;
  char* var;
#if DEBUG > 3  
  printf("Entering eval_expr::exp = ");
  print_exp(pe); printf("\n");
  printf(", g = %lu, infunc = %d, result = ",g,infunc);
  print_value(result,(*pgn)->from); printf("\n");
#endif
   switch (pe->tag) {
  case INTEGER_EXP:
    from = ((*pgn) == NULL)?0:getGroupRep(*pgn)->from;
    for (i=from;i<nrPoints;i++) {
      result[i] = add(result[i], mult(g, pmod(pe->u.integer)));
    } 
    break;
  case PLUS_EXP:
    eval_expr(pe->u.exps.e1, g, infunc, result, pgn);
    eval_expr(pe->u.exps.e2, g, infunc, result, pgn);
    break;
  case MINUS_EXP:
    eval_expr(pe->u.exps.e1, g, infunc, result, pgn);
    eval_expr(pe->u.exps.e2, sub(0,g), infunc, result, pgn);
    break;
  case TIMES_EXP:
    if (pe->u.exps.e1->tag == INTEGER_EXP) {
      int g1 = mult(g, pmod(pe->u.exps.e1->u.integer));
      eval_expr(pe->u.exps.e2, g1, infunc, result, pgn);
    }
    else if (pe->u.exps.e2->tag == INTEGER_EXP) {
      int g1 = mult(g, pmod(pe->u.exps.e2->u.integer));
      eval_expr(pe->u.exps.e1, g1, infunc, result, pgn);
    }
    else {
      _ERROR0("Unimplemented: Perhaps a Non-linear expression\n");
    }
    break;
  case VAR_EXP:
    // Get value for the variable, and update NewTerm
    var = pe->u.var;
    phe = HashLookupOrAdd(var_assignment, var, &found);
    if (found == 0) {
      v = rand_NewValue();
      if (*pgn == NULL) *pgn = newGroup();
      v->group = getGroupRep(*pgn);
      phe->key = var; phe->data = v;
      if (factor == 0 && !infunc) {
	factor = g; newTerm.tag = VAR_TERM; newTerm.u.str = var;
      }
    }
    else {
      v = getValueRep((NewValue*)(phe->data));
      if (*pgn == NULL) *pgn = getGroupRep(v->group);
      else setGroup(v->group,*pgn);
      if (factor != 0 && newTerm.tag == VAR_TERM
          && newTerm.u.str == var) { 
	if (infunc == 1) { factor = 0; }
	else { factor = factor + g; }
      }
    }
    from = getGroupRep(*pgn)->from;
    // Update result
    for (i=from;i<nrPoints;i++) {
      result[i]= add(result[i],mult(g,v->data[i]));
    };
    break;
  case FUN_EXP:
    _ASSERT(pe->u.func.elist != NULL);
    /* Construct the argument for the single unary function term
       from the real function term f(arg1,...argt) as
       arg = rand("f") + rand(1)*arg1 + .... rand(t)*arg(t)
    */
    // get_randWeight_func("f") returns rand("f") in the above example
    var = pe->u.func.var;
    a = get_randWeight_func(var);
    arg=(value)malloc(nrPoints*sizeof(element));
    // arg cannot be freed unless we make a copy of it for key and newTerm.v below
    for (i=0;i<nrPoints;i++) {
      arg[i]=a;
    }
    j=0;
    el = pe->u.func.elist;
    gn_arg=NULL;
    while (el != NULL) {
      // randWeight_arg[j] contains rand(j) in the above example
      eval_expr(&(el->e),randWeight_arg[j++],1,arg,&gn_arg);
      el = el->next;
    }
    
    // Get value for the function term
    tmp=(NewValue*)malloc(sizeof(struct NewValue)+nrPoints*sizeof(element));
    from = (gn_arg==NULL)?0:gn_arg->from; 
    for (i=from;i<nrPoints;i++) (tmp->data)[i]=arg[i];
    phe = HashLookupOrAdd(funcHash, tmp, &found);
    if (found == 0) {
      v = rand_NewValue();
      if (*pgn == NULL) *pgn = newGroup();
      v->group = getGroupRep(*pgn);
      tmp=newValue();
      for (i=from;i<nrPoints;i++) (tmp->data)[i]=arg[i];
      tmp->rep=NULL;
      tmp->func=v;
      tmp->group = (gn_arg==NULL)?NULL:getGroupRep(gn_arg);
      phe->key = tmp;
      if (factor == 0 && !infunc) {
	newTerm.tag = FUNC_TERM;
	newTerm.u.v = tmp;
	factor = g;  
#if DEBUG > 3
	printf("newTerm is Func: ");print_value(newTerm.u.v);printf("\n");
#endif
      } 
    }
    else if (! (getValueRep((NewValue*)phe->key))->func) {
      v = rand_NewValue();
      if (*pgn == NULL) *pgn = newGroup();
      v->group = getGroupRep(*pgn);
      getValueRep((NewValue*)phe->key)->func=v;
      
      if (factor == 0 && !infunc) {
	newTerm.tag = FUNC_TERM;
	newTerm.u.v = phe->key;
	factor = g;  
      }
    }
    else {
      v = getValueRep((getValueRep((NewValue*)phe->key))->func);
      if (*pgn == NULL) *pgn = getGroupRep(v->group);
      else setGroup(v->group,*pgn); 
      if (factor != 0 && newTerm.tag == FUNC_TERM && is_equal_value(newTerm.u.v->data, arg, from)) {
	if (infunc) { factor = 0; } 
	else { factor = factor + g; }
      } 
    }
    from = getGroupRep(*pgn)->from;
    // Update result
    for (i=from;i<nrPoints;i++) {
      result[i]= add(result[i],mult(g,v->data[i]));
    }
    //free(tmp); free(arg);
    break;
  default: exit(UNIMP_EXP);
  }
}

/* Input argument: diff \neq [c,...c] for any c
   Output arguments:
   -> j and k such that diff.(j) \neq diff.(k)
   -> w such that p \new diff.(i) for all i,
       where p = (w * diff.(j)) + ((1-w) * diff.(k)).
   -> af: the sequence of w_i's, w_i = p/(p-diff.(i))
   We call (j,k,w,af) the adjustment factors.
*/
void compute_adjustment_factors(value diff, value af,
                                int* jp, int* kp, element* wp, int from)
{
  int k = from, j = -1;
  int i;
  element q;

  /* Make one scan over the diff array and we find (1) a q != diff[i] and 
   * (2) an index j such that diff[j] != diff[k] */
  {
  Redo: // It is extremly unlikely that we need to Redo
    q = get_rand();
    for(i=from;i<nrPoints;i++) {
      if(q == diff[i])
        goto Redo; // Break out and re-do
      if(diff[i] != diff[k])
        j = i; // Remember an index where the diff is not the same
    }
  }
  _ASSERT(j > from && j < nrPoints);

  
  for (i=from;i<nrPoints;i++) {
    af[i] = divide(q,sub(q,diff[i]));
  }
  
  *jp=j;
  *kp=k;
  *wp = divide(sub(q, diff[k]),
               sub(diff[j], diff[k]));
  
  _ASSERT(q == join(diff[j], diff[k], *wp));
  _ASSERT(*kp == from);
}

/* Adjusts one value. 
   Input: the value v to be adjusted and the adjustment factors (af,i1,i2,r),
   Output: returns the adjusted value in result.
*/
inline void adjust_value(value v, value af,
                         int j, int k, element w, value result)
{
  int i;
  element p;
#if DEBUG > 3  
  printf("adjust_value");print_value(v);
#endif
  p=join(v[j],v[k],w);
  for (i=k+1;i<nrPoints;i++) {
    result[i]=join(v[i],p,af[i]);
  }
#if DEBUG > 3  
  printf("=");print_value(result);printf("\n");
#endif
}

/* Given diff, the following function adjusts all the values in the 
 * universe, and collects any pendingDiffs which result from hashtable 
 * collisions. Returns 0 if diff = [c,c,\ldots,c] for some c \neq 0 */

// before calling adjust_values2, we need to set CurrentGroup

bool adjust_values2(value diff, struct groupNode* gn)
{
  int i,j,k,found,from,g;
  struct groupNode *gn1, *gn2;
  NewValue *nv, *nv1, *nv2, *nvkey;
  value af, v;
  HASH_ENTRY* phe;
  element w;
  struct difflist *diffs=pendingDiffs;
  from = (gn==NULL)?0:getGroupRep(gn)->from;
#if DEBUG > 1
  printf("Adjusting for diff = "); print_value(diff,from);printf("[%d]\n",getGroupRep(gn)->g);
#endif
  if (from >= nrPoints-4) {
    printf("#Optimized Adjusts = %d, #Adjusts = %d, #from=%d, #Points = %d\n",
	   nrAdjustNewTerms,nrAdjusts,from,nrPoints);
    printf("INCREASE nrPoints\n");
    exit(FEW_POINTS);
  }
  if (!is_equal_zero(diff, from)) {
    if (is_constant_value(diff, from) == 1) {
      if (debug) printf("Axioms are inconsistent\n");
      return 0;
    }
    else {
      FreeHash(funcHash);
      funcHash  = NewHash(NewValue2int, NewValueCompare);
      af = (value)malloc(nrPoints*sizeof(element));
      compute_adjustment_factors(diff,af,&j,&k,&w,from);
      // adjust only those values which belong to the group g
      g = getGroupRep(gn)->g;
      // adjust pendingDiffs
      while (diffs != NULL)
	{
	  if (getGroupRep(diffs->gn)->g == g) {
	    adjust_value(diffs->v,af,j,k,w,diffs->v);
	  }
	  diffs=diffs->next;
	}
      // adjust values
      for (i=0;i<nrUsedValues;i++) {
	nv = values[i];
	if (nv->used && (!nv->rep) && (((getGroupRep(nv->group))->g == g))) {
	  adjust_value(nv->data,af,j,k,w,nv->data);
	}
      }
      // after adjustment, now eliminate duplicate values by appropriately setting their representatives, and
      // handle collisions of function arguments by adding difference of their values to pendingDiffs
      for (i=0;i<nrUsedValues;i++) {
	nv = values[i];
	if (  nv->used && (!nv->rep) ) {
	  phe = HashLookupOrAdd(funcHash,nv,&found);
	  if (found == 0) { // nv is a fresh adjusted value 
	    phe->key=nv;
	  }
	  else { // nv already exists
	    nvkey=getValueRep((NewValue*)phe->key);
	    if (nv->func) { // nv is some function argument
	      if (nvkey->func) {
		// some function argument that collides with another function argument
		nv1 = getValueRep(nv->func);
		nv2 = getValueRep(nvkey->func);
		gn1 = getGroupRep(nv1->group);
		gn2 = getGroupRep(nv2->group);
		setGroup(gn1,gn2);
		v=(value)malloc(nrPoints*sizeof(element));
		sub_value(nv1->data,nv2->data,v,gn2->from);
		addPendingDiffs(v,gn2);
#if DEBUG>1
		printf("nv1=");print_value(nv1->data,gn2->from);
		printf("nv2=");print_value(nv2,gn2->from);printf("\n");
		printf("Adding to pendingDiffs v1-v2="); print_value(v,gn2->from);
		printf("[%d]",gn2->g); printf("\n");
#endif
	      }
	      else { // a func. arg. clashes with a non func. arg. 
		nvkey->func=getValueRep(nv->func);
	      }
	    }
	    nv->rep=nvkey; // nv already exists, so set its representative 
	  }
	}
      }
      free(af);
      getGroupRep(gn)->from = getGroupRep(gn)->from+1;
      maxFrom = max(maxFrom, getGroupRep(gn)->from);
      nrAdjusts++;
    }
  }
#if DEBUG > 1
  printf("\n\nAfter Adjust\n");dump_state();
#endif
  return 1;
}
/* Given "diff" of an assumption, the following function calls 
 * adjust_values2 as long as there no pendingDiffs which result from 
 * applying the congruence closure axiom (detected via collisions in the 
 * hashtable). Returns 0 if any values cannot be adjusted at any stage
 * because of failure of adjust_values2 */


bool adjust_values(value diff, struct groupNode* gn)
{
  value diff1;
  if (!adjust_values2(diff,gn)) return 0;  
  while (pendingDiffs != NULL) {
    diff1 = pendingDiffs->v;
    gn = pendingDiffs->gn;
    pendingDiffs = pendingDiffs->next;
    if (!adjust_values2(diff1,gn)) return 0;
    free(diff1); 
  }
  return 1;
}



void addPendingDiffs(value diff,struct groupNode* gn)
{
  struct difflist* diffs;
  
  diffs=(struct difflist*)malloc(sizeof(struct difflist));
  diffs->v = diff;
  diffs->gn = gn;
  diffs->next = pendingDiffs;
  pendingDiffs = diffs;
}


// prints the global list pendingDiffs
void print_pendingDiffs()
{
  struct difflist* diffs=pendingDiffs;
  struct groupNode* gn;
  printf("PendingDiffs=[");
  while (diffs != NULL)
    {
      gn = getGroupRep(diffs->gn);
      print_value(diffs->v,gn->from); printf("[Group=%d]",gn->g);
      diffs=diffs->next;
    }
  printf("]\n");
}


void verify_assumptions(struct pred* p)
{
  int k,i,from;
  struct groupNode* gn;
  value v;
  struct exp e;
  struct atom* pa;
  v = (value)malloc(nrPoints*sizeof(element));
  for (k=0;k<nrProcessed;k++) {
    pa = &((*(p->u.s2.p1)).u.a); 
    e.tag=MINUS_EXP;
    e.u.exps.e1 = &(pa->u.s1.e1); 
    e.u.exps.e2 = &(pa->u.s1.e2);
    
    for (i=0;i<nrPoints;i++) { 
      v[i] = 0;
    }
    gn = NULL;
    eval_expr(&e,1,0,v,&gn);
    from = (gn==NULL)?0:getGroupRep(gn)->from;
    if (is_equal_zero(v,from)==0) {
#ifdef DEBUG
      printf("expr ");print_exp(&e);printf("evaluates to"); print_value(v,from); printf("\n");
#endif
      printf("BUG: Axiom %d is not satisfied by the sample\n",k+1);
      _ASSERT(0);
    }
    p=p->u.s2.p2;
  }
  free(v);
}

bool process_assumption (struct exp* e) {
  value v,u;
  int i, found, from;
  struct groupNode* gn;
  bool result=1;
  HASH_ENTRY* phe;
  v = (value)malloc(nrPoints*sizeof(element));
  factor = 0;

  for (i=0;i<nrPoints;i++) {
    v[i] = 0;
  }
  gn=NULL;
  eval_expr(e,1,0,v,&gn);
  from = (gn==NULL)?0:getGroupRep(gn)->from;
#if DEBUG > 3
  printf("gn->g = %d\n",(gn==NULL)?(-1):getGroupRep(gn)->g);
#endif
  for (i=0;i<nrUsedValues;i++) {
    getGroupRep(values[i]->group);
  }

  
#if DEBUG > 1     
  if (firsttime) { firsttime =0; printf("\n\nInitial State\n"); dump_state(); }
#endif
  
#ifdef DEBUG
  printf("Now processing assumption %d :: ",nrProcessed+1); print_exp(e); printf("=0\n");
  //printf("expr evaluates to ");print_value(v,from);printf("\n");
#endif      
  if (is_constant_value(v,from) == 1) {
    if (is_equal_zero(v,from) == 1) {
#ifdef DEBUG
      printf("REDUNDANT\n");
#endif
    }
    else {
#ifdef DEBUG
      printf("INCONSISTENT\n");
#endif
      result = 0;
    }
  }
  else { // adjustment is required
    if (factor != 0) { // we can adjust cheaply
#ifdef DEBUG
      printf("Cheap adjustment%d\n",nrProcessed+1);
#endif
      if (newTerm.tag == VAR_TERM) {
	phe = HashLookupOrAdd(var_assignment, newTerm.u.str, &found);
	_ASSERT(found == 1);
	u = ((NewValue*)phe->data)->data;
	for (i=from;i<nrPoints;i++) {
	  u[i] = sub(u[i],divide(v[i],factor));
	}
      }
      else {
	phe = HashLookupOrAdd(funcHash, newTerm.u.v, &found);
	_ASSERT(found == 1);
	u = (getValueRep(getValueRep((NewValue*)phe->key))->func)->data;
	for (i=from;i<nrPoints;i++) {
	  u[i] = sub(u[i],divide(v[i],factor));
	}
      }
      nrAdjustNewTerms++;
#if DEBUG > 1
      printf("\n\nAfter Cheap Adjustment\n");dump_state();
#endif
    }
    else { // cheap adjustment cannot be performed
      result = adjust_values(v,gn);
    }
  }
  nrProcessed++;
  free(v);

  return result;
}


bool verify_goal(struct exp* e)
{
  bool res;
  int i,from;
  struct groupNode* gn;
  value v;
  v = (value)malloc(nrPoints*sizeof(element));
  
  for (i=0;i<nrPoints;i++) {
    v[i]=0;
  }
  gn = NULL;
  eval_expr(e,1,0,v,&gn);
  from = (gn==NULL)?0:getGroupRep(gn)->from;
  res = is_equal_zero(v,from);
  free(v);
  return res;
}




int prove_goal(struct pred *p)
{
  struct exp e;
  struct atom* pa;
  struct pred* pc=p;
  while (pc->tag != ATOM_PRED) {
    if (pc->tag == IMPL_PRED && (*(pc->u.s2.p1)).tag == ATOM_PRED
	&& (*(pc->u.s2.p1)).u.a.tag == EQ_ATOM) {
      pa = &((*(pc->u.s2.p1)).u.a); 
      e.tag=MINUS_EXP;
      e.u.exps.e1 = &(pa->u.s1.e1); 
      e.u.exps.e2 = &(pa->u.s1.e2);
      if (process_assumption(&e) == 0) { return INCONSISTENT_GOAL; } 
#ifdef DEBUG
      verify_assumptions(p);
#endif
    }
    else { printf("Unimplemented\n"); exit(UNIMP_PRED); }
    pc=pc->u.s2.p2;    
  }
  e.tag=MINUS_EXP;
  e.u.exps.e1=&(pc->u.a.u.s1.e1);
  e.u.exps.e2=&(pc->u.a.u.s1.e2);
  if (verify_goal(&e) == 1) return PROVED_GOAL;
  else return FALSE_GOAL;
}


/************************************************
 **
 **    MAIN
 **
 ************************************************/

int main(int argc, char **argv) {
  struct pred *goal;
  int i;
  int res = 1;
  int firstRes;
  int timesRun;
  double totalTime;
  
  // Get the clock speed
  int __clock_tick = sysconf(_SC_CLK_TCK);
  // Get the command-line options
  {
    static struct option fast_options[] = {
      {/*Name=*/"points", /*HasArg=*/1, 0, /*Code=*/1},
      {"test",1,0,2},
    };
    
    while(1) {
      int c = getopt_long(argc, argv, "", fast_options, NULL);
      if(c == -1) break; // Done with options
      switch(c) {
      case 1:
        nrPoints = atoi(optarg);
        if(nrPoints <= 3) {
          _ERROR1("Too few points: %d", nrPoints);
        }
        break;
      case 2:
	test = atoi(optarg);
	break;
      default:
        _ERROR0("Invalid option");
      }
    }
  }
  // Some initialization
  invPrime = 1.0/(double)randPrime;
#ifndef DEBUG
  srand((int)time(NULL));
#endif
  randWeight_func = NewStringPointerHash();
  for (i=0;i<16;i++) {
    randWeight_arg[i]=get_rand();
#if DEBUG > 2
    randWeight_arg[i]=1;
#endif
  }
  pendingDiffs=NULL;
  nrProcessed=0;
  
  // See if there are any file names passed as arguments
  if(optind >= argc)
    _ERROR0("No file name was given\n");
  goal = parseGoal(argv[optind]);
  fprintf(stderr, "Proving goal with %d points\n", nrPoints);  
  firstRes = -1; // Keep here the first result, to check consistency

  if (test == 0) { /* time it! */
    TIME_MILLISECONDS(1000, /* Run for 1 second */
		      timesRun, totalTime, 
    {
      initialize ();
      res = prove_goal(goal);
      if(firstRes >= 0) {
	if(res != firstRes) 
	  _ERROR2("Inconsistent results: res=%d, firstRes=%d\n", res, firstRes);
      } else {
	firstRes = res;
      }
      cleanup();
    }
		      );
  }
  else {
    timesRun=1;
    TIME(totalTime,
    {
      initialize ();
      res = prove_goal(goal);
    }
	 );
  }
  fprintf(stderr,
	    "Ran %d iterations\nTime %f ms\n", timesRun,
	    totalTime / (double)timesRun);
  fprintf(stderr,
          "#Optimized Adjusts=%d, #Adjusts=%d, #Points=%d\n",
	  nrAdjustNewTerms, nrAdjusts, nrPoints);
  switch(res) {
  case FALSE_GOAL: _ERROR0("Unable to prove goal\n"); break;
  case PROVED_GOAL: fprintf(stderr, "Proved Goal\n"); exit(0);
  case INCONSISTENT_GOAL: printf("Inconsistent\n"); exit(0);
  };
  return 0;
}

  
void print_NewValues()
{
  int i;
  for (i=0;i<nrUsedValues;i++) {
    printf("NewValue %3d group=%3d address=%3d rep=%3d func=%3d data=",
	   i,getGroupRep(values[i]->group)->g, ((int)values[i])%1000,
	   ((int)(getValueRep(values[i])))%1000,((int)(values[i]->func))%1000);
    print_value(values[i]->data,getGroupRep(values[i]->group)->from); printf("\n");
  }
  printf("\n");
}

