module E = Errormsg
module L = Logic
module P = Proof
module S = Satproc
module H = Hashtbl

(* INTEGER ARITHMETIC MODULO PRIME *)
module A = Integer_arithmetic.Integer_arithmetic

(* RATIONAL ARITHMETIC *)
(* module A = Rational_arithmetic.Rational_arithmetic *)

(* FLOATING POINT ARITHMETIC BASED ON RICHARD FATEMAN'S IDEA *)
(* module A = Floating2_arithmetic.Floating2_arithmetic *)

(* FLOATING POINT ARITHMETIC *)
(* module A = Floating_arithmetic.Floating_arithmetic *)

open Pretty

let ccid = "RAND"
let debug = ref true

let proveIt _ = E.s (E.unimp "Rand.proveIt")


exception Failed


(* tunable parameters *)
(* Greater are these parameter, smaller is the error probability *)
let randBound = 1000
let nrEval = 6
(* nrEval should be at least one greater than maximum number of variables 
   that are dependent on each other in the assumptions, 
   otherwise "inconsistent assumptions" will be signalled *)

(* We collect all the disEqualities encountered in assumptions to check their validity at the end 
 * when proving the goal *)
let disEqualities = ref ([]) 


type value = A.elt array
type variable = string * value list

(* some useful functions on values *)
let randValue () = Array.init nrEval (fun i -> A.randElt () randBound)

let is_equal_value v1 v2 =
  let l1 = Array.length v1 in
  let l2 = Array.length v2 in
  if (l1 != l2) then false
  else  
    let el1 = Array.to_list v1 in
    let el2 = Array.to_list v2 in
    List.for_all2 (fun e1 e2 -> A.eq e1 e2) el1 el2


let is_constant_value v =
  let c = v.(0) in
  try
    for i = 1 to nrEval - 1 do
      if (not (A.eq c v.(i))) then raise Failed
    done;
    true
  with Failed -> false

let docValue () (v: value) : Pretty.doc =
  dprintf "[%a]" (docList (fun n -> dprintf "%a" A.docElt n)) (Array.to_list v)

let valueI i = Array.create nrEval (A.int2elt i) 
let valueZero = valueI 0
let valueBinOp (f: A.elt -> A.elt -> A.elt) (v1: value) (v2: value): value = Array.init nrEval (fun i -> (f v1.(i) v2.(i)))
let valueUnOp (f: A.elt -> A.elt) (v: value): value = Array.init nrEval (fun i -> f v.(i))
let add_value = valueBinOp (fun i1 i2 -> A.add i1 i2) 
let sub_value = valueBinOp (fun i1 i2 -> A.sub i1 i2)
let mult_value = valueBinOp (fun i1 i2 -> A.mult i1 i2) 

(* ---- Hashtable Stuff begins ---- *)

(* Some arithmetic modules use the type Num.num and unfortunately *)
(* the '=' operator in ocaml does not work for types like Num.num *)
(* Hence, we need to define our own 'equal' and 'hash' operation for use with hashtables *)

let hash_of_value v = 
  let n = Array.fold_left (fun result n' -> A.add result n') (A.zero) v in
  A.hashElt n

let equal_for_variable (s, vl) (s', vl') = (s = s') & (List.for_all2 (fun v v' -> is_equal_value v v') vl vl')

let hash_of_variable (s, vl) = Hashtbl.hash (s, List.map hash_of_value vl)

module VariableHashType = 
  struct type t = (string * value list)
    let equal = equal_for_variable
    let hash = hash_of_variable
  end
module Variable_Hashtbl = Hashtbl.Make (VariableHashType)
(* ----- Hashtbl Stuff ends -----*) 


let store : (value) Variable_Hashtbl.t ref = ref (Variable_Hashtbl.create 53) 

(* get the value of the variable var1 from store *)
let get_value ((s,vl) as var1: variable): value = 
  try 
    Variable_Hashtbl.find !store var1
  with Not_found ->
    let newValue = randValue () in
    Variable_Hashtbl.add !store var1 newValue;
    newValue

let rec evalExp (e: L.exp) : value = 
  let currentFuncDebug = false & !debug in
  let result = 
    match e with
      L.NVar v -> get_value (v,[])
    | L.Int i -> valueI (L.to_int i)
    | L.BinOp(L.Plus,e1,e2) -> 
	let v1 = evalExp e1 in
	let v2 = evalExp e2 in
	add_value v1 v2
    | L.BinOp(L.Minus,e1,e2) ->
	let v1 = evalExp e1 in
	let v2 = evalExp e2 in
	sub_value v1 v2
    | L.BinOp(L.Times,e1,e2) ->
	let v1 = evalExp e1 in
	let v2 = evalExp e2 in
	mult_value v1 v2
    | L.Fun (s,el) -> 
	let vl = List.map evalExp el in
	get_value (s,vl)
    | _ -> E.s (E.unimp "evalExp: Unimplemented")
  in
  if (currentFuncDebug) then ignore (E.log "RAND: evalExp(%a) = %a \n" 
                                       L.d_exp e docValue result);
  result


(****************************************************************************************)
(* The following function is called with an argument diff \neq [0,...0] *)
(* It computes the adjustment factors such that diff goes to [0,...,0] after adjust_values *)
(* adjusting factors is the tuple (i1,i2,r,[r0, ..., rn]) such that
   (i) diff.(i1) \neq diff.(i2)  
   (ii) Let p = (r * diff.(i1)) + ((1-r) * diff.(i2)). 
        r is such that p \neq diff.(i) 
   (ii) ri = p/(p-diff.(i))
*)

let compute_adjustment_factors diff =
  let currentFuncDebug = false & !debug in
  let af = Array.create nrEval (A.zero) in   (* the adjusting factors array *)

  (* we first find indices i1 and i2 such that diff.(i1) \neq diff.(i2) *)
  let i1 = 0 in 
  let tmp = ref (1) in
  let found = ref false in 
  while (not !found) do
    if A.eq diff.(!tmp) diff.(i1) then incr tmp
    else found := true;
  done;
  let i2 = !tmp in

  (* we now find an r such that p = (r * diff.(i1)) + ((1-r) * diff.(i2)) \neq diff.(i) *)
  let r = ref (A.one) in 
  let p = ref (A.zero) in  
  let found = ref (false) in
  while (not !found) do 
    (* we guess r and then test whether it satisfies the required property *)
    r := A.add !r A.one; (* our new guess for r is the next bigger number *)
    found := true;
    p := A.join diff.(i1) diff.(i2) !r; 
    let i = ref (0) in
    while (!found) & (!i < nrEval) do
       if A.eq !p diff.(!i) then found := false
       else incr i;
    done;
  done;
 
  (* Now, we compute ri *)
  for i = 0 to nrEval - 1 do    
    let ri = A.div !p (A.sub !p diff.(i)) in 
    af.(i) <- ri; 
  done;

  if (currentFuncDebug) then begin 
    ignore (E.log "RAND: adjustment_factors(%a) := i1 = %d, i2 = %d, r = %a, af = %a \n" 
	      docValue diff i1 i2 A.docElt !r 
              (docArray 
                 (fun i r -> dprintf "(%d,%a)" i A.docElt r)) af);
  end;

  (i1,i2,!r,af)



(* adjust_values is a recursive function because one call may trigger several 
 * other calls in case of hash-table conflicts *)

let rec adjust_values (diff: A.elt array) : unit = 

 (* dumping the state before adjustment *)
  let currentFuncDebug = false & !debug in
  if (!debug) then begin
    ignore (E.log "RAND: Before adjustment, the state is: \n");
    Variable_Hashtbl.iter (fun (s,vl) v ->
      ignore (E.log "RAND: %s(%a) = %a \n" s (docList (fun v' -> dprintf "%a" docValue v')) vl docValue v)) !store;
  end;
  
  if (not (is_equal_value diff valueZero)) then begin (* adjustment is needed such that diff' becomes zero *)
    if (is_constant_value diff) then begin (* adjustment cannot be done *)
      ignore (E.log "RAND: Inconsistent Axioms \n");
      assert (false);
    end;    
    let diffList = ref ([]) in (* for collecting the differences that need to be adjusted because of hash-table clashes *)
    let (i1,i2,r,af) = compute_adjustment_factors diff in 
   
    (* a function to adjust one value *)
    let adjust_value v = 
      let p = A.join v.(i1) v.(i2) r in   
      Array.init nrEval (fun i -> A.join v.(i) p af.(i))
    in

    (* we need to adjust all the values in the store and re-hash them *)        
    let store' = Variable_Hashtbl.create 53 in
    Variable_Hashtbl.iter (fun (s, vl) v ->
      let vl' = List.map adjust_value vl in
      let v' = adjust_value v in
      if (currentFuncDebug) then ignore (E.log "RAND: adjust_value(%a) = %a \n" docValue v docValue v');
      try
	let vo' = Variable_Hashtbl.find store' (s,vl') in
	let diff' = sub_value v' vo' in 
        (* There is a clash in the hash-table and the difference diff' needs to be adjusted *)
	diffList := diff'::!diffList;
      with Not_found ->
	Variable_Hashtbl.add store' (s,vl') v') !store;
    store := store';

    (* dumping the state after adjustment *)
    if (!debug) then begin
      ignore (E.log "RAND: After adjustment, the state is: \n");
      Variable_Hashtbl.iter (fun (s,vl) v ->
	ignore (E.log "RAND: %s(%a) = %a \n" s (docList (fun v' -> dprintf "%a" docValue v')) vl docValue v)) !store;
    end;

    (* more adjust_values triggered because of hash-table conflicts *)
    List.iter adjust_values !diffList;  
  end
      
      
let process_assumption e1 e2 = 
  if (!debug) then ignore (E.log "RAND: Processing Axiom %a = %a \n" 
                             L.d_exp e1 L.d_exp e2);
  let v1 = evalExp e1 in
  let v2 = evalExp e2 in
  let diff = sub_value v1 v2 in 
  adjust_values diff;
  ignore (E.log "RAND: Successfully processed the axiom %a = %a \n" 
            L.d_exp e1 L.d_exp e2);


type undoelem = 
    SnapShot  (* This is the last snapshot *)
      
      
let undoList: undoelem list ref = ref []


let verify_disEqualities () = 
  List.iter (fun (e1,e2) -> 
    let v1 = evalExp e1 in
    let v2 = evalExp e2 in
    if (is_equal_value v1 v2) then begin
      ignore (E.log "RAND: The disequality %a != %a was found to be Inconsistent at the end \n" L.d_exp e1 L.d_exp e2);
      assert (false);
    end
    else ignore (E.log "RAND: The disequality %a != %a was found to be Consistent at the end \n" L.d_exp e1 L.d_exp e2);
	    ) !disEqualities


let assume acc msg = 
  match msg.S.body with
    L.PNez(L.BinOp(L.SetEq, e1, e2)) ->
      if msg.S.goal then begin
	verify_disEqualities ();
	let v1 = evalExp e1 in
	let v2 = evalExp e2 in
	let diff = sub_value v1 v2 in
	if (is_equal_value v1 v2) then begin 
	  ignore (E.log "RAND: Proved Goal %a = %a \n" L.d_exp e1 L.d_exp e2);
	  raise (S.Proved (P.G("eqid", [P.E e1])));
	end
	else begin 
	  ignore (E.log "RAND: Unable to prove the goal %a = %a \n" L.d_exp e1 L.d_exp e2);
	  ignore (E.log "RAND: diff = %a\n" docValue diff);
	end;
	Stats.print stderr "Timings:\n";	
      end
      else Stats.time "process_assumption" process_assumption e1 e2;
      acc 
      (* See if we try to prove an identity *)

  | L.PNez(L.BinOp(L.SetNe, e1, e2)) -> 
      disEqualities := (e1, e2) :: !disEqualities; (* collect disEqualities *)
      acc

  | L.Bool false -> ignore (E.log "RAND: FalseAtom \n"); 
      E.s (E.error "false atom")

  | L.Bool true -> acc
  | _ -> acc
	
let initialize () =
  (* Comment the following initialization of random number generator when debugging *)  
  Random.self_init ();
  disEqualities := []; 
  undoList := []

let entries = 
  { S.init = initialize;
    S.snapshot = (fun () -> ());
    S.undo = (fun () -> ());
    S.assume = assume;
    S.proveIt = proveIt;
    S.debug = debug;
    S.ident  = ccid;
    S.descr  = "randomized congruence closure";
  } 







