(* A library of functions for the VCGen Graphical User Interface
 * Implemented with LablTk ToolKit
 * COOLAID Project: 2003
 *)


val defaultSearchDirectories: string list

(** Set this to the semicolon separated list of directories to search for 
 * source files *)
val searchDirectories: string list ref


(** These are the options for the source window *)
val sourcewinOptions: Options.optionDescr list

(** Private information we keep about an instruction *)
type privateInstrInfo


(** Information about one instruction *)
type instrInfo = { 
    mutable text: string; 
    (** The text corresponding to this instruction *)

    mutable count: int; 
    (** The number of lines it takes *)

    mutable lineno: int;
    (** The line number *)

    mutable supLine: int;
    (** The line in the super-source that it corresponds to. May be 0, to 
     * denote that it does not correspond to a super-line. *)

            iPrivate: privateInstrInfo;
    (** Some private information that we keep about an instruction *)
  }

(** Make some new instruction information *)
val newInstrInfo: string -> instrInfo

class type sourceStream = object

  method name: string
    (** A name that differentiates this stream from other streams *)

  method show: bool ref   
    (** Whether this view is enabled *)

    (** Keep an array with information about each instruction. The first 
     * entry is reserved. *)
  method info: instrInfo GrowArray.t
 
    (** The number of instructions. Instruction 0 is reserved *)
  method nrInstrs: int

  method insertInstr: int -> instrInfo 
                      -> Widget.text Widget.widget -> Tk.textIndex -> int
    (** For each instruction, insert it into the textbox and return the 
     * number of window lines it takes *)

end


class simpleSourceStream :
    name:string ->
    show:bool ref ->
    insertInstr:(int -> instrInfo -> 
                        Widget.text Widget.widget -> Tk.textIndex -> int) 
      -> sourceStream

(** Define a class that can collect a number of source files. Make it take a 
 * parameter to ensure that each "new" gets its own array. *)
class concatenatedSourcesStream :
    name:string ->
    show:bool ref ->
    insertInstr:(int -> instrInfo -> 
                        Widget.text Widget.widget -> Tk.textIndex -> int)  ->
  object
    inherit sourceStream

    method reset: unit

        (** Load a file if necessary, and return the index for a given line. 
         * Also return "trye" if it was just loaded *)
    method sourceLocationToIndex: string -> int -> int * bool
        
        (** Load a file if necessary, and return the start and last indices 
         * for it. Also return "true" if it was just loaded *)
    method sourceFileRange: string -> int * int * bool
  end


(** This is the main class exported here. Create a window that can show 
 * source files. Must set the sources with the setSources method *)
class sourceWindow:
    ?title:string -> (** The tile of the window *)
    ?textwidth:int ->
    ?textheight:int ->
    ?expand:bool ->  (** See Guilib.uiComponent *)
    ?fill:Tk.fillMode ->  (** See Guilib.uiComponent *)
    ?pagingSize:int -> (** The source files are paged in lazily. The size of 
                        * a page *)
    ?name:string ->  (** THe name of this window *)
    ?saveID:string -> (** The ID under which to save configuration data *)
    'b Widget.widget (** The top widget *) -> 
      object
        inherit Guilib.uiComponent
            
            (** Get a handle to the textbox *)
        method private textbox: Widget.text Widget.widget
            
            (** Get the current sources *)
        method private sources: sourceStream array

            (** Set the sources. The streams are ordered in the SUP->INF 
             * order. The supLine of some entries must be set *)
        method private setSources: sourceStream list -> (unit -> unit) -> unit

            (** Call this method in Bind to record an event in the textbox *)
        method private markEvent: Tk.eventInfo -> unit

            (** markEvent will leave here the window line number where the 
             * event occurred *)
        val mutable eventWindowLine: int

            (** Convert a window line number to an index into the given 
             * stream index. If the window line shows a line into an inferior 
             * stream, we use supLine. If the window line shows a line into 
             * a superior stream, we use the inverse of supLine. Raises 
             * Not_found if we cannot find the line. *)
        method private windowLineToStream: int -> stridx:int -> int

            (** Call this method to select a given line in a certain source. 
             * We walk the supLine until we find a line that is displayed *)
        method private selectLine: srcidx:int -> line:int -> unit

      end
