// c2vml.cc
// implementation of c2vml methods declared in c2vml.ast

#include "c2vml.h"              // this module
#include "cc.ast.gen.h"         // C++ AST: source language
#include "vml.ast.gen.h"        // VML AST: target language
#include "c2v.ast.gen.h"        // C2V AST: transformation spec
#include "exttype.h"            // EFunctionType, etc.
#include "trace.h"              // tracingSys
#include "vml_build.h"          // makeId, etc.
#include "strtable.h"           // StringTable
#include "variable.h"           // Variable
#include "cc_print.h"           // PrintEnv
#include "c2v_parse.h"          // parse_C2V_file

#include <ctype.h>              // isalnum

// STUBEMIT: #include "cc.ast.gen.h"

// can I dispense with this altogether?
#define USE_META 0


#if USE_META
// true if the Variable also has a 'meta' variable for its state
static bool hasMetaVar(Variable const *_v)
{
  EVariable const *v = extC(_v);
  return !v->hasVAttr(VA_PARAMBYREF) &&
         v->type->isPointerType() &&
         !v->type->asPointerType()->atType->isVoid();   // for malloc/free ..
}
#endif


// print first line
template <class T>
string printFirstLine(T &t)
{
  stringBuilder sb;
  StringPrintEnv penv(sb);
  t.print(penv);
  penv.finish();

  // truncate at the first newline; this ensures in general that I
  // don't print beyond the end of the line (b/c if I did then it
  // wouldn't be treated as part of the comment), and also makes sure
  // I abbreviate large compound statements and loop constructs
  char const *newline = strchr(sb.c_str(), '\n');
  if (newline) {
    return substring(sb.c_str(), newline-sb.c_str());
  }
  else {
    return string(sb);    // no newline, print whole thing
  }
}


// -------------------------- C2VScope -----------------------------
C2VScope::C2VScope()
  : map()
{}

C2VScope::~C2VScope()
{}


// ----------------------- C2VScopePushPop -------------------------
C2VScopePushPop::C2VScopePushPop(C2VML &e)
  : env(e)
{
  scope = env.pushScope();
}

C2VScopePushPop::~C2VScopePushPop()
{
  env.popScope(scope);
}


// ---------------------------- C2VML ------------------------------
C2VML::C2VML(StringTable &s, CCLang &lang)
  : env(*this),
    str(s),
    scopes(),
    c2vfile(NULL),
    inPredicate(0),
    tempNameCounters(),
    functionAddressCounter(1),  // avoid confusion with NULL
    typeTagCt(1),               // not 0, since 0 means "freed"
    automaticTypingAxioms(true),
    modelUsesObjct(true),
    vmlfile(new VMLFile(NULL /*forms*/)),
    currentFunc(NULL),
    stmtStack(),
    #define STR(x) str_##x(s(#x))
    STR(malloc),
    STR(memset),
    STR(memmove),
    STR(memcpy),
    STR(free),
    STR(kfree),
    STR(metaVar),
    STR(fresh_value),
    STR(CLT_between),
    STR(declare_partial_order)
    #undef STR
{
  // add a global scope
  scopes.prepend(new C2VScope);

  // for now, just parse a specifically hardcoded spec file
  c2vfile = parse_C2V_file(str, lang, "basic.c2v");
}

C2VML::~C2VML()
{
  delete c2vfile;

  if (vmlfile) {
    delete vmlfile;
  }
}


void C2VML::beginFunc(Function *f)
{
  currentFunc = f;
  pushCompound(f->getLoc());
  
  // by resetting this each time, we make the names used in any
  // particular function independent of what other functions are
  // translated at the same time
  tempNameCounters.empty();

  // add assumptions about parameters based on their type
  SFOREACH_OBJLIST(Variable, f->funcType->params, iter) {
    Variable const *v = iter.data();
    if (v->type->isSimple(ST_UNSIGNED_INT)) {               
      // unsigned vars are at least 0
      comment(stringc << v->name << " is unsigned");
      addStmt(new VS_pred(f->getLoc(), PS_ASSUME,
        p_compare(intLit(0), CO_LESSEQ, makeVar(v->name))));
    }
  }
}

VS_compound *C2VML::endFunc()
{
  currentFunc = NULL;
  xassert(inPredicate == 0);
  return popCompound();
}


void C2VML::addTopForm(VTopForm * owner vtf)
{
  vmlfile->forms.append(vtf);
}


void C2VML::addStmt(VStatement * owner stmt)
{
  stmtStack.top()->body.append(stmt);
}

void C2VML::foldStmt(VS_compound * owner compound)
{
  stmtStack.top()->body.concat(compound->body);
  delete compound;
}


void C2VML::pushCompound(SourceLoc loc)
{
  stmtStack.push(new VS_compound(loc, NULL /*body*/));
}

VS_compound * owner C2VML::popCompound()
{
  return stmtStack.pop();
}


void C2VML::prove(rostring  /*why*/, Pred * owner pred)
{
  // TODO: do something with 'why'
  addStmt(new VS_pred(SL_UNKNOWN, PS_ASSERT, pred));
}


void C2VML::define(rostring varName, VExpr * owner init)
{
  VDeclaration *decl = new VD_const(new Identifier(SL_UNKNOWN, str(varName)), init);
  addDecl(decl);
}

void C2VML::declare(rostring varName)
{
  VDeclaration *decl = new VD_var(new Identifier(SL_UNKNOWN, str(varName)), NULL /*init*/);
  addDecl(decl);
}

void C2VML::addDecl(VDeclaration *decl)
{
  if (env.atTopLevel()) {
    env.addTopForm(new VTF_decl(SL_UNKNOWN, decl));
  }
  else {
    env.addStmt(new VS_decl(SL_UNKNOWN, decl));
  }
}


void C2VML::assign(rostring targetVar, VExpr * owner value)
{
  env.addStmt(new VS_assign(SL_UNKNOWN, env.makeId(targetVar), value));
}


void C2VML::assignLive(Variable *var, VExpr * owner value)
{
  assign(var->name, value);
                                         
  #if USE_META
  if (var->type->isPointerType() &&
      var->hasFlag(DF_GLOBAL)) {     // ack..
    assign(stringc << var->name << "_meta", makeVar("live"));
  }
  #endif
}


void C2VML::assignUpd(rostring varName, VExpr * owner index, VExpr * owner val)
{
  env.assign(varName,
    env.upd(env.makeVar(varName), index, val));
}


void C2VML::comment(rostring text)
{
  if (atTopLevel()) {
    addTopForm(new VTF_comment(SL_UNKNOWN, text));
  }
  else {
    addStmt(new VS_comment(SL_UNKNOWN, text));
  }
}


StringRef C2VML::makeTempVarName(rostring basename, rostring why)
{
  // for now, I ignore the 'why' argument..

  // find the current value of counter 'basename', or create one
  int *counter = tempNameCounters.queryif(basename.c_str());
  if (!counter) {
    counter = new int(0);
    tempNameCounters.add(basename.c_str(), counter);
  }

  return str(stringc << "tmp___" << basename << "_" << (*counter)++);
}

static bool isWordChar(char c)
{
  return c=='_' || isalnum(c);
}

StringRef C2VML::makeTempVarName(rostring why)
{
  // use 'why' to obtain tempvar base name
  char const *start = toCStr(why);
  while (*start && !isWordChar(*start)) {
    start++;
  }
  char const *end = start;
  while (*end && isWordChar(*end)) {
    end++;
  }
  if (*start == 0 || start == end) {
    xfailure("makeTempVarName: 'why' must include at least one word");
  }
  string basename = substring(start, end-start);

  return makeTempVarName(basename.c_str(), why);
}

StringRef C2VML::makeTempVarName(Expression const *e, rostring why)
{
  // derive the tempvar base name from 'e' if possible
  e = e->skipGroupsC();
  if (e->isE_variable()) {
    return makeTempVarName(e->asE_variableC()->var->name, why);
  }
  else {
    // derive it from 'why'
    return makeTempVarName(why);
  }
}

StringRef C2VML::makeTempVarName(Condition const *c, rostring why)
{
  if (c->isCN_expr()) {
    return makeTempVarName(c->asCN_exprC()->expr->expr, why);
  }
  else {
    return makeTempVarName(why);
  }
}


StringRef C2VML::makeLabelName(rostring why)
{
  return makeTempVarName("label", why);
}


Identifier * owner C2VML::returnLabelId()
{
  return makeId("return_label");
}


Identifier * owner C2VML::makeId(rostring id)
{                          
  // rather unfortunate that I don't have a good way of enforcing
  // the distinction between 'char const *' and 'StringRef' ... the
  // way C++ wants me to do it is to make the latter into a class,
  // but I don't trust gcc to do that efficiently....
  //
  // 2005-03-01: actually, now with 'rostring', there *is* a
  // difference; but it's too early to tell if that will help much
  return ::makeId(str(id.c_str()));
}

VE_var * owner C2VML::makeVar(rostring id)
{
  return new VE_var(makeId(id));
}


StringRef C2VML::fieldSlotName(Variable const *field) const
{
  xassert(field->scope && field->scope->curCompound);
  return extC(field)->modelVar->name;
}


EVariable *C2VML::getSynthField(EVariable *v, Variable *field)
{
  FOREACH_OBJLIST_NC(EVariable, v->synthFields, sfiter) {
    if (sfiter.data()->originalStructField == field) {
      // found the right synth field
      return sfiter.data();
    }
  }

  return NULL;     // not found
}


Pred * owner C2VML::hasClassTag(ECompoundType *ct, VExpr * owner objIndex)
{
  // base case
  if (ct->subclasses.isEmpty()) {
    return p_equal(sel(makeVar("tag"), objIndex),
                   makeVar(stringc << ct->name << "_tag"));
  }

  Pred * owner p =
    p_equal(sel(makeVar("tag"), objIndex->clone()),
            makeVar(stringc << ct->name << "_tag"));

  SFOREACH_OBJLIST_NC(ECompoundType, ct->subclasses, iter) {
    // recursive case
    p = p_or(p, hasClassTag(iter.data(), objIndex->clone()));
  }

  delete objIndex;
  return p;
}


VExpr * owner C2VML::sel(VExpr * owner array, VExpr * owner index)
{
  return apply2(makeId("sel"), array, index);
}

VExpr * owner C2VML::upd(VExpr * owner array, VExpr * owner index, VExpr * owner value)
{
  return apply3(makeId("upd"), array, index, value);
}

VExpr * owner C2VML::ptr(VExpr * owner object, VExpr * owner offset)
{
  return apply2(makeId("ptr"), object, offset);
}


// this handles a range of situations where the value denoted by the
// lvalue expression 'target' is being overwritten, by the value
// denoted by 'srcValue'; it's among the most complicated because it
// is after all what makes the language "imperative"
//
// TODO: there are a number of places where I return 'srcValue',
// but that is wrong because the assignment might have changed
// the meaning of 'srcValue', so I instead need to be constructing
// an expression (based on 'target') that properly names that value
// (or else make lots of temporaries...)
VExpr * owner C2VML::assignmentExpr
  (Expression *target, VExpr * owner srcValue)
{
  target = target->skipGroups();

  if (target->isE_variable()) {
    Variable *var = target->asE_variable()->var;

    if (ext(var)->hasVAttr(VA_PARAMBYREF)) {
      env.error("write to `parambyref' variable address");
    }

    if (ext(var)->hasVAttr(VA_ADDRTAKEN)) {
      env.error("write to `addrtaken' raw variable");
    }

    env.assignLive(var, srcValue);

    return env.makeVar(var->name);
  }

  if (target->isE_deref()) {
    E_deref *deref = target->asE_deref();
    //string syntax = target->asString();

    if (deref->ptr->isE_variable()) {
      EVariable *v = ext(deref->ptr->asE_variable()->var);
      if (v->hasVAttr(VA_PARAMBYREF)) {
        // match_expr
        //   *$n = $e
        // where
        //   $n has attribute VA_PARAMBYREF
        // translate {
        //   $n := $e;
        //   return $e;
        // }
        env.assign(v->name, srcValue->clone());
        return srcValue;
      }
    }

    if (deref->ptr->isBinary(BIN_PLUS)) {
      E_binary *bin = deref->ptr->asE_binary();

      // match_expr
      //   ($e1->$f)[$e2] = $e3        // *( ( (*$e1).$f + $e2 ) = $e3
      // where {
      //   isArrayType($e1->$f);
      //   match_type $e1 with struct $n *;
      // }
      // translate {
      //   prove tag[$e1] == tagName(struct $n);
      //   prove 0 <= $e2 < arraySize($f);
      //   mem := mem{$e1,($f + $e2) := $e3};
      //   meta := meta{$1,($f + $e2) := live};
      //   return $e3;                 // implicitly, each expr eval'd only once
      // }
      if (bin->e1->isE_fieldAcc()) {
        E_fieldAcc *acc = bin->e1->asE_fieldAcc();
        if (acc->obj->isE_deref()) {
          if (acc->field->type->isArrayType() &&
              acc->obj->type->asRval()->isCompoundType()) {
            EVariable *field = ext(acc->field);
            ArrayType *at = field->type->asArrayType();
            ECompoundType *ct = ext(acc->obj->type->asRval()->asCompoundType());

            // I ensure that each expression is evaluated only once,
            // and the result of c2vml of course has no remaining side
            // effects, so could be freely used multiple times in the
            // translated code.. I'd just have to make sure my
            // translator was smart enough to only evaluate each
            // expression once (and perhaps emit a warning if an
            // expression isn't evaluated at all)
            VExpr * owner e1 = acc->obj->asE_deref()->ptr->c2vml(env);
            VExpr * owner e2 = bin->e2->c2vml(env);
            // 'srcValue' is e3

            env.prove(stringc << "structure array field write: " 
                              << target->exprToString() << " = "
                              << srcValue->print(),
              p_and(env.hasClassTag(ct, e1->clone()),
                    p_bounds(e2->clone(), e_lit(at->size))));

            if (ct->flatMap) {
              env.assign("mem",
                env.upd(env.makeVar("mem"),
                        env.ptr(e1->clone(),
                                e_plus(env.makeVar(field->modelVar->name), e2->clone())
                               ),
                        srcValue->clone()));
              env.assign("meta",
                env.upd(env.makeVar("meta"),
                        env.ptr(e1,
                                e_plus(env.makeVar(field->modelVar->name), e2)
                               ),
                        env.makeVar("live")));
            }
            else {
              // fieldName($f) := fieldName($f){$e1 := 
              //   fieldName($f)[$e1]{$e2 := $e3}};
              string fieldName = env.fieldSlotName(field);
              env.assign(fieldName,
                env.upd(env.makeVar(fieldName),
                        e1->clone(),
                        env.upd(env.sel(env.makeVar(fieldName), e1),
                                e2,
                                srcValue->clone())
                       ));
            }

            return srcValue;
          }
        }
      }

      // match_expr
      //   $n[$e2] = $e3                // *($n + $e2) = $e3
      // where
      //   typeof($n) == $t [$eSz]      // directly-modeled array
      // translate {
      //   prove "array var write"
      //     0 <= $e2 < $eSz;
      //   $n := $n{$e2 := $e3};
      //   return $e3;
      // }
      if (bin->e1->isE_variable() &&
          bin->e1->type->asRval()->isArrayType()) {
        Variable *n = bin->e1->asE_variable()->var;
        int eSz = n->type->asArrayType()->size;
        if (eSz != ArrayType::NO_SIZE) {
          VExpr * owner e2 = bin->e2->c2vml(env);

          env.prove(stringc << "array var write: " << target->exprToString(),
            p_bounds(e2->clone(), e_lit(eSz)));
          env.assign(n->name,
            env.upd(env.makeVar(n->name), e2, srcValue->clone()));

          return srcValue;
        }
      }

      // match_expr
      //   $e1[$e2] = $e3            // *($e1 + $e2) = $e3
      // translate {
      //   prove "array ptr write"
      //     0 <= $e2 < objsize[$e1];
      //   mem := mem{$e1,$e2 := $e3};
      //   meta := meta{$e1,$e2 := live};
      //   return $e3;
      // }
      VExpr * owner e1 = bin->e1->c2vml(env);
      VExpr * owner e2 = bin->e2->c2vml(env);
      
      env.prove(stringc << "array ptr write: " << target->exprToString(),
        p_bounds(e2->clone(), env.sel(env.makeVar("objsize"), e1->clone())));
      env.assign("mem",
        env.upd(env.makeVar("mem"),
                env.ptr(e1->clone(), e2->clone()),
                srcValue->clone()));
      env.assign("meta",
        env.upd(env.makeVar("meta"),
                env.ptr(e1->clone(), e2->clone()),
                env.makeVar("live")));

      return srcValue;
    }

    // get a symbolic value for the address being modified
    VExpr * owner addr = deref->ptr->c2vml(env);


    // mem := mem{addr := srcValue};
    env.assignUpd("meta", addr->clone(), env.makeVar("live"));
    env.assignUpd("mem", addr, srcValue->clone());

    return srcValue;
  }

  if (target->isE_fieldAcc()) {
    E_fieldAcc *fldAcc = target->asE_fieldAcc();
    string f = env.fieldSlotName(fldAcc->field);
    if (fldAcc->obj->type->asRval()->isCompoundType()) {
      ECompoundType *ct = ext(fldAcc->obj->type->asRval()->asCompoundType());

      // interpret field accesses to objects in memory
      VExpr * owner objid = NULL;
      if (fldAcc->obj->isE_deref()) {
        objid = fldAcc->obj->asE_deref()->ptr->c2vml(env);
      }
      else if (fldAcc->obj->isE_variable()) {
        EVariable *v = ext(fldAcc->obj->asE_variable()->var);
        if (v->hasVAttr(VA_ADDRTAKEN)) {
          objid = env.makeVar(v->name);
        }
        else {
          // field of non-addr-taken structure instance
          // match_expr
          //   $n.$f = $e
          // where
          //   $n does not have attr `addrtaken'
          // translate {
          //   instanceFieldName($n,$f) := $e;
          //   return $e;
          // }
          EVariable *synth = env.getSynthField(v, fldAcc->field);
          if (synth) {
            // found the right synth field
            env.assign(synth->name, srcValue->clone());
            return srcValue;
          }
        }
      }

      if (objid) {
        env.prove(stringc << "ptr->field write: " << target->exprToString(),
          env.hasClassTag(ct, objid->clone()));

        if (ct->flatMap) {
          // match_expr
          //   $e1->$f = $e2            // (*$e1).f = $e2
          // where
          //   match_type $e1 with struct $n *;
          //   struct $n uses the flat model;
          // translate {
          //   prove tag[$e1] == tagName(struct $n);
          //   mem := mem{$e1,$f := $e2};
          //   return $e2;
          // }
          env.assignUpd("mem",
                        env.ptr(objid, env.makeVar(f)),
                        srcValue->clone());
        }

        else {
          // match_expr
          //   $e1->$f = $e2            // (*$e1).f = $e2
          // where
          //   match_type $e1 with struct $n *;
          //   struct $n uses the split model;
          // translate {
          //   prove tag[$e1] == tagName(struct $n);
          //   $f := $f{$e1 := $e2};
          //   return $e2;
          // }
          env.assignUpd(f, objid, srcValue->clone());
        }
          
        // also covers:
          // match_expr
          //   $n.$f = $e
          // where
          //   $n has attribute `addrtaken'

        return srcValue;
      }
    }
  }

  return todo(stringc 
    << "unhandled: assignment to " << target->exprToString());
}


ASTList<VD_var> *C2VML::funcTypeToParams(EFunctionType *ft,
                                         ASTList<VD_var> * nullable returns)
{
  ASTList<VD_var> *params = new ASTList<VD_var>;

  if (returns && ft->result) {
    returns->append(new VD_var(new Identifier(ft->result->loc, ft->result->name), NULL /*init*/));
  }

  SFOREACH_OBJLIST_NC(Variable, ft->params, iter) {
    EVariable *v = ext(iter.data());
    //env.params.append(v);

    VD_var *vv = new VD_var(new Identifier(v->loc, v->name), NULL /*init*/);
    if (!v->hasVAttr(VA_PARAMBYREF)) {
      params->append(vv);
    }
    else if (returns) {
      returns->append(vv);
    }

    // this is causing a problem in gc_copy.c, and apparently isn't
    // being used in other examples (at the moment), so I am temporarily
    // turning it off
    #if USE_META
    if (hasMetaVar(v)) {
      StringRef metaName = env.str(stringc << v->name << "_meta");
      vv = new VD_var(new Identifier(v->loc, metaName), NULL /*init*/);
      params->append(vv);
    }
    #endif
  }

  return params;
}


// given the 'changes' set of the C-language annotation, add elements
// to the VML 'writes' list
void convertChangesToWrites(ASTList<Identifier> *writes,
                            ASTList<E_variable> const &changes)
{
  FOREACH_ASTLIST(E_variable, changes, iter) {
    EVariable const *var = extC(iter.data()->var);
    SourceLoc loc = iter.data()->name->loc;
    StringRef name = iter.data()->name->getName();

    if (var->hasFlag(DF_TYPEDEF)) {
      // the 'changes' variable refers to an entire type; add all of
      // that type's fields
      ECompoundType *ct = ext(var->type->asCompoundType());
      FOREACH_OBJLIST(Variable, ct->fieldSlots, slotIter) {
        writes->append(new Identifier(loc, slotIter.data()->name));
      }
    }

    else if (var->changeGroup) {
      // refers to a change group, add them
      convertChangesToWrites(writes, var->changeGroup->group);
    }

    else {
      // simple variable (common case)
      writes->append(new Identifier(loc, name));
    }
  }
}


VProcedure *C2VML::buildProcedure(Identifier *id, EFunctionType *ft,
                                  VS_compound *vbody)
{
  // parameters, return values
  ASTList<VD_var> *returns = new ASTList<VD_var>;
  ASTList<VD_var> *params = env.funcTypeToParams(ft, returns);

  // pre/post
  // NOTE: we don't call getPostcondition() b/c that was for the version
  // that had VML and vcgen in the same pass
  env.inPredicate++;     // not using PredBracket b/c it would limit scope of pre/post
  Pred *pre = ft->precondition? ft->precondition->c2vmlPredPure(env) : new P_lit(SL_UNKNOWN, true);
  Pred *post = ft->postcondition? ft->postcondition->c2vmlPredPure(env) : new P_lit(SL_UNKNOWN, true);
  env.inPredicate--;

  // reads (for now, they're always empty)
  ASTList<Identifier> *reads = new ASTList<Identifier>;

  // writes
  ASTList<Identifier> *writes = new ASTList<Identifier>;
  if (ft->changes) {
    convertChangesToWrites(writes, ft->changes->vars);
  }

  // build the complete result procedure
  return new VProcedure(
    id,
    params,
    returns,

    pre,
    post,
    reads,
    writes,

    vbody
  );
}


VExpr * owner C2VML::intLit(int i)
{
  return e_lit(i);
}


void C2VML::error(rostring msg)
{
  // punt
  throw_XFatal(msg);
}

void C2VML::error(SourceLoc loc, rostring msg)
{
  throw_XFatal(stringc << toString(loc) << ": " << msg);
}


VExpr * owner C2VML::todo(rostring msg)
{
  error(stringc << "unimplemented: " << msg);
  return NULL;
}


// Build a statement with the following semantics:
//   if (<cond>) {
//     <sTrue>;
//   }
//   else {
//     <sFalse>;
//   }
//
// Note that 'cond' has not been c2vml'd yet, and this function will
// do so, putting side effects into the current innermost open compound.
//
// This is part of an effort to "optimize" the generated VML, in the
// sense of trying to emit a minimal number of conditional expressions
// and statements, since those turn into case splits when the VML hits
// the theorem prover.  Right now, the optimization effort is ad-hoc
// and not very sophisticated.
VStatement * owner C2VML::buildConditional(SourceLoc loc, Condition *cond,
                                           VStatement * owner sTrue,
                                           VStatement * owner sFalse)
{
  if (cond->isCN_decl()) {
    // no short-circuit possibilities, do the straightforward thing
    Pred * owner guard = cond->c2vmlPred(env);
    return new VS_if(loc, guard, sTrue, sFalse);
  }
  else {
    return buildConditional(loc, cond->asCN_expr()->expr->expr, sTrue, sFalse);
  }
}

VStatement * owner C2VML::buildConditional(SourceLoc loc, Expression *expr,
                                           VStatement * owner sTrue,
                                           VStatement * owner sFalse)
{
  expr = expr->skipGroups();

  if (expr->isBinary(BIN_AND) || expr->isBinary(BIN_OR)) {
    E_binary *ebin = expr->asE_binary();

    // construct what to do if e2 must be evaluated
    pushCompound(loc);              // confine side effects
    VStatement * owner evalE2 = buildConditional(loc, ebin->e2, sTrue, sFalse);
    env.addStmt(evalE2);
    evalE2 = env.popCompound();

    // now wrap that with a test of e1
    if (ebin->op == BIN_AND) {
      // if e1 is true, then evaluate e2, otherwise goto sFalse
      return buildConditional(loc, ebin->e1, evalE2, buildGoto(loc, sFalse));
    }
    else {
      // if e1 is true, then goto sTrue, otherwise evaluate e2
      return buildConditional(loc, ebin->e1, buildGoto(loc, sTrue), evalE2);
    }
  }

  if (expr->isE_unary() && expr->asE_unary()->op == UNY_NOT) {
    // reverse the sense of the branches
    return buildConditional(loc, expr->asE_unary()->expr, sFalse, sTrue);
  }

  // no obvious optimization opportunities, so do generic thing
  Pred * owner guard = expr->c2vmlPred(env);
  return new VS_if(loc, guard, sTrue, sFalse);
}


// Build a statement with semantics
//   goto fresh_label;
// where we add
//   fresh_label:
// to front of
//   <target>;
VStatement * owner C2VML::buildGoto(SourceLoc loc, VStatement *target)
{
  if (target->isVS_label()) {
    // the target is already a label
    return new VS_goto(loc, target->asVS_label()->id->clone());
  }

  if (target->isVS_goto()) {
    // target is itself a goto; go same place
    return new VS_goto(loc, target->asVS_goto()->id->clone());
  }

  if (target->isVS_compound()) {
    VS_compound *cpd = target->asVS_compound();

    // skip down to the bottom of any VS_compound trees
    while (cpd->body.isNotEmpty() && cpd->body.first()->isVS_compound()) {
      cpd = cpd->body.first()->asVS_compound();
    }

    // is the first statement one of the easy cases above?
    if (cpd->body.isNotEmpty()) {
      VStatement *first = cpd->body.first();
      if (first->isVS_label() || first->isVS_goto()) {
        // handle it using the code above
        return buildGoto(loc, first);
      }
    }

    // no, so insert a label so we can jump to it
    StringRef label = env.makeLabelName("target for short-circuit eval");
    Identifier * owner id = new Identifier(loc, label);
    cpd->body.prepend(new VS_label(loc, id));

    // and yield a jump to that label
    return new VS_goto(loc, id->clone());
  }

  // if this fails, I'll need to pass more context into this function
  xfailure("buildGoto: statement type cannot be targeted");
  return NULL;    // silence warning
}


void C2VML::emitTypingInvariant(SourceLoc loc, ECompoundType *ct, EVariable *slot,
                                ECompoundType *pointee)
{
  string ct_tag = stringc << ct->name << "_tag";

  // tag[p.<slot>]==<pointee>_tag
  Pred * owner tagEquality =
    p_equal(
      sel(makeVar("tag"),
        sel(makeVar(slot->name),
          makeVar("p"))),
      makeVar(stringc << pointee->name << "_tag"));

  Pred * owner implRHS;
  if (modelUsesObjct) {
    // 0 < p.<slot> < objct && tag[p.<slot>]==<pointee>_tag
    implRHS =
      p_and(
        p_compare3(e_lit(0), CO_LESS,
                   sel(makeVar(slot->name), makeVar("p")), CO_LESS,
                   makeVar("objct")),
        tagEquality
      );
  }
  else {
    implRHS = tagEquality;
  }

  // forall(<ct> *p) PATS(p.<slot>).
  //   tag[p]==<ct>_tag && p.<slot>!=0 ==>
  //     0 < p.<slot> < objct && tag[p.<slot>]==<pointee>_tag;
  //     ^^^^^^^^^^^^^^^^^^^^^^^
  //             optional
  addTopForm(new VTF_inv(loc,
    p_forall1(loc,
              str(stringc << "typing for " << slot->name),
              str("p"),
              apply1(makeId("PATS"),
                sel(makeVar(slot->name), makeVar("p"))),
              p_impl(
                p_and(
                  p_equal(
                    sel(makeVar("tag"), makeVar("p")),
                    makeVar(ct_tag)),
                  p_notEqual(
                    sel(makeVar(slot->name), makeVar("p")),
                    e_lit(0))),
                implRHS
              )
             )));
}


void C2VML::defineTypeTag(ECompoundType *ct)
{
  ct->typeTag = env.typeTagCt++;
  define(stringc << ct->name << "_tag", new VE_lit(ct->typeTag));
}


// ------------------------ TranslationUnit ------------------------
#if 0      // not used
void TranslationUnit::c2vml(C2VML &env)
{
  FOREACH_ASTLIST_NC(TopForm, topForms, iter) {
    iter.data()->c2vml(env);
  }
}
#endif // 0


// ------------------------ TopForm ------------------------
// STUBPATTERN: ^void.*TF_.*::c2vml\b

// print a location to a string for the comments in the VML output
string vmlLocToString(SourceLoc loc)
{
  // TODO: only print the file= if it is different from the last
  // one printed

  // the "file=", "line=" is for the benefit of the Kettle GUI
  char const *fname;
  int line, col;
  sourceLocManager->decodeLineCol(loc, fname, line, col);
  return stringc << "file=" << fname << " line=" << line;
}

void TF_decl::c2vml(C2VML &env)
{
  env.comment(stringc << vmlLocToString(loc) << ": " << printFirstLine(*this));

  FAKELIST_FOREACH_NC(Declarator, decl->decllist, iter) {
    EVariable *v = ext(iter->var);
    Type *t = v->type;

    // exclude some global declarations
    if (t->isFunctionType()) {
      EFunctionType *ft = ext(t->asFunctionType());

      if (v->hasVAttr(VA_TRIGGER)) {          // trigger declaration
        // don't emit; we don't need a declaration for __pat
      }
      else if (v->hasVAttr(VA_FUNCSYM) ||     // function symbol
               v->hasVAttr(VA_PREDSYM)) {     // predicate symbol
        env.addTopForm(new VTF_decl(SL_UNKNOWN,
          new VD_funcsym(env.makeId(v->name), 
                         env.funcTypeToParams(ft, NULL /*returns*/))
        ));
      }
      else {                                  // prototype
        env.addTopForm(new VTF_proc(SL_UNKNOWN,
          env.buildProcedure(new Identifier(v->loc, v->name) /*id*/,
                             ft, NULL /*vbody*/)));
      }

      continue;
    }

    if (v->hasFlag(DF_TYPEDEF)) {             // typedef
      continue;
    }

    if (t->isConst()) {                       // (symbolic) constants
      if (v->value) {
        env.addTopForm(new VTF_decl(SL_UNKNOWN,
          new VD_const(new Identifier(v->loc, v->name),
                       v->value->c2vmlPure(env))));
      }
      else {
        // not sure what might be here...
      }
      continue;
    }

    if (t->isCompoundType() &&
        !v->hasVAttr(VA_ADDRTAKEN)) {
      // global instance of a struct, whose addr is not taken:
      // model as a bunch of separate globals, one for each field
      //
      // TODO: what if there are two instances?  I think this code
      // will emit colliding names ...
      FOREACH_OBJLIST_NC(EVariable, v->synthFields, fieldIter) {
        EVariable *synth = fieldIter.data();
           
        env.addTopForm(new VTF_decl(SL_UNKNOWN,
          new VD_var(new Identifier(synth->loc, synth->name), NULL /*init*/)));

        #if USE_META
        if (hasMetaVar(synth)) {
          StringRef metaName = env.str(stringc << synth->name << "_meta");
          env.addTopForm(new VTF_decl(SL_UNKNOWN,
            new VD_var(new Identifier(synth->loc, metaName), NULL /*init*/)));
        }
        #endif

        //env.globals.append(synth);
      }

      continue;
    }

    env.addTopForm(new VTF_decl(SL_UNKNOWN,
      new VD_var(new Identifier(v->loc, v->name), NULL /*init*/)));

    #if USE_META
    if (hasMetaVar(v)) {
      StringRef metaName = env.str(stringc << v->name << "_meta");
      env.addTopForm(new VTF_decl(SL_UNKNOWN,
        new VD_var(new Identifier(v->loc, metaName), NULL /*init*/)));
    }
    #endif

    //env.globals.append(v);
  }
}


void TF_func::c2vml(C2VML &env)
{
  // functions are handled directly through a visitor so that we
  // doesn't have to worry about nesting the function environments
}


void TF_template::c2vml(C2VML &env)
{}

void TF_explicitInst::c2vml(C2VML &env)
{}

void TF_linkage::c2vml(C2VML &env)
{}

void TF_one_linkage::c2vml(C2VML &env)
{}

void TF_asm::c2vml(C2VML &env)
{
  // the previous system used asms as a klunky include mechanism,
  // but I think it's better to let my C source syntax express
  // everything that was in the .vc includes, and just #include
  // them at the start
}

void TF_namespaceDefn::c2vml(C2VML &env)
{}

void TF_namespaceDecl::c2vml(C2VML &env)
{}

void TF_macro::c2vml(C2VML &env)
{
  FOREACH_ASTLIST_NC(Macro, macros, iter) {
    iter.data()->c2vml(env);
  }
}

void TF_changeGroup::c2vml(C2VML &env)
{}

void TF_axiom::c2vml(C2VML &env)
{
  PredBracket pb(env);
  env.comment(stringc << vmlLocToString(loc) << ": axiom");
  env.addTopForm(new VTF_axiom(loc, label, ax->c2vmlPred(env)));
}

void TF_option::c2vml(C2VML &env)
{
  if (0==strcmp(optionName, "no_objct")) {
    env.modelUsesObjct = false;
  }
  else if (0==strcmp(optionName, "no_automatic_typing")) {
    env.automaticTypingAxioms = false;
  }
  else {
    env.error(stringc << "unknown option: " << optionName);
  }
}


// ------------------------ Function ------------------------
void Function::c2vml(C2VML &env)
{
  StringRef name = nameAndParams->var->name;

  env.comment(stringc << "------------------ " << name << " --------------------");
  env.comment(stringc << vmlLocToString(getLoc()));

  if (0==strcmp(name, "global_inv")) {
    S_return *r = body->asS_compound()->stmts.first()->asS_return();
    PredBracket pb(env);
    Pred * owner pred = r->expr->c2vmlPred(env);
    env.addTopForm(new VTF_inv(SL_UNKNOWN, pred));
    return;
  }

  // make a variable to stand for the address
  env.addTopForm(new VTF_decl(SL_UNKNOWN,
    new VD_const(new Identifier(nameAndParams->var->loc,
                                env.str(stringc << name << "_address")),
                 new VE_lit(env.functionAddressCounter++))));

  EFunctionType *ft = ext(funcType);

  // body
  env.beginFunc(this);                  // pushes an empty VS_compound
  body->c2vmlScoped(env);
  env.addStmt(new VS_label(SL_UNKNOWN, env.returnLabelId()));
  VS_compound *vbody = env.endFunc();   // pop that VS_compound

  // build the complete result procedure
  VProcedure *vproc =
    env.buildProcedure(new Identifier(nameAndParams->var->loc, name) /*id*/,
                       ft, vbody);

  // add it to the running list of toplevel forms
  env.addTopForm(new VTF_proc(SL_UNKNOWN, vproc));
}


// ------------------------ Statement ------------------------
// STUBPATTERN: ^void.*S_.*::ic2vml\b

void Statement::c2vml(C2VML &env)
{
  // print information about this statement
  if (!isS_compound()) {           // don't print single '{' chars everywhere
    env.comment(stringc << vmlLocToString(loc) << ": " << printFirstLine(*this));
  }

  ic2vml(env);
}

void Statement::c2vmlScoped(C2VML &env)
{
  if (isS_compound()) {
    // dig down directly, and don't make a VS_compound
    S_compound *c = asS_compound();
    FOREACH_ASTLIST_NC(Statement, c->stmts, iter) {
      iter.data()->c2vml(env);
    }
  }
  else {
    c2vml(env);
  }
}


void S_skip::ic2vml(C2VML &env)
{}


void S_label::ic2vml(C2VML &env)
{
  env.addStmt(new VS_label(loc, makeId(name)));
  xfailure("this is wrong!  I do not translate 's'!");
}


// S_case and S_default appear below S_switch


void S_expr::ic2vml(C2VML &env)
{
  // evaluate and discard
  delete expr->c2vml(env);
}


void S_compound::ic2vml(C2VML &env)
{
  // reflect nesting for scoping purposes
  env.pushCompound(loc);
  FOREACH_ASTLIST_NC(Statement, stmts, iter) {
    iter.data()->c2vml(env);
  }
  env.addStmt(env.popCompound());
}


void S_if::ic2vml(C2VML &env)
{
  // delete me
  //Pred *guard = cond->c2vmlPred(env);

  // then-branch
  env.pushCompound(thenBranch->loc);
  thenBranch->c2vmlScoped(env);
  VStatement * owner sThen = env.popCompound();

  // else-branch
  env.pushCompound(elseBranch->loc);
  elseBranch->c2vmlScoped(env);
  VStatement * owner sElse = env.popCompound();
  
  // delete me
  //env.addStmt(new VS_if(loc, guard, sThen, sElse));

  env.addStmt(env.buildConditional(loc, cond, sThen, sElse));
}


// --------------------- BEGIN: switch ------------------------
struct SwitchInfo {
  // name of the temporary that holds the condition value
  StringRef condVar;

  // 'if' ladder we're building for the initial jump
  VS_compound *ladder;        // (serf)

  // if a 'default' label appears, this is its label name (else NULL)
  StringRef defaultLabel;
};


void S_switch::ic2vml(C2VML &env)
{
  // push information about this statement
  SwitchInfo info;
  info.defaultLabel = NULL;
  env.switchStack.push(&info);

  // evaluate, save the condition value
  info.condVar = env.makeTempVarName(cond, "switch condition value");
  env.define(info.condVar, cond->c2vml(env));

  // make, register the break label
  StringRef breakLabel = env.makeLabelName("break target in switch");
  env.breakLabelStack.push(breakLabel);

  // begin building an 'if' ladder that will determine which
  // label to jump to; the S_case and S_default c2vml functions
  // will fill most of it in
  info.ladder = new VS_compound(loc, new ASTList<VStatement>());

  // insert the ladder now into the main statement sequence, even
  // though it is empty for the moment, because the ladder eval
  // happens first
  env.addStmt(info.ladder);

  // now evaluate the body "normally" (note that by the time I see it,
  // 'branches' has been turned into an S_compound by tcheck, if it
  // wasn't already)
  branches->c2vml(env);

  // after the main body comes the 'break' target
  env.addStmt(new VS_label(loc, makeId(breakLabel)));

  // final 'goto' in the ladder: if no 'default' label appears, jump
  // to the 'break' target, otherwise jump to the 'default' target
  info.ladder->body.append(new VS_goto(loc, env.makeId(
    info.defaultLabel? info.defaultLabel : breakLabel)));

  // clean up
  env.switchStack.pop();
  env.breakLabelStack.pop();
}


void S_case::ic2vml(C2VML &env)
{
  // get info about the active switch statement
  if (env.switchStack.isEmpty()) {
    env.error("`case' statement only allowed inside switch statement");
    return;
  }
  SwitchInfo *info = env.switchStack.top();

  // make up a label for this 'case' target
  StringRef label = env.makeLabelName("for 'case' target");

  // insert into the active 'switch' statement ladder something like:
  //   if (<condVar> == <expr>) {
  //     goto this_case_label;
  //   }
  info->ladder->body.append(
    new VS_if(loc, p_equal(env.makeVar(info->condVar), expr->c2vmlPure(env)),
              new VS_goto(loc, env.makeId(label)),
              NULL /*sElse*/));

  // and insert into the main statement
  //   this_case_label:
  //   <this_case_stmt>;
  env.addStmt(new VS_label(loc, env.makeId(label)));
  s->c2vml(env);
}


void S_default::ic2vml(C2VML &env)
{
  // get info about the active switch statement
  if (env.switchStack.isEmpty()) {
    env.error("`default' statement only allowed inside switch statement");
    return;
  }
  SwitchInfo *info = env.switchStack.top();

  // make up a label for this 'default' target
  StringRef label = env.makeLabelName("for 'default' target");

  // install it into the 'info'
  if (info->defaultLabel) {
    env.error("duplicate `default' target in switch statement");
  }
  info->defaultLabel = label;    // do it regardless of error

  // the S_switch will take care of adding the jump to this label

  // insert target and code into main statement sequence
  env.addStmt(new VS_label(loc, env.makeId(label)));
  s->c2vml(env);
}


// --------------------- END: switch ------------------------



// generic c2vml for loop statements, generating pattern:
//
//   while (true) {
//     if (!(<cond>)) {
//       goto breakLabel;
//     }
//     else {
//       <body>;
//       continueLabel:
//       <after>;               (optional)
//     }
//   }
//   breakLabel:
//
// if 'alwaysOne' is true, then the conditional appears
// at the end of the loop instead of the beginning
template <class CONDTYPE>     // CONTYPE is either Condition or Expression
void c2vmlLoopStatement(C2VML &env, SourceLoc loc, CONDTYPE *cond, Statement *body,
                        FullExpression * /*nullable*/ after, bool alwaysOne)
{
  // create and register break/continue targets
  StringRef breakLabel = env.makeLabelName("break label in loop");
  StringRef continueLabel = env.makeLabelName("continue label in loop");
  env.breakLabelStack.push(breakLabel);
  env.continueLabelStack.push(continueLabel);

  // build up the loop body
  //       <body>;
  //       continueLabel:
  //       <after>;               (optional)
  env.pushCompound(loc);

  // <body>;
  body->c2vmlScoped(env);

  // next-to-last thing in loop: target for 'continue'
  env.addStmt(new VS_label(loc, makeId(continueLabel)));

  if (after) {
    // finally: the 'after' statement (discarding its value)
    delete after->c2vml(env);
  }

  // common fragment of both alternatives below
  VS_goto * owner jumpToBreakLabel = new VS_goto(loc, makeId(breakLabel));

  if (!alwaysOne) {
    //   while (true) {
    //     if (<cond>) {
    //       <body>;
    //       continueLabel:
    //       <after>;               (optional)
    //     }
    //     else {
    //       goto breakLabel;
    //     }
    //   }
    //   breakLabel:

    // pull the body out so we can stick it into a conditional
    VS_compound * owner vbody = env.popCompound();
    
    // open a new compound to enclose the whole thing and confine
    // the side effects of <cond>
    env.pushCompound(loc);

    // based on guard, either do the body or jump to breakLabel
    env.addStmt(env.buildConditional(loc, cond, vbody, jumpToBreakLabel));
  }

  else /*alwaysOne*/ {
    //   while (true) {
    //     <body>;
    //     continueLabel:
    //     <after>;               (optional)
    //
    //     if (<cond>) {
    //       // nothing
    //     }
    //     else {
    //       goto breakLabel;
    //     }
    //   }
    //   breakLabel:

    // based on guard, either do nothing or jump to breakLabel
    VS_compound * owner nop = new VS_compound(loc, NULL /*body*/);
    env.addStmt(env.buildConditional(loc, cond, nop, jumpToBreakLabel));
  }

  // wrap it all in a loop
  env.addStmt(new VS_loop(loc, env.popCompound()));

  // add the goto target after the loop
  env.addStmt(new VS_label(loc, makeId(breakLabel)));

  env.breakLabelStack.pop();
  env.continueLabelStack.pop();
}

void S_while::ic2vml(C2VML &env)
{
  c2vmlLoopStatement(env, loc, cond, body, NULL /*after*/, false /*alwaysOne*/);
}

void S_doWhile::ic2vml(C2VML &env)
{
  c2vmlLoopStatement(env, loc, expr->expr, body, NULL /*after*/, true /*alwaysOne*/);
}

void S_for::ic2vml(C2VML &env)
{
  // initializing statement
  init->c2vml(env);

  // now generic loop code
  c2vmlLoopStatement(env, loc, cond, body, after, false /*alwaysOne*/);
}


void S_break::ic2vml(C2VML &env)
{
  if (env.breakLabelStack.isEmpty()) {
    env.error("`break' statement must appear in a loop or switch statement");
  }
  else {
    env.addStmt(new VS_goto(loc, makeId(env.breakLabelStack.top())));
  }
}

void S_continue::ic2vml(C2VML &env)
{
  if (env.continueLabelStack.isEmpty()) {
    env.error("`continue' statement must appear in a loop statement");
  }
  else {
    env.addStmt(new VS_goto(loc, makeId(env.continueLabelStack.top())));
  }
}


void S_return::ic2vml(C2VML &env)
{
  if (expr) {
    // get return value variable
    Variable *result = ext(env.currentFunc->funcType)->result;

    // evaluate the expression, and assign the return value to it
    VExpr *val = expr->c2vml(env);
    env.addStmt(new VS_assign(loc, makeId(result->name), val));
  }

  // jump to final label
  env.addStmt(new VS_goto(loc, env.returnLabelId()));
}


void S_goto::ic2vml(C2VML &env)
{
  env.addStmt(new VS_goto(loc, new Identifier(loc, target)));
}


VExpr * owner c2vml_if(C2VML &env, Type *type, Initializer *init)
{
  if (init) {
    return init->c2vml(env, type);
  }
  else {
    return NULL;
  }
}

void S_decl::ic2vml(C2VML &env)
{
  FAKELIST_FOREACH_NC(Declarator, decl->decllist, iter) {
    EVariable *v = ext(iter->var);
    Type *t = v->type;

    // exclude some things
    if (t->isFunctionType() ||         // prototypes
        v->hasFlag(DF_TYPEDEF)) {      // typedef
      continue;
    }

    env.addStmt(new VS_decl(loc,
      new VD_var(new Identifier(v->loc, v->name), c2vml_if(env, t, iter->init))));

    #if USE_META
    if (hasMetaVar(v)) {
      StringRef metaName = env.str(stringc << v->name << "_meta");

      env.declare(metaName);
      env.assign(metaName, env.makeVar(iter->init? "live" : "dead"));
    }
    #endif
  }
}


void S_try::ic2vml(C2VML &env)
{ env.todo("S_try"); }
void S_asm::ic2vml(C2VML &env)
{ /* I'm going to pretend I don't see it... */ }
void S_namespaceDecl::ic2vml(C2VML &env)
{ /* should be safe to ignore */ }


void S_assert::ic2vml(C2VML &env)
{                                    
  PredBracket pb(env);

  env.addStmt(new VS_pred(loc, pure? PS_PURE_ASSERT : PS_ASSERT, 
              expr->c2vmlPred(env)));
}

void S_assume::ic2vml(C2VML &env)
{
  PredBracket pb(env);
  env.addStmt(new VS_pred(loc, PS_ASSUME, expr->c2vmlPred(env)));
}

void S_invariant::ic2vml(C2VML &env)
{
  PredBracket pb(env);
  env.addStmt(new VS_pred(loc, standalone? PS_STANDALONE_INVARIANT : PS_INVARIANT,
                          expr->c2vmlPred(env)));
}


// ------------------------ Condition ------------------------
// STUBPATTERN: ^VExpr.*CN_.*::c2vml\b

VExpr * owner CN_expr::c2vml(C2VML &env)
{
  return expr->c2vml(env);
}

VExpr * owner CN_decl::c2vml(C2VML &env)
{ return env.todo("CN_decl::c2vml"); }


// STUBPATTERN: ^Pred.*CN_.*::c2vmlPred\b

Pred * owner CN_expr::c2vmlPred(C2VML &env)
{
  return expr->c2vmlPred(env);
}

Pred * owner CN_decl::c2vmlPred(C2VML &env)
{ env.todo("CN_del::c2vmlPred"); return NULL; }


// ------------------------ Expression::c2vml ------------------------
// STUBPATTERN: ^VExpr.*E_.*::ic2vml\b

VExpr * owner Expression::c2vml(C2VML &env)
{
  return ic2vml(env);
}


VExpr * owner E_boolLit::ic2vml(C2VML &env)
{
  return new VE_lit(b? 1 : 0);
}

VExpr * owner E_intLit::ic2vml(C2VML &env)
{
  return new VE_lit(i);
}

VExpr * owner E_floatLit::ic2vml(C2VML &env)
{ return env.todo("floating literal"); }
VExpr * owner E_stringLit::ic2vml(C2VML &env)
{ return env.todo("string literal"); }

VExpr * owner E_charLit::ic2vml(C2VML &env)
{
  return new VE_lit(c);
}


VExpr * owner E_this::ic2vml(C2VML &env)
{ return env.todo("this"); }

VExpr * owner E_variable::ic2vml(C2VML &env)
{
  return new VE_var(new Identifier(var->loc, var->name));
}


// evaluate an argument list
ASTList<VExpr> *c2vmlArgs(C2VML &env, FakeList<ArgExpression> *args,
                          ASTList<Identifier> *targets = NULL)
{
  ASTList<VExpr> *vargs = new ASTList<VExpr>;

  FAKELIST_FOREACH(ArgExpression, args, iter) {
    Expression *e = iter->expr;
    
    // passing byref argument?
    if (e->isE_addrOf() &&
        e->asE_addrOf()->expr->isE_variable()) {
      E_variable *evar = e->asE_addrOf()->expr->asE_variable();
      if (ext(evar->var)->hasVAttr(VA_ARGBYREF)) {
        if (targets) {
          targets->append(new Identifier(evar->name->loc, evar->var->name));
        }
        else {
          env.error(evar->name->loc, "passing byref argument in predicate?");
        }
        continue;
      }
    }

    vargs->append(iter->expr->c2vml(env));
  }
  
  return vargs;
}


// lots of special-case translations depending on which
// function is being called
VExpr * owner E_funCall::c2vmlSpecialFunctions(C2VML &env)
{
  if (!tracingSys("no_c2v_patterns")) {
    // consult c2v spec first
    FOREACH_ASTLIST(C2VpatternSpec, env.c2vfile->specs, iter) {
      C2VpatternSpec const *spec = iter.data();
      if (0!=strcmp(spec->funcName, "c2vmlExpr")) { continue; }

      // open a new C2V scope
      C2VScopePushPop scope(env);

      // bind 'this' to mainThing, and bind the return value to nothing
      env.bind(spec->mainThing, new C2VValue(VK_C_EXPRESSION, this));
      C2VValue *retv = new C2VValue(VK_VML_EXPRESSION, (VExpr*)NULL);
      env.bind(env.str("$__retval"), retv);

      try {
        // evaluate; if it returns true then this is the matching pattern
        if (spec->body->eval(env) != SER_FALSE) {
          TRACE("c2vml", "applied pattern " << spec->ruleName);
          if (retv->getVExpr()) {
            return retv->getVExpr();
          }
          else {
            // user did not supply a return value
            return e_lit(0);
          }
        }
      }
      catch (xBase &x) {
        x.addContext(stringc << "applying rule: " << spec->ruleName);
        x.addContext(stringc << "translating E_funCall: " << this->exprToString());
        throw;
      }

      // didn't match, try the next one
    }

    // no spec matched
    return NULL;
  }

  #if 1      // obsolete (temporarily re-enabling)
  // match_expr
  //   metaVar($n)     // the user-invoked metaVar function on a variable
  // translate {
  //   metaVar($n)
  // }
  #if USE_META
  if (func->isE_variable() &&
      func->asE_variable()->var->name == env.str_metaVar &&
      args->nth(0)->expr->isE_variable()) {
    Variable *var = args->nth(0)->expr->asE_variable()->var;

    return env.makeVar(stringc << var->name << "_meta");
  }
  #endif

  // match_expr
  //   fresh_value()
  // translate {
  //   var $n;    // new name, fresh value
  //   return $n;
  // }
  if (func->isE_variable() &&
      func->asE_variable()->var->name == env.str_fresh_value) {
    StringRef fresh = env.makeTempVarName("fresh_value");
    env.declare(fresh);
    return env.makeVar(fresh);
  }

  // match_expr
  //   CLT_between(x,y)
  // translate {
  //   return CLT_between(x,y);
  // }
  if (func->isE_variable() &&
      func->asE_variable()->var->name == env.str_CLT_between) {
    return new VE_apply(env.makeId("CLT_between"), c2vmlArgs(env, args));
  }

  if (func->isE_variable() &&
      func->asE_variable()->var->name == env.str_memset) {
    // TODO: this really only makes sense when 'e_val' is 0 ...

    PointerType *pt = args->nth(0)->expr->type->asRval()->asPointerType();

    VExpr * owner e_ptr = args->nth(0)->expr->c2vml(env);
    VExpr * owner e_val = args->nth(1)->expr->c2vml(env);

    if (args->nth(2)->expr->isBinary(BIN_MULT)) {
      VExpr * owner e_length = args->nth(2)->expr->asE_binary()->e1->c2vml(env);

      if (pt->atType->isCompoundType()) {
        ECompoundType *ct = ext(pt->atType->asCompoundType());
        if (!ct->flatMap) {
          // match_expr
          //   memset($e_ptr, $e_val, $e_length * sizeof(struct $n))
          // where
          //   type_eq($e_ptr, struct $n *);
          //   struct $n uses the split model;
          // translate {
          //   prove forall(i). 0<=i<$e_length ==> tag[$e_ptr+i] == tagName(struct $n);
          //   for_each_field $f of struct $n {
          //     $f := updflatrange($f, $e_ptr, $e_ptr+$e_length, $e_val);
          //   }
          // }
          env.prove("tag checks on memset[]",
            // TODO: potential name capture problem!
            new P_quant(SL_UNKNOWN, true /*universal*/, env.str("\"memset range\""),
              new ASTList<VD_var>(new VD_var(env.makeId("i"), NULL /*init*/)),
              NULL /*patterns*/,
              p_impl(p_bounds(env.makeVar("i"), e_length->clone()),
                     env.hasClassTag(ct, e_plus(e_ptr->clone(), env.makeVar("i")))
                    )));
          FOREACH_OBJLIST(Variable, ct->fieldSlots, iter) {
            StringRef f = iter.data()->name;
            env.assign(f, 
              apply4(env.makeId("updflatrange"),
                     env.makeVar(f),
                     e_ptr->clone(),
                     e_plus(e_ptr->clone(), e_length->clone()),
                     e_val->clone()));
          }

          delete e_length;
          delete e_val;
          delete e_ptr;
          return e_lit(0);      // no return value
        }
      }

      // match_expr
      //   memset($e_ptr, $e_val, $e_length * sizeof($t_elt))
      // where
      //   type_eq($e_ptr, $t_elt *)
      // translate {
      //   mem := mem{$e_ptr, [0 ... $e_length] := $e_val};
      //   meta := meta{$e_ptr, [0 ... $e_length] := live};
      // }

      //Expression *width = args->nth(2)->expr->asE_binary()->e2;
      // TODO: check that type in 'width' matches what I expect

      env.assign("mem",
        apply5(env.makeId("updrange"),
               env.makeVar("mem"),
               e_ptr->clone(),
               env.intLit(0),
               e_length->clone(),
               e_val));
      env.assign("meta",
        apply5(env.makeId("updrange"),
               env.makeVar("meta"),
               e_ptr,
               env.intLit(0),
               e_length,
               env.makeVar("live")));

      // e_ptr, e_val, and e_length consumed above
      return env.intLit(0);      // no return value
    }

    if (args->nth(2)->expr->isE_sizeofType() &&
        args->nth(0)->expr->type->asRval()->isPointerType()) {
      if (pt->atType->isCompoundType()) {
        ECompoundType *ct = ext(pt->atType->asCompoundType());

        env.prove("tag check on memset",
          env.hasClassTag(ct, e_ptr->clone()));

        if (ct->flatMap) {
          // match_expr
          //   memset($e_ptr, $e_val, sizeof(struct $n))
          // where
          //   type_eq($e_ptr, struct $n *);
          //   struct $n uses the flat model;
          // translate {
          //   mem := mem{$e_ptr, [0 ... $e_length] := $e_val};
          //   meta := meta{$e_ptr, [0 ... $e_length] := live};
          // }

          string size = stringc << ct->name << "_size";

          env.assign("mem",
            apply5(env.makeId("updrange"),
                   env.makeVar("mem"),
                   e_ptr->clone(),
                   env.intLit(0), env.makeVar(size),
                   e_val));
          env.assign("meta",
            apply5(env.makeId("updrange"),
                   env.makeVar("meta"),
                   e_ptr,
                   env.intLit(0), env.makeVar(size),
                   env.makeVar("live")));
        }

        else {
          // match_expr
          //   memset($e_ptr, $e_val, sizeof(struct $n))
          // where
          //   type_eq($e_ptr, struct $n *);
          //   struct $n uses the split model;
          // translate {
          //   prove "memset" tag[$e1] == tagName(struct $n);
          //   for_each_field $f of struct $n {
          //     $f := $f{$e_ptr := $e_val};
          //   }
          // }
          FOREACH_OBJLIST(Variable, ct->fieldSlots, iter) {
            StringRef f = iter.data()->name;
            env.assignUpd(f, e_ptr->clone(), e_val->clone());
          }
          delete e_val;
          delete e_ptr;
        }

        return env.intLit(0);      // no return value
      }
    }
    delete e_val;
    delete e_ptr;
  }

  // match_expr
  //   free($e_arr[$e_idx])
  // translate {
  //   prove "precondition of free" {
  //     mem[$e_arr,$e_idx] != 0;
  //     meta[$e_arr,$e_idx] == live;
  //   }
  //   meta := meta{$e_arr,$e_idx := dead};
  // }
  if (func->isE_variable() &&
      func->asE_variable()->var->name == env.str_free &&
      args->nth(0)->expr->isE_deref()) {
    Expression *ptr = args->nth(0)->expr->asE_deref()->ptr;
    if (ptr->isBinary(BIN_PLUS)) {
      Expression *arr = ptr->asE_binary()->e1;
      Expression *idx = ptr->asE_binary()->e2;
      env.prove("precondition of free",
        p_and(
          p_notEqual(apply2(env.makeId("sel"),
                            env.makeVar("mem"),
                            apply2(env.makeId("ptr"),
                                   arr->c2vml(env),
                                   idx->c2vml(env))),
                     env.intLit(0)),
          p_equal(apply2(env.makeId("sel"),
                         env.makeVar("meta"),
                         apply2(env.makeId("ptr"),
                                arr->c2vml(env),
                                idx->c2vml(env))),
                  env.makeVar("live")
                 )));
      env.assign("meta",
        apply3(env.makeId("upd"),
               env.makeVar("meta"),
               apply2(env.makeId("ptr"),
                      arr->c2vml(env),
                      idx->c2vml(env)),
               env.makeVar("dead")));
      return env.intLit(0);     // no return value
    }
  }

  if (func->isE_variable() &&
      func->asE_variable()->var->name == env.str_free) {
    Expression *arg = args->nth(0)->expr;

    // the following patterns are *not* exclusive, so track
    // whether we've used at least one
    bool ok = false;

    // match_expr
    //   free($n)     // free called on a variable
    // translate {
    //   prove "precondition of free" {
    //     $n != 0;
    //     metaVar($n) == live;
    //   }
    //   if (hasAttr($n, "array")) {
    //     objsize[$n] := 0;
    //   }
    //   metaVar($n) := dead;
    // }
    if (arg->isE_variable()) {
      EVariable *var = ext(arg->asE_variable()->var);

      env.prove("precondition of free",
        p_and(
          p_notEqual(env.makeVar(var->name),
                     env.intLit(0)),
          #if USE_META
          p_equal(env.makeVar(stringc << var->name << "_meta"),
                  env.makeVar("live"))
          #else
          new P_lit(SL_UNKNOWN, true)
          #endif
        ));

      #if USE_META
      env.assign(stringc << var->name << "_meta",
                 env.makeVar("dead"));
      #endif

      if (var->hasVAttr(VA_ARRAY)) {
        env.assignUpd("objsize", env.makeVar(var->name), env.intLit(0));
      }

      ok = true;
    }

    // match_expr
    //   free($e)
    // where
    //   typeof($e) is struct $n *
    // translate {
    //   prove "precondition of free(struct ptr)" {
    //     $e != 0;
    //     tag[$e] == tagName($n);
    //   }
    //   tag := tag{$e := 0};
    // }
    if (arg->type->asRval()->isPointerType() &&
        arg->type->asRval()->asPointerType()->atType->isCompoundType()) {
      ECompoundType *ct = ext(arg->type->asRval()->asPointerType()->atType->asCompoundType());
      VExpr * owner e = arg->c2vml(env);

      env.prove("precondition of free(struct ptr)",
        p_and(p_notEqual(e->clone(), e_lit(0)),
              env.hasClassTag(ct, e->clone())
             ));
      env.assignUpd("tag", e->clone(), e_lit(0));

      delete e;
      ok = true;
    }

    if (ok) {
      // matched at least one of the patterns
      return env.intLit(0);     // no return value
    }
  }

  // match_expr
  //   kfree($e)
  // translate {
  //   prove "precondition of kfree"
  //     tag[$e] != 0;
  //   tag := tag{$e := 0};
  // }
  if (func->isE_variable() &&
      func->asE_variable()->var->name == env.str_kfree) {
    VExpr * owner e = args->nth(0)->expr->c2vml(env);

    env.prove("precondition of kfree",
      p_notEqual(env.sel(env.makeVar("tag"), e->clone()), e_lit(0)));
    env.assignUpd("tag", e->clone(), e_lit(0));

    delete e;
    return env.intLit(0);     // no return value
  }

  // match_expr
  //   memmove($eDestObj->$fDest + $eDest,       // (*$eDestObj).$fDest + $eDest
  //           $eSrcObj->$fSrc + $eSrc,          // (*$eSrcObj).$fSrc + $eSrc
  //           $eElts * sizeof($t))
  // or
  //   memcpy(...)
  // where
  //   type of $fDest is $t[], and
  //   type of $fSrc is $t[]
  // translate {
  //   prove "memmove: dest range check"
  //     0 <= $eDest <= $eDest+$eElts <= arraySize($fDest);
  //   prove "memmove: src range check"
  //     0 <= $eSrc  <= $eSrc+$eElts  <= arraySize($fSrc);
  //   if (func is memcpy && $fDest==$fSrc) {
  //     prove "memcpy: non-overlapping"
  //       $eDestObj != $eSrcObj       ||
  //       $eDest+$eElts <= $eSrc      ||
  //       $eSrc+$eElts <= $eDest      ;
  //   }
  //   $fDest := $fDest{$eDestObj :=
  //     copyrange($fDest[$eDestObj], $eDest, $fSrc[$eSrcObj], $eSrc, $eElts)};
  // }
  if (func->isE_variable() &&
      (func->asE_variable()->var->name == env.str_memmove ||
       func->asE_variable()->var->name == env.str_memcpy) &&
      args->count() == 3 &&
      args->nth(0)->expr->isBinary(BIN_PLUS) &&
      args->nth(1)->expr->isBinary(BIN_PLUS) &&
      args->nth(2)->expr->isBinary(BIN_MULT)) {
    E_binary *arg0 = args->nth(0)->expr->asE_binary();
    E_binary *arg1 = args->nth(1)->expr->asE_binary();
    E_binary *arg2 = args->nth(2)->expr->asE_binary();

    if (arg0->e1->isE_fieldAcc() &&
        arg1->e1->isE_fieldAcc() &&
        arg2->e2->isE_sizeofType()) {
      E_fieldAcc *fieldAcc0 = arg0->e1->asE_fieldAcc();
      E_fieldAcc *fieldAcc1 = arg1->e1->asE_fieldAcc();
      E_sizeofType *sizeof2 = arg2->e2->asE_sizeofType();

      if (fieldAcc0->obj->isE_deref() &&
          fieldAcc1->obj->isE_deref() &&
          fieldAcc0->field->type->isArrayType() &&
          fieldAcc0->field->type->isArrayType()) {
        E_deref *deref0 = fieldAcc0->obj->asE_deref();
        E_deref *deref1 = fieldAcc1->obj->asE_deref();
        ArrayType *at0 = fieldAcc0->field->type->asArrayType();
        ArrayType *at1 = fieldAcc1->field->type->asArrayType();

        if (at0->eltType->equals(at1->eltType) &&
            at0->eltType->equals(sizeof2->atype->getType())) {
          StringRef funcName = func->asE_variable()->var->name;

          VExpr * owner eDestObj = deref0->ptr->c2vml(env);
          string fDest = env.fieldSlotName(fieldAcc0->field);
          VExpr * owner eDest = arg0->e2->c2vml(env);

          VExpr * owner eSrcObj = deref1->ptr->c2vml(env);
          string fSrc = env.fieldSlotName(fieldAcc1->field);
          VExpr * owner eSrc = arg1->e2->c2vml(env);

          VExpr * owner eElts = arg2->e1->c2vml(env);

          // debugging aid
          env.comment(stringc << funcName
            << ": eDestObj=" << eDestObj->print()
            << " fDest=" << fDest
            << " eDest=" << eDest->print()
            << " eSrcObj=" << eSrcObj->print()
            << " fSrc=" << fSrc
            << " eSrc=" << eSrc->print()
            << " eElts=" << eElts->print()
          );

          env.prove(stringc << funcName << ": dest range check",
            p_compare4(e_lit(0), CO_LESSEQ,
                       eDest->clone(), CO_LESSEQ,
                       e_plus(eDest->clone(), eElts->clone()), CO_LESSEQ,
                       e_lit(at0->size)));
                       
          env.prove(stringc << funcName << ": src range check",
            p_compare4(e_lit(0), CO_LESSEQ,
                       eSrc->clone(), CO_LESSEQ,
                       e_plus(eSrc->clone(), eElts->clone()), CO_LESSEQ,
                       e_lit(at1->size)));

          if (funcName == env.str_memcpy &&
              fieldAcc0->field == fieldAcc1->field) {
            env.prove("memcpy: non-overlapping",
              p_or3(p_notEqual(eDestObj->clone(), eSrcObj->clone()),
                    p_compare(e_plus(eDest->clone(), eElts->clone()), CO_LESSEQ,
                              eSrc->clone()),
                    p_compare(e_plus(eSrc->clone(), eElts->clone()), CO_LESSEQ,
                              eDest->clone())
                   ));
          }
          
          env.assignUpd(fDest,
            eDestObj->clone(),
            apply5(env.makeId("copyrange"),
                   env.sel(env.makeVar(fDest), eDestObj->clone()),
                   eDest->clone(),
                   env.sel(env.makeVar(fSrc), eSrcObj->clone()),
                   eSrc->clone(),
                   eElts->clone()
                  ));

          delete eDestObj;
          delete eDest;
          delete eSrcObj;
          delete eSrc;
          delete eElts;

          return env.intLit(0);
        }
      }
    }
  }

  // check for specific functions that we expect to model directly
  if (func->isE_variable()) {
    Variable *v = func->asE_variable()->var;
    if (v->name == env.str_malloc ||
        v->name == env.str_free ||
        v->name == env.str_kfree) {
      cout << "warning: unmodeled call to " << v->name << "\n";
    }
  }

  // not a special function
  return NULL;
  #endif // 0; obsolete
}


VExpr * owner E_funCall::ic2vml(C2VML &env)
{
  // check for special functions
  {
    VExpr * owner ret = c2vmlSpecialFunctions(env);
    if (ret) {
      return ret;
    }
  }

  // get the function being called
  if (!func->isE_variable()) {
    // I actually already know how to vcgen indirect calls, but VML
    // doesn't yet have a syntax for them!
    return env.todo("indirect call to function");
  }
  EVariable *funcVar = ext(func->asE_variable()->var);
  FunctionType *ft = funcVar->type->asFunctionType();

  // list of target variables
  ASTList<Identifier> *targets = new ASTList<Identifier>;

  // evaluate the arguments
  ASTList<VExpr> *vargs = c2vmlArgs(env, args, targets);

  if (funcVar->hasVAttr(VA_PATS_MPAT)) {
    // This function (e.g., PATS or MPAT) takes pattern terms as
    // arguments.  As syntactic sugar, we allow BIN_EQUAL and
    // BIN_NOTEQUAL to appear, and translate them into TRIGGEREQ
    // and TRIGGERNEQ, respectively.  Since we just did c2vmlArgs,
    // they will appear as VE_apply("logical_equal",...) and
    // VE_apply("logical_notequal",...).
    FOREACH_ASTLIST_NC(VExpr, *vargs, iter) {
      if (!iter.data()->isVE_apply()) { continue; }
      VE_apply *ap = iter.data()->asVE_apply();
      
      if (0==strcmp("logical_equal", ap->func->name)) {
        ap->func->name = env.str("TRIGGEREQ");
      }
      if (0==strcmp("logical_notequal", ap->func->name)) {
        ap->func->name = env.str("TRIGGERNEQ");
      }
    }
  }

  // Why is VA_PREDSYM included in the conditional below?  Typically,
  // predicate symbols get handed in E_funcall::c2vmlPred().  However,
  // if a predicate symbol is used as the topmost symbol in a trigger
  // pattern, then we get here, and want the trigger to be a simple
  // VE_apply (as opposed to P_compare, as c2vmlPred would make), so
  // go ahead.  I could conceivably pass in a special flag allowing
  // predicate symbols here only in that one circumstance, but it
  // doesn't seem worth it (maybe there are other cases where I want
  // to treat it as a function symbol, and it does not hurt anything).

  if (funcVar->hasVAttr(VA_FUNCSYM) || funcVar->hasVAttr(VA_PREDSYM)) {
    if (funcVar->hasVAttr(VA_TRIGGER)) {
      if (vargs->count() != 1) {
        env.error("trigger pattern function takes 1 argument");
        return NULL;
      }

      // make an E_pat
      VExpr *e = vargs->removeFirst();
      delete vargs;
      return new VE_pat(e);
    }

    else {
      // make an E_apply
      return new VE_apply(
        new Identifier(funcVar->loc, funcVar->name),   // function
        vargs /*implicitly deallocated*/               // arguments
      );
    }

    // not needed after all
    delete targets;
  }

  else {
    if (env.inPredicate) {
      env.error(func->asE_variable()->name->loc,
                stringc << "call to non-function-symbol `" << funcVar->name
                        << "' from within a predicate");
    }

    StringRef tempName = NULL;
    if (!ft->retType->isVoid()) {
      // make a variable to hold the return value
      tempName = env.makeTempVarName(func, "returned value in E_funCall");
      env.addStmt(new VS_decl(SL_UNKNOWN,
        new VD_var(makeId(tempName), NULL /*init*/)));
      targets->prepend(makeId(tempName));
    }

    // build the call node
    env.addStmt(new VS_procCall(SL_UNKNOWN,
      targets,                                       // targets
      new Identifier(funcVar->loc, funcVar->name),   // procedure
      vargs /*implicitly deallocated*/               // arguments
    ));

    if (tempName) {
      // result of the call is the return value
      return new VE_var(makeId(tempName));
    }
    else {
      return e_lit(0);     // arbitrary
    }
  }
}


VExpr * owner E_constructor::ic2vml(C2VML &env)
{ return env.todo("E_constructor"); }


VExpr * owner E_fieldAcc::ic2vml(C2VML &env)
{
  if (obj->type->asRval()->isCompoundType()) {
    ECompoundType *ct = ext(obj->type->asRval()->asCompoundType());
    if (obj->isE_deref() ) {
      VExpr * owner e = obj->asE_deref()->ptr->c2vml(env);

      if (!env.inPredicate) {
        env.prove(stringc << "ptr->field read: " << exprToString(),
          env.hasClassTag(ct, e->clone()));
      }

      if (ct->flatMap) {
        // match_expr
        //   $e->$f                       // (*$e1).$f
        // where
        //   match_type $e with struct $n *
        //   struct $n uses the flat model
        // translate {
        //   prove tag[$e] == tagName(struct $n);
        //   return mem[$e, fieldName($f)];
        // }
        return apply2(env.makeId("sel"),
                      env.makeVar("mem"),
                      apply2(env.makeId("ptr"),
                             e,
                             env.makeVar(env.fieldSlotName(field))
                            ));
      }
      else {
        // match_expr
        //   $e->$f                       // (*$e1).$f
        // where
        //   match_type $e with struct $n *
        //   struct $n uses the split model
        // translate {
        //   prove tag[$e] == tagName(struct $n);
        //   return fieldName($f)[$e];
        // }
        return apply2(env.makeId("sel"),
                      env.makeVar(env.fieldSlotName(field)),
                      e);
      }
    }

    if (obj->isE_variable()) {
      EVariable *v = ext(obj->asE_variable()->var);
      if (v->hasVAttr(VA_ADDRTAKEN)) {
        // match_expr
        //   $n.$f
        // where
        //   $n has attribute `addrtaken'
        // translate {
        //   return $f[$n];
        // }
        return apply2(env.makeId("sel"),
                      env.makeVar(env.fieldSlotName(field)),
                      env.makeVar(v->name));
      }
      else {
        // match_expr
        //   $n.$f
        // where
        //   $n does not have attribute `addrtaken'
        // translate {
        //   return instanceFieldName($n,$f)
        // }
        EVariable *synth = env.getSynthField(v, field);
        if (synth) {
          return env.makeVar(synth->name);
        }
      }
    }
  }

  return env.todo("unhandled E_fieldAcc");
}


VExpr * owner E_sizeof::ic2vml(C2VML &env)
{
  // expression is not evaluated, so no side effects
  return new VE_lit(size);
}


VExpr * owner E_unary::ic2vml(C2VML &env)
{
  VExpr * owner e = expr->c2vml(env);
  
  switch (op) {
    default: xfailure("bad");

    case UNY_PLUS:   return e;
    case UNY_MINUS:  return e_neg(e);

    // plan: Functions of the form "logical_<foo>" will be defined by
    // axioms supplied to Simplify that give the condition for a 0
    // value and the condition for a 1 value.  If Simplify needs to do
    // the case analysis, it can; but I won't force it by splitting
    // the path.
    case UNY_NOT:    return apply1(env.makeId("logical_not"), e);

    case UNY_BITNOT: return apply1(env.makeId("bitwise_not"), e);
  }
}


// raw combination, ignoring side effects, short ciruit, etc.
VExpr * owner convertBinaryOperator
  (C2VML &env, VExpr * owner expr1, BinaryOp op, VExpr * owner expr2)
{
  switch (op) {
    default:
    case BIN_BRACKETS:
    case BIN_ASSIGN:
      xfailure("convertBinaryOperator: bad operator");

    case BIN_EQUAL:      return apply2(env.makeId("logical_equal"), expr1, expr2);
    case BIN_NOTEQUAL:   return apply2(env.makeId("logical_notequal"), expr1, expr2);
    case BIN_LESS:       return apply2(env.makeId("logical_less"), expr1, expr2);
    case BIN_GREATER:    return apply2(env.makeId("logical_greater"), expr1, expr2);
    case BIN_LESSEQ:     return apply2(env.makeId("logical_lesseq"), expr1, expr2);
    case BIN_GREATEREQ:  return apply2(env.makeId("logical_greatereq"), expr1, expr2);

    case BIN_MULT:       return new VE_arith(expr1, AO_TIMES, expr2);
    case BIN_DIV:        return apply2(env.makeId("div"), expr1, expr2);
    case BIN_MOD:        return apply2(env.makeId("mod"), expr1, expr2);
    case BIN_PLUS:       return new VE_arith(expr1, AO_PLUS, expr2);
    case BIN_MINUS:      return new VE_arith(expr1, AO_MINUS, expr2);
    case BIN_LSHIFT:     return apply2(env.makeId("lshift"), expr1, expr2);
    case BIN_RSHIFT:     return apply2(env.makeId("rshift"), expr1, expr2);
    case BIN_BITAND:     return apply2(env.makeId("bitand"), expr1, expr2);
    case BIN_BITXOR:     return apply2(env.makeId("bitxor"), expr1, expr2);
    case BIN_BITOR:      return apply2(env.makeId("bitor"), expr1, expr2);
    case BIN_AND:        return apply2(env.makeId("logical_and"), expr1, expr2);
    case BIN_OR:         return apply2(env.makeId("logical_or"), expr1, expr2);
    case BIN_COMMA:      delete expr1; return expr2;

    case BIN_DOT_STAR:
    case BIN_ARROW_STAR:
      return env.todo("convertBinaryOperator: .* or ->*");

    case BIN_IMPLIES:    return apply2(env.makeId("logical_implies"), expr1, expr2);
    case BIN_EQUIVALENT: return apply2(env.makeId("logical_equivalent"), expr1, expr2);
  }
}


VExpr * owner E_effect::ic2vml(C2VML &env)
{
  // ++ -> +, -- -> -
  BinaryOp bop = (op==EFF_POSTINC || op==EFF_PREINC)? BIN_PLUS : BIN_MINUS;
  
  // compute original value
  VExpr * owner orig = expr->c2vml(env);

  if (op == EFF_POSTINC || op == EFF_POSTDEC) {
    // save original in a temporary
    StringRef tmp = env.makeTempVarName(expr, "postfix ++/-- operator");
    env.define(tmp, orig->clone());

    // increment
    VExpr * owner inc = convertBinaryOperator(env, orig, bop, env.intLit(1));
    delete env.assignmentExpr(expr, inc);

    // yield temporary
    return env.makeVar(tmp);
  }
  else {
    // increment
    VExpr * owner inc = convertBinaryOperator(env, orig, bop, env.intLit(1));
    return env.assignmentExpr(expr, inc);    // yield modified
  }
}


VExpr * owner E_binary::ic2vml(C2VML &env)
{
  VExpr * owner expr1 = e1->c2vml(env);

  // isolate side effects of expr2
  env.pushCompound(SL_UNKNOWN);
  VExpr * owner expr2 = e2->c2vml(env);
  VS_compound * owner expr2effects = env.popCompound();

  // short circuit operator with side effects?
  if ((op == BIN_AND || op == BIN_OR) && expr2effects->body.isNotEmpty()) {
    // don't do this until I've had a chance to test the similar behavior
    // in c2vmlPred
    return env.todo("c2vml: && or || with side effects in RHS");

    #if 0
    // temporary to hold the result of the evaluation
    StringRef tempName = env.makeTempVarName();

    // if we evaluate expr2, the last thing is to set the temporary
    expr2effects->body.append(new VS_assign(env.makeId(tempName),
      apply1(env.makeId("logical_value" /*zero or one*/), expr2)));

    // conditionally evaluate expr2's side effects:
    //   - for BIN_AND, evaluate expr2 if expr1 is not zero
    //   - for BIN_OR, evaluate expr2 if expr1 *is* zero
    env.addStmt(new VS_if(
      p_compare(expr1, op==BIN_AND? CO_NOTEQUAL : CO_EQUAL, new VE_lit(0)),
      expr2effects /*sThen*/,

      // sElse: set the temporary to a determined value, since we've
      // tested expr1 against zero
      new VS_assign(env.makeId(tempName), new VE_lit(op==BIN_AND? 1 : 0))
    ));

    return env.makeVar(tempName);
    #endif // 0
  }

  // both sides are evaluated
  env.foldStmt(expr2effects);

  return convertBinaryOperator(env, expr1, op, expr2);
}


VExpr * owner E_addrOf::ic2vml(C2VML &env)
{
  // match_expr
  //   &*$e
  // translate {
  //   $e
  // }
  if (expr->isE_deref()) {
    return expr->asE_deref()->ptr->c2vml(env);
  }

  if (expr->isE_variable()) {
    EVariable *v = ext(expr->asE_variable()->var);
    if (v->hasVAttr(VA_ADDRTAKEN)) {
      // match_expr
      //   &$n
      // where
      //   $n has attribute `addrtaken'
      // translate {
      //   $n            // name is treated like an object id in memory
      // }
      return env.makeVar(v->name);
    }

    //env.warning(stringc << "address of " << v->name);
  }

  return env.todo(stringc << "address of: " << exprToString());
}


VExpr * owner E_deref::ic2vml(C2VML &env)
{
  if (ptr->isE_variable()) {
    EVariable *v = ext(ptr->asE_variable()->var);
    if (v->hasVAttr(VA_PARAMBYREF)) {
      // match_expr
      //   *$n
      // where
      //   $n has attribute VA_PARAMBYREF
      // translate {
      //   return $n;
      // }
      return env.makeVar(v->name);
    }
  }

  if (ptr->isBinary(BIN_PLUS)) {
    E_binary *bin = ptr->asE_binary();

    // match_expr
    //   ($e1->$f)[$e2]               // *( (*$e1).$f + $e2 )
    // where {
    //   isArrayType($e1->$f);
    //   match_type $e1 with struct $n *;
    // }
    // translate {
    //   prove tag[$e1] == tagName(struct $n);
    //   prove 0 <= $e2 < arraySize($f);
    //   return mem[$e1,($f + $e2)]
    // }
    if (bin->e1->isE_fieldAcc()) {
      E_fieldAcc *acc = bin->e1->asE_fieldAcc();
      if (acc->obj->isE_deref()) {
        if (acc->field->type->isArrayType() &&
            acc->obj->type->asRval()->isCompoundType()) {
          EVariable *field = ext(acc->field);
          ArrayType *at = field->type->asArrayType();
          ECompoundType *ct = ext(acc->obj->type->asRval()->asCompoundType());

          VExpr * owner e1 = acc->obj->asE_deref()->ptr->c2vml(env);
          VExpr * owner e2 = bin->e2->c2vml(env);

          env.prove(stringc << "structure array field read: " << exprToString(),
            p_and(env.hasClassTag(ct, e1->clone()),
                  p_bounds(e2->clone(), e_lit(at->size))));

          if (ct->flatMap) {
            return env.sel(env.makeVar("mem"),
                           apply2(env.makeId("ptr"),
                                  e1,
                                  e_plus(env.makeVar(field->modelVar->name), e2)
                                 )
                          );
          }
          else {
            // return fieldName($f)[$e1][$e2];
            VExpr * owner objectArray =
              env.sel(env.makeVar(env.fieldSlotName(field)), e1);
            return env.sel(objectArray, e2);
          }
        }
      }
    }

    // match_expr
    //   $n[$e2]             // *($n + $e2)
    // where
    //   typeof($n) == $t [$eSz]      // directly-modeled array
    // translate {
    //   prove "array var read"
    //     0 <= $e2 < $eSz;
    //   return $n[$e2];
    // }
    if (bin->e1->isE_variable() &&
        bin->e1->type->asRval()->isArrayType()) {
      Variable *n = bin->e1->asE_variable()->var;
      int eSz = n->type->asArrayType()->size;
      if (eSz != ArrayType::NO_SIZE) {
        VExpr * owner e2 = bin->e2->c2vml(env);

        env.prove(stringc << "array var read: " << exprToString(),
          p_bounds(e2->clone(), e_lit(eSz)));

        return apply2(env.makeId("sel"),
                      env.makeVar(n->name),
                      e2);
      }
    }

    // match_expr
    //   $e1[$e2]             // *($e1 + $e2)
    // translate {
    //   prove "array ptr read"
    //     0 <= $e2 < objsize[$e1];
    //   return mem[$e1,$e2];
    // }
    VExpr * owner e1 = bin->e1->c2vml(env);
    VExpr * owner e2 = bin->e2->c2vml(env);
    
    env.prove(stringc << "array ptr read: " << exprToString(),
      p_bounds(e2->clone(), env.sel(env.makeVar("objsize"), e1->clone())));

    return env.sel(env.makeVar("mem"),
                   env.ptr(e1, e2));
  }

  VExpr * owner addr = env.rval(ptr->c2vml(env));

  return env.sel(env.makeVar("mem"), addr);
}


VExpr * owner E_cast::ic2vml(C2VML &env)
{
  // is this a call to malloc with a typecast in front?
  if (ctype->getType()->isPointerType() &&
      expr->isE_funCall()) {
    Type *eltType = ctype->getType()->asPointerType()->atType;
    E_funCall *f = expr->asE_funCall();
    if (f->func->isE_variable() &&
        f->func->asE_variable()->var->name == env.str_malloc) {
      Expression *a = f->args->first()->expr;
      if (a->isE_binary() &&
          a->asE_binary()->op == BIN_MULT) {
        Expression *lhs = a->asE_binary()->e1;
        Expression *rhs = a->asE_binary()->e2;
        if (rhs->isE_sizeofType() &&
            rhs->asE_sizeofType()->atype->getType()->equals(eltType)) {
          if (eltType->isCompoundType() &&
              0==strcmp(eltType->asCompoundType()->name, "Scull_Dev")) {
            // match: (Scull_Dev *)malloc($e * sizeof(Scull_Dev))
            // replacement:
            //   define $fresh1 = objct;
            //   objct := objct + $e;
            //   tag := updflatrange(tag, $fresh1, $fresh1+$e, tagName(struct Scull_Dev));
            //   return $fresh1;      // becomes the yielded expression
            StringRef fresh = env.makeTempVarName("malloc effective result");
            VExpr * owner e = lhs->c2vml(env);

            env.define(fresh, env.makeVar("objct"));

            env.assign("objct",
              e_plus(env.makeVar("objct"), e->clone()));

            env.assign("tag",
              apply4(env.makeId("updflatrange"),
                     env.makeVar("tag"),
                     env.makeVar(fresh),
                     e_plus(env.makeVar(fresh), e),
                     env.makeVar("Scull_Dev_tag")
                    ));

            return env.makeVar(fresh);
          }

          // match: ($t *)malloc($e * sizeof($t))
          // replacement:
          //   define $fresh1 = objct;
          //   objct := objct + 1;
          //   meta := meta{$fresh1,[0..$e] := dead};
          //   objsize := objsize{$fresh1 := $e};
          //   return $fresh1;      // becomes the yielded expression
          StringRef fresh = env.makeTempVarName("malloc effective result");
          VExpr * owner e = lhs->c2vml(env);
          env.define(fresh, env.makeVar("objct"));
          env.assign("objct", e_plus(env.makeVar("objct"), e_lit(1)));
          env.assign("meta",
            apply5(env.makeId("updrange"),
                   env.makeVar("meta"),
                   env.makeVar(fresh),
                   new VE_lit(0),
                   e->clone(),
                   env.makeVar("dead")));
          env.assign("objsize",
            env.upd(env.makeVar("objsize"),
                    env.makeVar(fresh),
                    e));
          return env.makeVar(fresh);
        }
      }

      if (a->isE_sizeofType() &&
          eltType->isCompoundType()) {
        Type *typeInSizeof = a->asE_sizeofType()->atype->getType();
        if (typeInSizeof->equals(eltType)) {
          ECompoundType *ct = ext(eltType->asCompoundType());
          string numFields = stringc << ct->name << "_size";
          string tagName = stringc << ct->name << "_tag";

          // match_expr
          //   (struct $n *)malloc(sizeof(struct $n))
          // translate {
          //   define $fresh1 = objct;
          //   objct := objct + 1;
          //   tag := tag{$fresh1 := tagName(struct $n)};
          //   static_if (struct $n uses flat model) {
          //     meta := meta{$fresh1,[0 ... numFields(struct $n)] := dead};
          //   }
          //   objsize := objsize{$fresh1 := numFields(struct $n)};
          //   return $fresh1;
          // }
          StringRef fresh = env.makeTempVarName("malloc effectie result");
          env.define(fresh, env.makeVar("objct"));
          env.assign("objct", e_plus(env.makeVar("objct"), e_lit(1)));
          env.assign("tag",
            env.upd(env.makeVar("tag"),
                    env.makeVar(fresh),
                    env.makeVar(tagName)));
          if (ct->flatMap) {
            env.assign("meta",
              apply5(env.makeId("updrange"),
                     env.makeVar("meta"),
                     env.makeVar(fresh),
                     new VE_lit(0),
                     env.makeVar(numFields),
                     env.makeVar("dead")));
          }
          env.assign("objsize",
            env.upd(env.makeVar("objsize"),
                    env.makeVar(fresh),
                    env.makeVar(numFields)));
          return env.makeVar(fresh);
        }
      }
    }
  }

  VExpr * owner v = expr->c2vml(env);

  // cast to pointer?
  if (type->isPointerType() &&
      type->asPointerType()->atType->isCompoundType()) {
    // to demonstrate that the cast is valid, prove that either the
    // pointer is NULL or else that the tag matches
    //
    // UPDATE: George points out this is probably redundant, since
    // what really matters is the tag check when the pointer is
    // *used*; being passed around doesn't really affect anything.
    // But for now I'll keep it.

    ECompoundType *ct = ext(type->asPointerType()->atType->asCompoundType());

    env.prove("type cast to a structure pointer",
      p_or(p_equal(v->clone(), e_lit(0)),      // NULL, or
           env.hasClassTag(ct, v->clone())));  // has expected tag
  }

  return v;
}


VExpr * owner E_cond::ic2vml(C2VML &env)
{
  return new VE_cond(cond->c2vmlPred(env), 
                     th->c2vml(env), el->c2vml(env));
}


VExpr * owner E_sizeofType::ic2vml(C2VML &env)
{
  return e_lit(size);
}


VExpr * owner E_assign::ic2vml(C2VML &env)
{
  // source (RHS) value
  VExpr * owner srcValue = src->c2vml(env);

  if (op != BIN_ASSIGN) {
    // combine with LHS
    srcValue = convertBinaryOperator(env, target->c2vml(env), op, srcValue);
  }

  return env.assignmentExpr(target, srcValue);
}


VExpr * owner E_new::ic2vml(C2VML &env)
{ return env.todo("E_new"); }
VExpr * owner E_delete::ic2vml(C2VML &env)
{ return env.todo("E_delete"); }
VExpr * owner E_throw::ic2vml(C2VML &env)
{ return env.todo("E_throw"); }
VExpr * owner E_keywordCast::ic2vml(C2VML &env)
{ return env.todo("E_keywordCast"); }
VExpr * owner E_typeidExpr::ic2vml(C2VML &env)
{ return env.todo("E_typeidExpr"); }
VExpr * owner E_typeidType::ic2vml(C2VML &env)
{ return env.todo("E_typeidType"); }


VExpr * owner E_grouping::ic2vml(C2VML &env)
{
  return expr->c2vml(env);
}


VExpr * owner E_arrow::ic2vml(C2VML &env)
{
  xfailure("E_arrow::ic2vml: should not get here");
  return NULL;
}


VExpr * owner E_quantifier::ic2vml(C2VML &env)
{
  xfailure("E_quantifier used as an expression, not a predicate");
  return NULL;
}


VExpr * owner E_let::ic2vml(C2VML &env)
{
  // I think to support this, VML should have a 'let' construct too
  return env.todo("E_let");
}


VExpr * owner E_pre::ic2vml(C2VML &env)
{
  return new VE_pre(expr->c2vml(env));
}


VExpr * owner E_vc_expression::ic2vml(C2VML &env)
{
  PredBracket pb(env);
  return expr->c2vml(env);
}


// ------------------------ Expression::c2vmlPred ------------------------
// STUBPATTERN: ^Pred.*E_.*::c2vmlPred\b

Pred * owner Expression::c2vmlPredDefault(C2VML &env)
{
  // evaluate like an expression
  VExpr * owner e = c2vml(env);

  // true if not zero
  return p_compare(e, CO_NOTEQUAL, e_lit(0));
}


Pred * owner E_boolLit::c2vmlPred(C2VML &env)
{
  return new P_lit(SL_UNKNOWN, b);
}


Pred * owner E_intLit::c2vmlPred(C2VML &env)
{ return c2vmlPredDefault(env); }
Pred * owner E_floatLit::c2vmlPred(C2VML &env)
{ return c2vmlPredDefault(env); }
Pred * owner E_stringLit::c2vmlPred(C2VML &env)
{ return c2vmlPredDefault(env); }
Pred * owner E_charLit::c2vmlPred(C2VML &env)
{ return c2vmlPredDefault(env); }
Pred * owner E_this::c2vmlPred(C2VML &env)
{ return c2vmlPredDefault(env); }
Pred * owner E_variable::c2vmlPred(C2VML &env)
{ return c2vmlPredDefault(env); }


static StringRef get_dpo_sym(C2VML &env, ArgExpression *aexp)
{
  if (!aexp->expr->isE_variable()) {
    env.error("declare_partial_order: both arguments must be variable names");
  }
  return aexp->expr->asE_variableC()->var->name;
}

Pred * owner E_funCall::c2vmlPred(C2VML &env)
{
  if (func->isE_variable()) {
    EVariable *vfunc = ext(func->asE_variable()->var);
    if (vfunc->name == env.str_declare_partial_order) {
      if (args->count() != 2) {
        env.error("declare_partial_order requires exactly two arguments");
      }
      return new P_declare_partial_order(SL_UNKNOWN,
                   get_dpo_sym(env, args->nth(0)),
                   get_dpo_sym(env, args->nth(1)));
    }

    if (vfunc->hasVAttr(VA_PREDSYM)) {
      // make an E_apply
      VE_apply *app = new VE_apply(
        new Identifier(vfunc->loc, vfunc->name),       // function
        c2vmlArgs(env, args)                           // arguments
      );

      // see note in vml_vcgen.cc regarding '|@true|'
      return p_equal(app, env.intLit(1));
    }
  }

  return c2vmlPredDefault(env);
}


Pred * owner E_constructor::c2vmlPred(C2VML &env)
{ return c2vmlPredDefault(env); }
Pred * owner E_fieldAcc::c2vmlPred(C2VML &env)
{ return c2vmlPredDefault(env); }
Pred * owner E_sizeof::c2vmlPred(C2VML &env)
{ return c2vmlPredDefault(env); }          


Pred * owner E_unary::c2vmlPred(C2VML &env)
{
  if (op == UNY_NOT) {
    return p_not(expr->c2vmlPred(env));
  }
  else {
    return c2vmlPredDefault(env);
  }
}


Pred * owner E_effect::c2vmlPred(C2VML &env)
{ return c2vmlPredDefault(env); }


Pred * owner E_binary::c2vmlPred(C2VML &env)
{
  switch (op) {
    // comparison operators (could abstract with a single case, but
    // this is simpler and I don't care about code size here)
    case BIN_EQUAL:     return p_compare(e1->c2vml(env), CO_EQUAL,     e2->c2vml(env));
    case BIN_NOTEQUAL:  return p_compare(e1->c2vml(env), CO_NOTEQUAL,  e2->c2vml(env));
    case BIN_LESS:      return p_compare(e1->c2vml(env), CO_LESS,      e2->c2vml(env));
    case BIN_GREATER:   return p_compare(e1->c2vml(env), CO_GREATER,   e2->c2vml(env));
    case BIN_LESSEQ:    return p_compare(e1->c2vml(env), CO_LESSEQ,    e2->c2vml(env));
    case BIN_GREATEREQ: return p_compare(e1->c2vml(env), CO_GREATEREQ, e2->c2vml(env));

    // non-short-circuit predicate combinators
    case BIN_IMPLIES:
    case BIN_EQUIVALENT:
      return new P_binary(SL_UNKNOWN, 
                          e1->c2vmlPred(env), op==BIN_IMPLIES? BP_IMPLIES : BP_IFF,
                          e2->c2vmlPred(env));

    // short-circuit predicate combinators
    case BIN_AND:
    case BIN_OR: {
      // eval e1, e2 as predicates
      Pred * owner pred1 = e1->c2vmlPred(env);

      // isolate side effects of expr2
      env.pushCompound(SL_UNKNOWN);
      Pred * owner pred2 = e2->c2vmlPred(env);
      VS_compound * owner expr2effects = env.popCompound();

      // RHS has side effects?
      if (expr2effects->body.isNotEmpty()) {
        // temporary to hold the result of the evaluation: 1=true, 0=false
        StringRef tempName = env.makeTempVarName(e2, "rhs for short-circuit boolean");
        env.declare(tempName);

        // if we evaluate expr2, the last thing is to set the temporary;
        // use E_cond to avoid a second path split
        expr2effects->body.append(new VS_assign(SL_UNKNOWN,
          env.makeId(tempName),
          new VE_cond(pred2, e_lit(1), e_lit(0))));

        // conditionally evaluate expr2's side effects
        if (op == BIN_AND) {
          env.addStmt(new VS_if(SL_UNKNOWN,
            // is pred1 true?
            pred1,

            // yes: evaluate expr2, setting the temporary to its result
            expr2effects,

            // no: set the temporary to 0
            new VS_assign(SL_UNKNOWN, env.makeId(tempName), new VE_lit(0))
          ));
        }
        else {
          env.addStmt(new VS_if(SL_UNKNOWN,
            // is pred1 true?
            pred1,

            // yes: set the temporary to 1
            new VS_assign(SL_UNKNOWN, env.makeId(tempName), new VE_lit(1)),

            // no: evaluate expr2, setting the temporary to its result
            expr2effects
          ));
        }

        // final predicate is the nonzero-ness of 'tempName'
        return p_compare(env.makeVar(tempName), CO_NOTEQUAL, e_lit(0));
      }

      else {
        // and make a simple VML combinator
        return new P_binary(SL_UNKNOWN, pred1, op==BIN_AND? BP_AND : BP_OR, pred2);
      }
    }

    // everything else: evaluate them, then compare with zero
    default:
      return c2vmlPredDefault(env);
  }
}


Pred * owner E_addrOf::c2vmlPred(C2VML &env)
{ return c2vmlPredDefault(env); }          
Pred * owner E_deref::c2vmlPred(C2VML &env)
{ return c2vmlPredDefault(env); }          
Pred * owner E_cast::c2vmlPred(C2VML &env)
{ return c2vmlPredDefault(env); }

Pred * owner E_cond::c2vmlPred(C2VML &env)
{ env.todo("E_cond::c2vmlPred"); return NULL; }

Pred * owner E_sizeofType::c2vmlPred(C2VML &env)
{ return c2vmlPredDefault(env); }
Pred * owner E_assign::c2vmlPred(C2VML &env)
{ return c2vmlPredDefault(env); }
Pred * owner E_new::c2vmlPred(C2VML &env)
{ return c2vmlPredDefault(env); }
Pred * owner E_delete::c2vmlPred(C2VML &env)
{ return c2vmlPredDefault(env); }
Pred * owner E_throw::c2vmlPred(C2VML &env)
{ return c2vmlPredDefault(env); }
Pred * owner E_keywordCast::c2vmlPred(C2VML &env)
{ return c2vmlPredDefault(env); }
Pred * owner E_typeidExpr::c2vmlPred(C2VML &env)
{ return c2vmlPredDefault(env); }
Pred * owner E_typeidType::c2vmlPred(C2VML &env)
{ return c2vmlPredDefault(env); }             

Pred * owner E_grouping::c2vmlPred(C2VML &env)
{
  return expr->c2vmlPred(env);
}

Pred * owner E_arrow::c2vmlPred(C2VML &env)
{ return c2vmlPredDefault(env); }


Pred * owner E_quantifier::c2vmlPred(C2VML &env)
{                                        
  // build list of VML quantifier variables
  ASTList<VD_var> *qvars = new ASTList<VD_var>;
  FAKELIST_FOREACH(ASTTypeId, vars, iter) {
    Variable *v = iter->decl->var;
    qvars->append(new VD_var(new Identifier(v->loc, v->name), NULL /*init*/));
  }

  static bool enablePats = !tracingSys("disablePatterns");

  VExpr *pat = NULL;
  if (patterns && enablePats) {
    pat = patterns->c2vml(env);
  }

  return new P_quant(loc, isForall /*universal*/, label, qvars, pat, 
                     body->c2vmlPred(env));
}


Pred * owner E_let::c2vmlPred(C2VML &env)
{
  env.todo("E_let::c2vmlPred");
  return NULL;
}


Pred * owner E_pre::c2vmlPred(C2VML &env)
{
  // I think the typechecker should catch and prevent this, though
  // at the moment it does not
  xfailure("E_pre::c2vmlPred");
  return NULL;
}  


Pred * owner E_vc_expression::c2vmlPred(C2VML &env)
{
  return expr->c2vmlPred(env);
}


// COPIED to below
VExpr * owner Expression::c2vmlPure(C2VML &env)
{
  // isolate the effects
  env.pushCompound(SL_UNKNOWN);
  VExpr * owner ret = c2vml(env);
  VS_compound *com = env.popCompound();

  // had better be empty
  if (com->body.isNotEmpty()) {
    env.error(stringc << "expression `" << exprToString()
                      << "' was supposed to be side-effect-free, "
                      << "but it is not");
  }

  return ret;
}

// COPIED from above
Pred * owner Expression::c2vmlPredPure(C2VML &env)
{
  // isolate the effects
  env.pushCompound(SL_UNKNOWN);
  Pred * owner ret = c2vmlPred(env);
  VS_compound *com = env.popCompound();

  // had better be empty
  if (com->body.isNotEmpty()) {
    env.error(stringc << "expression `" << exprToString()
                      << "' was supposed to be side-effect-free, "
                      << "but it is not");
  }

  return ret;
}


// ------------------------ Initializer ------------------------
// STUBPATTERN: ^VExpr.*IN_.*::c2vml\b

VExpr * owner IN_expr::c2vml(C2VML &env, Type *)
{
  return e->c2vml(env);
}

VExpr * owner IN_compound::c2vml(C2VML &env, Type *)
{ return env.todo("IN_compound::c2vml"); }
VExpr * owner IN_ctor::c2vml(C2VML &env, Type *)
{ return env.todo("IN_compound::c2vml"); }


// ------------------------- Macro --------------------------
void Macro::c2vml(C2VML &env)
{
  env.todo("Macro");
}


// EOF
