// c2vvalue.h
// object to represent values in the C2V imperative fragment

#ifndef C2VVALUE_H
#define C2VVALUE_H

#include "strtable.h"     // StringRef
#include "xassert.h"      // xassert

#include <string.h>       // memcpy

class Statement;          // cc.ast
class Expression;         // cc.ast
class ASTTypeId;          // cc.ast
class Variable;           // variable.h
class Type;               // cc_type.h

class Identifier;         // vml.ast
class VExpr;              // vml.ast
class Pred;               // vml.ast
class VStatement;         // vml.ast


// kinds of C2V values
enum C2VValueKind {
  // primitives
  VK_INT,
  VK_IDENTIFIER,
  VK_STRING,

  // C AST fragments
  VK_C_STATEMENT,
  VK_C_EXPRESSION,
  VK_C_TYPE,
  VK_C_VARIABLE,
  VK_C_FIELD,

  // VML AST fragments
  //VK_VML_VARIABLE,
  VK_VML_EXPRESSION,
  VK_VML_PREDICATE,
  VK_VML_STATEMENT,

  NUM_C2V_VALUEKINDS
};

char const *toString(C2VValueKind k);     // e.g. "int"


// a C2V variable can be one of these, which is essentially just
// a tagged union; all contained pointers are serfs
class C2VValue {
public:      // data
  C2VValueKind kind;

  union {
    int i;                     // VK_INT
    StringRef str;             // VK_IDENTIFIER, VK_STRING

    Statement const *c_stmt;   // VK_C_STATEMENT
    Expression const *c_expr;  // VK_C_EXPRESSION
    Type *c_type;              // VK_C_TYPE
    Variable *c_var;           // VK_C_VARIABLE, VK_C_FIELD

    //Identifier *v_var;         // VK_VML_VARIABLE
    VExpr *v_expr;             // VK_VML_EXPRESSION
    Pred *v_pred;              // VK_VML_PREDICATE
    VStatement *v_stmt;        // VK_VML_STATEMENT
  } val;                       // (all pointers are serfs)

private:     // disallowed
  bool operator==(C2VValue const &obj) const;

public:      // funcs
  C2VValue(int i = 0) : kind(VK_INT) { val.i = i; }
  C2VValue(C2VValue const &obj)
    { *this = obj; }
  C2VValue& operator= (C2VValue const &obj)
    { memcpy(this, &obj, sizeof(*this)); return *this; }

  // equality is structural, which requires recursive traversal, and
  // is implemented in C2VML::equalValues

  C2VValue(C2VValueKind k, int i)
    { xassert(k == VK_INT); kind = k; val.i = i; }
  C2VValue(C2VValueKind k, StringRef s)
    { xassert(k == VK_IDENTIFIER || k == VK_STRING); kind = k; val.str = s; }
  C2VValue(C2VValueKind k, Expression const *e)
    { xassert(k == VK_C_EXPRESSION); kind = k; val.c_expr = e; }
  C2VValue(C2VValueKind k, Type *t)
    { xassert(k == VK_C_TYPE); kind = k; val.c_type = t; }
  C2VValue(C2VValueKind k, Variable *v)
    { xassert(k == VK_C_VARIABLE || k == VK_C_FIELD); kind = k; val.c_var = v; }
  C2VValue(C2VValueKind k, VExpr *e)
    { xassert(k == VK_VML_EXPRESSION); kind = k; val.v_expr = e; }
  // TODO: other ctor cases

  int getInt() const
    { xassert(kind == VK_INT); return val.i; }
  StringRef getIdentifier() const
    { xassert(kind == VK_IDENTIFIER); return val.str; }
  StringRef getString() const
    { xassert(kind == VK_STRING); return val.str; }
  Expression const *getCExpr() const
    { xassert(kind == VK_C_EXPRESSION); return val.c_expr; }
  Type *getCType() const
    { xassert(kind == VK_C_TYPE); return val.c_type; }
  Variable *getCVariable() const
    { xassert(kind == VK_C_VARIABLE); return val.c_var; }
  Variable *getCField() const
    { xassert(kind == VK_C_FIELD); return val.c_var; }
  VExpr *getVExpr() const
    { xassert(kind == VK_VML_EXPRESSION); return val.v_expr; }
  // TODO: other extractor cases
};


#endif // C2VVALUE_H
