#!/usr/bin/perl -w
# codepatch.pl
#
# This is sort of like 'patch' but a little higher-level.  I plan
# to extend the script language with constructs as I need them.

if (@ARGV != 1) {
  print("usage: $0 script.patch <input >output\n");
  exit(0);
}

$script = $ARGV[0];
open(SCRIPT, "<$script") or die("cannot open $script: $!\n");

# read all input
@lines = <STDIN>;

# parse the script
$slineNum = 0;
while ($sline = <SCRIPT>) {
  chomp($sline);
  $slineNum++;

  # skip blank lines, comments
  if ($sline =~ m|^\s*(//.*)?$|) { next; }

  # syntax:
  #   replace_line
  #   <source line>
  #   with
  #   <destination line>
  #
  # This is essentially just sed s/^<source>$/<dest>/ except we
  # require that <source> occur exactly once in the input.
  if ($sline eq "replace_line") {
    my $src = <SCRIPT>; chomp($src);
    my $with = <SCRIPT>; chomp($with);
    my $dest = <SCRIPT>; chomp($dest);
    if (!defined($dest) || $with ne "with") {
      fail("malformed replace_line");
    }

    replaceLine($src, $dest);

    $slineNum += 3;
    next;
  }

  # syntax:
  #   comment_function
  #   <function name>
  #
  # Find the given function, and comment it out.
  if ($sline eq "comment_function") {
    my $name = <SCRIPT>;
    chomp($name);

    surroundFunction($name,
                     "#if 0     // comment_function: $name",
                     "#endif // 0");

    $slineNum += 1;
    next;
  }

  # syntax:
  #   add_before_function
  #   <function name>
  #   the_line
  #   <line to add before>
  if ($sline eq "add_before_function") {
    my $name = <SCRIPT>; chomp($name);
    my $the_line = <SCRIPT>; chomp($the_line);
    my $toAdd = <SCRIPT>; chomp($toAdd);
    if (!defined($toAdd) || $the_line ne "the_line") {
      fail("malformed add_before_function");
    }

    surroundFunction($name, $toAdd, "");

    $slineNum += 3;
    next;
  }

  # syntax:
  #   add_after_function
  #   <function name>
  #   the_line
  #   <line to add after>
  if ($sline eq "add_after_function") {
    my $name = <SCRIPT>; chomp($name);
    my $the_line = <SCRIPT>; chomp($the_line);
    my $toAdd = <SCRIPT>; chomp($toAdd);
    if (!defined($toAdd) || $the_line ne "the_line") {
      fail("malformed add_after_function");
    }

    surroundFunction($name, "", $toAdd);

    $slineNum += 3;
    next;
  }

  # syntax:
  #   comment_line
  #   <exact line text to comment out>
  if ($sline eq "comment_line") {
    my $text = <SCRIPT>;
    chomp($text);

    replaceLine($text, "//$text");

    $slineNum += 1;
    next;
  }

  fail("unknown directive");
}

close(SCRIPT);


# emit the modified input
print(@lines);
exit(0);


# ---------------------- subroutines --------------------------
sub fail {
  my ($msg) = @_;
  print STDERR ("$script:$slineNum: $msg\n");
  close(SCRIPT);
  exit(4);
}


sub replaceLine {
  my ($src, $dest) = @_;

  my $found = 0;
  for (my $i=0; $i < @lines; $i++) {
    my $tmp = $lines[$i];
    chomp($tmp);
    if ($tmp eq $src) {
      $lines[$i] = $dest . "\n";
      $found++;
    }
  }

  if ($found == 0) {
    fail("replaceLine: didn't find any matching lines");
  }
  if ($found > 1) {
    fail("replaceLine: found $found matches (only want to find one)");
  }
}


# Find a function definition and bracket it with two given lines.
# Either line may be empty, indicating no insertion to be done.
# Heuristics for finding the start and end are a little fragile,
# but are sufficient for my present needs.
sub surroundFunction {
  my ($name, $startLine, $endLine) = @_;

  my $i=0;
  my @destLines = ();

  # find the function header
  while ($i < @lines && !atFuncHeader($i, $name)) {
    push @destLines, ($lines[$i++]);
  }
  if ($i == @lines) {
    fail("surroundFunction $name: did not find function header");
  }

  # insert begin-comment
  if ($startLine) {
    push @destLines, ($startLine . "\n");
  }

  # and function header as before
  push @destLines, ($lines[$i++]);

  # the next line should have an open-brace, with some indentation
  my ($braceInd) = ($lines[$i] =~ m/^(\s*)\{/);
  if (!defined($braceInd)) {
    # should not happen now that 'atFuncHeader' checks for this too
    fail("surroundFunction $name: did not find opening brace");
  }
  push @destLines, ($lines[$i++]);

  # look for the close brace
  while ($i < @lines &&
         $lines[$i] !~ m/^$braceInd\}/) {
    push @destLines, ($lines[$i++]);
  }
  if ($i == @lines) {
    fail("surroundFunction $name: did not find closing brace");
  }

  # emit the closing brace
  push @destLines, ($lines[$i++]);

  # close the comment
  if ($endLine) {
    push @destLines, ($endLine . "\n");
  }

  # copy rest of file
  while ($i < @lines && !atFuncHeader($i, $name)) {
    push @destLines, ($lines[$i++]);
  }
  if ($i != @lines) {
    fail("surroundFunction $name: found second function header on line $i");
  }

  # put it all back into 'lines'
  @lines = @destLines;
}


sub atFuncHeader {
  my ($i, $name) = @_;
  my $line = $lines[$i];

  # if name is followed by '(', say yes.
  # don't allow line to begin with '#'.
  if ($line =~ m/^[^\#]*$name\s*\(/) {
    # check next line for open-brace
    if ($lines[$i+1] =~ m/^\s*\{/) {
      return 1;
    }
  }

  return 0;
}


# EOF
