// compchain.cc
// code for compchain.h

#include "compchain.h"       // this module
#include "vml_build.h"       // p_and, etc.
#include "exc.h"             // throw_XFatal


CompareChain::Elt::~Elt()
{
  if (exp) {
    delete exp;
  }
}


CompareChain::CompareChain(P_compare *_first)
  : first(_first),
    links()
{}

CompareChain::~CompareChain()
{
  if (first) {
    delete first;
  }
}


static ComparisonOp normalizedOp(ComparisonOp op)
{
  if (op == CO_LESSEQ) {
    return CO_LESS;
  }
  else if (op == CO_GREATEREQ) {
    return CO_GREATER;
  }       
  else {
    return op;
  }
}


Pred * owner CompareChain::convertToPredicate()
{
  // kind of comparison chain, from among:
  //   CO_EQUAL:     ==
  //   CO_NOTEQUAL:  ==, !=
  //   CO_LESS:      <, <=, ==
  //   CO_GREATER:   >, >=, ==
  ComparisonOp kind = normalizedOp(first->op);

  // need this just after I nullify 'first'...
  VExpr *first_e1 = first->e1;

  // expression most recently seen; this must be cloned before
  // being used
  VExpr *prevExp = first->e2;

  // predicate to return
  Pred *ret = first;
  first = NULL;        // ownership transfer
  if (links.isEmpty()) {
    return ret;
  }

  // for CO_NOTEQUAL chains, this list contains one representative
  // for each of the equality equivalence classes; each of these
  // expressions must be cloned before use
  SObjList<VExpr> classes;
  if (kind == CO_NOTEQUAL) {
    classes.prepend(first_e1);
    classes.prepend(prevExp);
  }

  FOREACH_ASTLIST_NC(Elt, links, iter) {
    Elt *elt = iter.data();

    // check for compatibility with previous operator
    if (elt->op == CO_EQUAL) {
      // ok
    }
    else if (kind == CO_EQUAL) {
      kind = normalizedOp(elt->op);
      if (kind == CO_NOTEQUAL) {
        // transitioning from == to != chain, need to grab the
        // first equivalence class' representative
        classes.prepend(prevExp);
      }
    }
    else if (kind == normalizedOp(elt->op)) {
      // ok
    }
    else {
      throw_XFatal(stringc << toString(elt->loc)
                           << ": incompatible comparisons in chain: "
                           << toString(kind) << " and "
                           << toString(elt->op));
    }

    // continue building 'ret'
    if (elt->op != CO_NOTEQUAL) {
      P_compare *p = new P_compare(elt->loc, prevExp->clone(), elt->op, elt->exp);
      elt->exp = NULL;      // ownership transfer
      prevExp = p->e2;

      ret = p_and(ret, p);
    }

    else {
      // the disequality asserts that 'elt->exp' is distinct from all
      // of the previously seen equivalence classes
      VExpr *eltExp = elt->exp;     // this is a serf pointer
      SFOREACH_OBJLIST(VExpr, classes, cIter) {
        VExpr const *c = cIter.data();
        
        // get the RHS exp
        VExpr *rhsExp;
        if (elt->exp) {
          rhsExp = elt->exp;
          elt->exp = NULL;    // ownership transfer
        }
        else {
          rhsExp = eltExp->clone();
        }

        // build the !=
        P_compare *p = new P_compare(elt->loc, c->clone(), CO_NOTEQUAL, rhsExp);

        // conjoin
        ret = p_and(ret, p);
      }

      prevExp = eltExp;
      classes.prepend(prevExp);
    }
  }

  return ret;
}    


// EOF
