(* nreach.v *)
(* Nelson's list reachability axioms *)

Section Nelson_Reach.

(* type of memory addresses, and values stored in memory *)
Variable Addr : Set.

(* type of memories *)
Definition Mem := Addr -> Addr.

(* the function that updates memories *)
(* syntax: (upd mem addr value) *)
Variable upd : Mem -> Addr -> Addr -> Mem.

(*                                    { value          if addr1=addr2  *)
(* ((upd oldMem addr1 value) addr2) = {                                *)
(*                                    { (oldMem addr2) o.w.            *)
Axiom read_same : (m:Mem)(addr,value:Addr)
                  ((upd m addr value) addr) = value.
Axiom read_diff : (m:Mem)(addr1,addr2,value:Addr)
                  ~(addr1=addr2) -> ((upd m addr1 value) addr2)=(m addr2).

(* test the memory-read axioms *)
Goal (m1,m2,m3:Mem)(a1,a2,a3:Addr)
  ~(a1=a2) ->
  ~(a2=a3) ->
  ~(a1=a3) ->
  m2 = (upd m1 a1 a2) ->        (* a1->next = a2 *)
  m3 = (upd m2 a2 a3) ->        (* a2->next = a3 *)
  (m3 a1) = a2.
Intros.
Rewrite H3.
Rewrite H2.
Cut ((upd (upd m1 a1 a2) a2 a3) a1)=((upd m1 a1 a2) a1).
Intro.
Rewrite H4.
Apply read_same.
Apply read_diff.
Apply sym_not_eq.
Assumption.
Save some_random_test.
      
(* read the value written one step back (read_same is zero steps) *)
Lemma read_step1 : (m:Mem)(a1,v1,a2,v2:Addr)
  ~(a1=a2) ->
  ((upd (upd m a1 v1) a2 v2) a1) = v1.
Intros.
Cut ((upd (upd m a1 v1) a2 v2) a1) = ((upd m a1 v1) a1).
Intros.
Rewrite H0.
Apply read_same.
Apply read_diff.
Apply sym_not_eq.
Assumption.
Save.

(* read the value written two steps back *)
Lemma read_step2 : (m:Mem)(a1,v1,a2,v2,a3,v3:Addr)
  ~(a1=a2) ->
  ~(a1=a3) ->
  ((upd (upd (upd m a1 v1) a2 v2) a3 v3) a1) = v1.
Intros.
Cut ((upd (upd (upd m a1 v1) a2 v2) a3 v3) a1) = ((upd (upd m a1 v1) a2 v2) a1).
Intros.
Rewrite H1.
Apply read_step1.
Assumption.
Apply read_diff.
Apply sym_not_eq.
Assumption.
Save.

(* repeated application of a function *)
(* (repeat n f x) means f^n(x) *)
Fixpoint repeat [n:nat] : (Addr->Addr) -> Addr -> Addr :=
  [f:Addr->Addr][init:Addr]
  Cases n of
    O => init
  | (S p) => (f (repeat p f init))
end.

(* f^n(f(x)) = f(f^n(x)) *)
Lemma repeat_commute : (f:Addr->Addr)(n:nat)(x:Addr)
  (repeat n f (f x)) = (f (repeat n f x)).
Intros.
Induction n.
Unfold repeat.
Reflexivity.
Unfold repeat.
Fold repeat.
Rewrite Hrecn.
Reflexivity.
Save.

(* f^n(f(x)) = f^(n+1)(x) *)
Lemma repeat_plus_one : (f:Addr->Addr)(n:nat)(x:Addr)
  (repeat n f (f x)) = (repeat (S n) f x).
Intros.
Unfold 2 repeat.
Fold repeat.
Apply repeat_commute.
Save.

(* f(f^n(x)) = f^(n+1)(x) *)
Lemma repeat_one_plus : (f:Addr->Addr)(n:nat)(x:Addr)
  (f (repeat n f x)) = (repeat (S n) f x).
Intros.
Unfold 2 repeat.
Fold repeat.
Reflexivity.
Save.

(* apply an equality, don't worry about a=b or b=a *)
Tactic Definition ApplyEq rule :=             
  Solve [ Apply rule | Symmetry; Apply rule ].


(* f^m(f^n(x)) = f^(n+m)(x) *)
Lemma repeat_plus : (f:Addr->Addr)(n,m:nat)(x:Addr)
  (repeat m f (repeat n f x)) = (repeat (plus n m) f x).
Intros.
Induction m.

(* m=0 *)
Assert H : (plus n O)=n.
Symmetry.
Apply plus_n_O.
Rewrite H.
Unfold 1 repeat.
Reflexivity.

(* m>0 *)
Replace (repeat (S m) f (repeat n f x)) with (f (repeat m f (repeat n f x))).
2:ApplyEq repeat_one_plus.
Replace (plus n (S m)) with (S (plus n m)).
2:ApplyEq plus_n_Sm.
Replace (repeat (S (plus n m)) f x) with (f (repeat (plus n m) f x)).
2:ApplyEq repeat_one_plus.
Rewrite Hrecm.
Reflexivity.
Save.



(* no indirections *)
Theorem read_repeat_zero : (m:Mem)(a:Addr)
  (repeat O m a)=a.
Intros.
Unfold repeat.
Reflexivity.
Save.

(* follow one indirection *)
Theorem read_repeat_one : (m:Mem)(a,v:Addr)
  (repeat (S O) (upd m a v) a)=v.
Intros.
Unfold repeat.
Apply read_same.
Save.

(* follow two indirections *)
Theorem read_repeat_two : (m:Mem)(a1,a2,v:Addr)
  ~(a1=a2) ->
  (repeat (S (S O)) (upd (upd m a1 a2) a2 v) a1) = v.
Intros.
Unfold repeat.
Cut ((upd (upd m a1 a2) a2 v) a1) = a2.
Intros.
Rewrite H0.
Apply read_same.
Apply read_step1.
Assumption.
Save.

(* follow three indirections *)
Theorem read_repeat_three : (m:Mem)(a1,a2,a3,v:Addr)                 
  ~(a1=a2) ->
  ~(a1=a3) ->
  ~(a2=a3) ->
  (repeat (S (S (S O))) (upd (upd (upd m a1 a2) a2 a3) a3 v) a1) = v.
Intros.  
Unfold repeat.
Cut ((upd (upd (upd m a1 a2) a2 a3) a3 v) a1) = a2.
Intro.
Rewrite H2.
Cut ((upd (upd (upd m a1 a2) a2 a3) a3 v) a2) = a3.
Intro.
Rewrite H3.
Apply read_same.
Apply read_step1.
Assumption.
Apply read_step2.
Assumption.
Assumption.
Save.


(* need reasoning about naturals for what follows,
 * particularly le and lt *)
Require Arith.

(* also need excluded middle so I can do case analysis:
 * m<n, m=n, or m>n *)
Require Classical.

(* experimental axiom: substitutability of propositions *)
Axiom subst_prop : (P,Q:Prop)
  (P<->Q) -> (P==Q).


(* ---------- exists_min ------------ *)
(* property stating that a given predicate has a minimal element
 * that satisfies the predicate *)
Definition exists_min := [P:nat->Prop]
  (EX n:nat | (P n) /\ (m:nat)(lt m n)->~(P m)).

(* strong-inductive lemma to help prove 'smallest' theorem *)
Lemma strong_smallest : (P:nat->Prop) (n,k:nat)
  (le k n) ->       (* this accomplishes strong induction *)
  (P k) ->          (* if we have a witness to P ... *)
  (exists_min P).   (* ... then there's a minimal element satisfying P *)
Intro.
Intro.              (* don't introduce k before inducting! *)
Induction n.

(* case: n=0 *)
Intro k.
Intro k_O.
Intro Pk.
Cut k=O.
2:Symmetry.
2:Apply le_n_O_eq.
Clear k_O.
Intro k_O.
Unfold exists_min.
Exists k.
Split.
Assumption.
Intro m.
Intro Hmk.
Absurd (lt m k).
Rewrite k_O.
Apply lt_n_O.
Assumption.
Assumption.

(* case: n>0 *)
Intro k.
Intro Hk.
Intro Pk.
Cut k=(S n) \/ ~(k=(S n)).
2:Apply classic.
Intro Hchoice.
Elim Hchoice.
Clear Hchoice.
Intro Heq.
Cut ((i:nat)(lt i k)->~(P i)) \/ ~((i:nat)(lt i k)->~(P i)).
2:Apply classic.
Intro Hchoice.
Elim Hchoice.
Clear Hchoice.
Intro Hnone.
Unfold exists_min.
Exists k.
Split.
Assumption.
Assumption.
Clear Hchoice.
Intro H_one.

(* ----- lemma: not_all_P_not_Q ----- *)
Lemma not_all_P_not_Q : (U:Set)(P,Q:U->Prop)
  (~(x:U)((P x)->~(Q x))) <-> (EX x | (P x) /\ (Q x)).
Intros.
Split.

(*  (~(x:U)((P x)->~(Q x))) -> (EX x | (P x) /\ (Q x))  *)
Intro.
(*
  unify
    (EX n:U | (P n))
  with
    (EX x:U | (P x) /\ (Q x))
*)
Apply not_all_not_ex with P := ([x:U](P x) /\ (Q x)).
Cut ((x:U)(P x)->~(Q x)) == ((n:U)~((P n)/\(Q n))).
Intro.
Rewrite <- H0.
Assumption.
Apply subst_prop.
Split.
Intros.
EAuto.     (* finally worked *)
Intro.
Intro.
Apply or_to_imply.
Apply not_and_or.
Auto.

(*  (EX x | (P x) /\ (Q x)) -> (~(x:U)((P x)->~(Q x)))   *)
Intro.
(*
  unify
    ~(n:U)(P n)
  with
    ~((x:U)(P x)->~(Q x))
*)
Apply ex_not_not_all with P := ([x:U](P x)->~(Q x)).
Elim H.
Intros.
Exists x.
Tauto.
Save.
(* ----- end lemma: not_all_P_not_Q ----- *)


Cut (~((i:nat)(lt i k)->~(P i))) == (EX x | (lt x k) /\ (P x)).
2:Apply subst_prop.
2:Apply not_all_P_not_Q with
  P := ([x:nat](lt x k))
  Q := ([x:nat](P x)).
Intro H_equiv.
Rewrite H_equiv in H_one.
Elim H_one.
Intro x.
Intro H_and.
Apply Hrecn with k := x.
Elim H_and.
Intro Hxk.
Intro Px.
Rewrite Heq in Hxk.
Apply lt_n_Sm_le.
Assumption.
Elim H_and.
Intros.
Assumption.


(* ----- le_defn ----- *)
Lemma le_defn : (m,n:nat)
  (le m n) <-> ((lt m n) \/ (m=n)).
Split.
Intro.
Apply le_lt_or_eq.
Assumption.
Intro.
Elim H.
Intro.
Apply lt_le_weak.
Assumption.
Intro.
Rewrite H0.

Lemma le_refl : (n:nat) (le n n).
Intro.
Induction n.
Apply le_O_n.
Apply le_n_S.
Assumption.
Save.

Apply le_refl.
Save.
(* ----- le_defn ----- *)

(* ----- le_but_not_eq ----- *)
Lemma le_but_not_eq : (m,n:nat)
  (le m n)->(~(m=n))->(lt m n).
Intros.
Cut (le m n) == ((lt m n) \/ (m=n)).
2:Apply subst_prop.
2:Apply le_defn.
Intro.
Rewrite H1 in H.
Elim H.
Trivial.
Intro.
Absurd m=n.
Assumption.
Assumption.
Save.
(* ----- le_but_not_eq ----- *)


Intro H_neq.
Cut (le k n).
Intro Hkn.
Apply Hrecn with k := k.
Assumption.
Assumption.
Apply lt_n_Sm_le.
Apply le_but_not_eq.
Assumption.
Assumption.
Save.


(* finally, what I wanted to prove: in essence, that every non-empty
 * subset of naturals has a least element *)
Theorem smallest : (P:nat->Prop)
  (EX n:nat | (P n)) -> (exists_min P).
Intros.
Elim H.
Intros.
Apply strong_smallest with
  n := x
  k := x.
Apply le_n.
Assumption.
Save.

(* need a version that has 'exists_min' expanded, and I don't see
 * how to get Coq to do that on the fly.. *)
(* actually, it will do it on the fly automatically.. *)
Lemma smallest2 : (P:nat->Prop)
  (EX n:nat | (P n)) ->
  (EX n:nat | (P n) /\ (m:nat)(lt m n)->~(P m)).
Apply smallest.
Save.


(* now, use that theorem to prove a fact related to reachability *)
Theorem smallest_example : (f:Mem)(p,q:Addr)
  (EX i:nat | (repeat i f p)=q) ->
  (EX i:nat | (repeat i f p)=q /\ (j:nat)(lt j i)->~((repeat j f p)=q)).
Intros.
Apply smallest with
  P := [i:nat](repeat i f p)=q.
Assumption.
Save.


(* simple reachability: can reach p from q *)
Definition canReach := [f:Mem][p:Addr][q:Addr]
  (EX i:nat | (repeat i f p) = q).

(* reachability w/o deref'ing an address *)
Definition canReachWO := [f:Mem][p:Addr][q:Addr][x:Addr]
  (EX i:nat | (repeat i f p)=q /\
              (j:nat) (lt j i)->~((repeat j f p)=x)).


Theorem canReach_equiv : (f:Mem)(p,q:Addr)
  (canReach f p q) <-> (canReachWO f p q q).
Intros.
Split.

(*  (canReach f p q) -> (canReachWO f p q q)  *)
Intro.
Unfold canReach in H.
Unfold canReachWO.
Elim H.
Intros.
Apply smallest_example.
Assumption.

(*  (canReachWO f p q q)->(canReach f p q)  *)
Intro.
Unfold canReachWO in H.
Unfold canReach.
Elim H.
Intros.
Exists x.
Elim H0.
Auto.
Save.


(* annoying.. *)
Theorem canReach_equiv1 : (f:Mem)(p,q:Addr)
  (canReach f p q) -> (canReachWO f p q q).
Intros f p q.
Assert (canReach f p q) <-> (canReachWO f p q q).
Apply canReach_equiv.
Tauto.
Save.

(* maybe more useful *)
Theorem canReach_equal : (f:Mem)(p,q:Addr) 
  (canReach f p q) == (canReachWO f p q q).
Intros f p q.
Apply subst_prop.
Apply canReach_equiv.
Save.


(* -------- A1 -------- *)
Theorem A1 : (f:Mem)(u,v,x:Addr)
  (canReachWO f u v x) <->
  (u=v \/ (~(u=x) /\ (canReachWO f (f u) v x))).
Intros.
Split.

(* ---  ->  --- *)
Intro.
Unfold canReachWO.
Unfold canReachWO in H.
Elim H.
Clear H.
Intro n.        (* Nelson uses n here *)
Intro.
Induction n.    (* actually just want case split, n=0 or n>0 *)

(* n=0 *)
Left.
Elim H.
Clear H.
Intros.
Unfold repeat in H.
Assumption.

(* n>0 *)
Clear Hrecn.      (* don't actually need recursive assumption *)
Right.
Split.
Elim H; Clear H; Intros.
Cut ~(repeat O f u)=x.
2:Apply H0.
2:Apply lt_O_Sn.
Unfold repeat.
Trivial.
Exists n.
Rewrite repeat_plus_one.
Elim H; Clear H; Intros.
Split.
Assumption.
Intro j.
Intro Hjn.
Cut ~(repeat (S j) f u)=x.
2:Apply H0.
2:Apply lt_n_S.
2:Assumption.
Rewrite repeat_plus_one.
Trivial.


(* ---  <-  --- *)
Intro H_or.
Elim H_or.

(* u=v *)
Clear H_or.
Intro Huv.
Unfold canReachWO.
Exists O.
Unfold 1 repeat.
Split.
Assumption.
Intro j.
Intro.
Absurd (lt j O).
Apply lt_n_O.
Assumption.

(* u!=v *)
Clear H_or.
Intro.
Unfold canReachWO.
Elim H.
Clear H.
Intros.
Unfold canReachWO in H0.
Elim H0.
Clear H0.
Intro m.        (* Nelson's uses m here *)
Intro.
Exists (S m).
Split.
Elim H0.
Clear H0.
Intros.
Rewrite <- repeat_plus_one.
Assumption.
Elim H0.
Clear H0.
Intros.
Induction j.    (* case split, j=0 or j>0 *)
Unfold repeat.
Assumption.
Rewrite <- repeat_plus_one.
Apply H1.
Apply lt_S_n.
Assumption.
Save.

  
(* given a canReachWO as an assumption, instantiate the
 * variable saying how many times f is applied, and unfold
 * the definition *)
Tactic Definition ElimCanReachWO H n :=
  Unfold canReachWO in H;
  Elim H;
  Clear H;
  Intros n H.

(* given a conjunctive hypothesis, break it into two hypotheses *)
Tactic Definition BreakConj H H1 H2 :=
  Elim H;
  Clear H;
  Intros H1 H2.

(* do both of the above *)
(* why doesn't this work?? *)
Tactic Definition ElimBreakCanReachWO H1 H2 n :=
  ElimCanReachWO H1 n;
  BreakConj H1 H1 H2.


Lemma or_sym : (P,Q:Prop)
  (P\/Q) == (Q\/P).
Intros.
Apply subst_prop.
Tauto.
Save.


Lemma case_split_lt_Sn : (j,n:nat)
  (lt j (S n)) ->
  j=n \/ (lt j n).
Intros.
Assert H1 : (le j n).
Apply lt_n_Sm_le.
Assumption.
Replace j=n\/(lt j n) with (lt j n)\/j=n.
2:Apply or_sym.
Replace (lt j n)\/j=n with (le j n).
2:Apply subst_prop.
2:Apply le_defn.
Assumption.
Save.



Lemma case_split_lt_sum : (j,n,m:nat)
  (lt j (plus n m)) ->
  (lt j n) \/ (EX i:nat | (lt i m) /\ (j=(plus n i))).
Intros.
Induction m.
Left.
Replace n with (plus n O).
2:ApplyEq plus_n_O.
Assumption.
Assert (plus n (S m)) = (S (plus n m)).
ApplyEq plus_n_Sm.
Rewrite H0 in H.
Clear H0.
Assert j=(plus n m) \/ (lt j (plus n m)).
Apply case_split_lt_Sn.
Assumption.
Elim H0.
Clear H0.
Intro.
Right.
Exists m.
Split.
Apply lt_n_Sn.
Assumption.
Clear H0.
Intro.
Assert (lt j n)\/(EX i:nat|(lt i m)/\j=(plus n i)).
Apply Hrecm.
Assumption.
Clear Hrecm.
Elim H1.
Intro.
Left.
Assumption.
Clear H1.
Intro.
Right.
Elim H1.
Clear H1.
Intros.
Exists x.
BreakConj H1 H1a H1b.
Split.
Apply lt_S.
Assumption.
Assumption.
Save.


Lemma A1_equal : (f:Mem)(u,v,x:Addr)
  (canReachWO f u v x) ==
  (u=v \/ (~(u=x) /\ (canReachWO f (f u) v x))).
Intros.
Apply subst_prop.
Apply A1.
Save.


(* -------- A2 -------- *)
Theorem A2 : (f:Mem)(u,v,w,x:Addr)
  (canReachWO f u v x) ->
  (canReachWO f v w x) ->
  (canReachWO f u w x).
Intros f u v w x.
Intros Ruv Rvw.
ElimCanReachWO Ruv n.
ElimCanReachWO Rvw m.
BreakConj Ruv Ruv1 Ruv2.
Rewrite <- Ruv1 in Rvw.
BreakConj Rvw Rvw1 Rvw2.
Unfold canReachWO.
Exists (plus n m).
Split.
Rewrite <- Rvw1.
ApplyEq repeat_plus.
Intros.
Assert (lt j n) \/ (EX i:nat | (lt i m) /\ (j=(plus n i))).
Apply case_split_lt_sum.
Assumption.
Elim H0.
Clear H0.
Intro.
Apply Ruv2.
Assumption.
Clear H0.
Intro.
Elim H0.
Clear H0.
Intro i.
Intro Hi.
BreakConj Hi Hi1 Hi2.
Rewrite Hi2.
Replace (repeat (plus n i) f u) with (repeat i f (repeat n f u)).
2:ApplyEq repeat_plus.
Apply Rvw2.
Assumption.
Save.



(* -------- A3 -------- *)
Theorem A3 : (f:Mem)(u,v,x:Addr)
  (canReachWO f u v x) ->
  (canReach f u v).
Intros.
Unfold canReachWO in H.
Elim H.
Clear H.
Intro n.
Intros.
BreakConj H H Hdummy.
Clear Hdummy.
Assert H1 : (EX n:nat | (repeat n f u)=v).
Exists n.
Assumption.
Clear H.
Clear n.
Assert H2 : (EX n:nat | (repeat n f u)=v /\ (m:nat)(lt m n)->~(repeat m f u)=v).
Apply smallest with P := [n:nat](repeat n f u)=v.
Assumption.
Clear H1.
Assert H3 : (canReachWO f u v v).
Unfold canReachWO.
Assumption.
Replace (canReach f u v) with (canReachWO f u v v).
Assumption.
Symmetry.
Apply subst_prop.
Apply canReach_equiv.
Save.


(* ---- le_total ---- *)
(* le is a total order *)
Theorem le_total : (n,m:nat)
  (le n m) \/ (le m n).
Intros.
Induction m.

(* m=0 *)
Right.
Apply le_O_n.

(* m>0 *)
Elim Hrecm.

(* n<=m *)
Intro.
Left.
Apply le_S.
Assumption.

(* m<=n *)
Intro.
Assert Hle : (le m n) == ((lt m n) \/ (m=n)).
Apply subst_prop.
Apply le_defn.
Rewrite Hle in H.
Clear Hle.
Elim H.

(* m<n *)
Intro.
Right.
Apply lt_le_S.
Assumption.

(* m=n *)
Intro.
Left.
Rewrite H0.
Apply le_n_Sn.
Save.


(* we can restrict the range of quantification *)
Lemma smaller_range : (P:nat->Prop)(m,n:nat)
  ((j:nat)(lt j m)->(P j)) ->
  (le n m) ->
  ((j:nat)(lt j n)->(P j)).
Intros P m n.
Induction m.

(* m=0 *)
Intros.
Assert H2 : n=O.
Symmetry.
Apply le_n_O_eq.
Assumption.
Rewrite H2 in H1.
Absurd (lt j O).
Apply lt_n_O.
Assumption.

(* m>0 *)
Intros.
Assert H2 : (le n (S m)) == ((lt n (S m)) \/ (n=(S m))).
Apply subst_prop.
Apply le_defn.
Rewrite H2 in H0.
Clear H2.
Elim H0.

(* n<m+1 *)
Clear H0.
Intro.
Apply H.
Apply lt_trans with m := n.
Assumption.
Assumption.

(* n=m+1 *)
Clear H0.
Intro.
Rewrite <- H0 in H.
Apply H.
Assumption.
Save.


(* replace an existential hypothesis with a variable naming
 * the thing that exists *)
Tactic Definition ElimExist H x :=
  Elim H;
  Clear H;
  Intro x;
  Intro.


(* reachability w/o going through two possibly-distinct nodes *)
Definition canReachWO2 := [f:Mem][p,q,x,y:Addr]
  (canReachWO f p q x) /\ (canReachWO f p q y).

Lemma canReachWO2_implies : (f:Mem)(p,q,x,y:Addr)
  (canReachWO2 f p q x y) ->
  (EX n:nat | (repeat n f p)=q /\
              ((j:nat)(lt j n)->~(repeat j f p)=x) /\
              ((j:nat)(lt j n)->~(repeat j f p)=y)).
Intros.
Unfold canReachWO2 in H.
Unfold canReachWO in H.
BreakConj H Ha Hb.
ElimExist Ha a.
ElimExist Hb b.
BreakConj H a_eq a_neq.
BreakConj H0 b_eq b_neq.
Elim (le_total a b).     (* case analysis on a,b and le relation *)

(* a <= b *)
Intro.
Exists a.
Split.
Assumption.
Split.
Assumption.
Apply smaller_range with
  P := [j:nat]~(repeat j f p)=y
  m := b
  n := a.
Assumption.
Assumption.

(* b <= a *)
Intro.
Exists b.
Split.
Assumption.
Split.
Apply smaller_range with
  P := [j:nat]~(repeat j f p)=x
  m := a
  n := b.
Assumption.
Assumption.
Assumption.
Save.


(* if a property does not hold for all naturals up to n-1,
 * but does hold for some natural j, then j >= n *)
Lemma all_not_but_one : (P:nat->Prop)(n,j:nat)
  ((i:nat)(lt i n)->~(P i)) ->
  (P j) ->
  (le n j).
Intros P n j.
Induction n.

(* n=0 *)
Intros.
Apply le_O_n.

(* n>0 *)
Intro H_all.
Intro Hj.
Assert Hle : (le n j).
Apply Hrecn.
Intro i.
Intro.
Apply H_all.
Apply lt_S.
Assumption.
Assumption.
Assert Hlt : (lt n j).
Assert H1 : (le n j) == ((lt n j) \/ n=j).
Apply subst_prop.
Apply le_defn.
Rewrite H1 in Hle.
Clear H1.
Elim Hle.
Trivial.
Intro.
Absurd (P j).
Apply H_all.
Rewrite H.
Apply lt_n_Sn.
Assumption.
Apply lt_le_S.
Assumption.
Save.


(* replace proposition in a hypothesis *)
(* doesn't work if 'dest' contains "\/"??? *)
Tactic Definition ReplaceProp src dest rule H :=
  Assert Hequal : src == dest;
  Apply subst_prop;
  Apply rule;
  Rewrite Hequal in H;
  Clear Hequal.

(* -------- A4 -------- *)
Theorem A4 : (f:Mem)(u,x,y,z:Addr)
  (canReachWO f u x y) ->
  (canReachWO f u y z) ->
  (canReachWO f u x z).
Intros f u x y z.
Intro xBy.        (* x before y *)
Intro yBz.        (* y before z *)
ElimCanReachWO xBy n.
ElimCanReachWO yBz m.
BreakConj xBy xBy1 xBy2.
BreakConj yBz yBz1 yBz2.
Assert Hnm : (le n m).
Apply all_not_but_one with
  P := [n:nat](repeat n f u)=y.
Assumption.
Assumption.
Unfold canReachWO.
Exists n.
Split.
Assumption.
Apply smaller_range with
  P := [j:nat]~(repeat j f u)=z
  m := m.
Assumption.
Assumption.
Save.


(* I hate that I have to spell this out for Coq.. *)
Lemma not_all_P_not_Q_1 : (U:Set)(P,Q:U->Prop)
  (~(x:U)((P x)->~(Q x))) -> (EX x | (P x) /\ (Q x)).
Intros U P Q.
Assert (~(x:U)((P x)->~(Q x))) <-> (EX x | (P x) /\ (Q x)).
Apply not_all_P_not_Q.
Tauto.
Save.


(* -------- A5 -------- *)
Theorem A5 : (f:Mem)(u,x,y:Addr)
  ((canReach f u x) \/ (canReach f u y)) ->
  ((canReachWO f u x y) \/ (canReachWO f u y x)).

(* to exploit symmetry between x and y *)
Lemma A5_half : (f:Mem)(u,x,y:Addr)
  (canReach f u x) ->
  ((canReachWO f u x y) \/ (canReachWO f u y x)).
Intros f u x y.
Intro Hx.
Assert Heq : (canReach f u x) == (canReachWO f u x x).
Apply subst_prop.
Apply canReach_equiv.
Rewrite Heq in Hx.
Clear Heq.

ElimCanReachWO Hx n.
BreakConj Hx Hx1 Hx2.
Assert Hcases : ((j:nat)(lt j n)->~(repeat j f u)=y) \/
                ~((j:nat)(lt j n)->~(repeat j f u)=y).
Apply classic.
Elim Hcases.

(* y does not occur *)
Clear Hcases.
Intro H_all.
Left.
Unfold canReachWO.
Exists n.
Auto.

(* y does occur *)
Clear Hcases.
Intro H_notAll.
Assert H_ex : (EX j:nat | (lt j n) /\ (repeat j f u)=y).

Apply not_all_P_not_Q_1 with
  P := [j:nat](lt j n)
  Q := [j:nat](repeat j f u)=y.
Assumption.

ElimExist H_ex m.
BreakConj H mLTn Hy.
Clear H_notAll.
Right.
Unfold canReachWO.
Exists m.
Split.
Assumption.
Assert Hweak : (le m n).
Apply lt_le_weak.
Assumption.
Apply smaller_range with
  P := [j:nat]~(repeat j f u)=x
  m := n.
Assumption.
Assumption.
Save.

(* now use the symmetric lemma *)
Intros f u x y.
Intro Hcases.
Elim Hcases.
Intro.
Apply A5_half.
Assumption.
Intro.

Lemma sym_or : (P,Q:Prop)
  (P\/Q) -> (Q\/P).
Intros.
Tauto.
Save.

Apply sym_or.     (* reverse roles of x and y *)
Apply A5_half.
Assumption.
Save.



(* -------- L1 -------- *)
(* the other transitivity-like property *)
Theorem L1 : (f:Mem)(u,v,w:Addr)
  (canReach f u v) ->
  (canReach f v w) ->
  (canReach f u w).
Intros f u v w.
Intros Fuv Fvw.
Apply NNPP.
Unfold 1 not.
Intro.
Absurd (canReach f u w).
Assumption.

Lemma and_contrapositive : (P,Q,R:Prop)
  ((P/\Q)->R) -> (Q /\ ~R) -> ~P.
Tauto.
Save.

Assert Step1 : ~(canReachWO f u v w).
Apply and_contrapositive with
  P := (canReachWO f u v w)
  Q := (canReachWO f v w w)
  R := (canReachWO f u w w).
Intro.
BreakConj H0 H1 H2.
Apply A2 with v := v.
Assumption.
Assumption.
Split.

Apply canReach_equiv1.
Assumption.
Assert (canReach f u w) == (canReachWO f u w w).
Apply subst_prop.
Apply canReach_equiv.
Rewrite <- H0.
Assumption.

Assert Step2 : (canReachWO f u w v).

Lemma or_not : (P,Q:Prop)
  (P \/ Q) -> ~P -> Q.
Tauto.
Save.

Apply or_not with
  P := (canReachWO f u v w)
  Q := (canReachWO f u w v).
Apply A5.
Left.
Assumption.
Assumption.
Apply A3 with x := v.
Assumption.
Save.


(* -------- L2 -------- *)
Theorem L2 : (f:Mem)(u,v,w:Addr)
  (canReach f u v) ->
  ~(canReach f u w) ->
  (canReachWO f u v w).
Intros f u v w.
Intros Fuv nFuw.

Assert Step1 : (canReachWO f u v w) \/ (canReachWO f u w v).
Apply A5.
Left.
Assumption.

Assert Step2 : ((canReachWO f u v w) /\ ~(canReach f u w)) \/
               ((canReachWO f u w v) /\ ~(canReach f u w)).

Lemma distr_and_or : (P,Q,R:Prop)         
  P -> (Q \/ R) -> ((P /\ Q) \/ (P /\ R)).
Tauto.
Save.

Lemma distr_and_or2 : (P,Q,R:Prop)     (* doh *)
  P -> (Q \/ R) -> ((Q /\ P) \/ (R /\ P)).
Tauto.
Save.

Apply distr_and_or2.
Assumption.
Assumption.
Elim Step2.
Clear Step2.
Intro.
Elim H.
Trivial.
Clear Step2.
Intros.
BreakConj H Hyes Hno.
Absurd (canReach f u w).
Assumption.
Apply A3 with x := v.
Assumption.
Save.


(* if n <= m, then m = n+k for some k:nat *)
Lemma le_sum : (n,m:nat)                
  (le n m) -> (EX k:nat | m=(plus n k)).
Intros n m.
Induction m.

(* m=0 *)
Intro.
Assert H1 : n=O.
Symmetry.
Apply le_n_O_eq.
Assumption.
Rewrite H1.
Exists O.
Auto.

(* m>0 *)
Intro.
Elim le_defn with
  m := n
  n := (S m).
Intros.
Clear H1.
Assert H1 : (lt n (S m))\/n=(S m).
Apply H0.
Assumption.
Clear H.
Clear H0.
Elim H1.

(* n < m+1 *)
Clear H1.
Intro.
Assert H1 : (EX k:nat|m=(plus n k)).
Apply Hrecm.
Apply lt_n_Sm_le.
Assumption.
ElimExist H1 k0.
Rewrite H0.
Exists (S k0).
Auto.

(* n=m+1 *)
Clear H1.
Intro.
Rewrite H.
Exists O.
Auto.
Save.


(* -------- A6 -------- *)
Theorem A6 : (f:Mem)(u,x,y,z:Addr)
  (canReachWO f u x y) ->
  (canReachWO f u y z) ->
  (canReachWO f x y z).
Intros f u x y z.
Intros Hxy Hyz.
ElimCanReachWO Hxy n.
ElimCanReachWO Hyz m.
BreakConj Hxy Hxy1 Hxy2.
BreakConj Hyz Hyz1 Hyz2.

Assert nLEm : (le n m).
Apply all_not_but_one with
  P := [n:nat](repeat n f u)=y.
Assumption.
Assumption.

(* need to break m down as m = n + k; then Nelson's "m-n" is k *)
Assert k : (EX k:nat | m=(plus n k)).
Apply le_sum.
Assumption.
ElimExist k k.

Assert Step1 : (j:nat)(lt j k)->~(repeat j f (repeat n f u))=z.
Intro.
Rewrite (repeat_plus f n j u).
Intro.
Apply Hyz2.
Rewrite H.
Apply lt_reg_l.
Assumption.
Clear Hyz2.    (* the paper mistakenly drops Step1, instead of Hyz2 *)

Assert Step2 : (repeat k f (repeat n f u)) = y.
Rewrite (repeat_plus f n k u).
Rewrite <- H.
Assumption.
Clear Hyz1.
Clear Hxy2.

Assert Step3 : (repeat k f x) = y /\ 
               (j:nat)(lt j k)->~(repeat j f x)=z.
Rewrite Hxy1 in Step1.
Rewrite Hxy1 in Step2.
Auto.
Clear Step1.
Clear Step2.
Clear Hxy1.
BreakConj Step3 Step3a Step3b.

Unfold canReachWO.
Exists k.
Auto.
Save.




(* -------- L3 -------- *)
Theorem L3 : (f:Mem)(u,v,w:Addr)
  (canReachWO f u v w) ->
  (canReachWO f u w v) ->
  v = w.
Intros f u v w.
Intros Huvw Huwv.

Assert H1 : (canReachWO f v w v).
Apply A6 with
  u := u.
Assumption.
Assumption.
Clear Huvw.
Clear Huwv.

Elim (A1 f v w v).
Intros.
Clear H0.
Elim H.
Trivial.
Intro.
BreakConj H0 Ha Hb.
Absurd v=v.
Assumption.
Reflexivity.
Assumption.
Save.


(* -------- A7 -------- *)
Theorem A7_half : (f:Mem)(u,v:Addr)
  (canReach f (f u) v) ->
  (canReachWO f (f u) v u).
Intros f u v.
Intro H.
Assert H1 : (canReachWO f (f u) v v).
Apply canReach_equiv1.
Assumption.
Clear H.

(* Step 1 *)
Unfold canReachWO in H1.
ElimExist H1 n.
BreakConj H H1 H2.

Assert Step2a : (repeat (S n) f u) = v.
Rewrite <- (repeat_plus_one f n u).
Assumption.
Assert Step2b : (m:nat)(lt m n)->~(repeat (S m) f u)=(repeat (S n) f u).
Rewrite Step2a.
Intro.
Rewrite <- (repeat_plus_one f m u).
Generalize m.
Assumption.
Clear H1.
Clear H2.

Assert Step3 : (m:nat)(lt m n)->                                
  ~(repeat (S m) f u)=(repeat (S m) f (repeat (minus n m) f u)).
Intros.
Assert H2 : ~(repeat (S m) f u)=(repeat (S n) f u).
Apply Step2b.
Assumption.
Rewrite (repeat_plus f (minus n m) (S m) u).
Rewrite <- (plus_n_Sm (minus n m) m).
Rewrite (plus_sym (minus n m) m).
Assert H3 : (plus m (minus n m)) = n.
Apply le_plus_minus_r.
Apply lt_le_weak.
Assumption.
Rewrite H3.
Assumption.
Clear Step2b.

Assert Step4: (m:nat)(lt m n)->~u=(repeat (minus n m) f u).
Intros.
Assert H2 : ~(repeat (S m) f u)=(repeat (S m) f (repeat (minus n m) f u)).
Apply Step3.
Assumption.
Apply NNPP.
Unfold 1 not.
Intro.
Assert Heq : u=(repeat (minus n m) f u).
Apply NNPP.
Assumption.
Absurd (repeat (S m) f u)=(repeat (S m) f (repeat (minus n m) f u)).
Assumption.
Rewrite <- Heq.
Reflexivity.
Clear Step3.

Assert Step5a : (repeat n f (f u)) = v.
Rewrite -> (repeat_plus_one f n u).
Assumption.
Clear Step2a.
Assert Step5b : (j:nat)(lt j n)->~(repeat j f (f u))=u.
Intros.
Rewrite -> (repeat_plus_one f j u).
Apply sym_not_eq.
Assert H1 : (le (S j) n).
Apply lt_le_S.
Assumption.
Assert k : (EX k:nat | n=(plus (S j) k)).
Apply le_sum.
Assumption.
ElimExist k k.
Assert Heq : (S j) = (minus n k).
Apply plus_minus.
Rewrite (plus_sym k (S j)).
Assumption.
Rewrite Heq.
Apply Step4.
Rewrite H0.
Rewrite (plus_Sn_m j k).
Auto with arith.         (* interesting *)
Clear Step4.

Unfold canReachWO.
Exists n.
Auto.
Save.


Theorem A7_otherHalf : (f:Mem)(u,v:Addr)
  (canReachWO f (f u) v u) ->
  (canReach f (f u) v).
Intros f u v.
Intro.
Apply A3 with
  x := u.
Assumption.
Save.


Theorem A7 : (f:Mem)(u,v:Addr)
  (canReach f (f u) v) <->
  (canReachWO f (f u) v u).
Intros f u v.
Split.
Intro.
Apply A7_half.
Assumption.
Intro.
Apply A7_otherHalf.
Assumption.
Save.



(* -------- L4 -------- *)
Theorem L4 : (f:Mem)(u,x:Addr)
  (canReach f u x) ->
  (canReachWO f (f u) u x) ->
  u = x.
Intros f u x.
Intros H1 H2.
Apply NNPP.
Unfold 1 not.
Intro.

Assert Step1 : (canReach f (f u) x).
Rewrite (canReach_equal f (f u) x).
Rewrite (canReach_equal f u x) in H1.
Rewrite (A1_equal f u x x) in H1.
Elim H1.
Intro.
Absurd u=x.
Assumption.
Assumption.
Intro.
BreakConj H0 Ha Hb.
Assumption.
Clear H1.

Assert Step2 : (canReachWO f (f u) x u).
Apply A7_half.
Assumption.
Clear Step1.

Absurd u=x.
Assumption.
Apply L3 with
  f := f
  u := (f u).
Assumption.
Assumption.
Save.



(* -------- L5 -------- *)
(* NOTE: this proof does not appear in the paper; 
 * "left as exercise" :) 
 *)
Theorem L5 : (f:Mem)(u,x:Addr)
  (canReach f u x) ->
  (canReach f (f u) u) ->
  (canReach f x u).
Intros f u x.
Intros H1 H2.

Assert Step1 : (canReach f (f u) x).         (* L1 *)
Apply L1 with
  v := u.
Assumption.
Assumption.

Assert Step2 : (canReachWO f (f u) x u).     (* A7 *)
Apply A7_half.
Assumption.

Assert Step3 : (canReachWO f x u u).         (* A6 *)
Apply A6 with
  u := (f u).
Assumption.
Rewrite (canReach_equal f (f u) u) in H2.
Assumption.

Rewrite (canReach_equal f x u).
Assumption.
Save.


(* -------- Lemma 0 of A8 -------- *)
Lemma A8_0 : (f:Mem)(n:nat)(p,q,u:Addr)
  ((j:nat)(lt j n)->~(repeat j (upd f p q) u)=p) ->
  (repeat n (upd f p q) u) = (repeat n f u).
Intros f n p q u.
Induction n.

(* n=0 *)
Intros.
Unfold repeat.
Reflexivity.

(* n>0 *)
Intro H1.                                                (* "(1)" *)
Assert H2 : (repeat n (upd f p q) u)=(repeat n f u).     (* "(2)" *)
Apply Hrecn.
Intros.
Apply smaller_range with
  P := [j:nat]~(repeat j (upd f p q) u)=p
  n := n
  m := (S n).
Assumption.
Apply le_n_Sn.
Assumption.
Clear Hrecn.

Rewrite <- (repeat_one_plus (upd f p q) n u).

Assert H_notp : ~(repeat n (upd f p q) u)=p.
Apply H1.
Apply lt_n_Sn.

Assert H_skip : (upd f p q (repeat n (upd f p q) u)) = (f (repeat n (upd f p q) u)).
Apply read_diff.
Apply sym_not_eq.
Assumption.
Rewrite H_skip.
Clear H_skip.

Rewrite H2.

Rewrite (repeat_one_plus f n u).
Reflexivity.
Save.


Lemma A8_0_corr : (f:Mem)(n:nat)(p,q,u,x:Addr)
  ((j:nat)(lt j n)->~(repeat j (upd f p q) u)=p) ->
  ( (j:nat)(lt j n)->~(repeat j (upd f p q) u)=x ) <->
  ( (j:nat)(lt j n)->~(repeat j f u)=x ).
Intros.
Split.

(* -> *)
Intros.
Rewrite <- (A8_0 f j p q u).
Apply H0.
Assumption.
Apply smaller_range with
  P := [j:nat]~(repeat j (upd f p q) u)=p
  m := n
  n := j.
Assumption.
Apply lt_le_weak.
Assumption.

(* <- *)
Intros.
Rewrite -> (A8_0 f j p q u).
Apply H0.
Assumption.
Apply smaller_range with
  P := [j:nat]~(repeat j (upd f p q) u)=p
  m := n
  n := j.
Assumption.
Apply lt_le_weak.
Assumption.

Save.


(* -------- Lemma 1 of A8 -------- *)
(* proof largely copied from Lemma 0 *)
Lemma A8_1 : (f:Mem)(n:nat)(p,q,u:Addr)
  ((j:nat)(lt j n)->~(repeat j f u)=p) ->
  (repeat n (upd f p q) u) = (repeat n f u).
Intros f n p q u.
Induction n.

(* n=0 *)
Intros.
Unfold repeat.
Reflexivity.

(* n>0 *)
Intro H1.                                                (* "(1)" *)
Assert H2 : (repeat n (upd f p q) u)=(repeat n f u).     (* "(2)" *)
Apply Hrecn.
Intros.
Apply smaller_range with
  P := [j:nat]~(repeat j f u)=p
  n := n
  m := (S n).
Assumption.
Apply le_n_Sn.
Assumption.
Clear Hrecn.

Rewrite <- (repeat_one_plus (upd f p q) n u).

Assert H_notp : ~(repeat n f u)=p.
Apply H1.
Apply lt_n_Sn.

Rewrite H2.

Assert H_skip : (upd f p q (repeat n f u)) = (f (repeat n f u)).
Apply read_diff.
Apply sym_not_eq.
Assumption.
Rewrite H_skip.
Clear H_skip.

Rewrite (repeat_one_plus f n u).
Reflexivity.
Save.


(* proof largely copied from Lemma 0's corollary *)
Lemma A8_1_corr : (f:Mem)(n:nat)(p,q,u,x:Addr)
  ((j:nat)(lt j n)->~(repeat j f u)=p) ->
  ( (j:nat)(lt j n)->~(repeat j (upd f p q) u)=x ) <->
  ( (j:nat)(lt j n)->~(repeat j f u)=x ).
Intros.
Split.

(* -> *)
Intros.
Rewrite <- (A8_1 f j p q u).
Apply H0.
Assumption.
Apply smaller_range with
  P := [j:nat]~(repeat j f u)=p
  m := n
  n := j.
Assumption.
Apply lt_le_weak.
Assumption.

(* <- *)
Intros.
Rewrite -> (A8_1 f j p q u).
Apply H0.
Assumption.
Apply smaller_range with
  P := [j:nat]~(repeat j f u)=p
  m := n
  n := j.
Assumption.
Apply lt_le_weak.
Assumption.

Save.



(* -------- A8 -------- *)
Theorem A8 : (f:Mem)(u,v,x,p,q:Addr)
  (canReachWO2 (upd f p q) u v x p) <->
  (canReachWO2 f           u v x p).
Intros.
Split.

(* -> *)
Intro Step0.

Assert Step1 : (EX n:nat | (repeat n (upd f p q) u) = v /\
                           ((j:nat)(lt j n)->~(repeat j (upd f p q) u)=x) /\
                           ((j:nat)(lt j n)->~(repeat j (upd f p q) u)=p)).
Apply canReachWO2_implies.
Assumption.
Clear Step0.
ElimExist Step1 n.
BreakConj H H1 H2.
BreakConj H2 H2 H3.

Assert Step2a : (repeat n f u)=v.
Rewrite <- (A8_0 f n p q u).
Assumption.
Assumption.

Assert Step2b : (j:nat)(lt j n)->~(repeat j f u)=x.
Elim (A8_0_corr f n p q u x).
Intros.
Clear H0.
Apply H.
Assumption.
Assumption.
Assumption.

Assert Step2c : (j:nat)(lt j n)->~(repeat j f u)=p.
Elim (A8_0_corr f n p q u p).    (* use p for x *)
Intros.
Clear H0.
Apply H.
Assumption.
Assumption.
Assumption.

Clear H1.
Clear H2.
Clear H3.

Unfold canReachWO2.
Unfold canReachWO.
Split.
Exists n.
Auto.
Exists n.
Auto.

(* <- *)
(* this section works backwards from Step3 to Step0, since it's the
 * reverse-implication of the previous case *)
Intro Step3.

Assert Step2 : (EX n:nat | (repeat n f u) = v /\
                           ((j:nat)(lt j n)->~(repeat j f u)=x) /\
                           ((j:nat)(lt j n)->~(repeat j f u)=p)).
Apply canReachWO2_implies.
Assumption.
Clear Step3.
ElimExist Step2 n.
BreakConj H Step2a Step2b.
BreakConj Step2b Step2b Step2c.

Assert Step1a : (repeat n (upd f p q) u) = v.
Rewrite -> (A8_1 f n p q u).
Assumption.
Assumption.

Assert Step1b : ((j:nat)(lt j n)->~(repeat j (upd f p q) u)=x).
Elim (A8_1_corr f n p q u x).
Intros.
Clear H.
Apply H0.
Assumption.
Assumption.
Assumption.

Assert Step1c : ((j:nat)(lt j n)->~(repeat j (upd f p q) u)=p).
Elim (A8_1_corr f n p q u p).    (* use p for x *)
Intros.
Clear H.
Apply H0.
Assumption.
Assumption.
Assumption.

Clear Step2a Step2b Step2c.

Unfold canReachWO2.
Unfold canReachWO.
Split.
Exists n.
Auto.
Exists n.
Auto.

Save.     (* whew! *)










(* stuff for emacs coq-mode


Undo.
Undo.
Undo.
Undo.
Undo.
Undo.
Undo.
Abort.
Restart.

*)


End Nelson_Reach.

(* EOF *)
