// doublylinked.c
// doubly-linked list, modeled after pale/Examples/doublylinked.pale

// --------- verifier preamble ----------
#include "selupd.model"     // sel/upd
#include "state.model"      // metaVar

int fresh_value();


// --------------- code to verify -----------------
struct Node;

typedef struct Head {
  struct Node *first;   // NULL or valid
  struct Node *last;    // NULL or last element in list
} Head;

typedef struct Node {
  int value;
  struct Node *next;    // invariant: NULL or valid
  struct Node *prev;    // thing whose 'next' points at me
  struct Head *head;    // ghost: pointer to my list's header object
} Node;

// invariant: NULL or valid
Node *garbage = NULL;


// this disables the effect of the pattern specs
#if 0
  #define PATS(...) /*nothing*/
  #define MPAT(...) /*nothing*/
#endif // 0


#define VALID_NODE(n) \
  (0 < (n) < objct && tag[n]==Node_tag)
  
#define VALID_HEAD(h) \
  (0 < (h) < objct && tag[h]==Head_tag)


// global invariants
int global_inv() { return vc_expression(all_of {
  tag[NULL]==0;                         // zero tag at NULL, i.e. not an object

  // garbage
  garbage!=NULL ==> all_of {
    VALID_NODE(garbage);
    garbage->prev==NULL;
    garbage->head==NULL;
  };


  // Head

  #if 0    // not needed
  // 'first' is valid if not NULL
  forall(Head *h) PATS(h->first) .
    tag[h]==Head_tag && h->first!=NULL ==> tag[h->first] == Node_tag;
  #endif // 0

  // list labels are accurate; 'first' is injective
  forall(Head *h) PATS(h->first) .
    tag[h]==Head_tag && h->first!=NULL ==> h->first->head == h;

  // 'first' points to the first element (if not NULL)
  forall(Head *h) PATS(h->first->prev) .
    tag[h]==Head_tag && h->first!=NULL ==> h->first->prev == NULL;

  // similar for 'last'
  #if 0    // not needed
  forall(Head *h) PATS(h->last) .
    tag[h]==Head_tag && h->last!=NULL ==> tag[h->last] == Node_tag;
  #endif // 0
  forall(Head *h) PATS(h->last) .
    tag[h]==Head_tag && h->last!=NULL ==> h->last->head == h;
  forall(Head *h) PATS(h->last->next) .
    tag[h]==Head_tag && h->last!=NULL ==> h->last->next == NULL;

  // should have 'first' iff have 'next'; actually, I cannot ensure this
  // because my invariants about 'first' and 'last' are too weak
  //forall(Head *h).
  //  tag[h]==Head_tag ==> (h->first==NULL <==> h->last==NULL);

  // Node

  #if 0    // not needed
  // declared type of 'next' is accurate
  forall(Node *n) PATS(n->next) .
    tag[n]==Node_tag && n->next!=NULL ==> tag[n->next] == Node_tag;
  // declared type of 'prev' is accurate
  forall(Node *n) PATS(n->prev) .
    tag[n]==Node_tag && n->prev!=NULL ==> tag[n->prev] == Node_tag;
  #endif // 0

  // transitivity of list label
  forall(Node *n) PATS(MPAT(n->next, n->head), n->next->head) .
    tag[n]==Node_tag && n->next!=NULL ==> n->next->head == n->head;

  // injectivity of 'next'
  forall(Node *n) PATS(n->next) .
    tag[n]==Node_tag && n->next!=NULL ==> n->next->prev == n;

  // injectivity of 'prev' (seems to be needed...)
  forall(Node *n) PATS(n->prev) .
    tag[n]==Node_tag && n->prev!=NULL ==> n->prev->next == n;
});}


// ----------------------- delete -----------------------
void delete(Head *list, Node *p)
  pre(all_of {
        VALID_HEAD(list);         // implied by 'data' keyowrd
        VALID_NODE(p);            // p!=null
        p->head == list;          // list.first<next*>p
        /* ? */                   // list.last<prev*>p
        /* ? */                   // p<next*>list.last
        /* ? */                   // p<prev*>list.first
        /* ? */                   // list.first<next*>S
      })
  post(all_of {
         p->head == NULL;         // !L.first<next*>P
         forall(Node *q)          // allpos q of Node: q in S <=> (L.first<next*>q | q=P)
           PATS(pre(q->head), q->head).
             q!=p ==> pre(q->head) == q->head;
       })
  changes(Head, Node, garbage)
{
  if (p == list->first) {
    list->first = p->next;
  }
  if (p == list->last) {
    list->last = p->prev;
  }
  if (p->prev != NULL) {
    p->prev->next = p->next;
  }
  if (p->next != NULL) {
    p->next->prev = p->prev;
  }
  if (garbage != NULL) {
    garbage->prev = p;
  }
  p->next = garbage;
  p->prev = NULL;
  p->head = NULL;
  garbage = p;
}


Node *search(Head *list, int value)
  pre(VALID_HEAD(list))           // list!=null
  post(all_of {
    /* redundant! */              // L!=null
    result!=NULL ==> result->value == pre(value);     // approximates pale's postcondition...
  })
  changes()
{
  //int dummy; if (!dummy) { dummy++; }

  Node *t = list->first;
  while (t!=NULL && t->value != value) {

    invariant(all_of {
      tag[list] == Head_tag;      // list!=null
      pre(value) == value;        // V=value
      pre(list) == list;          // L=list
      t->head == list;            // list.first<next*>t
      // cannot state: allpos p of Node: p<next.next*>t => p.value!=value]
      
      // additional
      tag[t] == Node_tag;
    });

    t = t->next;
  }
  return t;
}


void insert(Head *list, Node *node)
  pre(all_of {
    VALID_HEAD(list);               // list!=null
    VALID_NODE(node);               // node!=null
    node->next == NULL;             // node.next=null
    node->prev == NULL;             // node.prev=null

    // additional
    node->head == NULL;
    node != garbage;
  })
  post(1==1)
  changes(Head, Node)
{
  if (list->first == NULL) {
    list->last = node;
  }
  else {
    list->first->prev = node;
  }
  node->head = list;
  node->next = list->first;
  list->first = node;
  node = NULL;
}


Head *concat(Head *list1, Head *list2)
  pre(all_of {
    VALID_HEAD(list1);       // list1!=null
    VALID_HEAD(list2);       // list2!=null
  })
  post(1==1)
  changes(Head, Node)
{
  if (list1->last != NULL) {
    list1->last->next = list2->first;
  }
  else {
    list1->first = list2->first;
  }
  if (list2->first != NULL) {
    list2->first->prev = list1->last;
    list1->last = list2->last;
  }
  list2->first = NULL;
  list2->last = NULL;
  
  // update all of 'head' in one shot (it is ghost)
  Node_head = fresh_value();
  assume(forall(Node *n) PATS(pre(n->head), n->head). all_of {
    pre(n->head)!=list2 ==> n->head == pre(n->head);
    pre(n->head)==list2 ==> n->head == list1;
  });
  
  return list1;
}






// EOF
