// redblack.c
// red-black trees, based on redblackinsert.pale (and related files)

// ------------------- preamble ----------------
#include "ptr.model"        // sel/upd/ptr
#include "state.model"      // metaVar

#if 0
 #define DLT(x,y) ((x)<(y))
 #define DLE(x,y) ((x)<=(y))
int DLT_between(int x, int y)
     pre(x < y)
     post(x < result && result < y)
     // We ought to say that the global invariant does not need to hold here.
{
  assume(x + 1 < y);
  return x + 1;
}
#else
 #include "dpo.model"        // DLT, DLT_between
#endif

// C functions that are modeled directly
void *malloc(int size);

// declarations of model functions
int fresh_value();


// experiment
//verifier_option((no_automatic_typing));


// ---------------- code to verify -------------
#define AND all_of
#define OR one_of


typedef int bool;
#define false 0
#define true 1


// node in tree
typedef struct Node {
  // primary neighbor pointers; mutually injective
  struct Node *left, *right;

  // inverse of left/right
  bool isLeft;           // (ghost)
  struct Node *parent;          // this one isn't ghost

  // color: false=red, true=black
  bool color;
  #define Red false
  #define Black true

  // level: an element of a continuous partial order; nodes
  // lower in the tree (further from the root) have a lower level
  int level;
} Node;

#define isNode(n) (0 < (n) < objct && tag[n]==Node_tag)
#define NODE_TYPE(field) ((field)!=NULL ==> isNode(field))
#define blackOrNull(n) ((n)!=NULL ==> (n)->color==Black)

// single global tree root
Node *root;

// a level that is greater than any in any tree
int top_level;

// also need one at the other end
int bottom_level;


// when a node is red, there are two implications:
//   - its parent (if any) must be black
//   - its children must be black
// during insertion, each of these consequences needs to be
// locally violated
Node *exception1;      // when red, this node need not have a black parent
Node *exception2;      // when red, this node need not have black childen

// pale's "almostinv1" is exception1
// pale's "almostinv2" is exception1 + exception2


// global invariant, codifying the field relationships
int global_inv() { return vc_expression(AND {
  0 < objct;
  tag[0] == 0;

  // types
  NODE_TYPE(root);
  NODE_TYPE(exception1);
  NODE_TYPE(exception2);

  // not really necessary..
  DLT(bottom_level, top_level);

  // invariant that redblackinsert.pale says is missing from "the book";
  // CLR allows red roots, but I'll follow what pale says; interestingly,
  // they only insist on a black root when there are no exception nodes
  exception1==NULL && exception2==NULL && root!=NULL ==>
    root->color==Black;

  // root is the root
  root!=NULL ==> root->parent==NULL;

  // declared types are accurate
  #if 0   // not needed
  forall(Node *n) PATS(n->left). tag[n]==Node_tag ==>
    NODE_TYPE(n->left);
  forall(Node *n) PATS(n->right). tag[n]==Node_tag ==>
    NODE_TYPE(n->right);
  forall(Node *n) PATS(n->parent). tag[n]==Node_tag ==>
    NODE_TYPE(n->parent);
  #endif // 0
  forall(Node *n) PATS(n->isLeft). tag[n]==Node_tag ==>
    n->isLeft==true || n->isLeft==false;
  forall(Node *n) PATS(n->color). tag[n]==Node_tag ==>
    n->color==Red || n->color==Black;

  // isLeft/parent is inverse of left/right
  forall(Node *n) PATS(n->left) . tag[n]==Node_tag ==>
    n->left!=NULL ==>  (n->left->isLeft == true &&
                        n->left->parent == n);

  forall(Node *n) PATS(n->right) . tag[n]==Node_tag ==>
    n->right!=NULL ==> (n->right->isLeft == false &&
                        n->right->parent == n);

  // assert the 'up' direction as well, to strengthen the
  // correspondence between (e.g.) n->isLeft and n->parent->left
  forall(Node *n) PATS(n->parent) . tag[n]==Node_tag ==>
    n->isLeft==true && n->parent!=NULL ==> n->parent->left == n;
  forall(Node *n) PATS(n->parent) . tag[n]==Node_tag ==>
    n->isLeft==false && n->parent!=NULL ==> n->parent->right == n;

  // red nodes have black children
  forall(Node *n) PATS(n->left) . tag[n]==Node_tag ==>
    (n!=exception2 && n->color==Red) ==>
      (n->left!=exception1 && n->left!=NULL) ==>
        n->left->color==Black;

  forall(Node *n) PATS(n->right) . tag[n]==Node_tag ==>
    (n!=exception2 && n->color==Red) ==>
      (n->right!=exception1 && n->right!=NULL) ==>
        n->right->color==Black;

  // variation: if I am the root then I am black, unless exception1
  // points at me (TODO: try removing the invariant above that
  // explicitly makes 'root' black, since this should imply that)
  forall(Node *n) PATS(n->parent). tag[n]==Node_tag ==>
    exception1!=n && n->parent==NULL ==> n->color==Black;

  // children are lower (TODO: let me use '<', and use the type
  // of 'level' to know to use DLT instead of '<' for Simplify)
  forall(Node *n) PATS(n->left) . tag[n]==Node_tag ==>
    n->left!=NULL ==> DLT(n->left->level, n->level);
  forall(Node *n) PATS(n->right) . tag[n]==Node_tag ==>
    n->right!=NULL ==> DLT(n->right->level, n->level);

  // all levels are between top and bottom; this is better than just
  // saying the root is lower than top, b/c the transitivity would
  // need induction to propagate the fact to the whole tree
  forall(Node *n) PATS(n->level). tag[n]==Node_tag ==>
    DLT(n->level, top_level);
  forall(Node *n) PATS(n->level). tag[n]==Node_tag ==>
    DLT(bottom_level, n->level);

});}


// just testing DLT, etc.
int foo(int x, int y)
  pre(DLT(x,y))
{
  assert(DLT(x, y));            // just the precondition

  assert(x != y);               // antireflexive
  assert(!DLT(y, x));           // antisymmetric

  assert(DLE(x, y));            // weakened form of precondition
  assert(DLE(x, x));            // nonstrict is reflexive

  int z = DLT_between(x,y);     // x < z < y
  assert(DLT(x, z));
  assert(DLT(z, y));
}


void leftrotate(Node *x)
  pre(AND {
    isNode(x);
    root != NULL;
    // TODO: pale claims rechability from root to x; I would handle
    // that with a 'header' pointer
    x->right != NULL;
    // almostinv1: x or x->right might have the wrong color
    exception1==x || exception1==x->right;
    exception2==NULL;
    x->color == Red;
    // binding pre-state names:
    //   B == pre(x->right->left)
    //   X == pre(x)
    //   Y == pre(x->right)
    //   P == pre(x->parent)
    //   Q == pre(root)
  })
  post(AND {
    // Y->left == X
    pre(x->right)->left == pre(x);
    // X->right == B
    pre(x)->right == pre(x->right->left);
    // P == Y->parent
    pre(x->parent) == pre(x->right)->parent;
    // TODO: X is reachable from (post) root
    // almostinv2: X or its children might have the wrong color
    exception1 == pre(x);
    exception2 == pre(x);
    // X->color == Red
    pre(x)->color == Red;
    // P==NULL ==> root==Y
    pre(x->parent)==NULL ==> root==pre(x->right);
    // P!=NULL ==> root==Q
    pre(x->parent)!=NULL ==> root==pre(root);
    
    // extra                 

    // x->left isn't modified
    pre(x)->left == pre(x->left);
    // x->right (Y) inherits x's left-ness
    pre(x->isLeft) == pre(x->right)->isLeft;
    // if x was the left child, then the parent's right child didn't change
    pre(x->isLeft)==true ==> pre(x->parent)->right == pre(x->parent->right);
    pre(x->isLeft)==false ==> pre(x->parent)->left == pre(x->parent->left);
    // only three nodes' parent links change
    pre(x->right->right->parent) == pre(x->right->right)->parent;
  })
  changes(Node_right, Node_left, Node_parent, Node_isLeft,
          Node_level,   // but *not* Node_color!
          root, exception1, exception2)          // extra
{
  Node *y;
  y = x->right;

  assert(x != y);    // got it, by using the partial order

  x->right = y->left;
  if (y->left != NULL) {
    y->left->parent = x;
    y->left->isLeft = false;                               // extra
  }
  y->parent = x->parent;
  y->isLeft = x->isLeft;                                   // extra
  if (x->parent == NULL) {
    y->level = DLT_between(x->level, top_level);           // extra
    root = y;
  }
  else {
    y->level = DLT_between(x->level, y->parent->level);    // extra
    if (/*x->isLeft == true*/ x == x->parent->left) {
      assert(x->isLeft == true);
      assert(x == x->parent->left);
      x->parent->left = y;
    }
    else {
      assert(x->isLeft != true);
      assert(x == x->parent->right);
      x->parent->right = y;
    }
  }
  y->left = x;                   //ERRORIFMISSING(3):
  x->parent = y;
  x->isLeft = true;                                        // extra

  // note exceptions
  exception1 = x;                                          // extra
  exception2 = x;                                          // extra
}


void rightrotate(Node *x)
  pre(AND {
    isNode(x);
    root != NULL;
    // TODO: x is reachable from root
    x->left != NULL;
    // almostinv1: x or x->left might have the wrong color
    exception1==x || exception1==x->left;
    exception2==NULL;
    x->color == Red;
    // binding pre-state names:
    //   B == pre(x->left->right)
    //   X == pre(x)
    //   Y == pre(x->left)
    //   P == pre(x->parent)
    //   Q == pre(root)
  })
  post(AND {
    // Y->right == X
    pre(x->left)->right == pre(x);
    // X->left == B
    pre(x)->left == pre(x->left->right);
    // P == Y->parent
    pre(x->parent) == pre(x->left)->parent;
    // TODO: X is reachable from (post) root
    // almostinv2: X or its children might have the wrong color
    exception1 == pre(x);
    exception2 == pre(x);
    // X->color == Red
    pre(x)->color == Red;
    // P==NULL ==> root==Y
    pre(x->parent)==NULL ==> root==pre(x->left);
    // P!=NULL ==> root==Q
    pre(x->parent)!=NULL ==> root==pre(root);
    
    // extra                 

    // x->left isn't modified
    pre(x)->right == pre(x->right);
    // x->left (Y) inherits x's right-ness
    pre(x->isLeft) == pre(x->left)->isLeft;
    // if x was the right child, then the parent's left child didn't change
    pre(x->isLeft)==false ==> pre(x->parent)->left == pre(x->parent->left);
    pre(x->isLeft)==true ==> pre(x->parent)->right == pre(x->parent->right);
    // only three nodes' parent links change
    pre(x->left->left->parent) == pre(x->left->left)->parent;
  })
  changes(Node_right, Node_left, Node_parent, Node_isLeft,
          Node_level,   // but *not* Node_color!
          root, exception1, exception2)          // extra
{
  Node *y;
  y = x->left;
  x->left = y->right;
  if (y->right != NULL) {
    y->right->parent = x;
    y->right->isLeft = true;                               // extra
  }
  y->parent = x->parent;
  y->isLeft = x->isLeft;                                   // extra
  if (x->parent == NULL) {
    y->level = DLT_between(x->level, top_level);           // extra
    root = y;
  }
  else {
    y->level = DLT_between(x->level, y->parent->level);    // extra
    if (x == x->parent->right) {
      x->parent->right = y;
    }
    else {
      x->parent->left = y;
    }
  }
  y->right = x;
  x->parent = y;                  //ERRORIFMISSING(2):
  x->isLeft = false;                                       // extra

  // note exceptions
  exception1 = x;                                          // extra
  exception2 = x;                                          // extra
}


void treeinsert(Node *z)
  pre(AND {
    isNode(z);
    z != root;                                             // extra
    z->left == NULL;
    z->right == NULL;
    z->parent == NULL;                                     // extra
    // binding pre-state names:
    //   Z == pre(z)
    //   Q == pre(root)
    // inv(root) is implied
    exception1 == NULL;
    exception2 == NULL;
  })
  post(AND {
    // TODO: Z is reachable from root
    // Z->left == NULL
    pre(z)->left == NULL;
    // Z->right == NULL
    pre(z)->right == NULL;
    // almostinv1: Z might have the wrong color
    exception1 == pre(z);
    exception2 == NULL;
    // root->color == Black || root == Z
    root->color == Black || root == pre(z);
    // Q!=NULL ==> root==Q
    pre(root)!=NULL ==> root==pre(root);
    // maybe this is supposed to be implied by root<(left+right)*>Z ?
    root!=NULL;
  })
  changes(Node, root, exception1)
{
  Node *y, *x;
  bool c;
  
  x = root;
  y = NULL;                  // missing from treeinsert.pale!
  while (x != NULL) {
    invariant {
      pre(z) == z;           // Z == z
      isNode(z);
      isNode(x);
      z != root;                                           // extra
      z->left == NULL;
      z->right == NULL;
      z->parent == NULL;                                   // extra
      x!=root ==> AND {
        // TODO: y is reachable from root
        isNode(y);
        (y->left==x && c!=false) || (y->right==x && c==false);
      };
      x==root ==> (y==NULL && x==root);
      exception1 == NULL;
      exception2 == NULL;
      pre(root) == root;     // Q == root
      root != NULL;
    };

    //assert(z != y);
    //assert(z != x);
    //assert(y != x);
    //assert(x!=root ==> x->parent!=NULL);

    y = x;
    c = fresh_value();       // nondeterministic choice
    if (c) {
      x = x->left;
    }
    else {
      x = x->right;
    }
    
    //ERROR(1):       if (fresh_value()) { break; }

    //assert(x!=NULL ==> x->parent!=NULL);
  }

  //assert(z != y);

  z->parent = y;
  if (y==NULL) {
    root = z;
    root->level = DLT_between(bottom_level, top_level);    // extra
  }
  else {
    if (c) {
      y->left = z;
      z->isLeft = true;                                    // extra
    }
    else {
      y->right = z;
      z->isLeft = false;                                   // extra
    }
    z->level = DLT_between(bottom_level, y->level);        // extra
  }
  
  exception1 = z;                                          // extra

  // the pale comment says "ensure disjointness"; I think this
  // means that they regard 'z' as being a root in the backbone
  // forest, and now that 'root' has acquired the node that 'z'
  // points to, 'z' itself needs to be nullified to maintain the
  // tree property (even though it's about to go out of scope)
  z = NULL;
}


void redblackinsert(Node *t)
  pre(AND {
    isNode(t);
    t != root;
    t->left == NULL;
    t->right == NULL;
    t->parent == NULL;                                     // extra
    // inv(root) is implied
    exception1 == NULL;
    exception2 == NULL;
  })
  post(AND {
    // inv(root) is implied
    exception1 == NULL;
    exception2 == NULL;
  })
  changes(Node, root, 
          exception1, exception2)  // not changed, but problems with meta...
{
  Node *x, *y;
  x = t;

  treeinsert(x);
  // [treeinsert.Z==x && treeinsert.Q=root]    // manually bind?

  x->color = Red;

  while (/*x!=root*/x->parent!=NULL && x->parent->color==Red) {
    exception1 = x;
    exception2 = NULL;
    invariant {
      // types
      isNode(x);

      // consequence of treeinsert
      root!=NULL;

      // loop guard
      x->parent!=NULL && x->parent->color==Red;

      // pale invariant
      x!=NULL;
      // TODO: x is reachable from root
      // almostinv1: x's color might be wrong
      exception1 == x;
      exception2 == NULL;
      root->color==Black || x==root;
      //x!=root && x->parent->color==Red ==>     // premise true since inside loop
      x->color==Red;
    };

    assert(x != NULL);                           // loop invariant
    assert(x->parent != NULL);                   // loop invariant
    assert(x->parent->parent != NULL);           // red nodes have parents 

    if (x->parent == x->parent->parent->left) {
      assert(x->parent->isLeft == true);
      y = x->parent->parent->right;
      if (y != NULL && y->color == Red) {
        x->parent->color = Black;
        y->color = Black;
        x->parent->parent->color = Red;
        x = x->parent->parent;
      }
      else {
        if (x == x->parent->right) {
          assert(x->isLeft == false);
          x = x->parent;

          // the precondition for 'leftrotate' says the exception can
          // be either one, but you have to figure out which it is
          // ahead of time, since saying the wrong thing opens up too
          // much uncertainty for the proof to go through
          exception1 = /*fresh_value()? x :*/ x->right;
          exception2 = NULL;
          leftrotate(x);
          // X == x
          // x is reachable from root
          // P->right is black
          // Y is red
          // X->left is black
          // B is black
          // Y->right is black
        }

        exception1 = x;
        exception2 = x;
        invariant {
          root!=NULL;
          isNode(x);
          isNode(x->parent);
          isNode(x->parent->parent);
          x->parent->left == x;
          x->parent->parent->left == x->parent;
          blackOrNull(x->parent->parent->right);
          // TODO: x is reachable from root
          // almostinv2: x or its children have the wrong color
          exception1 == x;
          exception2 == x;        // could also just be NULL
          blackOrNull(x->left);
          blackOrNull(x->right);
          blackOrNull(x->parent->right);
        };

        x->parent->color = Black;
        x->parent->parent->color = Red;
        
        // we've fixed all colors, except that xpp might have a red
        // parent (temporarily; about to fix that with a rotation)
        exception1 = x->parent->parent;

        exception2 = NULL;
        rightrotate(x->parent->parent);
        // Y->left == x
        // B is black
        // X->right is black
        // x is reachable from root
        // Y is black
        // X is red
        // Q == root
        // x != NULL

        root->color = Black;
      }
    }

    else {
      y = x->parent->parent->left;
      if (y!=NULL && y->color==Red) {
        x->parent->color = Black;
        y->color = Black;
        x->parent->parent->color = Red;
        x = x->parent->parent;
      }
      else {
        if (x == x->parent->left) {
          x = x->parent;

          exception1 = /*fresh_value()? x :*/ x->left;
          exception2 = NULL;
          rightrotate(x);
          // X == x
          // x is reachable from root
          // P->left is black
          // Y is red
          // X->right is black
          // B is black
          // Y->left is black
        }

        exception1 = x;
        exception2 = x;
        invariant {
          root!=NULL;
          isNode(x);
          isNode(x->parent);
          isNode(x->parent->parent);
          x->parent->right == x;
          x->parent->parent->right == x->parent;
          blackOrNull(x->parent->parent->left);
          // x is reachable from root
          // almostinv2: x or its children have the wrong color
          exception1 == x;
          exception2 == x;
          blackOrNull(x->right);
          blackOrNull(x->left);
          blackOrNull(x->parent->left);
        };

        x->parent->color = Black;
        x->parent->parent->color = Red;

        exception1 = x->parent->parent;
        exception2 = NULL;
        leftrotate(x->parent->parent);
        // Y->right == x
        // B is black
        // X->left is black
        // x is reachable from root
        // Y is black
        // X is red
        // Q == root
        // x != NULL
        
        root->color = Black;
      }
    }
  }
  
  root->color = Black;
  
  if (x->parent == NULL) {       // then x==root, but I don't know that
    x->color = Black;
  }

  exception1 = NULL;
  exception2 = NULL;
}


// EOF
