// fflow_main.cc            see license.txt for copyright and terms of use
// fact-flow vml->vml program

#include "vml_lexer.h"    // VMLLexer
#include "vml.gr.gen.h"   // VMLParse
#include "vml.ast.gen.h"  // VMLFile, etc.
#include "vml_env.h"      // VMLEnv
#include "vml_vcgen.h"    // VCEnv
#include "vml_tocnf.h"    // VML2CNF
#include "vml_fflow.h"    // vmlFFLowAnalyze
#include "glr.h"          // GLR
#include "trace.h"        // TRACE_ARGS
#include "srcloc.h"       // SourceLocManager
#include "exc.h"          // xbase
#include "strutil.h"      // replace

#include <iostream.h>     // cout


void doit(int argc, char **argv)
{
  char const *progname = argv[0];
  TRACE_ARGS()

  if (argc < 2) {
    cout << "usage: " << progname << " [options] [-o output.vml] input.vml\n"
         << "  options:\n"
         << "    -tr <flags>     turn on tracing (debugging) flags\n"
         ;
    return;
  }

  #define SHIFT argc--; argv++

  // non -tr arguments
  string outFname = "";

  while (argv[1][0] == '-') {
    if (0==strcmp(argv[1], "-o")) {
      if (!argv[2]) {
        xbase("-o requires an argument");
      }
      outFname = argv[2];
      SHIFT;
    }
    else {
      xbase(stringc << "unknown argument: " << argv[1]);
    }

    SHIFT;
  }

  SourceLocManager mgr;
  StringTable strTable;

  // ------------- lex/parse --------------
  VMLFile *file;
  {
    // lexer
    VMLLexer lexer(strTable, argv[1]);
    lexer.tokenFunc(&lexer);

    // parser
    VMLParse parse(strTable);
    GLR glr(&parse, parse.makeTables());

    // parse it
    SemanticValue result;
    if (!glr.glrParse(lexer, result) ||
        lexer.errors) {
      xbase("parse error");
    }

    file = (VMLFile*)result;
  }

  // -------------- tcheck ---------------
  {
    VMLEnv env(strTable);
    file->tcheck(env);

    if (env.numErrors()) {
      xbase("tcheck errors");
    }
  }

  // -------------- fflow ----------------
  {
    char const *selproc = getenv("SELPROC");

    // let "fflowd" imply "fflow"
    if (tracingSys("fflowd")) {
      traceAddSys("fflow");
    }

    FOREACH_ASTLIST_NC(VTopForm, file->forms, iter) {
      if (iter.data()->isVTF_proc()) {
        VProcedure *p = iter.data()->asVTF_proc()->p;

        if (selproc && 0!=strcmp(selproc, p->id->name)) {
          // skip this procedure
          return;
        }

        vmlFFlowAnalyze(p, strTable);
      }
    }
  }

  // write to file
  {
    if (outFname.empty()) {
      outFname = replace(argv[1], ".vml", ".ff.vml");
    }
    cout << "writing ff-VML file: " << outFname << "\n";

    ofstream out(outFname.c_str());
    if (!out) {
      perror(stringc << "open " << outFname);
      exit(2);
    }

    out << "// " << outFname << "\n"
        << "// ff-VML code generated from " << argv[1] << "\n"
        << "// automatically generated by " << progname << "\n"
        << "\n"
        ;
           
    string vmltext = file->print();
    out << vmltext;
  }
}


int main(int argc, char **argv)
{                                 
  // don't double-print error messages
  xBase::logExceptions = false;

  try {
    doit(argc, argv);
    return 0;
  }
  catch (xBase &x) {
    cout << x << endl;
    return 4;
  }
}


// EOF
