; pointer.thy
; theory of pointers and their relation to memories
; requires: memory.thy
    
(DEFPRED (isPointer ptr))

(BG_PUSH (AND

  ; (pointer region addr): construct a pointer to object at 'addr' in 'region'

  ; (pointer_region ptr): select region component
  (FORALL (region addr)
    (EQ
      (pointer_region (pointer region addr))
      region
    ))

  ; (pointer_addr ptr): select addr component
  (FORALL (region addr)
    (EQ
      (pointer_addr (pointer region addr))
      addr
    ))

  ; (isPointer ptr): ptr is a pair constructed with "pointer"; this is
  ; usually easy to prove, but useful nonetheless because it
  ; introduces the concepts to Simplify's semantic network
  (FORALL (ptr)         (PATS (isPointer ptr))
    (IFF
      (isPointer ptr)
      (EQ ptr (pointer (pointer_region ptr) (pointer_addr ptr)))
    ))

  ; (selptr mem ptr): select from 'mem' at 'ptr'
  ; pattern spec ensures one-directional triggering
  (FORALL (mem region addr)   (PATS (selptr mem (pointer region addr)))
    (EQ
      (selptr mem (pointer region addr))
      (sel (sel mem region) addr)
    ))

  ; (updptr mem ptr val): update 'mem' at 'ptr' to contain 'val'
  (FORALL (mem region addr val)   (PATS (updptr mem (pointer region addr) val))
    (EQ
      (updptr mem (pointer region addr) val)
      (upd mem region (upd (sel mem region) addr val))
    ))

  ; pointer distinctness
  (FORALL (r1 a1 r2 a2)
    (IMPLIES
      (NEQ r1 r2)
      (NEQ (pointer r1 a1) (pointer r2 a2))
    ))

  (FORALL (r1 a1 r2 a2)
    (IMPLIES       ; this used to be "IFF", see end of file
      (NEQ a1 a2)
      (NEQ (pointer r1 a1) (pointer r2 a2))
    ))

  (FORALL (r1 a1 r2 a2)    (PATS (MPAT (pointer r1 a1) (pointer r2 a2)))
    (IMPLIES
      (AND (EQ r1 r2) (EQ a1 a2))
      (EQ (pointer r1 a1) (pointer r2 a2))
    ))

  ; null pointer
  (EQ null 0)
          
  ; writes to null pointer are ignored
  (FORALL (mem val)
    (EQ
      (updptr mem null val)
      mem
    ))

  ; null is not a pointer
  (FORALL (region addr)
    (NEQ
      null
      (pointer region addr)
    ))

))


; The "IFF" in pointer-distinctness axiom:
;
; I originally had "IFF" because I wanted to make life easier for
; Simplify: if you have distinct pointers, just assume that means
; their addresses are different, so you don't have to do a case
; analysis, considering the possibility it's the region that makes
; them different.
;
; The problem is that
;   (IFF (NEQ a b) (NEQ c d))
; is of course equivalent to
;   (IFF (EQ a b) (EQ c d))
; so the axiom actually implied that the region component of pointers
; was irrelevant.  This led to subtle inconsistencies later on,
; eventually resulting in a proof that 1 == 2 (!).

