// btree.cc
// example of a btree


struct Interior;
struct Leaf;

// node in the btree
struct Node {
  // parent pointer, specifically identifying whether this is
  // the left child or the right child
  Interior *parent;      // ghost
  bool isLeft;           // ghost
  bool isLeaf;           // type of this node: true=Leaf, false=Interior

  invariant {
    tag[this]==Interior_tag || tag[this]==Leaf_tag;
  }
  
  Node() {
    parent = NULL;
    isLeft = false;      // doesn't matter until 'parent' set
  }
};

struct Interior : public Node {
  // binary tree links; a real B+-tree has an array of links,
  // interspersed with key values
  Node *left;
  int ikey;              // interior key; separates left and right subtrees
  Node *right;           // can be NULL

  // left/right breakdown:
  //   left: everything <= ikey
  //   right: everything > ikey

  invariant {
    isLeaf==false;

    0<left<objct && tag[left] <: Node_tag;
    left->parent==this && left->isLeft==true;

    right!=NULL ==> all_of {
      0<right<objct && tag[right] <: Node_tag;
      right->parent==this && right->isLeft==false;
    };
  }

  Interior(Node *L) {
    isLeaf=false;
    setLeft(L);
    right=NULL;
  }

  void setLeft(Node *L) {
    left = L;
    left->parent = this;
    left->isLeft = true;
  }

  void setRight(Node *R) {
    right = R;
    right->parent = this;
    right->isLeft = false;
  }
};

struct Leaf : public Node {
  // stored data; a real B+-tree leaf node contains an array of
  // key/data pairs
  int key;
  int data;

  // linked list, acyclic; made doubly-linked with ghost var 'prev'
  // (though some implementations actually materialize 'prev')
  Leaf *next;
  Leaf *prev;            // ghost

  invariant {
    isLeaf==true;

    next!=NULL ==> all_of {
      0<next<objct && tag[next]==Leaf_tag;
      next->prev == this;
    };

    prev!=NULL ==> all_of {
      0<prev<objct && tag[next]==Leaf_tag;
      prev->next = this;
    };
  }

  Leaf() { 
    isLeaf=true;
    next=NULL;
    prev=NULL;
  }
  
  void setPrev(Leaf *p) {
    prev = p;
    if (p) { p->next = this; }
  }
  
  void setNext(Leaf *n) {
    next = n;
    if (n) { n->prev = this; }
  }
};


// return the data associated with the given key
int find(Node *root, int key)
  pre(root==NULL || tag[root] <: Node_tag)
{
  if (!root) {
    return -1;                        // not found
  }

  if (root->isLeaf) {
    Leaf *leaf = (Leaf*)root;         // downcast
    if (leaf->key == key) {
      return leaf->data;              // found
    }
    else {
      return -1;                      // not found
    }
  }

  else {
    Interior *i = (Interior*)root;    // downcast
    if (key <= i->ikey) {
      return find(i->left, key);      // tail recursion
    }
    else {
      return find(i->right, key);     // tail recursion
    }
  }
}


// outer shell of insert routine
Node *insert(Node *root, int key, int data)
  pre(root==NULL || all_of {
        tag[root] <: Node_tag;
        parent[root]==NULL;      // really the root
      })
  changes(Node_tag, Interior_tag, Leaf_tag, object, tag)
{
  if (!root) {
    Leaf *leaf = new Leaf;
    leaf->key = key;
    leaf->data = data;
    leaf->next = NULL;
    leaf->prev = NULL;
    return leaf;
  }

  else {
    Node *rightHalf;
    Node *leftHalf = innerInsert(root, key, data, rightHalf);
    if (!leftHalf) {
      return root;            // no split, root unchanged
    }
    else {
      // make new root node
      Interior *intr = new Interior;
      intr->setLeft(leftHalf);
      intr->key = key;
      intr->setRight(rightHalf);
      return intr;
    }
  }
}


// insert a new key,data pair; if this returns non-NULL, then
// the tree was split, and 'key' and 'rightHalf' provide the
// other two values to make up the split triple: left/key/right
Node *innerInsert(Node *root, int &key, int data, Node *&rightHalf)
  pre(0<root<objct && tag[root] <: Node_tag)
  post(result==NULL || all_of {
         tag[result] <: Node_tag;
         tag[rightHalf] <: Node_tag;

         // two possible cases
         one_of {  
           all_of {
             result==root;
             rightHalf >= pre(objct);   // rightHalf fresh
             parent[rightHalf]==NULL;
           };
           all_of {
             rightHalf==root;
             result >= pre(objct);      // result fresh
             parent[result]==NULL;
           };
         };
         
         #if 0    // don't think I need this, though it's true
         // both returned nodes are either have their parent links
         // equal either to NULL, or to the original root->parent
         parent[result]==NULL || parent[result]==pre(parent)[root];
         parent[rightHalf]==NULL || parent[rightHalf]==pre(parent)[root];
         #endif // 0
       })
  changes(Node_tag, Interior_tag, Leaf_tag, object, tag)
{
  if (root->isLeaf) {
    Leaf *leaf = (Leaf*)root;

    // always split
    if (key <= leaf->key) {
      // new node
      Left *leftHalf = new Leaf;
      leftHalf->key = key;
      leftHalf->data = data;

      // splice it into the leaf list
      //   A  <->  leaf  <->  B
      leftHalf->setPrev(leaf->prev);     // A/leftHalf splice
      leftHalf->setNext(leaf);           // leftHalf/leaf splice
      //   A  <->  leftHalf  <->  leaf  <->  B

      // return them; 'key' remains unchanged
      rightHalf = leaf;
      return leftHalf;
    }
    else {
      // new node
      rightHalf = new Leaf;
      rightHalf->key = key;
      rightHalf->data = data;

      // splice it into the leaf list
      //   A  <->  leaf  <->  B
      rightHalf->setNext(leaf->next);    // rightHalf/B splice
      rightHalf->setPrev(leaf);          // leaf/rightHalf splice
      //   A  <->  leaf  <->  rightHalf  <->  B

      // return them
      key = leaf->key;        // split point key
      return leaf;
    }
  }

  else {
    Interior *intr = (Interior*)root;

    if (!intr->right) {
      // only one child
      Node *leftHalf = innerInsert(intr->left, key, data, rightHalf);
      if (!leftHalf) {
        return leftHalf;      // no split
      }
      else {
        // node below me split, but split stops here because I
        // have room to accomodate it
        intr->setLeft(leftHalf);
        intr->key = key;
        intr->setRight(rightHalf);
        return NULL;          // split does not propagate
      }
    }

    else {
      // two children, must pick one
      Node *chosen = (key <= intr->key? left : right);
      Node *leftHalf = innerInsert(chosen, key, data, rightHalf);
      if (!leftHalf) {
        return leftHalf;      // no split
      }
      else {
        // node below me split, I have to split also; I now have three
        // data pointers and two keys; strategy will be to yield the
        // existing node, plus a singleton node holding the extreme
        // half, as the split pair

        if (chosen==left) {
          // I have:
          //   <leftHalf, key, rightHalf, intr->key, intr->right>
          //   \       /   |   \                               /
          //    single   split  ------------ intr -------------

          assert(leftHalf != intr->right &&
                 rightHalf != intr->right);

          // make a singleton node for the extreme (left) half
          Interior *single = new Interior;
          single->setLeft(leftHalf);
          single->right = NULL;

          // update the existing node
          intr->setLeft(rightHalf);

          // yield them; 'key' is same
          rightHalf = intr;     // my split right half
          return single;        // my split left half
        }

        else {
          // I have:
          //  <intr->left, intr->key, leftHalf, key, rightHalf>
          //   \                             /   |   \       /
          //    ------------- intr ----------  split  single

          assert(leftHalf != intr->left &&
                 rightHalf != intr->left);

          // make a singleton node for the extreme (right) half
          Interior *single = new Interior;
          single->setLeft(rightHalf);
          single->right = NULL;

          // update the existing node
          intr->setRight(leftHalf);

          // yield them; 'key' is same
          rightHalf = single;   // my split right half
          return intr;          // my split left half
        }
      }
    }
  }
}





// EOF
