// btree2.c
// btree example, using child arrays instead of just two children


// ------------------- preamble ----------------
asm("verifier: include \"ptr.vc\";");
asm("verifier: include \"state.vc\";");

// C functions that are modeled directly
void *malloc(int size);
void memmove(void *dest, void const *src, int len);
void memcpy(void *dest, void const *src, int len);


// declarations of model functions
int sel(int mem, int p);
int fresh_value();

int tag;
int objct;
int objsize;

// ---------------- code to verify -------------
struct Interior;
struct Leaf;


#define AND all_of
#define OR one_of

#define PARAMBYREF __varattribute__((parambyref))
#define ARGBYREF __varattribute__((argbyref))


typedef int bool;
#define false 0
#define true 1

// node in the btree
typedef struct Node {
  // parent pointer, specifically identifying whether this is
  // the left child or the right child
  struct Interior *parent;      // (ghost) node that points at me as a child
  int parentIndex;              // (ghost) my index in parent->child[]
  bool isLeaf;                  // type of this node: true=Leaf, false=Interior
} Node;


typedef struct Interior {
  // fake inheritance
  Node *superclass;
  struct Interior *parent;
  int parentIndex;
  bool isLeaf;

  // B+-tree links: array of children, interspersed with key values
  #define MAXIKEYS 100   // makes no difference what the actual size is
  Node *child[MAXIKEYS+1];
  int ikey[MAXIKEYS];
  int numKeys;           // # of used entries in 'ikey'

  // indexing breakdown:
  //   child[0..i]: everything <= ikey[i]
  //   child[i+1..numKeys]: everything > ikey[i]

  // example, numKeys=5:
  //                +---+---+---+---+---+
  //       ikey     | 0 | 1 | 2 | 3 | 4 |
  //                +---+---+---+---+---+
  //              +---+---+---+---+---+---+
  //      child   | 0 | 1 | 2 | 3 | 4 | 5 |
  //              +---+---+---+---+---+---+

  #define newInterior(ret)                         \
  {                                                \
    (ret) = (Interior*)malloc(sizeof(Interior));   \
    (ret)->parent = NULL;                          \
    (ret)->isLeaf = false;                         \
    (ret)->numKeys = 0;                            \
    /* caller should set 'child[0]' */             \
  }

  #define setChild(ths, i, c)   \
  {                             \
    (ths)->child[i] = (c);      \
    (c)->parent = (ths);        \
    (c)->parentIndex = (i);     \
  }
} Interior;


typedef struct Leaf {
  // fake inheritance
  Node *superclass;
  Interior *parent;
  int parentIndex;
  bool isLeaf;

  // stored data pointers, as key/data pairs
  #define MAXLKEYS 50    // just to be different from MAXIKEYS
  int key[MAXLKEYS];
  int data[MAXLKEYS];
  int numKeys;

  // linked list, acyclic; made doubly-linked with ghost var 'prev'
  // (though some implementations actually materialize 'prev')
  struct Leaf *next;
  struct Leaf *prev;            // ghost

  #define newLeaf(ret)                    \
  {                                       \
    (ret) = (Leaf*)malloc(sizeof(Leaf));  \
    (ret)->parent = NULL;                 \
    (ret)->isLeaf = true;                 \
    (ret)->numKeys = 0;                   \
    (ret)->next = NULL;                   \
    (ret)->prev = NULL;                   \
  }

  #define setNext(ths, n)  \
  {                        \
    (ths)->next = (n);     \
    if (n) {               \
      (n)->prev = (ths);   \
    }                      \
  }

  // NOTE: I do not define 'setPrev', since doing so would imply
  // that 'prev' cannot be a ghost variable.
} Leaf;
  

#define hasNodeTag(p) (tag[p]==Interior_tag || tag[p]==Leaf_tag)

// global invariant
int global_inv() { return vc_expression(AND {
  0 < objct;
  tag[0] == 0;

  forall(Interior *i). tag[i]==Interior_tag ==> AND {
    i->isLeaf == false;

    0 <= i->numKeys <= MAXIKEYS;    // can have zero keys

    forall(int j). 0 <= j <= i->numKeys ==> AND {
      0 < i->child[j] < objct && hasNodeTag(i->child[j]);
      //i->child[j]->parent==i;
      Node_parent[i->child[j]]==i;
      //i->child[j]->parentIndex==j;
      Node_parentIndex[i->child[j]]==j;
    };
  };

  forall(Leaf *l). tag[l]==Leaf_tag ==> AND {
    l->isLeaf==true;

    0 < l->numKeys <= MAXLKEYS;     // at least one key

    l->next!=NULL ==> AND {
      0 < l->next < objct && tag[l->next]==Leaf_tag;
      l->next->prev == l;
    };

    l->prev!=NULL ==> AND {
      0 < l->prev < objct && tag[l->prev]==Leaf_tag;
      l->prev->next == l;
    };
  };
});}


// return the data associated with the given key
int find(Node *root, int key)
  pre(root==NULL || (0 < root < objct && hasNodeTag(root)))
{
  int k;

  if (!root) {
    return -1;                        // not found
  }

  if (root->isLeaf) {
    Leaf *leaf = (Leaf*)root;         // downcast
    for (k=0; k < leaf->numKeys; k++) {
      invariant {
        global_inv;
        0 <= k < leaf->numKeys;
        0 < leaf < objct && tag[leaf]==Leaf_tag;
      }

      if (leaf->key[k] == key) {
        return leaf->data[k];         // found
      }
    }
    return -1;                        // not found
  }

  else {
    Interior *i = (Interior*)root;    // downcast
    for (k=0; k < i->numKeys; k++) {
      invariant {
        global_inv;
        0 <= k < i->numKeys;
        0 < i < objct && tag[i]==Interior_tag;
      }

      if (key <= i->ikey[k]) {
        return find(i->child[k], key);
      }
    }
    // beyond last stored key
    return find(i->child[k], key);
  }
}


// insert key/data into 'intr'; 'iikey' defines the *left* edge of the
// subtree rooted at 'data'; if this causes a split, return 'iikey' and
// 'result' such that they form a left/key/right triple:
// intr / *iikey / result
Interior *interiorInsert(Interior *intr, int *iikey PARAMBYREF, Node *data)
  pre(AND {
    0 < intr < objct && tag[intr]==Interior_tag;
    0 < data < objct && hasNodeTag(data);
    data->parent == NULL;
  })
  post(AND {
    result==NULL || AND {
      intr == pre(intr);
      0 < intr < objct && tag[intr]==Interior_tag;
      0 < result < objct && tag[result]==Interior_tag;
      result != intr;
      result->parent == NULL;
    };

    // nothing deallocated
    forall(int id).  0 < id < pre(objct) ==> tag[id] == pre(tag)[id];

    // objct monotonic
    pre(objct) <= objct;
  })
  changes(Node_tag, Interior_tag, objct, tag, objsize)
{
  int k;

  if (intr->numKeys < MAXIKEYS) {
    // no split; find the right place to do the insertion
    for (k=0; k < intr->numKeys; k++) {
      invariant {
        // from above
        global_inv;
        0 < intr < objct && tag[intr]==Interior_tag;
        0 < data < objct && hasNodeTag(data);
        data->parent == NULL;
        intr == pre(intr);
        forall(int id).  0 < id < pre(objct) ==> tag[id] == pre(tag)[id];
        pre(objct) <= objct;
        intr->numKeys < MAXIKEYS;

        // this loop
        0 <= k < intr->numKeys;
      }

      if (*iikey <= intr->ikey[k]) break;
    }

    // insert before 'k'
    memmove(intr->ikey+(k+1), intr->ikey+k, (intr->numKeys - k) * sizeof(int));
    memmove(intr->child+(k+2), intr->child+(k+1), (intr->numKeys - k) * sizeof(Node*));
    intr->ikey[k] = *iikey;
    setChild(intr, k+1, data);
    intr->numKeys++;

    // fix the parent indices of the moved children
    Node_parentIndex = fresh_value();
    assume(AND {
      forall(Interior *i). tag[i]==Interior_tag ==> AND {
        forall(int j). 0 <= j <= i->numKeys ==> AND {
          Node_parentIndex[i->child[j]]==j;
        };
      };
    });

    return NULL;           // no split needs to propagate
  }

  else {
    // split
    assert(intr->numKeys == MAXIKEYS);
    Interior *leftHalf = intr;

    // make new interior node
    Interior *rightHalf;
    newInterior(rightHalf);

    // copy half of the children into rightHalf
    int numChildren = intr->numKeys+1;
    int splitPoint = numChildren / 2;
    assume(0 < splitPoint < MAXIKEYS+1);     // need a little knowledge about '/'

    // example, numKeys=5, splitPoint=3:
    //                        intr
    //                +---+---+---+---+---+
    //       ikey     | 0 | 1 | 2 | 3 | 4 |
    //                +---+---+---+---+---+
    //              +---+---+---+---+---+---+
    //      child   | 0 | 1 | 2 | 3 | 4 | 5 |
    //              +---+---+---+---+---+---+
    //
    // becomes:                 new *key
    //                 (intr)      ^     (result)
    //                  left       |      right
    //                +---+---+  +---+  +---+---+
    //       ikey     | 0 | 1 |  | 2 |  | 3 | 4 |
    //                +---+---+  +---+  +---+---+
    //              +---+---+---+     +---+---+---+
    //      child   | 0 | 1 | 2 |     | 3 | 4 | 5 |
    //              +---+---+---+     |---+---+---+

    memcpy(rightHalf->ikey+0, leftHalf->ikey+splitPoint, (numChildren-splitPoint-1) * sizeof(int));
    memcpy(rightHalf->child+0, leftHalf->child+splitPoint, (numChildren-splitPoint) * sizeof(Node*));
    leftHalf->numKeys = splitPoint-1;
    rightHalf->numKeys = numChildren-splitPoint-1;

    // fix the parent pointers of the moved children
    Node_parent = fresh_value();
    Node_parentIndex = fresh_value();
    assume(AND {
      // It would be possible to do this without an assumption, but a
      // programmer writing the code without concern for the verifier
      // would likely have used 'memcpy', and I'm thinking that this
      // kind of wholesale cleanup is a reasonable solution to
      // situations like this.  Of course, as with all assumptions,
      // some care is needed.  In particular, an inconsistent
      // assumption would nullify all further checking.
      forall(Interior *i). tag[i]==Interior_tag ==> AND {
        forall(int j). 0 <= j <= i->numKeys ==> AND {
          Node_parent[i->child[j]]==i;
          Node_parentIndex[i->child[j]]==j;
        };
      };

      // since the fresh_value() caused these to be forgotten..
      data->parent == NULL;
      rightHalf->parent == NULL;
    });

    int newKey = leftHalf->ikey[splitPoint-1];

    // insert key/data into one of the split halves; at this point,
    // such an insertion is guaranteed not to cause another split
    if (*iikey < newKey) {
      interiorInsert(leftHalf, iikey, data);
    }
    else {
      interiorInsert(rightHalf, iikey, data);
    }
    assume(rightHalf->parent == NULL);      // change isolation

    // finally, return right and its left-edge key
    *iikey = newKey;
    return rightHalf;
  }
}


// insert a new key,data pair; if this returns non-NULL, then the tree
// was split, and 'key' and 'result' provide the other two values to
// make up the split triple: left(root)/key/right(result)
Node *innerInsert(Node *root, int *key PARAMBYREF, int data)
  pre(0 < root < objct && hasNodeTag(root))
  post(AND {
    result==NULL || AND {
      root == pre(root);
      0 < root < objct && hasNodeTag(root);
      0 < result < objct && hasNodeTag(result);
      result != root;
      result->parent == NULL;
    };

    // nothing deallocated
    forall(int id).  0 < id < pre(objct) ==> tag[id] == pre(tag)[id];

    // objct monotonic
    pre(objct) <= objct;
  })
  changes(Node_tag, Interior_tag, Leaf_tag, objct, tag, objsize)
{
  // originally declared several times below, but lack of proper
  // alpha renaming makes me pull these out...
  Interior *single;
  int k;

  if (root->isLeaf) {
    Leaf *leaf = (Leaf*)root;

    if (leaf->numKeys < MAXLKEYS) {
      // insert into keys[] and data[] in sorted order
      for (k=0; k < leaf->numKeys; k++) {
        invariant {
          // from above
          global_inv;
          0 < leaf < objct && tag[leaf]==Leaf_tag;
          root == pre(root);
          root == leaf;
          forall(int id).  0 < id < pre(objct) ==> tag[id] == pre(tag)[id];
          pre(objct) <= objct;
          leaf->numKeys < MAXLKEYS;

          // this loop
          0 <= k < leaf->numKeys;
        }

        if (*key <= leaf->key[k]) break;
      }

      // now, insert before index 'k'
      memmove(leaf->key+(k+1), leaf->key+k, (leaf->numKeys - k) * sizeof(int));
      memmove(leaf->data+(k+1), leaf->data+k, (leaf->numKeys - k) * sizeof(int));
      leaf->key[k] = *key;
      leaf->data[k] = data;
      leaf->numKeys++;
      return NULL;
    }

    // must split
    assert(leaf->numKeys == MAXLKEYS);
    Leaf *leftHalfLeaf = leaf;

    // new node
    Leaf *rightHalfLeaf;
    newLeaf(rightHalfLeaf);

    // copy half of the keys and data to the right leaf
    int numForLeft = leftHalfLeaf->numKeys / 2;
    assume(0 < numForLeft < MAXLKEYS);
    int numForRight = leftHalfLeaf->numKeys - numForLeft;
    assert(numForLeft > 0 && numForRight > 0);
    memcpy(rightHalfLeaf->key + 0, leftHalfLeaf->key + numForLeft,
           numForRight * sizeof(int));
    memcpy(rightHalfLeaf->data + 0, leftHalfLeaf->data + numForLeft,
           numForRight * sizeof(int));

    // set new sizes
    leftHalfLeaf->numKeys = numForLeft;
    rightHalfLeaf->numKeys = numForRight;

    // split 'rightHalfLeaf' into the leaf list
    setNext(rightHalfLeaf, leftHalfLeaf->next);
    setNext(leftHalfLeaf, rightHalfLeaf);

    // now, pick a leaf into which to re-do the insertion; it will
    // succeed this time, without splitting (at the moment I do not
    // prove this, but the safety policy doesn't demand it since I'm
    // not checking for leaks)
    if (*key < rightHalfLeaf->key[0]) {
      innerInsert(leftHalfLeaf, key, data);
    }
    else {
      innerInsert(rightHalfLeaf, key, data);
    }
    assume(rightHalfLeaf->parent == NULL);     // change isolation

    // return the two leaves; 'key' remains unchanged
    return (Node*)rightHalfLeaf;
  }

  else {
    Interior *intr = (Interior*)root;

    // find the place in children[] into which the insertion will be made
    for (k=0; k < intr->numKeys; k++) {
      invariant {
        // from above
        global_inv;
        0 < intr < objct && tag[intr]==Interior_tag;
        root == pre(root);
        root == intr;
        forall(int id).  0 < id < pre(objct) ==> tag[id] == pre(tag)[id];
        pre(objct) <= objct;

        // this loop
        0 <= k < intr->numKeys;
      }

      if (*key <= intr->ikey[k]) break;
    }

    // situation looks like this:
    //
    //                     < *key <=
    //
    //             key[k-1]          key[k]
    //            /        \        /      \
    //  child[k-1]          child[k]        child[k+1]

    Node *leftHalfNode = intr->child[k];
    Node *rightHalfNode = innerInsert(leftHalfNode, key, data);
    if (!rightHalfNode) {
      return NULL;            // no split
    }

    // now things look like this:
    //
    //             key[k-1]          *key               key[k]
    //            /        \        /    \             /      \
    //  child[k-1]          child[k]      rightHalfNode        child[k+1]
    //                   (leftHalfNode)

    // so, insert *key/rightHalfNode into position 'k'; this might split
    return interiorInsert(intr, key, rightHalfNode);
  }
}


// outer shell of insert routine
Node *insert(Node *root, int key ARGBYREF, int data)
  pre(root==NULL || AND {
        0 < root < objct && hasNodeTag(root);
        root->parent==NULL;      // really the root
      })
  post(AND {
    0 < result < objct && hasNodeTag(result);
    pre(objct) <= objct;
  })
  changes(Node_tag, Interior_tag, Leaf_tag, objct, tag, objsize)
{
  if (!root) {
    Leaf *leaf;
    newLeaf(leaf);
    leaf->key[0] = key;
    leaf->data[0] = data;
    leaf->numKeys++;
    return leaf;
  }

  else {
    Node *rightHalf = innerInsert(root, &key, data);
    if (!rightHalf) {
      return root;            // no split, root unchanged
    }
    else {
      assume(root->parent == NULL);       // need more change isolation

      // make new root node
      Interior *intr;
      newInterior(intr);  
      setChild(intr, 0, root);
      intr->ikey = key;
      setChild(intr, 1, rightHalf);

      return intr;
    }
  }
}


#if 0    // trash
    // fix the parent pointers of the moved children
    rightHalf->
    for (k=0; k < numChildren-splitPoint; k++) {
      invariant {
        // from above
        global_inv;
        0 < intr < objct && tag[intr]==Interior_tag;
        0 < data < objct && hasNodeTag(data);
        data->parent == NULL;
        intr == pre(intr);
        forall(int id).  0 < id < pre(objct) ==> tag[id] == pre(tag)[id];
        pre(objct) <= objct;

        leftHalf == intr;

        0 < rightHalf < objct && tag[rightHalf]==Interior_node;
        rightHalf->parent == NULL;
        rightHalf->numKeys == 0;

        numChildren == MAXIKEYS+1;
        0 < splitPoint < MAXIKEYS+1;



        // this loop
        0 <= k < numChildren-splitPoint;

#endif // 0
