#!/usr/bin/perl -w
# create a _stubs.cc file from a .cc file by following the
# directives in the .cc file

# 2005-01-27: This is experimental.
# 2005-02-06: It seems to be working well, so will likely stay.

# This script creates a variant of an existing module that exports
# most of the same functions, but with empty implementations.
#                                                          
# The rationale for doing this is that I have a number of AST
# extensions (e.g. vcgen.ast), and for a variety of reasons I always
# run astgen with all extensions active.  But I do not want every
# executable to have the full implementations of the functions
# declared in those extensions.  Therefore I use this script to make
# stub versions of the implementations, and link against the stubs in
# executables that do not need the full implementations.

use strict 'subs';

if (@ARGV == 0) {
  print(<<"EOF");
usage: $0 file.cc

Produces file_stubs.cc
EOF
  exit(0);
}

$inFname = $ARGV[0];

my ($base, $ext) = ($inFname =~ m/^(.*)\.([^.]*)$/);
if (!defined($ext)) {
  die("source filename must have a period in it somewhere\n");
}

$outFname = "${base}_stubs.$ext";

# translation directives that come from the input file
$stubCode = "(auto)";
$stubPattern = "";
$patternUseCount = 1;

open(IN, "<$inFname") or die("cannot read $inFname: $!\n");
open(OUT, ">$outFname") or die("cannot write $outFname: $!\n");

print OUT (<<"EOF");
// $outFname
// created automatically by $0 from $inFname
// DO NOT EDIT MANUALLY

EOF

while (defined($line = <IN>)) {
  chomp($line);

  my $s;

  ($s) = ($line =~ m|//\s*STUBCODE:\s(.*)$|);
  if (defined($s)) {
    $stubCode = $s;
    next;
  }

  ($s) = ($line =~ m|//\s*STUBPATTERN:\s(.*)$|);
  if (defined($s)) {
    if ($patternUseCount == 0) {
      print("warning: unused pattern: $stubPattern\n");
    }

    $stubPattern = $s;
    $patternUseCount = 0;
    next;
  }

  ($s) = ($line =~ m|//\s*STUBEMIT:\s(.*)$|);
  if (defined($s)) {
    print OUT ($s . "\n");
    next;
  }

  # inline emit
  if ($line =~ m|^.*//\s*stubemit\s*$|) {
    print OUT ($line . "\n");
    next;
  }

  # explicit stubification request
  if ($line =~ m|^.*//\s*stubify\s*$|) {
    stubify($line);
    next;
  }

  # implicit stubification
  if ($stubPattern &&
      $line =~ m|$stubPattern|) {
    stubify($line);
    $patternUseCount++;
    next;
  }
  
  # retain preprocessor directives other than #include (which
  # can pull in exactly the dependencies the stubs are intended
  # to avoid)
  if ($line =~ m|^\s*\#| &&
      $line !~ m|^\s*\#\s*include|) {
    print OUT ($line . "\n");
    next;
  }
}

print OUT ("// EOF: $outFname\n");
exit(0);


sub stubify {
  my ($line) = @_;
  
  print OUT ("$line\n");
  
  if ($line =~ m/\}\s*$/) {
    # $line ends with a brace, so assume the implementation is already
    # there (for example, the original definition was itself a stub)
    return;
  }
  
  if ($stubCode eq "(auto)") {
    if ($line =~ m/^void\s/) {
      print OUT ("{}\n\n");
    }
    else {
      print OUT ("{ return 0; }\n\n");
    }
  }
  else {
    print OUT ("{ $stubCode }\n\n");
  }
}


# EOF
