// term.cc
// code for term.h

#include "term.h"      // Term
#include "exc.h"       // xformat

#include <ctype.h>     // isspace, isupper
#include <stdlib.h>    // exit


StringTable *Term::parseStrTable = NULL;


Term::Term(StringRef n)
  : name(n),
    args()
{}

Term::Term(StringRef n, Term *a1)
  : name(n)
{
  args.prepend(a1);
}

Term::Term(StringRef n, Term *a1, Term *a2)
  : name(n)
{
  args.prepend(a2);
  args.prepend(a1);
}

Term::Term(StringRef n, Term *a1, Term *a2, Term *a3)
  : name(n)
{
  args.prepend(a3);
  args.prepend(a2);
  args.prepend(a1);
}

Term::Term(StringRef n, Term *a1, Term *a2, Term *a3, Term *a4)
  : name(n)
{
  args.prepend(a4);
  args.prepend(a3);
  args.prepend(a2);
  args.prepend(a1);
}

Term::Term(StringRef n, Term *a1, Term *a2, Term *a3, Term *a4, Term *a5)
  : name(n)
{
  args.prepend(a5);
  args.prepend(a4);
  args.prepend(a3);
  args.prepend(a2);
  args.prepend(a1);
}

Term::~Term()
{}


bool Term::isVar() const
{
  // for now, any capitalized name is a unification variable
  return args.isEmpty() &&
         isupper(name[0]);
}


string Term::asString() const
{
  stringBuilder sb;
  asString(sb);
  return sb;
}


void Term::asString(stringBuilder &sb) const
{
  if (args.isEmpty()) {
    sb << name;
  }
  else {
    sb << "(" << name;

    FOREACH_OBJLIST(Term, args, iter) {
      sb << " ";
      iter.data()->asString(sb);
    }
    
    sb << ")";
  }
}


static char const *skipws(char const *s)
{
  while (isspace(*s)) {
    s++;
  }
  return s;
}


STATICDEF Term *Term::parse(char const *sexp)
{
  char const *endp;
  return parse(sexp, endp);
}

STATICDEF Term *Term::parse(char const *sexp, char const *&endp)
{
  sexp = skipws(sexp);

  if (*sexp == 0) {
    xformat("empty sexp");
  }

  bool paren = false;
  if (*sexp == '(') {
    paren = true;
    sexp++;
  }

  // function/variable/constant
  endp = sexp;
  while (*endp != 0 && *endp != ')' && *endp != ' ') {
    endp++;
  }

  Term *t = new Term(parseStrTable->add(substring(sexp, endp-sexp)));

  if (!paren) {
    // variable/constant; we're done
    return t;
  }

  // arguments
  for (;;) {
    sexp = skipws(endp);
    if (*sexp == ')') {
      endp = sexp+1;
      break;
    }

    // next arg
    Term *arg = parse(sexp, endp);
    t->args.prepend(arg);    // reverse order, initially
  }

  t->args.reverse();         // fix the order at the end
  return t;
}

     
// this is somewhat incompatible with using a string table
#if 0
Term *Term::clone(int id) const
{
  if (isVar()) {
    // rename the unification variables (tacitly assume the
    // name originally didn't have any underscores in it)
    return new Term(stringc << name << "_" << id);
  }
  else {
    Term *ret = new Term(name);

    FOREACH_OBJLIST(Term, args, iter) {
      ret->args.prepend(iter.data()->clone(id));
    }
    ret->args.reverse();

    return ret;
  }
}
#endif // 0


// --------------------------- test code ---------------------------
#ifdef TEST_TERM

void test(char const *sexp)
{
  cout << "testing: " << sexp << endl;

  Term *t = Term::parse(sexp);
  string s2 = t->asString();
  
  if (!s2.equals(sexp)) {
    cout << "vector failed!\n";
    cout << "  sexp: " << sexp << endl;
    cout << "    s2: " << s2 << endl;
    exit(4);
  }
  
  cout << "  name: " << t->name << endl;
  cout << "  args: " << t->args.count() << endl;
}

int main()
{
  StringTable str;
  Term::parseStrTable = &str;

  Term *t = new Term(str("t"));
  cout << t->asString() << endl;

  t = new Term(str("f"), new Term(str("x")), new Term(str("y")));
  cout << t->asString() << endl;

  t = new Term(str("f"), new Term(str("x")), 
                         new Term(str("f"), new Term(str("x")), 
                                            new Term(str("y"))));
  cout << t->asString() << endl;

  test("xyz");
  test("(a (b (c d)))");
  test("(AND (= x y) (= y z))");
  test("(= x z)");

  cout << "term ok\n";
  return 0;
}

#endif // TEST_TERM
