#!/usr/bin/perl -w
# send predicates and axioms to Simplify, and verify the
# responses are as expected

use strict 'subs';

use FileHandle;          # autoflush?
use IPC::Open2;          # open2 call

if (@ARGV != 1) {
  print("usage: $0 file1.sx\n");
  exit(0);
}

# open Simplify
my $pid = open2(*Reader, *Writer, "./run-Simplify -nosc");
Writer->autoflush();     # default here, actually

# open a log file so I can see what gets sent to Simplify
open(LOG, ">simp.log") or die;
LOG->autoflush();
#print LOG ("hi\n");

# turn off prompting
sendToSimplify ("(PROMPT_OFF)\n");

if (0) {
  # feed it the background predicate
  open(BG, "<bgpred.sx") or die("can't open bgpred.sx: $!\n");
  while (defined($line = <BG>)) {
    sendToSimplify ($line);
  }
  close(BG) or die;
}

# globals
$predicateCt = 0;
$errors = 0;
@blocks = (0);
@blockNames = ("toplevel");

if (0) {
  valid("true", "TRUE\n");
  notvalid("false", "FALSE\n");
  valid("false", "FALSE\n");
  notvalid("true", "TRUE\n");
  valid("true", "TRUE\n");
}


# work through experiments file
open(EXPER, "<$ARGV[0]") or die("can't open $ARGV[0]: $!\n");
while (defined($line = <EXPER>)) {
  chomp($line);

  # strip comments
  $line =~ s|;.*$||;

  if ($line =~ m|^\s*$|) {    # blank line
    next;
  }

  # proof obligation
  my ($skip, $inv, $val, $dummy2, $comment) =
    ($line =~ /^\s*(skip)?(not)?(valid|prove)\s+(\"(.*)\"\s+)?\(\s*$/);
  if (defined($val)) {     # predicate
    my @pred = collectPred();
    @pred = ("(AND\n", @pred, ")\n");
    if (!$skip) {          # "skip" is a way to comment-out
      if ($inv) {
        notvalid($comment, @pred);
      }
      else {
        valid($comment, @pred);
      }
    }
    next;
  }

  # lemma (proof, followed by assumption)
  ($dummy2, $comment) =
    ($line =~ /^\s*lemma\s+(\"(.*)\"\s+)?\(\s*$/);
  if (defined($comment)) {
    my @pred = collectPred();
    @pred = ("(AND\n", @pred, ")\n");
    lemma($comment, @pred);
    next;
  }

  if ($line =~ /^\s*(bg_push|assume)\s+\(\s*$/) {     # bg_push
    my @pred = collectPred();
    sendToSimplify ("(BG_PUSH (AND\n", @pred, "))\n");
    countAddedAssumption();
    next;
  }

  if ($line =~ /^\s*bg_pop\s*$/) {
    sendToSimplify ("(BG_POP)\n");
    next;
  }

  if ($line =~ /^\s*stop\s*$/) {
    # this is convenient when I insert some new proof obligations
    # into the middle of a file
    last;
  }

  my $name;
  ($name) =
    ($line =~ /^\s*begin_block\s+\"(.+)\"\s*$/);
  if (defined($name)) {
    # begin counting assumptions pushed
    print("----------- block: $name ----------\n");
    @blocks = (0, @blocks);
    @blockNames = ($name, @blockNames);
    next;
  }

  ($name) =
    ($line =~ /^\s*end_block\s+\"(.+)\"\s*$/);
  if (defined($name)) {
    # pop however many assumptions were pushed in this block
    if (@blocks == 1) {
      die("empty block stack\n");
    }
    if ($blockNames[0] ne $name) {
      die("mismatching block name; $blockNames[0] is at top\n");
    }
    my $i = $blocks[0];
    die if $i < 0;
    for (; $i > 0; $i--) {
      sendToSimplify("(BG_POP)\n");
    }
    shift @blocks;
    shift @blockNames;
    next;
  }

  # all on one line:
  #   prove <predicate>
  my ($predicate) =
    ($line =~ /^\s*prove (.*)$/);
  if (defined($predicate)) {
    valid("prove $predicate", $predicate);
    next;
  }

  #   notprove <predicate>
  ($predicate) =
    ($line =~ /^\s*notprove (.*)$/);
  if (defined($predicate)) {
    notvalid("notprove $predicate", $predicate);
    next;
  }

  my ($theory) =
    ($line =~ m|^\s*import \"(\S+)\"\s*$|);
  if (defined($theory)) {
    # import a theory
    open(TH, "<$theory") or die("can't open $theory: $!\n");
    my $line;
    while (defined($line = <TH>)) {
      sendToSimplify ($line);
    }
    close(TH) or die;

    # make sure the state it sync'd
    valid("check after $theory", "TRUE\n");
    next;
  }

  if ($line =~ m|^\s*consistency_check\s*$|) {
    notvalid("consistency check", "FALSE\n");
    next;
  }

  my ($pred) =
    ($line =~ m|^defpred\s*(.*)$|);
  if (defined($pred)) {
    sendToSimplify ("(DEFPRED $pred)\n");
    next;
  }

  my $args;
  ($name, $args) =
    ($line =~ m|^define\s+\((\S+)\s+(.*)\)\s+\($|);
  if (defined($args)) {
    my @body = collectPred();
    my $body = join("\n", @body);

    my $command = (<<"EOF");
(DEFPRED ($name $args))
(BG_PUSH
  (FORALL ($args)
    (IFF
      ($name $args)
      (AND
        $body
      )))
)
EOF
    #print ($command);
    sendToSimplify ($command);
    countAddedAssumption();
    next;
  }

  print("not understood: $line\n");
  pretendUsed($dummy2);
  exit(2);
}

# get all remaining responses
close(Writer);
@resp = <Reader>;
print(@resp);

pretendUsed($pid);

if ($errors > 0) {
  print("$errors errors\n");
  exit(2);
}
else {
  exit(0);
}


# ----------------------- subroutines ------------------------
sub collectPred {
  my @ret = ();
  my $line;

  while (defined($line = <EXPER>)) {
    if ($line =~ /^\)\s*$/) {   # done
      return @ret;
    }
    @ret = (@ret, $line);
  }

  print("unexpected EOF reading predicate:\n@ret\n");
  exit(2);
}


sub pretendUsed {
}

# give a single predicate to Simplify, and return 1 if it proves it
# and 0 if it cannot prove it
sub prove {
  my ($comment, @pred) = @_;

  # say what's going on
  $predicateCt++;
  printf("%3d: ", $predicateCt);
  if ($comment) {
    print("$comment... ");
  }
  else {
    print("predicate... ");
  }
  STDOUT->flush();
                       
  # hand predicate to Simplify
  sendToSimplify (@pred);
  $resp = <Reader>;
  if (defined($resp)) {
    chomp($resp);
    if ($resp =~ / Valid./) {
      print("proved");
      $blank = <Reader>;     # eat blank line
      return 1;
    }
    if ($resp =~ / Invalid./) {
      print("not proved");
      $blank = <Reader>;
      return 0;
    }

    print("\n  unexpected response: $resp\n");
  }
  else {
    print("\n  pipe EOF!\n");
  }
  exit(2);
}


# try to prove a predicate; expect success iff $resp is true
sub expect {
  my ($resp, $comment, @pred) = @_;
  if (!$resp != !prove($comment, @pred)) {    # normalize booleans with "!"
    # response is opposite what I expected
    print(" (UNEXPECTED)\n");
    $errors++;

    # print failing predicate
    my @lines = split('\n', join('', @pred));
    for (my $i=0; $i < @lines; $i++) {
      #chomp($lines[$i]);        # I think split removes newlines
      print("  $lines[$i]\n");
    }
    return 0;
  }
  else {
    # ok
    print("\n");
    return 1;
  }
}


# try to prove a predicate we think is valid
sub valid {
  my ($comment, @pred) = @_;
  print("valid:    ");
  return expect(1, $comment, @pred);
}

# try to prove a predicate we think is not valid or not provable
# (I avoid the term 'invalid', which connotes "provably false")
sub notvalid {
  my ($comment, @pred) = @_;
  print("notvalid: ");
  if (defined($ENV{SKIP_NOTVALID})) {
    print("(skipped)\n");
    return 1;
  }
  return expect(0, $comment, @pred);
}

# try to prove, then assume it
sub lemma {
  my ($comment, @pred) = @_;
                                                
  # try to prove it
  print("lemma:    ");
  expect(1, $comment, @pred);

  # then assume it
  sendToSimplify("(BG_PUSH ", @pred, ")\n");
  countAddedAssumption();
}


# send something to Simplify
sub sendToSimplify {
  my @data = @_;
  print LOG (@data);
  print Writer (@data);
}


# track that an assumption was added
sub countAddedAssumption {
  die if (@blocks == 0);
  $blocks[0]++;
}

