// rewrite.cc
// code for rewrite.h

#include "rewrite.h"         // this module
#include "unify.h"           // Unify
#include "trace.h"           // trace
#include "macros.h"          // Restorer
#include "owner.h"           // Owner


template <class T>
inline void swap(T &a, T &b)
{
  T tmp = a;
  a = b;
  b = tmp;
}


// ----------------------- order ------------------------
string toString(Order o)
{
  switch (o) {
    default: xfailure("bad order code");
    case LT: return "LT";
    case EQ: return "EQ";
    case NLE: return "NLE";
  }
}


// find smallest subterm, assuming unary and nullary functions only
Term const *findSmallest(Term const *t)
{
  if (t->args.isEmpty()) {
    return t;
  }
  
  xassert(t->args.count() == 1);
  return findSmallest(t->args.firstC());
}


// for now, assume unary and nullary functions only
Order compare(Term const *t1, Term const *t2)
{
  // get smallest subterm
  Term const *subt1 = findSmallest(t1);
  Term const *subt2 = findSmallest(t2);

  // get weights
  int wt1 = weight(t1);
  int wt2 = weight(t2);

  // check for variable mismatch
  if (subt1->isVar()) {       // main problem is when LHS is var
    if (subt2->isVar()) {
      if (subt1->name.equals(subt2->name)) {
        // ok, same variable
      }
      else {
        // different variables, can't compare
        return NLE;
      }
    }
    else {
      // LHS has var, RHS is constant: can't be LT nor EQ
      return NLE;
    }
  }
  else if (subt2->isVar()) {
    // RHS has var, LHS is constant: LHS can be smaller if
    // it has a strictly smaller weight
    if (wt1 < wt2) {
      return LT;
    }
    else {
      return NLE;
    }
  }

  // order by weight first
  if (wt1 < wt2) {
    return LT;
  }
  if (wt1 > wt2) {
    return NLE;
  }

  // now lexicographic, starting at outermost
  int cmp = strcmp(t1->name, t2->name);
  if (cmp < 0) {
    return LT;      // t1 is smaller b/c outermost symbol is smaller
  }
  if (cmp > 0) {
    return NLE;     // t2 is probably smaller
  }

  // same name; do they have children? (must both have or not since equal weight)
  xassert(t1->args.count() == t2->args.count());
  if (t1->args.isEmpty()) {
    // no children, they are equal
    return EQ;
  }
  else {
    return compare(t1->args.firstC(), t2->args.firstC());
  }
}


int weight(Term const *t)
{
  // assume all symbols have weight 1, which is then the minimum
  // weight if this is a unif. var
  int ret = 1;

  FOREACH_OBJLIST(Term, t->args, iter) {
    ret += weight(iter.data());
  }

  return ret;
}


// --------------------- Equality ------------------------
int Equality::nextId = 0;


Equality::Equality(Term *l, Term *r)
  : id(nextId++),
    lhs(l),
    rhs(r)
{}

Equality::~Equality()
{
  delete lhs;
  delete rhs;
}


bool Equality::orient()
{
  Order o = compare(rhs, lhs);
  if (o == LT) {
    return false;
  }
  if (o == EQ) {
    return true;
  }

  // try reversing order
  o = compare(lhs, rhs);
  if (o == LT) {
    // ok, swap
    swap(lhs, rhs);
    return false;
  }
  xassert(o != EQ);      // shouldn't now be equal...
  xfailure(stringc << "non-orientable: " << asString());
  return false;    // silence warning
}


string Equality::asString(bool disEq) const
{
  return stringc << "[" << id << "] "
                 << lhs->asString() << (disEq? " != " : " == ")
                 << rhs->asString();
}


// ---------------------- Axiom --------------------------
Axiom::Axiom(Term *l, Term *r)
  : Equality(l, r),
    positives(),
    negatives()
{}

Axiom::~Axiom()
{}


string Axiom::asString() const
{
  stringBuilder sb;

  int ct = 0;

  FOREACH_OBJLIST(Equality, positives, posIter) {
    if (ct++ > 0) {
      sb << " && ";
    }
    sb << posIter.data()->asString();
  }

  FOREACH_OBJLIST(Equality, negatives, negIter) {
    if (ct++ > 0) {
      sb << " && ";
    }
    sb << negIter.data()->asString(true /*disEq*/);
  }
  
  if (ct > 0) {
    sb << "  ==>  ";
  }
  sb << Equality::asString();
  
  return sb;
}


// --------------------- Rewriter ------------------------
Rewriter::Rewriter()
  : axioms(),
    equalities(),
    disequalities(),
    contradiction(false),
    changes(0)
{}

Rewriter::~Rewriter()
{}


bool Rewriter::reduce()
{
  changes = 0;

  // search for terms to reduce

  // search in axioms
  ObjListMutator<Axiom> mut(axioms);
  while (!mut.isDone()) {
    Axiom *ax = mut.data();

    reduceEqList(ax->positives, true /*positive*/);

    if (reduceEqList(ax->negatives, false /*positive*/)) {
      // a disequality has reduced to something like "x != x", which
      // is unsatisfiable; hence this rule is useless
      TRACE("reduce", "unsatisfiable premise, deleting rule " << ax->id);
      delete mut.remove();
      continue;
    }

    if (reduceEq(ax)) {
      // the consequence has reduced to a triviality
      TRACE("reduce", "trivial consequence, deleting rule " << ax->id);
      delete mut.remove();
      continue;
    }

    mut.adv();
  }

  // search in facts
  reduceEqList(equalities, true /*positive*/);

  if (reduceEqList(disequalities, false /*positive*/)) {
    TRACE("reduce", "found contradiction");
    contradiction = true;
  }

  return changes > 0;
}

bool Rewriter::reduceEqList(ObjList<Equality> &list, bool positive)
{
  bool triviality = false;

  ObjListMutator<Equality> mut(list);
  while (!mut.isDone()) {
    int id = mut.data()->id;
    if (reduceEq(mut.data())) {
      // equation reduced to triviality
      triviality = true;
      if (positive) {
        // for normal equalities, just remove it
        TRACE("reduce", "trivial fact, deleting equation " << id);
        delete mut.remove();
        continue;
      }
      else {
        // for disequalities, leave it there since some higher
        // level of processing is now necessary
        TRACE("reduce", "found reflexive disequality " << id);
      }
    }

    mut.adv();
  }

  return triviality;
}

// return true if reduces to triviality
bool Rewriter::reduceEq(Equality *reducee)
{
  int before = changes;

  reducee->lhs = reduceTerm(reducee->lhs);
  reducee->rhs = reduceTerm(reducee->rhs);

  if (before != changes) {
    bool triv = reducee->orient();
    TRACE("reduce", "reduced: " << reducee->asString() << (triv? " (triv)" : ""));
    return triv;
  }
  else {
    return false;
  }
}

Term *Rewriter::reduceTerm(Term *t)
{
  // search for axioms or rules that can reduce this term

  // iterate over ground equalities
  FOREACH_OBJLIST(Equality, equalities, iter1) {
    t = reduceTermVia(t, iter1.data(), NULL);
  }

  // iterate over axioms
  FOREACH_OBJLIST(Axiom, axioms, iter2) {
    Axiom const *ax = iter2.data();
    t = reduceTermVia(t, ax, ax);
  }

  return t;
}

// if 'ax' is non-NULL, it equals 'e'
Term *Rewriter::reduceTermVia(Term *t, Equality const *e, Axiom const *ax)
{
  xassert(!ax || ax==e);

  if (t == e->lhs) {
    // don't use an equation to reduce its own LHS
    return t;
  }

  Unify u;
  u.matchMode = true;
  if (u.unify(e->lhs, t)) {
    // if this is an axiom, we need to check for satisfied premises; this
    // is a recurive call, so we require (not currently checked!) that
    // terms in the condition be smaller than the rule's LHS term
    if (!ax || satisfiedPremises(u, ax)) {
      // replace 't' with 'e->rhs'
      changes++;

      // NOTE: must compute the substitution *before* deleting
      // 't' because the substitution in 'u' points into 't'!
      Term *ret = u.substitute(e->rhs);

      TRACE("reduceTerm", "reduced term via " << e->id << ": "
           << t->asString() << " to " << ret->asString());

      delete t;
      return ret;
    }
  }

  // reduce in subterms
  for (ObjListMutator<Term> iter(t->args); !iter.isDone(); iter.adv()) {
    Term *&arg = iter.dataRef();
    arg = reduceTermVia(arg, e, ax);
  }

  // but yield same term object container
  return t;
}

bool Rewriter::satisfiedPremises(Unify &u, Axiom const *ax)
{
  TRACE("axiom", "attempting to satisfy premises of axiom " << ax->id);

  // this is just a check, so remember the value of 'changes'
  // so we can reset it
  Restorer<int> restorer(changes);

  // check equalities
  {
    FOREACH_OBJLIST(Equality, ax->positives, iter) {
      Equality const *e = iter.data();

      // this is satisfied if the two terms, under the substitution by
      // which the axiom matched the object term, reduce to the same
      // normal form
      Owner<Term> t1; t1 = u.substitute(e->lhs);
      Owner<Term> t2; t2 = u.substitute(e->rhs);

      t1 = reduceTerm(t1.xfr());
      t2 = reduceTerm(t2.xfr());

      Order o = compare(t1, t2);

      TRACE("axiom", "  premise " << e->id << ": " << t1->asString()
                                  << " " << toString(o) << " "
                                  << t2->asString());

      if (o != EQ) {
        TRACE("axiom", "  premise " << e->id << " not satisfied");
        return false;
      }
    }
  }

  // check disequalities
  {
    FOREACH_OBJLIST(Equality, ax->negatives, iter) {
      Equality const *e = iter.data();

      // this is satisfied if, after reducing the terms, we can
      // find an identical disequality among the facts
      Owner<Term> t1; t1 = u.substitute(e->lhs);
      Owner<Term> t2; t2 = u.substitute(e->rhs);

      t1 = reduceTerm(t1.xfr());
      t2 = reduceTerm(t2.xfr());

      TRACE("axiom", "  premise " << e->id << ": " << t1->asString()
                                  << " !=? " << t2->asString());

      Order o = compare(t1, t2);
      if (o == LT) {
        // make it so t1 > t2
        swap(t1, t2);
      }
      else if (o == EQ) {
        // if we have a reflexive disequality, we'll have already
        // noticed the contradiction
        TRACE("axiom", "  premise " << e->id << " reflexively not satisfied");
        return false;
      }

      if (!haveDisequality(t1, t2)) {
        // this is not a known fact
        TRACE("axiom", "  premise " << e->id << " not satisfied");
        return false;
      }
    }
  }

  // ok, premises are satisfied!
  TRACE("axiom", "premises of axiom " << ax->id << " are satisfied!");
  return true;
}

bool Rewriter::haveDisequality(Term const *lhs, Term const *rhs) const
{
  xassert(compare(rhs, lhs) == LT);

  FOREACH_OBJLIST(Equality, disequalities, iter) {
    Equality const *d = iter.data();
    if (EQ==compare(lhs, d->lhs) &&
        EQ==compare(rhs, d->rhs)) {
      return true;
    }
  }
  return false;    // not found
}


bool Rewriter::mix()
{
  // generated equalities
  ObjList<Equality> generated;

  // for each axiom...
  FOREACH_OBJLIST(Axiom, axioms, axIter) {
    Axiom const *ax = axIter.data();

    // get the set of nontrivial proper subterms of the LHS
    SObjList<Term const> subterms;
    getNontrivProperSubterms(subterms, ax->lhs);

    // for each ground equality...
    FOREACH_OBJLIST(Equality, equalities, eqIter) {
      Equality const *eq = eqIter.data();

      // attempt to match each of the axiom LHS subterms with the
      // equality LHS
      SFOREACH_OBJLIST(Term const, subterms, termIter) {
        Term const *subterm = termIter.data();

        Unify u;
        if (u.match(subterm, eq->lhs)) {
          TRACE("mix", "LHS of equality " << eq->id << ", namely "
            << eq->lhs->asString() << ", unifies with subterm "
            << subterm->asString () << " of "
            << ax->Equality::asString());

          // are premises satisfied?  (is it complete to just
          // check them early like this, instead of splitting
          // cases?  I think it might be due to the term order..)
          if (satisfiedPremises(u, ax)) {
            TRACE("mix", "and the axiom premises are satisfied");

            // make the new fact: axiom RHS under substitution 'u'
            // is equal to axiom LHS under substitution 'u' but with
            // 'subterm' replaced with the fact RHS
            Term *axRHS = u.substitute(ax->rhs);
            Term *axLHSrepl = substituteWithReplacement(u, ax->lhs,
                                subterm /*slot*/, eq->rhs /*replacement*/);
            Equality *newFact = new Equality(axRHS, axLHSrepl);
            if (newFact->orient()) {
              // it's a triviality
              TRACE("mix", "generated fact is trivial: " << newFact->asString());
              delete newFact;
            }
            else {
              // don't actually add it to 'equalities' since we are
              // in the middle of iterating over that list; instead,
              // save it for later
              generated.prepend(newFact);

              TRACE("mix", "generated fact: " << newFact->asString());
            }
          }
          else {
            TRACE("mix", "but the axiom premises are NOT satisfied");
          }
        }

      } // subterm loop
    } // equality loop
  } // axiom loop

  // put all the generated facts into the main 'equalities' list
  bool ret = generated.isNotEmpty();
  equalities.concat(generated);
  return ret;
}

void Rewriter::getNontrivProperSubterms(SObjList<Term const> &dest,
                                        Term const *t)
{
  // look at subterms
  FOREACH_OBJLIST(Term, t->args, iter) {
    gnpsHelper(dest, iter.data());
  }
}

void Rewriter::gnpsHelper(SObjList<Term const> &dest, Term const *t)
{
  // ignore trivial term leaves
  if (t->args.isNotEmpty()) {
    dest.prepend(t);
    getNontrivProperSubterms(dest, t);
  }
}

// create 'u.substitute(t)', except when we get to 'slot' in 't',
// pretend it actually contained 'replacement'
Term *Rewriter::substituteWithReplacement
  (Unify const &u, Term const *t,
   Term const *slot, Term const *replacement) const
{
  // is this the replacement spot?
  if (t == slot) {
    return u.substitute(replacement);
  }

  // leaf?
  if (t->args.isEmpty()) {
    return u.substitute(t);
  }

  // clone the term with substituted children
  Term *ret = new Term(t->name);
  FOREACH_OBJLIST(Term, t->args, iter) {
    ret->args.prepend(
      substituteWithReplacement(u, iter.data(), slot, replacement));
  }
  ret->args.reverse();

  return ret;
}


void Rewriter::print(char const *title)
{
  cout << title << ":\n";
  cout << "  axioms:\n";
  FOREACH_OBJLIST(Axiom, axioms, axIter) {
    cout << "    " << axIter.data()->asString() << "\n";
  }

  cout << "  equalities:\n";
  FOREACH_OBJLIST(Equality, equalities, eqIter) {
    cout << "    " << eqIter.data()->asString() << "\n";
  }

  cout << "  disequalities:\n";
  FOREACH_OBJLIST(Equality, disequalities, diseqIter) {
    cout << "    " << diseqIter.data()->asString(true /*disEq*/) << "\n";
  }

  if (contradiction) {
    cout << "  the above set is UNSATISFIABLE\n";
  }
}


// ------------------------ main -------------------------
#include <stdlib.h>      // exit

void testCompare(char const *s1, char const *s2, Order expect1, Order expect2)
{
  Term *t1 = Term::parse(s1);
  Term *t2 = Term::parse(s2);

  Order o1 = compare(t1, t2);
  Order o2 = compare(t2, t1);

  if (o1 != expect1 || o2 != expect2) {
    cout << t1->asString() << " " << toString(o1)
         << " " << t2->asString() << "\n";
    cout << t2->asString() << " " << toString(o2)
         << " " << t1->asString() << "\n";

    cout << "UNEXPECTED: wrong comparison result\n";
    exit(2);
  }

  delete t1;
  delete t2;
}


Equality *makeEq(char const *lhs, char const *rhs)
{
  Equality *e = new Equality(Term::parse(lhs), Term::parse(rhs));
  bool triv = e->orient();
  xassert(!triv);
  return e;
}

Axiom *makeAx(char const *lhs, char const *rhs)
{
  Axiom *e = new Axiom(Term::parse(lhs), Term::parse(rhs));
  bool triv = e->orient();
  xassert(!triv);
  return e;
}


int main(int argc, char *argv[])
{
  TRACE_ARGS();

  if (tracingSys("all")) {
    traceAddSys("reduce");
    traceAddSys("reduceTerm");
    traceAddSys("axiom");
    traceAddSys("mix");
  }

  testCompare("(f x)", "(f x)", EQ, EQ);
  testCompare("(f x)", "(f y)", LT, NLE);
  testCompare("(f (f x))", "(f y)", NLE, LT);

  testCompare("(f X)", "(f X)", EQ, EQ);
  testCompare("(f (f X))", "(f X)", NLE, LT);

  testCompare("(f X)", "(f Y)", NLE, NLE);
  testCompare("(f (f X))", "(f Y)", NLE, NLE);
  testCompare("(f X)", "(f y)", NLE, NLE);
  testCompare("(f (f X))", "(f y)", NLE, LT);

  testCompare("(f X)", "(g X)", LT, NLE);
  testCompare("(f (f X))", "(g X)", NLE, LT);
  testCompare("(f (f X))", "(f (g X))", LT, NLE);
  testCompare("(f (g X))", "(f (g X))", EQ, EQ);
  testCompare("(g (g X))", "(f (g X))", NLE, LT);
  testCompare("(g (f X))", "(g (g X))", LT, NLE);


  // injective list
  if (1) {
    Rewriter r;

    // P.next!=0 ==> P.next.prev==P
    Axiom *ax = makeAx("(prev (next P))", "P");
    ax->negatives.append(makeEq("(next P)", "0"));
    r.axioms.append(ax);

    r.disequalities.append(makeEq("x", "y"));
    r.disequalities.append(makeEq("(next x)", "0"));
    r.equalities.append(makeEq("(next x)", "(next y)"));

    // this fact is generated by mixing
    //r.equalities.append(makeEq("(prev (next y))", "y"));

    r.print("before");

    r.mix();
    r.print("after mix");

    r.reduce();
    r.print("after reduce 1");

    r.reduce();
    r.print("after reduce 2");
  }


  // satisfaction via normalization vs. contradiction
  if (0) {
    Rewriter r;

    // forall(p).  p.a != 3  ==>  p.b == 4
    Axiom *ax1 = makeAx("(b P)", "4");
    ax1->negatives.append(makeEq("(a P)", "3"));
    r.axioms.append(ax1);

    // forall(p).  p.b != 5  ==>  p.c == 6
    Axiom *ax2 = makeAx("(c P)", "6");
    ax2->negatives.append(makeEq("(b P)", "5"));
    r.axioms.append(ax2);

    r.equalities.append(makeEq("(a x)", "2"));      // x.a == 2
    r.disequalities.append(makeEq("2", "3"));       // 2 != 3
    r.disequalities.append(makeEq("4", "5"));       // 4 != 5
    r.disequalities.append(makeEq("(c x)", "6"));   // x.c != 6 (negated goal)

    r.print("before");

    r.mix();
    r.print("after mix");

    r.reduce();
    r.print("after reduce 1");

    r.reduce();
    r.print("after reduce 2");
  }


  // need contrapositives
  if (0) {
    Rewriter r;

    // forall(p).  p.a != 3  ==>  p.b == 4
    Axiom *ax1 = makeAx("(b P)", "4");
    ax1->negatives.append(makeEq("(a P)", "3"));
    r.axioms.append(ax1);

    // forall(p).  p.a == 3  ==>  p.c == 5
    Axiom *ax2 = makeAx("(c P)", "5");
    ax2->positives.append(makeEq("(a P)", "3"));
    r.axioms.append(ax2);

    r.disequalities.append(makeEq("(b x)", "4"));     // b.x != 4
    r.disequalities.append(makeEq("(c x)", "5"));     // c.x != 5

    r.print("before");

    r.mix();
    r.print("after mix");

    r.reduce();
    r.print("after reduce 1");

    r.reduce();
    r.print("after reduce 2");
  }



  cout << "seems ok\n";
  return 0;
}


// EOF
