// tocnf.cc
// code for tocnf.h

#include "tocnf.h"        // this module


CNFStack::CNFStack(StringTable &s)
  : str_AND(s("AND")),
    str_OR(s("OR")),
    str_IMPLIES(s("IMPLIES")),
    str_NOT(s("NOT")),
    str_FORALL(s("FORALL")),
    str_PATS(s("PATS")),
    str_EQ(s("EQ")),
    str_NEQ(s("NEQ")),
    str_FALSE(s("FALSE")),
    str_TRUE(s("TRUE")),
    alpha(),
    alphaInv(),
    hypotheses(),
    str(s),
    facts()
{}

CNFStack::~CNFStack()
{}


void CNFStack::convert(Term const *formula, bool sense)
{
  if (sense) {
    convertPos(formula);
  }
  else {
    convertNeg(formula);
  }
}

void CNFStack::convertPos(Term const *formula)
{
  StringRef n = formula->name;

  if (n == str_FALSE) {
    xfailure("todo");
  }

  else if (n == str_TRUE) {
    // ignore
  }

  else if (n == str_AND) {
    FOREACH_OBJLIST(Term, formula->args, iter) {
      convertPos(iter.data());
    }
  }

  else if (n == str_OR) {
    //            A || (C && D)
    //        --> (A || C) && (A || D)

    //            (A && B) || (C && D)
    //        --> ((A && B) || C) && ((A && B) || D)
    //        --> ((A || C) && (B || C)) && ((A || D) && (B || D))
    //        --> (A || C) && (B || C) && (A || D) && (B || D)

    // entries from earlier disjuncts
    ObjArrayStack<Term> earlier;

    FOREACH_OBJLIST(Term, formula->args, iter) {
      // push a bunch of conjuncts from this disjunct
      int begin = facts.length();
      convertPos(iter.data());

      if (false was among the conjuncts just added) {
        // ignore this disjunct
        facts.deleteTopSeveral(end-begin);
        continue;
      }

      // pull the new entries out
      ObjArrayStack<Term> current;
      while (facts.length() > begin) {
        current.push(facts.pop());
      }
    
      if (earlier.isNotEmpty()) {
        // multiply 'earlier' by 'current', pushing results
        multiply(earlier, current);

        // throw away sources
        earlier.deleteAll();
        current.deleteAll();

        // put product into 'earlier'
        while (facts.length() > begin) {
          earlier.push(facts.pop());
        }
      }
      else {
        // just move results into 'earlier'
        earlier.swapWith(current);
      }
    }
    
    // push accumulated facts
    while (earlier.isNotEmpty()) {
      facts.push(earlier.pop());
    }
  }

  else if (n == str_IMPLIES) {
    checkNumArgs(formula, 2);

    // synthesize an equivalent form:
    //   P ==> Q      becomes     !P || Q
    Term *t = new Term(str_OR);
    Term *n = new Term(str_NOT);
    n->args.append(formula->args.nth(0));      // !P
    t->args.append(n);                         // !P ||
    t->args.append(formula->args.nth(1));      // !P || Q

    // convert the equivalent form
    convertPos(t);

    // now dismantle it so I can deallocate it without disturbing
    // the underlying terms
    t->args.removeFirst();
    t->args.removeFirst();
    delete t;
    n->args.removeFirst();
    delete n;
  }

  else if (n == str_NOT) {
    convertNeg(formula);
  }

  else if (n == str_FORALL) {
    // alpha map the quantification variable

    // remember the stack top

    // convert the body to CNF

    // replace all the constructed disjunctions with quantified
    // disjunction (this postpones the issue of picking one of
    // the literals as the consequence)
  }

  else if (n == str_EQ) {
    checkNumArgs(formula, 2);
    facts.push(convertTerm(formula));
  }

  else if (n == str_NEQ) {
    checkNumArgs(formula, 2);
    facts.push(convertTerm(formula));
  }

  else {
    xformat(stringc << "unknown connective or predicate: " << n);
  }
}

void CNFStack::convertNeg(Term const *formula)
{
  StringRef n = formula->name;

  if (n == str_FALSE) {
    xfailure("todo");
  }
  else if (n == str_TRUE) {
  }
  if (n == str_AND) {

  }
  else if (n == str_OR) {
  }
  else if (n == str_IMPLIES) {
  }
  else if (n == str_NOT) {
  }
  else if (n == str_FORALL) {
  }
  else if (n == str_EQ) {
  }
  else if (n == str_NEQ) {
  }
  else {
    xformat(stringc << "unknown connective or predicate: " << n);
  }
}


void CNFStack::checkNumArgs(Term const *formula, int ct)
{
  if (formula->args.count() != ct) {
    xformat(stringc << "IMPLIES must have exactly " << ct
                    << " " << plural(ct, "argument")
                    << ", not " << formula->args.count());
  }
}


void CNFStack::multiply(ObjArrayStack<Term> const &left,
                        ObjArrayStack<Term> const &right)
{
  for (int i=0; i < left.length(); i++) {
    for (int j=0; j < right.length(); j++) {
      pushDisjoin(left[i], right[i]);
    }
  }
}

void CNFStack::pushDisjoin(Term const *left, Term const *right)
{
  Term *t = new Term(str_OR);

  if (left->name != str_OR) {
    t->args.prepend(left->clone());
  }
  else {
    FOREACH_OBJLIST(Term, left->args, iter) {
      t->args.prepend(iter.data()->clone());
    }
  }

  if (right->name != str_OR) {
    t->args.prepend(right->clone());
  }
  else {
    FOREACH_OBJLIST(Term, right->args, iter) {
      t->args.prepend(iter.data()->clone());
    }
  }

  t->args.reverse();
  facts.push(t);
}


// EOF
