// tovc.cc
// implementation of Expression::tovc(), declared in vcgen.ast

#include "cc.ast.gen.h"      // C++ AST
#include "trace.h"           // tracingSys


// -------------------- Macro -------------------
string astTypeIdList_tovc(FakeList<ASTTypeId> *vars)
{
  stringBuilder sb;
  sb << "(";

  int ct=0;
  FAKELIST_FOREACH(ASTTypeId, vars, iter) {
    if (ct++ > 0) {
      sb << ", ";
    }
    // might not have been tcheck'd, since we sometimes get this from a clone() ...
    sb << iter->decl->decl->getDeclaratorId()->getName();
  }

  sb << ")";
  return sb;
}


string Macro::tovc() const
{
  stringBuilder sb;
  sb << "define " << name.str;
  if (!params) {
    sb << " = " << body->tovc() << ";\n";
  }
  else {
    sb << astTypeIdList_tovc(params) << " {\n";
    sb << body->tovcConjunctiveLines("  ");
    sb << "}\n\n";
  }
  return sb;
}


// ------------------ BindingSeq --------------
#if 0
string BindingSeq::tovc() const
{
  stringBuilder sb;

  sb << "(";
  int ct=0;
  FOREACH_ASTLIST(Binding, names, iter) {
    if (ct++ > 0) {
      sb << ", ";
    }
    sb << iter.data()->name.str;
  }
  sb << ")";

  return sb;
}
#endif // 0


// ------------------- Expression -----------------
string E_boolLit::tovc() const
{
  return string(b? "true" : "false");
}

string E_intLit::tovc() const
{
  return text;
}

string E_floatLit::tovc() const
{
  return text;
}

string E_stringLit::tovc() const
{
  return text;
}

string E_charLit::tovc() const
{
  return text;
}

string E_this::tovc() const
{
  return "this";
}

//string E_structLit::tovc() const

string E_variable::tovc() const
{
  return name->getName();
}

string E_funCall::tovc() const
{
  #define ARG(n) (args->nth(n)->expr->tovc())

  StringRef name = func->asE_variable()->name->getName();
  
  // write some common forms in a more compact, readable syntax
  if (0==strcmp(name, "updrange")) {
    return stringc
      << ARG(0) << "{" << ARG(1) << ",[" << ARG(2) << " ... " << ARG(3)
      << "] := " << ARG(4) << "}";
  }
  if (0==strcmp(name, "upd")) {
    xassert(args->count() == 3);
    if (args->nth(1)->expr->isE_funCall()) {
      E_funCall *arg1 = args->nth(1)->expr->asE_funCall();
      StringRef arg1name = arg1->func->asE_variable()->name->getName();
      if (0==strcmp(arg1name, "ptr")) {
        string arg1exp0 = arg1->args->nth(0)->expr->tovc();
        string arg1exp1 = arg1->args->nth(1)->expr->tovc();
        return stringc
          << ARG(0) << "{" << arg1exp0 << "," << arg1exp1 << " := " << ARG(2) << "}";
      }
    }
    return stringc
      << ARG(0) << "{" << ARG(1) << " := " << ARG(2) << "}";
  }
  if (0==strcmp(name, "sel")) {
    xassert(args->count() == 2);
    if (args->nth(1)->expr->isE_funCall()) {
      E_funCall *arg1 = args->nth(1)->expr->asE_funCall();
      StringRef arg1name = arg1->func->asE_variable()->name->getName();
      if (0==strcmp(arg1name, "ptr")) {
        string arg1exp0 = arg1->args->nth(0)->expr->tovc();
        string arg1exp1 = arg1->args->nth(1)->expr->tovc();
        return stringc
          << ARG(0) << "[" << arg1exp0 << "," << arg1exp1 << "]";
      }
    }
    return stringc
      << ARG(0) << "[" << ARG(1) << "]";
  }

  // general format: named function call
  stringBuilder sb;
  sb << name << "(";

  int ct=0;
  FAKELIST_FOREACH(ArgExpression, args, a) {
    if (ct++ > 0) {
      sb << ", ";
    }
    sb << a->expr->tovc();
  }

  sb << ")";
  return sb;
}


// "illegal" means this construct can't be generated by the vcgen
// process, so we should never be asked to convert it to a textual VC
string E_constructor::tovc() const
  { return "illegal"; }
string E_fieldAcc::tovc() const
  { return "illegal_E_fieldAcc"; }
string E_sizeof::tovc() const
  { return "illegal"; }


string E_unary::tovc() const
{
  return stringc << toString(op) << "(" << expr->tovc() << ")";
}


string E_effect::tovc() const
  { return "illegal"; }


string E_binary::tovc() const
{
  char const *opstr = toString(op);
  if (op == BIN_MINUS) {
    opstr = " - ";   // work around quirk in vclang
  }
  return stringc << "(" << e1->tovc() << opstr << e2->tovc() << ")";
}


string E_addrOf::tovc() const
  { return "illegal"; }
string E_deref::tovc() const
  { return "illegal"; }
string E_cast::tovc() const
  { return "illegal"; }


string E_cond::tovc() const
  { return "illegal"; }


//string E_gnuCond::tovc() const


string E_sizeofType::tovc() const
  { return "illegal"; }
string E_assign::tovc() const
  { return "illegal"; }
string E_new::tovc() const
  { return "illegal"; }
string E_delete::tovc() const
  { return "illegal"; }
string E_throw::tovc() const
  { return "illegal"; }
string E_keywordCast::tovc() const
  { return "illegal"; }
string E_typeidExpr::tovc() const
  { return "illegal"; }
string E_typeidType::tovc() const
  { return "illegal"; }


string E_grouping::tovc() const
{
  return stringc << " ( " << expr->tovc() << " ) ";
}


string E_arrow::tovc() const
  { return "illegal"; }


string E_quantifier::tovc() const
{
  stringBuilder sb;
  // quantifier
  sb << "(" << (isForall? "forall" : "exists");

  // bound variables
  sb << astTypeIdList_tovc(vars);

  static bool enablePats = !tracingSys("disablePatterns");

  // pattern
  if (patterns && enablePats) {
    sb << " " << patterns->tovc();
  }

  // body
  sb << ". " << body->tovc();

  sb << ")";

  return sb;
}


string E_let::tovc() const
{
  return stringc
    << "(let " << name.str << " = " << binding->tovc()
    << " in " << body->tovc() << ")";
}


string E_pre::tovc() const
  { return "illegal"; }
string E_vc_expression::tovc() const
  { return "illegal"; }


string Expression::tovcConjunctiveLines(char const *indent) const
{
  if (!isConjunction()) {
    // emit it as a single line
    return stringc << indent << tovc() << ";\n";
  }
  else {
    // break it, and print recursively
    E_binary const *b = asE_binaryC();
    return b->e1->tovcConjunctiveLines(indent) &
           b->e2->tovcConjunctiveLines(indent);
  }
}



// EOF
