// vc_env.cc
// code for vc_env.h

#include "vc_env.h"      // this module
#include "astbuild.h"    // apply*
#include "exttype.h"     // ext
#include "strutil.h"     // replace

#include <stdio.h>       // fprintf
#include <stdlib.h>      // exit


VCEnv::VCEnv(TypeFactory &t, StringTable &s, rostring outFname)
  : tfac(t),
    str(s),
    output(toCStr(outFname), "w"),
    memVar(t.makeVariable(SL_UNKNOWN, s("mem"),
                          t.getSimpleType(ST_INT),   // arbitrary
                          DF_NONE, NULL /*tunit*/)),
    tempCt(0),
    typeTagCt(1),
    #define STR(x) str_##x(s(#x))
    STR(malloc),
    STR(memset),
    STR(memmove),
    STR(memcpy),
    STR(free),
    STR(kfree),
    STR(metaVar),
    STR(fresh_value),
    STR(CLT_between),
    #undef STR
    globals(),
    currentFunc(NULL),
    result(NULL),
    params(),
    locals(),
    // how does one initialize array members?
    lastStmtLoc(SL_UNKNOWN),
    inPredicate(false),
    inPreState(false)
{
  globalsAndParams[0] = &globals;
  globalsAndParams[1] = &params;

  globalsAndParamsAndLocals[0] = &globals;
  globalsAndParamsAndLocals[1] = &params;
  globalsAndParamsAndLocals[2] = &locals;
}

VCEnv::~VCEnv()
{
  delete memVar;
}


void VCEnv::beginFunc(Function *f)
{
  xassert(inPredicate == false);

  currentFunc = f;

  // create a variable to stand for the returned value
  result = tfac.makeVariable(f->body->loc, str("result"),
                             f->funcType->retType, DF_NONE, NULL /*tunit*/);

  xassert(params.isEmpty());
  xassert(locals.isEmpty());
}

void VCEnv::endFunc()
{
  xassert(inPredicate == false);

  currentFunc = NULL;

  delete result;
  result = NULL;

  params.removeAll();
  locals.removeAll();
}


Expression *VCEnv::todo()
{
  // hopefully stands out in debug output
  return makeVar("VCGEN_TODO");
}


E_intLit * owner VCEnv::intLit(int i)
{
  return new E_intLit(str(stringc << i));
}


E_variable * owner VCEnv::makeVar(rostring name)
{
  E_variable *ret = new E_variable(new PQ_name(SL_UNKNOWN, str(name)));
  
  // hack: sometimes I'm constructing these things and sending them
  // to vcgen without typechecking them first...
  ret->type = tfac.getSimpleType(ST_INT);

  return ret;
}


E_variable * owner VCEnv::makePVar(rostring name)
{
  if (inPreState) {
    return makeVar(stringc << name << "_pre");
  }
  else {
    return makeVar(name);
  }
}


void VCEnv::assume(Expression * owner pred, rostring why)
{
  if (why[0]) {
    *this << "    // " << why << "\n";
  }

  char const *ind = (why[0]? "    " : "  ");   // bit of a hack

  if (!pred->isConjunction()) {
    // compact form
    *this << ind << "assume " << pred->tovc() << ";\n";
  }
  else {
    // elaborated form
    *this << ind << "assume {\n";
    printConjunctiveLines(stringc << ind << "  ", pred);
    *this << ind << "}\n";
  }

  delete pred;
}


void VCEnv::assumeNot(Expression * owner pred, rostring why)
{
  *this << "    // " << why << "\n";
  *this << "    assume !( " << pred->tovc() << " );\n";
  delete pred;
}


void VCEnv::assumeSense(Expression * owner pred, bool sense,
                        rostring why)
{
  if (sense) {
    assume(pred, why);
  }
  else {
    assumeNot(pred, why);
  }
}


void VCEnv::assign(rostring name, Expression * serf target)
{
  *this << "    " << name << " := " << target->tovc() << ";\n";
}

void VCEnv::assign(Variable *var, Expression * serf target)
{
  assign(var->name, target);
}

void VCEnv::assignLive(Variable *var, Expression * serf target)
{
  assign(var, target);
  if (var->type->isPointerType()) {
    assignDel(stringc << var->name << "_meta", makeVar("live"));
  }
}


void VCEnv::prove(Expression * owner pred, rostring why)
{
  proveOrLemma("prove", pred, why);
}

void VCEnv::lemma(Expression * owner pred, rostring why)
{
  proveOrLemma("lemma", pred, why);
}

void VCEnv::proveOrLemma(rostring which, Expression * owner pred, rostring why)
{
  *this << "    " << which << " \"" << why << "\"";
  if (!pred->isConjunction()) {
    // emit a compact form
    *this << "\n"
          << "      " << pred->tovc() << ";\n";
  }
  else {
    // emit a more elaborated form, with each conjunct on its own line
    *this << " {\n";
    printConjunctiveLines("      ", pred);
    *this << "    }\n";
  }
  delete pred;
}


void VCEnv::printConjunctiveLines(rostring indent, Expression *pred)
{
  *this << pred->tovcConjunctiveLines(indent.c_str());
}


void VCEnv::discard(Expression * owner expr)
{
  delete expr;
}


void VCEnv::define(rostring name, Expression * owner val, rostring why)
{                                                                 
  *this << "    define " << name << " = " << val->tovc() << ";"
        << "      // " << why << "\n";
  delete val;
}


FakeList<ArgExpression> *VCEnv::globalVarList()
{
  FakeList<ArgExpression> *list = FakeList<ArgExpression>::emptyList();

  SFOREACH_OBJLIST(Variable, globals, iter) {
    Variable const *v = iter.data();
    list = list->prepend(new ArgExpression(makeVar(v->name)));
    if (v->type->isPointerType()) {
      list = list->prepend(new ArgExpression(makeVar(stringc << v->name << "_meta")));
    }
  }

  return list->reverse();
}

string VCEnv::globalVarNameList(bool preState) const
{
  stringBuilder sb;

  char const *suffix = (preState? "_pre" : "");

  int ct=0;
  SFOREACH_OBJLIST(Variable, globals, iter) {
    Variable const *v = iter.data();
    if (ct++ > 0) {
      sb << ", ";
    }
    sb << v->name << suffix;
    if (v->type->isPointerType()) {
      sb << ", " << v->name << "_meta" << suffix;
    }
  }

  return sb;
}


string VCEnv::fieldSlotName(Variable const *field) const
{
  xassert(field->scope && field->scope->curCompound);
  return extC(field)->modelVar->name;
}


string VCEnv::maybePre(rostring name) const
{
  if (inPreState) {
    return stringc << name << "_pre";
  }
  else {
    return string(name);
  }
}


string VCEnv::hasClassTag(ECompoundType *ct, rostring objIndex) const
{
  stringBuilder sb;
                                                          
  // base case
  sb << "tag[" << objIndex << "]==" << ct->name << "_tag";
  
  SFOREACH_OBJLIST_NC(ECompoundType, ct->subclasses, iter) {
    // recursive case
    sb << " || " << hasClassTag(iter.data(), objIndex);
  }
  
  return sb;
}


Expression *VCEnv::ptr(Expression *addr, Expression *field)
{
  return apply2(makeVar("ptr"), addr, field);
}

Expression *VCEnv::upd(Expression *mem, Expression *ptr, Expression *value)
{
  return apply3(makeVar("upd"), mem, ptr, value);
}

Expression *VCEnv::updrange(Expression *mem, Expression *addr, Expression *low,
                            Expression *high, Expression *value)
{
  return apply5(makeVar("updrange"), mem, addr, low, high, value);
}


void VCEnv::out(rostring str)
{
  string tmp = replace(str, "$L", toLCString(lastStmtLoc));
  fprintf(output, "%s", tmp.c_str());
}


void VCEnv::error(rostring msg)
{
  error(lastStmtLoc, msg);
}

void VCEnv::error(SourceLoc loc, rostring msg)
{
  if (loc != SL_UNKNOWN) {
    fprintf(stderr, "%s: error: %s\n", toString(loc).c_str(), msg.c_str());
  }
  else {
    fprintf(stderr, "error: %s\n", msg.c_str());
  }
  exit(2);
}


void VCEnv::warning(rostring msg)
{
  warning(lastStmtLoc, msg);
}

void VCEnv::warning(SourceLoc loc, rostring msg)
{
  if (loc != SL_UNKNOWN) {
    fprintf(stderr, "%s: warning: %s\n", toString(loc).c_str(), msg.c_str());
  }
  else {
    fprintf(stderr, "warning: %s\n", msg.c_str());
  }
}


// EOF
