// vml_fflow.cc
// code for vml_fflow.h

#define VML_FFLOW_IMPLEMENTATION

#include "vml_fflow.h"      // this module
#include "vml_pred.h"       // equivalent, etc.
#include "vml_build.h"      // p_equal, etc.
#include "trace.h"          // TRACE
#include "autofile.h"       // AutoFILE

// "fflowd" = "fact flow detail"
#define FFLOWD "fflowd"
#define DETAIL(msg) TRACE(FFLOWD, msg)

// "fflow" = "fact flow"; this is used for final results
#define FFLOW "fflow"


// ----------------------- StmtFacts -----------------------
StmtFacts::StmtFacts(VStatement *s, StringTable &str)
  : stmt(s),
    facts(),
    genPreds(),
    negIfPred(NULL)
{
  computeGenPreds(str);
}

StmtFacts::~StmtFacts()
{
  if (negIfPred) {
    delete negIfPred;
  }
}


Pred *p_varEquals(Identifier *id, VExpr *e)
{
  return p_equal(new VE_var(id->clone()),
                 e->clone());
}


void detailPrintPreds(ObjList<Pred> const &list, SourceLoc loc,
                      char const *context)
{
  #ifndef NDEBUG
    if (tracingSys(FFLOWD)) {
      if (list.isNotEmpty()) {
        stringBuilder sb;

        sb << context << " at " << toLCString(loc) << ":\n";
        FOREACH_OBJLIST(Pred, list, iter) {
          sb << "  " << iter.data()->print() << "\n";
        }
        trace(FFLOWD) << sb;
      }
      else {
        trace(FFLOWD) << context << " at " << toLCString(loc) << ": empty\n";
      }
    }
  #endif // !NDEBUG
}

void detailPrintPreds(SObjList<Pred const> const &list, SourceLoc loc,
                      char const *context)
{
  // cast is ugly but safe as long as ObjList and SObjList have same
  // layout.. in modern C++ I'd use the template< template ..>
  // construction to write detailPrintPreds, but I want this code to
  // work with gcc-2
  detailPrintPreds(reinterpret_cast<ObjList<Pred> const &>(list),
                   loc, context);
}


void StmtFacts::computeGenPreds(StringTable &str)
{
  ASTSWITCHC(VStatement, stmt) {
    ASTCASEC(VS_decl, d) {
      if (d->d->isVD_var()) {
        VD_var const *v = d->d->asVD_varC();

        // note that we kill the old value here as well, but that is
        // implemented below

        if (v->init) {
          // a declaration with an initializer: gives rise to the
          // knowledge that the variable is equal to this value
          genPreds.append(p_varEquals(v->id, v->init));
        }
      }
    }

    ASTNEXTC(VS_assign, a) {
      // similar to VS_decl, except we need to make sure the
      // assigning expression does not refer to the target variable
      if (!refersTo(a->source, a->target->name)) {
        genPreds.append(p_varEquals(a->target, a->source));
      }

      // from the fact "ARR := ARR{i := v}", generate "ARR[i] == v"
      if (a->source->isVE_apply()) {
        VE_apply *srcApp = a->source->asVE_apply();
        if (0==strcmp(srcApp->func->name, "upd")) {
          VExpr *arr = srcApp->args.nth(0);
          VExpr *i = srcApp->args.nth(1);
          VExpr *v = srcApp->args.nth(2);

          if (arr->isVE_var() &&
              arr->asVE_var()->id->var == a->target->var) {
            genPreds.append(p_equal(apply2(makeId(str("sel")),
                                           arr->clone(),
                                           i->clone()),
                                    v->clone()));
          }
        }
      }
    }

    // TODO: look at the postcondition of a VS_procCall and introduce
    // those facts, but with proper substitution of call site args

    ASTNEXTC(VS_if, i) {
      // this is not a genPred, but it's as good a place as any
      // to create it
      if (i->p->isP_not()) {
        // negative form will be found simply by digging inside
        // the P_not of the guard
      }
      else {
        // wrap a P_not around the guard
        negIfPred = p_not(i->p->clone());
      }
    }

    ASTNEXTC(VS_pred, p) {
      if (p->psk != PS_PURE_ASSERT) {
        // in all cases, add the predicate; in the case of PS_INVARIANT
        // this is a little weaker than possible, since the global
        // invariant is implicitly conjoined; at some point I might
        // strengthen this to include the global invariant
        genPreds.append(p->p->clone());
      }
    }

    ASTENDCASECD
  }

  // debug-print the genPreds
  detailPrintPreds(genPreds, stmt->loc, "genPred");
}


// return negated form of the 'if' guard predicate
Pred const *StmtFacts::getNegIfPred()
{                          
  // this function only makes sense for 'if' statements
  xassert(stmt->isVS_if());
  VS_if *i = stmt->asVS_if();
  
  if (i->p->isP_not()) {
    // dig inside the P_not
    return i->p->asP_not()->p;
  }
  else {
    // get the wrapped form previously computed
    xassert(negIfPred);
    return negIfPred;
  }
}


// ----------------------- VMLFFlow ------------------------
VMLFFlow::VMLFFlow(StringTable &s)
  : str(s),
    map(VMLFFlow::getKey, 
        HashTable::lcprngHashFn, 
        HashTable::pointerEqualKeyFn)
{}

VMLFFlow::~VMLFFlow()
{}


// for 'map' hash table
STATICDEF void const *VMLFFlow::getKey(StmtFacts *data)
{
  return data->stmt;
}


void VMLFFlow::reset()
{
  map.empty();
  worklist.empty();
}


// DFS from 'node'; 'seen' is those nodes either currently being
// considered somewhere in the call chain ("gray"), or else finished
// entirely ("black")
void VMLFFlow::rp_dfs(SObjSet<VStatement*> &seen, VStatement *stmt)
{
  // we're now considering 'node'
  seen.add(stmt);

  // consider each of this node's successors
  ArrayStack<VStatement*> successors(2);
  getSuccessors(successors, stmt);
  for (int i=0; i < successors.length(); i++) {
    VStatement *succ = successors[i];
    xassert(succ);

    if (seen.contains(succ)) {
      // we're already considering, or have already considered, this node;
      // do nothing with it
    }
    else {
      // visit this new child
      rp_dfs(seen, succ);
    }
  }

  // since we're finished with this node, we append it to compute the
  // postorder
  worklist.enqueue(stmt);
}

void VMLFFlow::reversePostorder(VProcedure *proc)
{
  xassert(worklist.isEmpty());

  // DFS from the procedure start, computing the spanning tree implicitly,
  // and the postorder explicitly
  SObjSet<VStatement*> seen;
  rp_dfs(seen, proc->body);

  // reverse the list
  worklist.reverse();
}


// put the successor statements of 'stmt' into 'dest'
void VMLFFlow::getSuccessors(ArrayStack<VStatement*> &dest, VStatement *stmt)
{
  // I guess it's a bit heaviweight to put the results into an array
  // when it's always either 0, 1 or 2 elements.. oh well.

  // Decode successor according to comment in vml_tcheck.ast that
  // explains semantics of 'VStatement::next' field.

  if (stmt->isVS_if()) {
    VS_if *i = stmt->asVS_if();
    dest.push(i->sThen);
    if (i->sElse) {
      dest.push(i->sElse);
      return;
    }
    else {
      // fall through and push 'stmt->next', if it is non-NULL
    }
  }

  if (stmt->next) {
    dest.push(stmt->next);
  }
  else {
    // last statement in procedure, nothing to do
  }
}


// enqueue *if* it isn't already present
void VMLFFlow::enqueueToWorklist(VStatement *stmt)
{
  if (!worklist.contains(stmt)) {
    worklist.enqueue(stmt);
  }
}


bool predInSet(PredSet &set, Pred const *p)
{
  SFOREACH_OBJLIST(Pred const, set, iter) {
    if (equivalent(iter.data(), p)) {
      return true;
    }
  }
  return false;         // not equivalent to any in 'set'
}


// conjoin the existing set of facts with another fact; if 'f'
// is a conjunction, it will be broken down as needed
void VMLFFlow::addFact(PredSet &dest, Pred const *f)
{
  // recursively break down f's conjunctive structure
  if (f->isBinary(BP_AND)) {
    P_binary const *b = f->asP_binaryC();
    addFact(dest, b->p1);
    addFact(dest, b->p2);
    return;
  }

  // check to see if the given predicate is already among those in 'dest'
  if (predInSet(dest, f)) {
    return;    // equivalent predicate already present
  }

  // add it
  dest.append(f);
}


// intersect 'dest' and 'src', putting the result into 'dest'; return
// true if the result is different that 'dest' was originally
bool VMLFFlow::intersectFacts(PredSet &dest, PredSet &src)
{ 
  bool ret = false;

  SObjListMutator<Pred const> mut(dest);
  while (!mut.isDone()) {
    if (predInSet(src, mut.data())) {
      mut.adv();     // keep it
    }
    else {
      mut.remove();  // remove it
      ret = true;
    }
  }

  return ret;
}


// remove from 'facts' any that refer to 'var'
void VMLFFlow::killReferencesTo(PredSet &facts, StringRef var)
{
  SObjListMutator<Pred const> mut(facts);
  while (!mut.isDone()) {
    if (refersTo(mut.data(), var)) {
      mut.remove();        // remove
    }
    else {
      mut.adv();           // keep
    }
  }
}


// given 'stmtFacts' the set of facts at the beginning of a statement,
// compute the effects of flowing through that statement, so we arrive
// at the set of facts at the end of the statement
void VMLFFlow::computeOutFacts(PredSet &outFacts, StmtFacts *stmtFacts)
{          
  // initialize 'outFacts' to be the same as the "in" facts
  outFacts = stmtFacts->facts;                               // propagate

  // kill or modify facts based on the statement's effects; the
  // genPreds are then added after this 'switch'
  ASTSWITCHC(VStatement, stmtFacts->stmt) {
    ASTCASEC(VS_decl, d)
      if (d->d->isVD_var()) {
        VD_var const *v = d->d->asVD_varC();
        killReferencesTo(outFacts, v->id->name);             // kill
      }

    ASTNEXTC(VS_assign, a)
      killReferencesTo(outFacts, a->target->name);           // kill

    ASTNEXTC(VS_procCall, p)
      // kill variables updated in return
      FOREACH_ASTLIST(Identifier, p->targets, iter) {
        killReferencesTo(outFacts, iter.data()->name);       // kill
      }

      // kill variables modified as side-effects of the procedure call
      FOREACH_ASTLIST(Identifier, p->proc->writes, iter2) {
        killReferencesTo(outFacts, iter2.data()->name);      // kill
      }

    ASTENDCASECD
  }

  // now add the genPreds
  FOREACH_OBJLIST(Pred, stmtFacts->genPreds, iter) {
    addFact(outFacts, iter.data());                          // gen
  }

  detailPrintPreds(outFacts, stmtFacts->stmt->loc, "outFacts");
}


// add to 'facts' the consequences of control flowing from 'start' to 'end'
void VMLFFlow::addEdgeFacts(PredSet &facts, StmtFacts *startFacts, VStatement *end)
{
  VStatement *start = startFacts->stmt;
  if (start->isVS_if()) {
    VS_if *i = start->asVS_if();
    if (end == i->sThen) {
      addFact(facts, i->p);                         // positive form of guard
    }
    else {
      addFact(facts, startFacts->getNegIfPred());   // negative form of guard
    }
  }

  detailPrintPreds(facts, start->loc, 
                   stringc << "edge facts to " << toLCString(end->loc));
}


void VMLFFlow::analyze(VProcedure *proc)
{
  // Initially, 'map' does not map any node, which means they are all
  // initialized to the universe of all predicates.  Then, we will
  // iteratively approach the fixpoint closest to that universe.  This
  // ensures that in cases where there is more than one solution to
  // the nominal set of dataflow equations (as happens in loops, for
  // example) that our computed fact sets will be as
  // informative/useful as possible.
  reset();

  // initialize worklist with a reverse postorder enumeration
  reversePostorder(proc);
                       
  // debug-print the initial worklist
  #ifndef NDEBUG
    if (tracingSys(FFLOWD)) {
      stringBuilder sb;
      sb << "initial worklist:";
      for (int i=0; i < worklist.length(); i++) {
        sb << " " << toLCString(worklist[i]->loc);
      }
      trace(FFLOWD) << sb << "\n";
    }
  #endif // !NDEBUG

  // inititialize the start node with the precondition
  {
    VStatement *start = proc->body;
    StmtFacts *startFacts = new StmtFacts(start, str);
    addFact(startFacts->facts, proc->pre);
    map.add(start, startFacts);
  }

  while (worklist.isNotEmpty()) {
    // dequeue next statement to process
    VStatement *stmt = worklist.dequeue();
    DETAIL("processing from worklist: " << toLCString(stmt->loc));

    // retrieve associated dataflow info
    StmtFacts *stmtFacts = map.get(stmt);

    // reverse postorder should guarantee that no node gets visited
    // before at least one of its predecessors is visited, which will
    // create a NodeFacts if necessary
    xassert(stmtFacts);

    // compute Out(stmt):
    //   - remove any facts jeopardized by state changes in stmt
    //   - add facts asserted, assumed, or invariant'd in stmt
    PredSet outFacts;
    computeOutFacts(outFacts, stmtFacts);

    // consider each successor
    ArrayStack<VStatement*> successors;
    getSuccessors(successors, stmt);
    for (int i=0; i < successors.length(); i++) {
      VStatement *succ = successors[i];

      // add facts implied by the path used to exit the stmt
      PredSet succInFacts(outFacts);
      addEdgeFacts(succInFacts, stmtFacts, succ);
      
      // we will copy 'succInFacts' into the successor's StmtFacts
      // structure; does it have a StmtFacts yet?
      StmtFacts *succFacts = map.get(succ);
      if (!succFacts) {
        // no: create something to hold knowledge about it
        succFacts = new StmtFacts(succ, str);
        succFacts->facts = succInFacts;
        map.add(succ, succFacts);

        // the successor node has changed, but since it's the first
        // time anybody's seen it, it's already on the worklist somewhere
        xassert(worklist.contains(succ));
      }
      else {
        // yes: intersect its existing facts with 'succInFacts'
        if (intersectFacts(succFacts->facts, succInFacts)) {
          // the intersection is different from the old set, so we need
          // to reconsider it at some point
          enqueueToWorklist(succ);
          
          detailPrintPreds(succFacts->facts, succFacts->stmt->loc, 
                           "new intersection with predecessor");
        }
      }
    }
  }

  // we have reached the fixpoint; apply the computed information
  strengthenInvariants();

  // may as well eagerly reset..
  reset();
}


void VMLFFlow::strengthenInvariants()
{
  // iterate over all statements' computed info
  OwnerHashTableIter<StmtFacts> iter(map);
  for (; !iter.isDone(); iter.adv()) {
    StmtFacts *stmtFacts = iter.data();
    
    // only the VS_pred statements are of interest
    if (!stmtFacts->stmt->isVS_pred()) {
      continue;
    }
    VS_pred *vsp = stmtFacts->stmt->asVS_pred();

    // furthermore, only the invariant statements should be strengthened
    if (!vsp->isInvariant()) {
      continue;
    }

    // break down the existing invariant
    PredSet set;
    addFact(set, vsp->p);
    
    // now, any predicate in 'stmtFacts->facts' that is not also in 'set'
    // should be added to 'vsp->p' to strengthen it
    SFOREACH_OBJLIST(Pred const, stmtFacts->facts, iter2) {
      Pred const *pred = iter2.data();
      
      if (!predInSet(set, pred)) {
        // conjoin 'pred' with 'vsp->p'
        TRACE(FFLOW, "adding to inv at " << toLCString(vsp->loc) <<
                     ": " << pred->print());
        vsp->p = p_and(vsp->p, pred->clone());
      }
    }
  }
}


void vmlFFlowAnalyze(VProcedure *proc, StringTable &str)
{
  VMLFFlow f(str);
  if (proc->body) {
    f.analyze(proc);
  }
  
  char const *fflowOutput = getenv("FFLOW_OUTPUT");
  if (fflowOutput) {
    AutoFILE fp(fflowOutput, "a");
    fprintf(fp, "--------------- %s -------------\n", proc->id->name);

    BoxPrint bp;
    proc->iprint(bp);

    BPRender bpr;
    string s = bpr.takeAndRender(bp);

    fwrite(s.c_str(), s.length(), 1, fp);
    fprintf(fp, "\n");
  }
}


// EOF
