// vml_pred.cc
// code for vml_pred.h
   
#include "vml_pred.h"       // this module


// for now it is simply structural equality
bool equivalent(Pred const *p, Pred const *q)
{
  // TODO: This is a clear candidate for an AST feature that builds
  // visitors that simultaneously traverse two trees.

  if (p->kind() != q->kind()) {
    return false;
  }

  switch (p->kind()) {
    default: xfailure("bad kind");

    case Pred::P_LIT: {
      P_lit const *pl = p->asP_litC();
      P_lit const *ql = q->asP_litC();
      return pl->b == ql->b;
    }

    case Pred::P_NOT: {
      P_not const *pn = p->asP_notC();
      P_not const *qn = q->asP_notC();
      return equivalent(pn->p, qn->p);
    }

    case Pred::P_QUANT: {
      P_quant const *pq = p->asP_quantC();
      P_quant const *qq = q->asP_quantC();
      // I don't even tolerate different quantification variables ...
      return pq->universal == qq->universal &&
             equalVarLists(pq->vars, qq->vars) &&
             equivalent(pq->body, qq->body);
    }

    case Pred::P_BINARY: {
      P_binary const *pb = p->asP_binaryC();
      P_binary const *qb = q->asP_binaryC();
      return pb->con == qb->con &&
             equivalent(pb->p1, qb->p1) &&
             equivalent(pb->p2, qb->p2);
    }

    case Pred::P_COMPARE: {
      P_compare const *pc = p->asP_compareC();
      P_compare const *qc = q->asP_compareC();
      return pc->op == qc->op &&
             equalExprs(pc->e1, qc->e1) &&
             equalExprs(pc->e2, qc->e2);
    }
  }
}


bool equalVarLists(ASTList<VD_var> const &list1, ASTList<VD_var> const &list2)
{
  ASTListIter<VD_var> iter1(list1);
  ASTListIter<VD_var> iter2(list2);

  while (!iter1.isDone() && !iter2.isDone()) {
    // only the names are relevant; I ignore source locs and init exprs
    if (iter1.data()->id->name != iter2.data()->id->name) {
      return false;
    }
    
    iter1.adv();
    iter2.adv();
  }

  return iter1.isDone() == iter2.isDone();
}


bool equalExprs(VExpr const *e1, VExpr const *e2)
{
  if (e1->kind() != e2->kind()) {
    return false;
  }

  switch (e1->kind()) {
    default: xfailure("bad tag kind");

    case VExpr::VE_LIT: {
      VE_lit const *el1 = e1->asVE_litC();
      VE_lit const *el2 = e2->asVE_litC();
      return el1->n == el2->n;
    }

    case VExpr::VE_VAR: {
      VE_var const *ev1 = e1->asVE_varC();
      VE_var const *ev2 = e2->asVE_varC();
      return ev1->id->name == ev2->id->name;
    }
    
    case VExpr::VE_NEG: {
      VE_neg const *en1 = e1->asVE_negC();
      VE_neg const *en2 = e2->asVE_negC();
      return equalExprs(en1->e, en2->e);
    }
    
    case VExpr::VE_APPLY: {
      VE_apply const *ea1 = e1->asVE_applyC();
      VE_apply const *ea2 = e2->asVE_applyC();
      return ea1->func->name == ea2->func->name &&
             equalExprLists(ea1->args, ea2->args);
    }
    
    case VExpr::VE_ARITH: {
      VE_arith const *ea1 = e1->asVE_arithC();
      VE_arith const *ea2 = e2->asVE_arithC();
      return ea1->op == ea2->op &&
             equalExprs(ea1->e1, ea2->e1) &&
             equalExprs(ea1->e2, ea2->e2);
    }
    
    case VExpr::VE_PRE: {
      VE_pre const *ev1 = e1->asVE_preC();
      VE_pre const *ev2 = e2->asVE_preC();
      return equalExprs(ev1->e, ev2->e);
    }
    
    case VExpr::VE_COND: {
      VE_cond const *ev1 = e1->asVE_condC();
      VE_cond const *ev2 = e2->asVE_condC();
      return equivalent(ev1->p, ev2->p) &&
             equalExprs(ev1->a, ev2->a) &&
             equalExprs(ev1->b, ev2->b);
    }
    
    case VExpr::VE_PAT: {
      // it would be quite reasonable to not require that pattern
      // specs match up.. but I suspect this won't be an issue anyway
      VE_pat const *ep1 = e1->asVE_patC();
      VE_pat const *ep2 = e2->asVE_patC();
      return equalExprs(ep1->e, ep2->e);
    }
  }
}


bool equalExprLists(ASTList<VExpr> const &list1, ASTList<VExpr> const &list2)
{
  ASTListIter<VExpr> iter1(list1);
  ASTListIter<VExpr> iter2(list2);

  while (!iter1.isDone() && !iter2.isDone()) {
    if (!equalExprs(iter1.data(), iter2.data())) {
      return false;
    }

    iter1.adv();
    iter2.adv();
  }

  return iter1.isDone() == iter2.isDone();
}


// ---------------------- refersTo -----------------------
class RefersToVis : public VMLVisitor {
public:      // data
  StringRef var;   // variable of interest
  bool ret;        // true if we find a reference to 'var'

public:      // funcs
  RefersToVis(StringRef v) : var(v), ret(false) {}
    
  // NOTE: These visitor functions should regard 'obj' as a pointer to
  // *const*.  Unfortunately, providing a const version of the visitor
  // interface appears nontrivial without entirely duplicating the
  // traversal mechanism, so I make do with some casts and this note.

  virtual bool visitPred(Pred /*const*/ *obj);
  virtual bool visitVExpr(VExpr /*const*/ *obj);
};

bool RefersToVis::visitPred(Pred /*const*/ *obj)
{
  // quantification can hide a variable
  if (obj->isP_quant()) {
    P_quant const *q = obj->asP_quantC();

    // do any of the quantification variables match 'var'?
    FOREACH_ASTLIST(VD_var, q->vars, iter) {
      if (iter.data()->id->name == var) {
        // this qvar hides 'var', so we should not look inside the body
        return false;
      }
    }
  }

  // dig down normally
  return true;
}

bool RefersToVis::visitVExpr(VExpr /*const*/ *obj)
{
  if (obj->isVE_var()) {
    VE_var const *v = obj->asVE_var();
    if (v->id->name == var) {
      // found a reference to the variable in question!
      ret = true;

      // at this point, further traversal is unnecessary; but my
      // visitor interface doesn't provide a way to bail out
      // instantly, and throwing an exception is a bit heavy handed
      // for my tastes, so we just finish off the traversal anyway
    }
  }

  return true;
}


bool refersTo(VExpr const *e, StringRef var)
{
  RefersToVis vis(var);
  const_cast<VExpr*>(e)->traverse(vis);
  return vis.ret;
}

bool refersTo(Pred const *p, StringRef var)
{
  RefersToVis vis(var);
  const_cast<Pred*>(p)->traverse(vis);
  return vis.ret;
}


// EOF
