// vml_tosexp.cc
// code for vml_tosexp.ast

#include "vml.ast.gen.h"      // VML AST


// ------------------------ SexpEnv ------------------------
SexpEnv::SexpEnv()
{}

SexpEnv::~SexpEnv()
{}


SexpEnv& SexpEnv::operator<< (Pred const &p)
{
  p.sexp(*this);
  return *this;
}

SexpEnv& SexpEnv::operator<< (VExpr const &e)
{
  e.sexp(*this);
  return *this;
}


// ------------------------- Pred --------------------------
string Pred::toSexp() const
{
  SexpEnv se;
  sexp(se);
  return se.sb;
}

void P_lit::sexp(SexpEnv &se) const
{
  se << (b? "TRUE" : "FALSE");
}

void P_not::sexp(SexpEnv &se) const
{
  se << "(NOT " << *p << ")";
}

void P_quant::sexp(SexpEnv &se) const
{
  se << "(" << (universal? "FORALL" : "EXISTS") << " (";
  int ct=0;
  FOREACH_ASTLIST(VD_var, vars, iter) {
    if (ct++) { se << " "; }
    se << iter.data()->id->name;
  }
  se << ") ";
  
  if (patterns) {
    patterns->sexp(se);
    se << " ";
  }

  se << *body << ")";
}

static char const *opmap(BPropCon con)
{
  static char const * const map[] = {
    "AND",
    "OR",
    "IMPLIES",
    "IFF"
  };
  ASSERT_TABLESIZE(map, NUM_BPROPCONS);

  xassert((unsigned)con < NUM_BPROPCONS);
  return map[con];
}

// if 'p' matches 'con', print its elements directly instead of
// nesting them inside what would be a redudant sexp layer
void printBinary(SexpEnv &se, BPropCon con, Pred const *p)
{
  if ((con == BP_AND || con == BP_OR) && p->isBinary(con)) {
    P_binary const *b = p->asP_binaryC();
    printBinary(se, con, b->p1);
    se << " ";
    printBinary(se, con, b->p2);
  }
  else {
    // don't fold into parent; emit a self-contained sexp
    p->sexp(se);
  }
}

void P_binary::sexp(SexpEnv &se) const
{
  se << "(" << opmap(con) << " ";

  printBinary(se, con, p1);
  se << " ";
  printBinary(se, con, p2);

  se << ")";
}

static char const *opmap(ComparisonOp op)
{
  static char const * const map[] = {
    "EQ",
    "NEQ",
    "<",
    "<=",
    ">",
    ">="
  };
  ASSERT_TABLESIZE(map, NUM_COMPARISONOPS);

  xassert((unsigned)op < NUM_COMPARISONOPS);
  return map[op];
}

void P_compare::sexp(SexpEnv &se) const
{
  se << "(" << opmap(op) << " " << *e1 << " " << *e2 << ")";
}

void P_declare_partial_order::sexp(SexpEnv &se) const
{
  se << "(ORDER " << lt << " " << le << ")";
}

#ifdef VML_PAT_INCLUDED
void P_patvar::sexp(SexpEnv &se) const
{
  xfailure("E_patvar::sexp should not be called");
}
#endif // VML_PAT_INCLUDED


// -------------------------- VExpr ----------------------------
string VExpr::toSexp() const
{
  SexpEnv se;
  sexp(se);
  return se.sb;
}

void VE_lit::sexp(SexpEnv &se) const
{
  se << n;
}

void VE_var::sexp(SexpEnv &se) const
{
  se << id->name;
}

void VE_neg::sexp(SexpEnv &se) const
{
  // negation is sometimes flaky
  se << "(- 0 " << *e << ")";
}

void VE_apply::sexp(SexpEnv &se) const
{
  if (0==strcmp(func->name, "TRIGGEREQ") ||
      0==strcmp(func->name, "TRIGGERNEQ")) {
    // sort of a hack, just pass the first term as the pattern term,
    // ignoring the equality/disequality part of the constraint
    VExpr const *term = args.firstC();
    
    // argh.. it seems that Simplify doesn't like patterns like 'n',
    // so (hack hack hack) just suppress this altogether if it
    // is not a function application
    if (term->isVE_apply()) {
      se << *term;
    }
    else {
      // suppressed as degenerate
    }

    return;
  }

  se << "(" << func->name;
  FOREACH_ASTLIST(VExpr, args, iter) {
    se << " " << *(iter.data());
  }
  se << ")";
}

void VE_arith::sexp(SexpEnv &se) const
{
  se << "(" << opString(op) << " " << *e1 << " " << *e2 << ")";
}

// vcgen should never yield these
void VE_pre::sexp(SexpEnv &se) const
{ xfailure("should not happen: VE_pre::sexp"); }
void VE_cond::sexp(SexpEnv &se) const
{ xfailure("should not happen: VE_cond::sexp"); }


void VE_pat::sexp(SexpEnv &se) const
{
  // skip the 'pat' designator; a TODO is to pull these out and
  // make Simplify trigger pattern syntax with them
  se << *e;
}

#ifdef VML_PAT_INCLUDED
void VE_patvar::sexp(SexpEnv &se) const
{
  xfailure("VE_patvar::sexp should not be called");
}
#endif // VML_PAT_INCLUDED


// EOF
