// vml_vcgen.cc
// code for vml_vcgen.h and vml_vcgen.ast

#include "vml_vcgen.h"        // this module
#include "vml_build.h"        // VML AST, p_and, etc.
#include "vml_paths.h"        // isRoot
#include "trace.h"            // tracingSys
#include "strutil.h"          // replace
#include "strtable.h"         // StringTable
#include "nonport.h"          // getMilliseconds
#include "exc.h"              // xbase
#include "vml_fflow.h"        // vmlFFLowAnalyze
#include "prv_simp.h"         // SimplifyInterface
#include "prv_ket.h"          // KettleInterface
#include "strtokp.h"          // StrtokParse

// STUBEMIT: #include "vml.ast.gen.h"

#include <stdlib.h>           // getenv


// -------------------------- VCEnv --------------------------
VCEnv::VCEnv(StringTable &s, VMLFile *f)
  : prover(NULL),
    env(*this),
    str(s),
    file(f),
    globals(),
    invariants(),
    callsite(0),
    callSiteParams(false),
    preState(0),
    assumptions(0),
    tempNameCount(0),
    ptype(PT_SIMPLIFY),         // just a default
    dryRun(NULL),
    breakApartBigFacts(false),
    breakApartSmallFacts(false),
    errors(0)
{}

VCEnv::~VCEnv()
{
  if (prover) {
    delete prover;
  }
  if (dryRun) {
    delete dryRun;
  }
}

// I delay creating the Prover object until the first time it is
// actually needed because if I am generating CVC then I never need
// it, and it is silly to require Simplify to be present just to
// generate some CVC.
//
// 2005-02-02: Above isn't quite true anymore, but design is still ok.
ProverInterface *VCEnv::getProver()
{
  if (!prover) {
    if (ptype == PT_SIMPLIFY) {
      prover = new SimplifyInterface;
      prover->openLog("simp.log");
    }
    else {
      xassert(ptype == PT_CVC);
      prover = new KettleInterface;
      prover->openLog("kettle.log");
    }
  }
  return prover;
}


void VCEnv::doDryRun(char const * /*nullable*/ fname)
{
  if (dryRun) {
    delete dryRun;
  }
  
  if (!fname) {
    fname = ptype==PT_SIMPLIFY? "simp.log" : "kettle.log";
  }

  dryRun = new ofstream(fname);
}


void VCEnv::beginPath(VProcedure *proc, bool procStart)
{
  // reset globals
  SFOREACH_OBJLIST_NC(VD_var, globals, iter1) {
    iter1.data()->beginPath(procStart);
  }

  // reset locals
  SFOREACH_OBJLIST_NC(VD_var, proc->locals, iter2) {
    iter2.data()->beginPath(procStart);
  }

  // reset params/returns
  FOREACH_ASTLIST_NC(VD_var, proc->params, iter3) {
    iter3.data()->beginPath(procStart);
  }
  FOREACH_ASTLIST_NC(VD_var, proc->returns, iter4) {
    iter4.data()->beginPath(procStart);
  }

  callsite = 0;
  tempNameCount = 0;
}


void VCEnv::endPath()
{
  // pop all assumptions from this path
  popAssumptions(assumptions, "path", "assumptions");
  assumptions = 0;
}


Pred * owner VCEnv::vcgenGlobalInv()
{
  Pred * owner ret = NULL;

  FOREACH_ASTLIST_NC(VTopForm, file->forms, iter) {
    if (iter.data()->isVTF_inv()) {
      Pred *inv = iter.data()->asVTF_inv()->p;

      Pred * owner vcinv = inv->vcgen(env);
      if (!ret) {
        ret = vcinv;
      }
      else {
        ret = p_and(ret, vcinv);
      }
    }
  }

  return ret? ret : new P_lit(SL_UNKNOWN, true);
}
    

char const *VCEnv::commentStart() const
{
  return ptype==PT_SIMPLIFY? "; " : "% ";
}

char const *VCEnv::nlCommentStart() const
{
  return ptype==PT_SIMPLIFY? "\n; " : "\n% ";
}


void VCEnv::comment(rostring c)
{
  if (c[0]) {
    feed(stringc << commentStart() << c << "\n");
  }
  else {
    // let "" mean to add a blank line
    feed("\n");
  }
}

void VCEnv::commentLoc(SourceLoc loc, rostring c)
{
  feed(stringc << commentStart() << toString(loc) << ": " << c << "\n");
}

void VCEnv::multilineComment(rostring c)
{
  feed(stringc << commentStart() << replace(c, "\n", nlCommentStart()) << "\n");
}


void VCEnv::commentLocBlock(SourceLoc loc, rostring c)
{
  if (ptype == PT_SIMPLIFY) {
    commentLoc(loc, c);
  }
  else {
    char const *fname;
    int line, col;
    sourceLocManager->decodeLineCol(loc, fname, line, col);
    feed(stringc << "BLOCK [file:\"" << fname
                 << "\",line:\"" << line
                 << "\",name:\"" << c << "\"]\n");
  }
}

void VCEnv::endBlock()
{
  if (ptype == PT_CVC) {
    feed("ENDBLOCK\n\n");
  }
}


// true if we are to print the various details of paths, proof
// obligations, etc.
bool VCEnv::verboseVerification()
{
  // only print these if we're actually running the prover
  return !dryRun;
}

void VCEnv::pcomment(rostring c)
{
  comment(c);

  if (verboseVerification()) {
    cout << c << "\n";
  }
}


void VCEnv::commentPred(Pred const *p)
{
  BoxPrint bp;
  printBigPred(bp, p);

  string c = BPRender().takeAndRender(bp);
  multilineComment(c);
}


void VCEnv::commentStmt(VStatement const *stmt)
{
  // we want to fill this nominally with 'stmt->print()', except
  // that for things like compound statements, just make it a summary
  string c;

  ASTSWITCHC(VStatement, stmt) {
    ASTCASEC(VS_if, i) {
      c = stringc << "if (" << i->p->print() << ") {...}";
    }
    ASTNEXTC1(VS_loop) {
      c = "while (true) {...}";
    }
    ASTNEXTC1(VS_compound) {
      c = "{...}";
    }
    ASTDEFAULTC {
      c = stmt->print();
    }
    ASTENDCASEC
  }

  // put the comment prefix into the internal linebreaks, if any
  string prefix = stringc << nlCommentStart() << toString(stmt->loc) << ": ";
  c = replace(c, "\n", prefix);
  commentLoc(stmt->loc, c);
}


string toProverLang(VCEnv &env, Pred *p)
{
  return env.ptype==env.PT_SIMPLIFY? p->toSexp() : p->toCVC();
}

string toProverLang(VCEnv &env, VExpr *e)
{
  return env.ptype==env.PT_SIMPLIFY? e->toSexp() : e->toCVC();
}


void VCEnv::assume(Pred * owner p, SourceLoc loc, rostring why)
{
  if (breakApartSmallFacts && p->isBinary(BP_AND)) {
    P_binary const *pbin = p->asP_binaryC();
    assume(pbin->p1, loc, why);
    assume(pbin->p2, loc, why);
    return;
  }

  commentLocBlock(loc, why);
  commentPred(p);

  string s = toProverLang(*this, p);
  innerAssume(s);
  delete p;

  // I think the endBlock should go here so that a linear sequence of
  // assumptions looks linear in the GUI, but main.ml assumes that
  // block/endblock and assert/retract will be properly nested w.r.t.
  // each other, so I cannot (get error message: "Input UnBlock occurs
  // at top-level")
  //
  // trying something else to fix it ...
  endBlock();
}

void VCEnv::innerAssume(rostring pred)
{
  assumptions++;
  feed(ptype==PT_SIMPLIFY? "(BG_PUSH " : "ASSERT (");
  feed(pred);
  feed(ptype==PT_SIMPLIFY? ")\n\n" : ");\n");
}

void VCEnv::assume(rostring pred, SourceLoc loc, rostring why)
{
  commentLocBlock(loc, why);
  innerAssume(pred);
  endBlock();
}


void VCEnv::printGoal(rostring label, Pred const *p)
{
  cout << label << ": AND {\n";

  BoxPrint bp;
  printSemiConj(bp, p);

  BPRender bpr;
  bpr.lineStartText = "  ";
  bpr.reset();
  cout << bpr.takeAndRender(bp);

  cout << "\n}\n";
}

                                
// dig down into 'p' to find a conjunction that can be broken apart
// to form simpler predicates whose conjunction is equivalent to 'p';
// such breaking is the main operation of goal simplification; returns
// NULL if no such conjunction can be found
Pred **VCEnv::findConjunction(Pred **p)
{
  ASTSWITCH(Pred, *p) {
    ASTCASE(P_binary, b) {
      if (b->con == BP_AND) {
        // this is a good conjunction right here
        return p;
      }
      if (b->con == BP_IMPLIES) {
        // try to break down the RHS
        return findConjunction(&(b->p2));
      }
    }
    ASTNEXT(P_quant, q) {
      if (q->universal) {
        // break down body
        return findConjunction(&(q->body));
      }
    }
    // TODO: I could break apart P_comparison, but I'd need to
    // generalize this function's interface...
    ASTENDCASED
  }
              
  // nothing to break
  return NULL;       
}


string VCEnv::proofObligation(Pred *p)
{
  if (ptype==PT_SIMPLIFY) {
    return p->toSexp() & "\n\n";
  }
  else {
    return stringc << "QUERY " << p->toCVC() << ";\n";
  }
}

void VCEnv::prove(Pred * owner p, SourceLoc loc, rostring why)
{
  if (breakApartSmallFacts && p->isBinary(BP_AND)) {
    P_binary const *pbin = p->asP_binaryC();
    prove(pbin->p1, loc, why);
    prove(pbin->p2, loc, why);
    return;
  }

  commentLocBlock(loc, why);
  commentPred(p);

  string s = proofObligation(p);
  string counterexample;

  if (verboseVerification()) {
    cout << "  " << why << " ...";
    cout.flush();
  }

  long start = getMilliseconds();
  bool ok = prove(s, &counterexample);
  long end = getMilliseconds();

  endBlock();

  if (verboseVerification()) {
    cout << " " << (ok? "" : "NOT") << " proved ("
         << (end-start) << " ms)\n";
  }

  if (!ok) {
    errors++;

    printGoal("failed proof", p);

    // break apart 'p' to assist diagnosis of failure
    bool simplified = false;
    Pred **conjptr = findConjunction(&p);
    while (conjptr) {
      P_binary *conj = (*conjptr)->asP_binary();

      // temporarily replace the conjunction with its LHS
      *conjptr = conj->p1;

      // try to prove the simplified goal
      s = proofObligation(p);
      if (!prove(s, &counterexample)) {
        // LHS is not provable, throw away RHS
        conj->p1 = NULL;    // don't delete LHS
        delete conj;
      }
      else {
        // LHS is provable, so switch to RHS
        *conjptr = conj->p2;
        conj->p2 = NULL;    // don't delete RHS
        delete conj;
      }

      simplified = true;
      conjptr = findConjunction(&p);
    }

    if (simplified) {
      printGoal("simplified", p);
    }
    
    if (counterexample.length() > 0) {
      cout << counterexample;
    }
  }

  delete p;

  if (!ok && !tracingSys("continue")) {
    // unless requested, stop on the first failure
    xbase("failed proof");
  }
}


void VCEnv::feed(rostring text)
{
  if (!dryRun) {
    getProver()->feed(text);
  }
  else {
    *dryRun << text;
  }
}

bool VCEnv::prove(rostring text, string *counterexample)
{
  if (!dryRun) {
    return getProver()->prove(text, counterexample);
  }
  else {
    *dryRun << text;
    return true;      // assume it worked
  }
}


void VCEnv::popAssumptions(int howMany, rostring context, rostring kind)
{
  comment("");
  comment(stringc << "end of " << context
                  << ", popping " << howMany << " " << kind);
  for (int i=0; i < howMany; i++) {
    feed(ptype==PT_SIMPLIFY? "(BG_POP) " : "RETRACT; ");
  }
  feed("\n\n");
}


static bool isParam(VDeclaration *d)
{
  if (!d->isVD_var()) { return false; }

  VarKind vk = d->asVD_var()->vkind;
  return vk == VK_PARAM ||
         vk == VK_RETVAL ||
         vk == VK_IOPARAM;
}

VE_var* owner VCEnv::declVar(VDeclaration *d)
{
  if (callSiteParams && isParam(d)) {
    return paramVar(d);
  }

  if (d->isVD_var()) {
    VD_var *v = d->asVD_var();
    xassert(v->alpha);
                         
    // what modct to use?
    int modct = v->modct;
    if (preState) {
      if (callSiteParams && v->vkind == VK_GLOBAL) {
        // we're asking for the pre-state of a global at a call site
        modct = v->callSitePre;
      }
      else {
        // initial value?  this is probably not right in all cases...
        modct = 0;
      }
    }

    if (modct == 0) {
      // initial value gets no number (suggestion from George)
      return new VE_var(makeId(v->alpha));
    }
    else {
      // others get a modification number
      return new VE_var(makeId(str(stringc << v->alpha << "_" << modct)));
    }
  }
  else {
    // assume no alpha needed
    return new VE_var(makeId(d->id->name));
  }
}


VE_var* owner VCEnv::paramVar(VDeclaration *d)
{
  // only makes sense if the declaration is a param/return
  VD_var *v = d->asVD_var();

  char const *prefix;
  switch (v->vkind) {
    default: xfailure("paramVar on a non-param?");
    case VK_PARAM:   prefix = "param";   break;
    case VK_RETVAL:  prefix = "retval";  break;
    case VK_IOPARAM: prefix = "ioparam"; break;
  }

  return new VE_var(makeId(str(stringc << prefix << callsite << "_" << d->id->name)));
}


void VCEnv::assumeGlobalInvariant(SourceLoc loc, rostring context)
{
  if (breakApartBigFacts) {
    foreachGlobalInvariantConjunct(loc, context, true /*assume*/);
  }
  else {
    env.assume(env.vcgenGlobalInv(), loc,
               stringc << "assume global invariant " << context);
  }
}


void VCEnv::proveGlobalInvariant(SourceLoc loc, rostring context)
{
  if (breakApartBigFacts) {
    foreachGlobalInvariantConjunct(loc, context, false /*assume*/);
  }
  else {
    env.prove(env.vcgenGlobalInv(), loc,
              stringc << "prove global invariant " << context);
  }
}


void VCEnv::foreachGlobalInvariantConjunct(SourceLoc loc, rostring context,
                                           bool doAssume)
{
  commentLocBlock(loc, stringc << "global invariant block " << context);
  char const *proveAssume = (doAssume? "assume" : "prove");

  // break the invariant apart into separate conjuncts to make it
  // easier for a human to read the CVC file
  int counter = 0;
  FOREACH_ASTLIST_NC(VTopForm, file->forms, iter) {
    if (iter.data()->isVTF_inv()) {
      Pred *inv = iter.data()->asVTF_inv()->p;

      foreachConjunct(inv, stringc << proveAssume << " global invariant " << context,
                      doAssume, counter);
    }
  }

  endBlock();
}


void VCEnv::foreachConjunct(Pred *pred, rostring context, bool doAssume,
                            int &counter)
{
  if (pred->isBinary(BP_AND)) {
    P_binary *bin = pred->asP_binary();
    foreachConjunct(bin->p1, context, doAssume, counter);
    foreachConjunct(bin->p2, context, doAssume, counter);
  }
  else {
    string desc = stringc << context << ", conjunct " << counter++;
    if (doAssume) {
      env.assume(pred->vcgen(env), pred->loc, desc);
    }
    else {
      env.prove(pred->vcgen(env), pred->loc, desc);
    }
  }
}


void VCEnv::proveBigFact(Pred *pred, SourceLoc loc, rostring context)
{
  proveOrAssumeBigFact(pred, loc, stringc << "prove " << context,
                       false /*doAssume*/);
}

void VCEnv::assumeBigFact(Pred *pred, SourceLoc loc, rostring context)
{
  proveOrAssumeBigFact(pred, loc, stringc << "assume " << context,
                       true /*doAssume*/);
}

void VCEnv::proveOrAssumeBigFact(Pred *pred, SourceLoc loc, rostring context,
                                 bool doAssume)
{
  if (breakApartBigFacts) {            
    commentLocBlock(loc, stringc << context << " block");

    int counter = 0;
    foreachConjunct(pred, context, doAssume, counter);

    endBlock();
  }
  else {
    if (doAssume) {
      assume(pred->vcgen(env), loc, context);
    }
    else {
      prove(pred->vcgen(env), loc, context);
    }
  }
}


// Mainly, this function is just doing the "modct++".  However, if that
// is all that is done, then it is not clear in the VC just where the
// variable got scrambled.  So the idea here is to emit something like
//
//   foo_3 = modified(foo_2, 11, 1)
//
// to record where the modification happened.  The final number is
// never the same as any other emitted 'modified' application (on this
// path), and 'modified' is itself uninterpreted, so the theorem
// prover will never use this to draw any conclusion about the new
// value.
void VCEnv::scrambleGlobal(VD_var *v, SourceLoc loc)
{
  // wrap the old value in a 'modified' application
  VExpr* owner mod = apply3(makeId(str("modified")),
                            env.declVar(v),
                            e_lit(callsite),
                            e_lit(tempNameCount++));

  // bump the variable
  v->modct++;

  // assign the new value to the new 'modified' app
  env.assume(p_equal(env.declVar(v), mod), loc,
             stringc << "scrambled variable " << v->id->name);
}


// ------------------------- VMLFile -------------------------
void VMLFile::vcgen(VCEnv &env)          // stubify
{
  env.commentLocBlock(forms.first()->loc, "whole file");

  // pass 1:
  //  - make a list of globals
  //  - make a list of invariant predicates
  //  - emit list of constants to Simplify
  //  - emit axioms as global assumptions
  //  - do nothing for the function symbols
  int emittedAxioms = 0;
  {
    SObjList<VD_const> constants;
    FOREACH_ASTLIST_NC(VTopForm, forms, iter) {
      ASTSWITCH(VTopForm, iter.data()) {
        ASTCASE(VTF_decl, d) {
          ASTSWITCH(VDeclaration, d->d) {
            ASTCASE(VD_var, v) {
              env.globals.prepend(v);
            }
            ASTNEXT(VD_const, c) {
              constants.prepend(c);
            }
            ASTENDCASED
          }
        }

        ASTNEXT(VTF_inv, i) {
          env.invariants.prepend(i->p);
        }

        ASTNEXT(VTF_axiom, a) {    
          if (env.ptype==env.PT_CVC &&
              a->label &&
              0==strcmp(a->label, "\"sel/upd\"")) {
            // this axiom is to be suppressed because it is already
            // built in to Kettle
            env.comment("suppressing sel/upd axioms");
            break;
          }

          string label = a->label? 
            stringc << "axiom: " << parseQuotedString(a->label) :
            string("axiom") ;
          env.assume(a->p, a->loc, label);
          emittedAxioms++;
        }

        ASTENDCASED
      }
    }

    env.globals.reverse();
    constants.reverse();
    env.invariants.reverse();

    // emit constants
    if (constants.isNotEmpty()) {
      stringBuilder sb;

      // Note: Equality of 1 and |@true|:
      //
      // When a function application f(x) is used like a predicate,
      // c2vml typically translates this as "f(x)!=0", is accordance
      // with the usual C conventions that nonzero is true.  However,
      // if a predicate has been declared to be a partial order using
      // declare_partial_order, then Simplify will want to see
      // "f(x)==|@true|".  But that quoting syntax is not legal in
      // VML, so c2vml instead emits "f(x)==1", and then (right here)
      // I tell Simplify that 1 is an alias for |@true|.

      SourceLoc loc = constants.firstC()->id->loc;

      if (env.ptype == env.PT_SIMPLIFY) {
        sb << "(AND\n"
           << "  (EQ 1 |@true|)\n";
        SFOREACH_OBJLIST(VD_const, constants, iter) {
          VD_const const *c = iter.data();
          sb << "  (EQ " << c->id->name << " " << c->value->toSexp() << ")\n";
        }
        sb << ")";   // final newline added by env.assume()

        env.assume(sb, loc, "symbolic constants");
        emittedAxioms++;
        env.comment("");
      }
      else {
        // It might be nice to emit this block as a single assertion
        // when breakApartBigFacts is false, but oh well.

        env.commentLocBlock(loc, "symbolic constants");

        SFOREACH_OBJLIST(VD_const, constants, iter) {
          VD_const const *c = iter.data();
          env.assume(stringc << c->id->name << " = " << c->value->toCVC(),
                     c->id->loc, stringc << "constant: " << c->id->name);
        }
        
        env.endBlock();
      }
    }
  }

  // all assumptions added above are *not* path-local
  env.assumptions = 0;

  // pass 2: for each procedure, vcgen it
  {
    FOREACH_ASTLIST_NC(VTopForm, forms, iter) {
      if (iter.data()->isVTF_proc()) {
        VProcedure *p = iter.data()->asVTF_proc()->p;
        p->vcgen(env);
      }
    }
  }

  // pop the axioms
  env.popAssumptions(emittedAxioms, "file", "axioms");

  env.comment("end of file");
  env.endBlock();
}


// --------------------- VProcedure -----------------------
void VProcedure::vcgen(VCEnv &env)           // stubify
{
  {
    if (getenv("SELFUNC")) {
      throw_XFatal("Variable SELFUNC has been renamed to SELPROC");
    }

    char const *selproc = getenv("SELPROC");
    if (selproc && 0!=strcmp(selproc, id->name)) {
      // skip this procedure
      return;
    }
  }

  if (!body) {
    return;      // skip prototypes
  }

  env.pcomment(stringc << "-------------------- " 
                       << id->name << " (" << pluraln(roots.count(), "root")
                       << ") --------------------");

  if (!tracingSys("disableFFlow")) {
    vmlFFlowAnalyze(this, env.str);
  }

  char const *selpath = getenv("SELPATH");
  int selectedRoot=0, selectedPath=0;     // only meaningful if 'selPath!=NULL'
  if (selpath) {
    StrtokParse tok(selpath, "/");
    if (tok != 2) {
      xbase(stringc << "SELPATH must be of the form X/Y (it is \"" << selpath << "\")");
    }
    selectedRoot = atoi(tok[0]);
    selectedPath = atoi(tok[1]);
  }

  if (env.ptype == VCEnv::PT_CVC) {
    env.commentLocBlock(id->loc, id->name);
  }

  // iterate over path roots
  int rootCt = -1;    // so first root is 0
  SFOREACH_OBJLIST_NC(VStatement, roots, rootIter) {
    VStatement *root = rootIter.data();
    rootCt++;

    if (selpath && selectedRoot!=rootCt) {
      continue;        // skip this root, including print its BLOCK
    }

    env.comment(stringc << id->name << ", root " << rootCt
                        << ", " << pluraln(root->paths, "path"));
    if (env.ptype == VCEnv::PT_CVC) {
      env.commentLocBlock(root->loc, stringc << "root " << rootCt);
    }

    // iterate over paths from the root
    for (int path=0; path < root->paths; path++) {
      int localPath = path;

      if (selpath && selectedPath!=path) {
        continue;      // skip this path
      }

      string pathDesc = stringc << rootCt << "/" << path;
      env.pcomment(stringc << id->name << ", path " << pathDesc);
      if (env.ptype == VCEnv::PT_CVC) {
        env.commentLocBlock(root->loc, stringc << "path " << pathDesc);
      }
      env.comment("");

      if (root == body) {
        // reset all of the counters associated with variables
        env.beginPath(this, true /*procStart*/);

        // beginning of the procedure: assume precondition and global inv
        env.assumeGlobalInvariant(id->loc, "at procedure start");
        env.assumeBigFact(pre, id->loc,
                          "precondition at procedure start");
      }
      else {
        // assume the function precondition held on entry
        env.beginPath(this, true /*procStart*/);    // set modct's to 0
        env.assumeBigFact(pre, id->loc,
                          "the precondition held at proc start");

        // reset all of the counters associated with variables
        env.beginPath(this, false /*procStart*/);   // set modct's to 1

        // assume invariant
        VS_pred *vsp = root->asVS_pred();
        if (vsp->psk != PS_STANDALONE_INVARIANT) {
          env.assumeGlobalInvariant(id->loc, "at invariant point");
        }
        env.assumeBigFact(root->asVS_pred()->p, root->loc,
                          "invariant at path start");
      }

      // walk down the path
      VStatement *s = root;
      do {
        // make sure our path count didn't go haywire
        xassert(0 <= localPath && localPath < s->paths);

        // emit the statement as a comment
        env.commentStmt(s);

        // vcgen this statement
        s->vcgen(env);

        // move to the next statement in the path
        if (!s->isVS_if()) {
          s = s->next;
        }
        else {
          VS_if *i = s->asVS_if();

          // compute guard predicate
          Pred * owner guard = i->p->vcgen(env);

          if (localPath < i->sThen->paths) {
            // 'then' branch
            s = i->sThen;
            env.assume(guard, i->loc, "true branch of 'if' statement");
          }
          else {
            // taking the 'else' branch: decrement 'localPath', so
            // it now indicates which path to take among those that
            // use the else branch, and move 's' along that branch
            localPath -= i->sThen->paths;
            s = i->sElse? i->sElse : i->next;
            env.assume(p_not(guard), i->loc, "false branch of 'if' statement");
          }
        }
      } while (s && !isRoot(s));

      if (!s) {
        // prove postcondition and global inv
        env.proveBigFact(post, id->loc, "procedure postcondition");
        env.proveGlobalInvariant(id->loc, "at end of procedure");
      }
      else {
        // prove invariant
        VS_pred *vsp = s->asVS_pred();
        env.proveBigFact(vsp->p, s->loc, "invariant at path end");
        if (vsp->psk != PS_STANDALONE_INVARIANT) {
          env.proveGlobalInvariant(id->loc, "at invariant point");
        }
      }

      env.endPath();
      env.comment(stringc << "end of path " << pathDesc);
      env.endBlock();
    }
    env.comment(stringc << "end of root " << rootCt);
    env.endBlock();
  }
  env.comment(stringc << "end of " << id->name);
  env.endBlock();
}


// ------------------- VDeclaration --------------------
void VD_var::beginPath(bool procStart)               // stubify
{
  // the idea here is that pre(x) refers to x_0, i.e. x's value at the
  // beginning of the procedure, while x_1 will be the value at the
  // start of the path
  modct = procStart? 0 : 1;
}


// -------------------- VStatement ---------------------
// STUBPATTERN: ^void.*VS_.*::vcgen\b

void VS_comment::vcgen(VCEnv &env)
{
  env.comment(c);
}


void VS_decl::vcgen(VCEnv &env)
{
  // VDeclarations are largely handled by scans in advance that collect
  // the relevant names, but where there are initializers we need to
  // include that info, and now is as good a time as any.

  ASTSWITCH(VDeclaration, d) {
    ASTCASE(VD_var, v) {
      // 2005-03-06: Fix a soundness bug.  I was not bumping this before,
      // which meant that vcgen was in essence assuming the variable's
      // value did not change when it went out of scope and came back in.
      v->modct++;

      if (v->init) {
        env.assume(p_equal(env.declVar(v), v->init->vcgen(env)), loc,
                   stringc << "initial value of variable " << v->id->name);
      }
    }
    ASTNEXT(VD_const, c) {
      env.assume(p_equal(env.declVar(c), c->value->vcgen(env)), loc,
                 stringc << "value of constant " << c->id->name);
    }
    ASTENDCASED
  }
}


void VS_pred::vcgen(VCEnv &env)
{
  if (isInvariant()) {
    // invariants are handled separately, by VProcedure::vcgen(), as
    // the path endpoints; nothing needs to be done here
    return;
  }

  // vcgen the predicate
  Pred * owner vcp = p->vcgen(env);

  // do something with it
  switch (psk) {
    default: xfailure("bad code");
    case PS_ASSUME: 
      env.assume(vcp, loc, "user assumption"); 
      break;

    case PS_ASSERT:
      env.prove(vcp->clone(), loc, "user assertion");
      env.assume(vcp, loc, "assume a just-proved fact");
      break;

    case PS_PURE_ASSERT:
      env.prove(vcp, loc, "user assertion");
      break;
  }
}


void VS_assign::vcgen(VCEnv &env)
{
  // evaluate the expression; this will use the target variable's old
  // value, if the expression refers to it
  VExpr* owner evalSrc = source->vcgen(env);

  // bump the count on the target so it can refer to the new value
  target->var->asVD_var()->modct++;

  // assign the new value to the new variable
  env.assume(p_equal(env.declVar(target->var), evalSrc), loc,
             stringc << "assignment to variable " << target->name);
}


void VS_procCall::vcgen(VCEnv &env)
{
  // increment call site index so we don't confuse params/returns
  // from other call sites
  env.callsite++;
  env.comment(stringc << "call site " << env.callsite 
                      << ", procedure " << procName->name);

  // global invariant must hold
  env.proveGlobalInvariant(loc, "before procedure call");

  // bind the procedure's in-parameters to the argument expression values
  {
    int argIndex = 0;
    FOREACH_ASTLIST_NC(VD_var, proc->params, iter) {
      VD_var *param = iter.data();
      VExpr * owner argExpr = args.nth(argIndex)->vcgen(env);

      env.assume(p_equal(env.paramVar(param), argExpr), loc,
                 stringc << "binding parameter " << param->id->name
                         << " at call site");

      argIndex++;
    }
  }

  // relative to this call site, this is the pre-state for globals
  {
    SFOREACH_OBJLIST_NC(VD_var, env.globals, iter) {
      VD_var *g = iter.data();
      g->callSitePre = g->modct;
    }
  }

  // prove the procedure's precondition holds for these arguments
  env.callSiteParams = true;
  env.proveBigFact(proc->pre, loc, "procedure call precondition");
  env.callSiteParams = false;

  // *** conceptually, the procedure call happens here ***

  // freshen the globals that are potentially written by the procedure
  {
    if (proc->writes.count() > 1) {
      env.commentLocBlock(loc, "scrambling global variables");
    }

    FOREACH_ASTLIST_NC(Identifier, proc->writes, iter) {
      VD_var *v = iter.data()->var->asVD_var();
      env.scrambleGlobal(v, loc);
    }

    if (proc->writes.count() > 1) {
      env.endBlock();
    }
  }

  // assume the global invariant holds for the freshened values
  env.assumeGlobalInvariant(loc, "after procedure call");

  // assume the postcondition, effectively binding the values of
  // the procedure's out-parameters
  env.callSiteParams = true;
  env.assumeBigFact(proc->post, loc, "callee's postcondition");
  env.callSiteParams = false;

  // now, assign the local result variables to those out-parameters
  {
    int retIndex = 0;
    FOREACH_ASTLIST_NC(VD_var, proc->returns, iter) {
      VD_var *ret = iter.data();
      VD_var *tgt = targets.nth(retIndex)->var->asVD_var();

      env.assume(p_equal(env.paramVar(ret), env.declVar(tgt)), loc,
                 stringc << "assigning " << tgt->id->name 
                         << " to retval " << ret->id->name);

      retIndex++;
    }
  }
}


// the rest of these have nothing to do, because VProcedure::vcgen()
// takes care of all control flow and implications thereof
void VS_if::vcgen(VCEnv &env) {}
void VS_loop::vcgen(VCEnv &env) {}
void VS_label::vcgen(VCEnv &env) {}
void VS_goto::vcgen(VCEnv &env) {}
void VS_compound::vcgen(VCEnv &env) {}

#ifdef VML_PAT_INCLUDED
void VS_patvar::vcgen(VCEnv &env)
{
  xfailure("VS_patvar::vcgen should not be called");
}
#endif // VML_PAT_INCLUDED


// ---------------------- Pred ------------------------
// STUBPATTERN: ^Pred.*P_.*::vcgen\b

Pred* owner P_lit::vcgen(VCEnv &env)
{
  return clone();
}

Pred* owner P_not::vcgen(VCEnv &env)
{
  return new P_not(loc, p->vcgen(env));
}

Pred* owner P_quant::vcgen(VCEnv &env)
{
  // for this to work, declVar needs to not do alpha conversions
  // on quantifier variables...
  return new P_quant(loc, universal, label, cloneASTList(vars),
                     patterns? patterns->vcgen(env) : NULL, body->vcgen(env));
}

Pred* owner P_binary::vcgen(VCEnv &env)
{
  return new P_binary(loc, p1->vcgen(env), con, p2->vcgen(env));
}

Pred* owner P_compare::vcgen(VCEnv &env)
{                    
  return new P_compare(loc, e1->vcgen(env), op, e2->vcgen(env));
}

Pred* owner P_declare_partial_order::vcgen(VCEnv &env)
{
  return clone();
}

#ifdef VML_PAT_INCLUDED
Pred* owner P_patvar::vcgen(VCEnv &env)
{
  xfailure("P_patvar::vcgen should not be called");
  return NULL;
}
#endif // VML_PAT_INCLUDED


// ------------------------ VExpr -------------------------
// STUBPATTERN: ^VExpr.*VE_.*::vcgen\b

VExpr* owner VE_lit::vcgen(VCEnv &env)
{
  return clone();
}

VExpr* owner VE_var::vcgen(VCEnv &env)
{
  return env.declVar(id->var);
}

VExpr* owner VE_neg::vcgen(VCEnv &env)
{
  return new VE_neg(e->vcgen(env));
}

VExpr* owner VE_apply::vcgen(VCEnv &env)
{
  VE_apply* owner ret = new VE_apply(func->clone(), NULL /*args*/);
  FOREACH_ASTLIST_NC(VExpr, args, iter) {
    VExpr *arg = iter.data();
    ret->args.append(arg->vcgen(env));
  }
  return ret;
}

VExpr* owner VE_arith::vcgen(VCEnv &env)
{
  return new VE_arith(e1->vcgen(env), op, e2->vcgen(env));
}

VExpr* owner VE_pre::vcgen(VCEnv &env)
{
  env.preState++;
  VExpr* owner ret = e->vcgen(env);
  env.preState--;
  return ret;
}

VExpr* owner VE_cond::vcgen(VCEnv &env)
{
  StringRef temp = env.str(stringc << "condtmpvalue" << env.tempNameCount++);

  // p ==> (temp == a)
  env.assume(p_impl(p->vcgen(env),
                    p_equal(new VE_var(makeId(temp)), a->vcgen(env))),
             SL_UNKNOWN, "true case of VE_cond");

  // !p ==> (temp == b)
  env.assume(p_impl(p_not(p->vcgen(env)),
                    p_equal(new VE_var(makeId(temp)), b->vcgen(env))),
             SL_UNKNOWN, "false case of VE_cond");

  // use temp as the value for this expression
  return new VE_var(makeId(temp));
}

VExpr* owner VE_pat::vcgen(VCEnv &env)
{
  // triggers shouldn't occur in mainline code, only in axioms; the
  // type checker ought to be preventing this
  xfailure("VE_pat in mainline code");
  return NULL;
}

#ifdef VML_PAT_INCLUDED
VExpr* owner VE_patvar::vcgen(VCEnv &env)
{
  xfailure("VE_patvar::vcgen should not be called");
  return NULL;
}
#endif // VML_PAT_INCLUDED


// EOF
